// POOMTask.cpp: implementation of the POOMTask class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POOMTask.h"

#include "PimalLongHelper.h"
#include "PimalDateHelper.h"
#include "PimalBoolHelper.h"
#include "POOMRecurrence.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// extractViewData
// extracts only the object id from this task and puts it into a String formatted like this:
// [ID on device]+DATASET_SEP
// ...
//////////////////////////////////////////////////////////////////////

CString POOMTask::extractViewData()
{
	CString retVal;
	// get the information needed out of the actual item and store it into temp variables
	long id;	
	task->get_Oid(&id);
	PimalLongHelper idH(id);
		
	// attach a new line with the data accomplished to the return value
	retVal = idH.getCString() + DATASET_SEP;

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// extractAllData
// extracts all  datafrom this task puts it into a String formatted like this:
// [ID on device]+D1_SEP+[Subject]+D1_SEP+[Categories]+D1_SEP+[StartDate]+D1_SEP+[DueDate]+D1_SEP+[DateCompleted]+D1_SEP+[Importance]+D1_SEP+[Complete]+D1_SEP+[RecurrencePattern]+D1_SEP+[Sensitivity]+D1_SEP+[TeamTask]+D1_SEP+[ReminderOptions]+D1_SEP+[ReminderTime]+D1_SEP+[Note]+DATASET_SEP;
// ...
//////////////////////////////////////////////////////////////////////

CString POOMTask::extractAllData()
{
	CString retVal;
	// get the information needed out of the actual item and store it into temp variables
	LPTSTR subject, categories, reminderSoundFile, note;
	DATE startDate, dueDate, dateCompleted, reminderTime;
	VARIANT_BOOL complete, isRecurring, teamTask, reminderSet;
	long id, importance, sensitivity, reminderOptions;	
	IRecurrencePattern* recurrence;

	task->get_Oid(&id);
	PimalLongHelper idH(id);
	task->get_Subject(&subject);
	task->get_Categories(&categories);
	task->get_StartDate(&startDate);
	PimalDateHelper startDateH(startDate);
	task->get_DueDate(&dueDate);
	PimalDateHelper dueDateH(dueDate);
	task->get_DateCompleted(&dateCompleted);
	PimalDateHelper dateCompletedH(dateCompleted);
	task->get_Importance(&importance);
	PimalLongHelper importanceH(importance);
	task->get_Complete(&complete);
	PimalBoolHelper completeH(complete);
	task->get_IsRecurring(&isRecurring);
	task->get_Sensitivity(&sensitivity);
	PimalLongHelper sensitivityH(sensitivity);
	task->get_TeamTask(&teamTask);
	PimalBoolHelper teamTaskH(teamTask);
	task->get_ReminderSet(&reminderSet);
	PimalBoolHelper reminderSetH(reminderSet);
	task->get_ReminderSoundFile(&reminderSoundFile);
	task->get_ReminderOptions(&reminderOptions);
	PimalLongHelper reminderOptionsH(reminderOptions);
	task->get_ReminderTime(&reminderTime);
	PimalDateHelper reminderTimeH(reminderTime);
	task->get_Body(&note);

	task->GetRecurrencePattern(&recurrence);
	POOMRecurrence recurrenceH(isRecurring, recurrence);

	// attach a new line with the data accomplished to the return value
	retVal = idH.getCString() + D1_SEP + (CString)subject + D1_SEP + (CString)categories + D1_SEP + startDateH.getGenericDateString() + D1_SEP + dueDateH.getGenericDateString() + D1_SEP + dateCompletedH.getGenericDateString() + D1_SEP + importanceH.getCString() + D1_SEP + completeH.getCString() + D1_SEP + recurrenceH.getGenericRecurrenceString() + D1_SEP + sensitivityH.getCString() + D1_SEP + teamTaskH.getCString() + D1_SEP + reminderOptionsH.getCString() + D1_SEP + reminderTimeH.getGenericDateString() + D1_SEP + (CString)note + DATASET_SEP;

	return retVal;
}


//////////////////////////////////////////////////////////////////////
// save() saves the stored ITask (all changes made are stored on the device
// and refreshes the oid
//////////////////////////////////////////////////////////////////////

BOOL POOMTask::save()
{
	HRESULT h = task->Save();
	long id;
	task->get_Oid(&id);

	PimalLongHelper idH(id);
	this->oid = idH.getCString().AllocSysString();
   return h == 0;
}

//////////////////////////////////////////////////////////////////////
// getOid() returns the Oid of the IContact as a BSTR
//////////////////////////////////////////////////////////////////////

BSTR POOMTask::getOid()
{
	return this->oid;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

POOMTask::POOMTask(ITask* tmp)
{
	this->task = tmp;
}

POOMTask::POOMTask(ITask* tmp, BSTR subject)
{
	this->task = tmp;
   task->put_Subject(subject);
}

POOMTask::POOMTask(ITask* tmp, BSTR subject, BSTR categories, BSTR startDate, BSTR dueDate, BSTR importance, BSTR completed, 
                   BSTR isRecurring, BSTR duration, BSTR recurrenceType, BSTR occurrences, BSTR interval, BSTR dayOfWeek, BSTR dayOfMonth, 
                   BSTR weekOfMonth, BSTR monthOfYear, BSTR patternStartDate, BSTR patternEndDate, BSTR startTime, BSTR endTime, BSTR noEndDate, 
                   BSTR sensitivity, BSTR teamTask, BSTR reminderIsSet, BSTR reminderOptions, BSTR reminderTime, BSTR note)
{
	this->task = tmp;

	if (subject && subject != UNCHANGED)
      task->put_Subject(subject);
	if (categories && categories != UNCHANGED) 
      task->put_Categories(categories);
	if ((startDate || dueDate) && startDate != UNCHANGED) // when setting startDate or dueDate, the other cannot be empty
	{
		PimalDateHelper startDateH(startDate);
		task->put_StartDate(startDateH.getDATE());
	}
	if ((startDate || dueDate) && dueDate != UNCHANGED) 
	{
		PimalDateHelper dueDateH(dueDate);
		task->put_DueDate(dueDateH.getDATE());
	}
	if (importance && importance != UNCHANGED) 
	{
		PimalLongHelper importanceH(importance);
		task->put_Importance(importanceH.getLong());
	}
	if (completed && completed != UNCHANGED)
	{
		PimalBoolHelper completeH(completed);
		task->put_Complete(completeH.getVARIANT_BOOL());
	}
	if (sensitivity && sensitivity != UNCHANGED)
	{
		PimalLongHelper sensitivityH(sensitivity);
		task->put_Sensitivity(sensitivityH.getLong());
	}
	if (teamTask && teamTask != UNCHANGED)
	{
		PimalBoolHelper teamTaskH(teamTask);
		task->put_TeamTask(teamTaskH.getVARIANT_BOOL());
	}
	if (reminderIsSet && reminderIsSet != UNCHANGED)
	{
		PimalBoolHelper reminderSetH(reminderIsSet);
		task->put_ReminderSet(reminderSetH.getVARIANT_BOOL());
	}
	if (reminderOptions && reminderOptions != UNCHANGED)
	{
		PimalLongHelper reminderOptionsH(reminderOptions);
		task->put_ReminderOptions(reminderOptionsH.getLong());
	}
	if (reminderTime && reminderTime != UNCHANGED)
	{
		PimalDateHelper reminderTimeH(reminderTime);
		task->put_ReminderTime(reminderTimeH.getDATE());
	}
	if (note != UNCHANGED) task->put_Body(note);

	if (isRecurring && isRecurring != UNCHANGED)
	{
		PimalBoolHelper isRecurringH(isRecurring);
		if (isRecurringH.getVARIANT_BOOL())
		{
			IRecurrencePattern* recurrencePattern;
			task->GetRecurrencePattern(&recurrencePattern);
			POOMRecurrence recurrenceH(recurrencePattern, isRecurring, recurrenceType, occurrences, duration, interval, dayOfWeek, dayOfMonth, weekOfMonth, monthOfYear, patternStartDate, patternEndDate, startTime, endTime, noEndDate);
		}
		else
		{
			task->ClearRecurrencePattern();
		}		
	}
}
