/*
    Windows CE Symbol Scanner Support for SuperWaba
    Copyright (C) 2003   Dave Slaughter  dslaughter@safeway.co.uk

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    SuperWaba Virtual Machine is copyright (C) 2001-2004
    Guilherme Campos Hazan <support@superwaba.com.br>.
*/

// Note: to download the Scanner API, go to:
// http://www.symbol.com/services/downloads/ppt2700_pocket_pc_sdk.html

//----------------------------------------------------------------------------
// Symbol Scanner Support
//
// This file includes all the native functions required to support the
// SuperWaba Scanner class.
//
// This dll starts up a thread to monitor the scanner, and uses
// SCAN_ReadLabelEvent to be notified of scans.
//
// It will also stop scanning for a period of time when the unit is
// powered up - see readme-ce.txt for more information.
//
// The power-up Event name is taken from registry setting
// HKCU\Software\SuperWaba\Scanner\pwrEvent, and defaults to the definition
// POWER_ON_EVENT if the registry key does not exist.
//----------------------------------------------------------------------------

#define appErrorClass 0 // required to compile in Windows
#define BUILDING_CUSTOM_LIB // must be defined prior to include superwaba.h
#include "..\..\superwaba.h"

#ifdef __SWNATIVE_CUSTOM_LIB_H
You included the wrong CustomLib.h file. Be sure it is NOT the
SWNatives\CustomLib.h.

If you are using CodeWarrior, Specify in the "Access path"
the directory where your CustomLib.h file resides, moving it to the top.

This must be done because a project (like SuperWaba) can have multiple
CustomLib.h files in different directories, and as default the compiler
gets it from the first directory found.
#endif

#include "..\..\ce\SWNatives\SWLib.c"

#include "tchar.h"
// now Scanner specific includes
#include <Strucinf.h>
#include <ScanCAPI.H>

//----------------------------------------------------------------------------
// Definitions
//----------------------------------------------------------------------------
#define BUFFER_SIZE 2710

// name of the default event which the pwrdll.dll sets when the unit it powered on
// This program will first look in registry setting HKCU\Software\SuperWaba\Scanner\pwrEvent
#define POWER_ON_EVENT TEXT("SW_PowerOnEventName")

// message number to use by this application
// TODO: see how we are supposed to get this to avoid clashes!
enum tagUSERMSGS
{
   UM_SCAN = WM_USER + 0x100
      , UM_SCAN_ERROR
};

//----------------------------------------------------------------------------
// function prototypes
//----------------------------------------------------------------------------

// function prototype for the natives method in this dll
void ScannerActivate(Var stack[]);
void ScannerSetBarcodeParam(Var stack[]);
void ScannerSetParam(Var stack[]);
void ScannerSetBarcodeLength(Var stack[]);
void ScannerCommitBarcodeParams(Var stack[]);
void ScannerGetData(Var stack[]);
void ScannerGetScanManagerVersion(Var stack[]);
void ScannerGetScanPortDriverVersion(Var stack[]);
void ScannerDeactivate(Var stack[]);

// other internal functions
static BOOL TranslateDecoder(int32 decoder, LPDECODER translated);
static BOOL ScanStartMonitorThread();
LRESULT ScanMonitorThread() ;
static HANDLE CreatePowerUpEvent();
static BOOL ScanOpen();
static BOOL ScanSetup();
static void ScanCancelReads();
static void ScanClose();
static BOOL PostErrorMessage(LPTSTR szInMsg, DWORD dwError);

//----------------------------------------------------------------------------
// Global variables
//----------------------------------------------------------------------------

// SuperWaba required variables to map native methods
uint32 nativeHashes[10] =
{
   1261220242,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerActivate_()Z
   1261207899,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetBarcodeParam_(IZ)Z
   1261232981,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetParam_(III)Z
   1261223390,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetBarcodeLength_(IIII)Z
   1261225117,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerCommitBarcodeParams_()Z
   1261182306,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetData_()Ljava/lang/String;
   1261209712,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetScanManagerVersion_()Ljava/lang/String;
   1261231603,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetScanPortDriverVersion_()Ljava/lang/String;
   1261233108,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerDeactivate_()Z
   0  // Must end with 0
};
// build at runtime the nativeFuncs array
NativeFunc nativeFuncs[10];

// Scanner variables
static HANDLE hScanner = NULL; // handle of the scanner
static LPSCAN_BUFFER lpScanBuffer = NULL;  // pointer to the scan buffer
static HANDLE hThread = NULL;  // handle of ScanMonitorThread
// Handles for some of the events we are waiting for
static HANDLE hActivateScanner  = NULL;  // request to activate the scanner
static HANDLE hDeactivateScanner = NULL;  // request to deactivate the scanner
static HANDLE hAbortEvent = NULL;      // abort the thread which is monitoring the scanner

//----------------------------------------------------------------------------
// Called by VM when the library opens to get a set of pointers to functions
// implemented in this dll.
//
// We also open the scanner & start off the thread which monitors the scanner.
//----------------------------------------------------------------------------
void buildNativeFuncArray()
{
   nativeFuncs[  0] = ScannerActivate;
   nativeFuncs[  1] = ScannerSetBarcodeParam;
   nativeFuncs[  2] = ScannerSetParam;
   nativeFuncs[  3] = ScannerSetBarcodeLength;
   nativeFuncs[  4] = ScannerCommitBarcodeParams;
   nativeFuncs[  5] = ScannerGetData;
   nativeFuncs[  6] = ScannerGetScanManagerVersion;
   nativeFuncs[  7] = ScannerGetScanPortDriverVersion;
   nativeFuncs[  8] = ScannerDeactivate;
   nativeFuncs[  9] = 0;  // Must end with 0

   if (ScanOpen())
      ScanStartMonitorThread();
}


//----------------------------------------------------------------------------
// Called by the Lib when a public method is invoked by the vm or someone else.
// This defines a interface that lets you add custom methods to interface with
// the vm without modifying the basic architecture of SW Libs.
//
// There are no public methods in this dll.
//----------------------------------------------------------------------------
Err callPublic(int index, uint32 *ret, uint32 *args)
{
   return 0;
}

//----------------------------------------------------------------------------
// Called by VM when the library opens to get a set of ClassHooks.
// None are needed by this library.
//----------------------------------------------------------------------------
ClassHook *getClassHooks()
{
   return 0;
}

//----------------------------------------------------------------------------
// Called by the VM to let the native libraries handle particular events.
// Must return SWErrEventHandled or SWErrEventNotHandled.
//
// Handles scanning events and scan errors.
//----------------------------------------------------------------------------
Err handleEvent(EventPtr eventP)
{
   static WClassMethod *onEventMethod=0;
   static WClass *scannerClass=0;

   switch (eventP->message)
   {
      case UM_SCAN:
         if (!onEventMethod)
         {
            UtfString utf = (*vmGlobals->createUtfString)("superwaba/ext/xplat/io/scanner/Scanner");
            scannerClass = (*vmGlobals->getClass)(utf,1);
            if (!scannerClass)
               return SWErrEventNotHandled;

            onEventMethod = (*vmGlobals->getMethod)(scannerClass, (*vmGlobals->createUtfString)("_onEvent"), (*vmGlobals->createUtfString)("(I)V"), NULL);
         }

         if (onEventMethod)
         {
            Var params[1];
            params[0].intValue = 1; // SCANNED
            (*vmGlobals->executeMethod)(scannerClass, onEventMethod, params, 1);
            return SWErrEventHandled; // Hey, VM, don't let anyone process this event anymore!
         }
         break;
      case UM_SCAN_ERROR:
      {
         LPTSTR msg = (LPTSTR) eventP->lParam;
         TCHAR buf[350];
         wsprintf(buf, TEXT("Scanner error:\n%s\nPlease reboot this unit"), msg);
         MessageBox(eventP->hwnd, buf, TEXT("Program Error"), MB_ICONEXCLAMATION);
         (*vmGlobals->postStopEvent)(0);
         return SWErrEventHandled; // Hey, VM, don't let anyone process this event anymore!
      }
   }
   return SWErrEventNotHandled; // Hey, VM, this event was not handled by me, Continue with normal event processing...
}

//----------------------------------------------------------------------------
// Called when this lib is about to be closed.
// The library must free any allocated resources.
// In this case this simply means signalling the ScanMonitorThread to shut
// down, and waiting for it to complete.
//----------------------------------------------------------------------------
void onClose()
{
   if (hThread)  // if we managed to open the scanner
      SetEvent( hAbortEvent );
   // wait for the ScanMonitorThread to complete before we return
   WaitForSingleObject(hThread, INFINITE);
   CloseHandle(hThread);
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static boolean scannerActivate()
//
// Activate or deactivate the scanner by setting the appropriate event which is
// monitored by the ScanMonitorThread
//----------------------------------------------------------------------------
void ScannerActivate(Var stack[]) // ()Z
{
   if (hActivateScanner == NULL)   // if scanner failed to open
      RETURN32(0)

   SetEvent( hActivateScanner );
   RETURN32(1)
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static boolean scannerDeactivate()
//
// Deactivate or deactivate the scanner by setting the appropriate event which is
// monitored by the ScanMonitorThread
//----------------------------------------------------------------------------
void ScannerDeactivate(Var stack[]) // ()Z
{
   if (hActivateScanner == NULL)   // if scanner failed to open
      RETURN32(0)

   SetEvent( hDeactivateScanner );
   RETURN32(1)
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static boolean scannerSetBarcodeParam(int barcodeType, boolean enable)
//
// Enable/disable a decoder
//----------------------------------------------------------------------------
void ScannerSetBarcodeParam(Var stack[]) // (IZ)Z
{
   int32 enabled;
   DECODER_LIST DecoderList;
   DECODER Decoder;
   DWORD dwResult;

   enabled = stack[1].intValue;

   if (!TranslateDecoder(stack[0].intValue, (LPDECODER) &Decoder))
      RETURN32(0)

   // Mark entire buffer allocated, and not used
   SI_INIT(&DecoderList);

   // Get enabled decoder list
   dwResult = SCAN_GetEnabledDecoders(hScanner,&DecoderList);
   if ( dwResult != E_SCN_SUCCESS )
      RETURN32(0)

   if (enabled)
   {
      DWORD i;
      // is this decoder already enabled?
      for(i=0;i<DecoderList.Decoders.dwDecoderCount;i++)
      {
         if ( DecoderList.Decoders.byList[i] == (BYTE) Decoder )
            RETURN32(1)
      }

      SI_SET_FIELD(&DecoderList, Decoders.byList[DecoderList.Decoders.dwDecoderCount], (BYTE) Decoder);
      DecoderList.Decoders.dwDecoderCount++;
   }
   else
   {
      // disable
      DWORD i;
      DWORD dwCount = DecoderList.Decoders.dwDecoderCount;

      for(i=0;i<DecoderList.Decoders.dwDecoderCount;i++)
      {
         if ( DecoderList.Decoders.byList[i] == (BYTE) Decoder )
         {
            memcpy(DecoderList.Decoders.byList+i,
                  DecoderList.Decoders.byList+i+1,
                  dwCount-i-1);
            DecoderList.Decoders.dwDecoderCount--;
         }
      }

      // check if nothing was disabled
      if ( dwCount == DecoderList.Decoders.dwDecoderCount )
         RETURN32(1)
   }
   dwResult = SCAN_SetEnabledDecoders(hScanner,&DecoderList);

   if (dwResult != E_SCN_SUCCESS)
      RETURN32(0)

   RETURN32(1)
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/io/scanner/NativeMethods boolean scannerSetParam(int type, int barcodeType, int value)
//
// TODO: Since we don't use this, haven't implemented it yet.
//----------------------------------------------------------------------------
void ScannerSetParam(Var stack[]) // (III)Z
{
   RETURN32(0)
   /*
   int32 type;
   BarType barcode;
   uint16 value;
   type = stack[1].intValue;
   barcode = (BarType)stack[2].intValue;
   value = (uint16)stack[3].intValue;
   switch (type)
   {
      case TRIGGERING_PARAM            :
         v.intValue = ScanSetTriggeringModes(value) == 0;
         break;
      case LINEAR_SECURITY_LEVEL_PARAM :
         v.intValue = ScanSetLinearCodeTypeSecurityLevel(value) == 0;
         break;
      case SUPPLEMENTALS_PARAM         :
         v.intValue = ScanSetDecodeUpcEanSupplementals(value) == 0;
         break;
      case TRANSMIT_CHECKDIGIT_PARAM   :
         v.intValue = ScanSetTransmitCheckDigit(barcode, value) == 0;
         break;
      case PREAMBLE_PARAM              :
         v.intValue = ScanSetUpcPreamble(barcode, value) == 0;
         break;
      case MSI_PLESSEY_CHECKDIGIT_PARAM:
         v.intValue = ScanSetMsiPlesseyCheckDigits(value) == 0;
         break;
      case MSI_PLESSEY_OPTIONS_PARAM   :
         v.intValue = ScanSetMsiPlesseyCheckDigits(value) == 0;
         break;
      case MSI_PLESSEY_ALGORITHMS_PARAM:
         v.intValue = ScanSetMsiPlesseyCheckDigitAlgorithm(value) == 0;
         break;
      case TRANSMIT_CODEID_PARAM       :
         v.intValue = ScanSetTransmitCodeIdCharacter(value) == 0;
         break;
      case SCAN_DATA_TRANSMISSION_PARAM:
         v.intValue = ScanSetScanDataTransmissionFormat(value) == 0;
         break;
      case SCAN_ANGLE_PARAM            :
         v.intValue = ScanSetAngle(value) == 0;
         break;
   }
   */
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/io/scanner/NativeMethods boolean setBarcodeLength(int barcodeType, int lengthType, int length1, int length2 )
//----------------------------------------------------------------------------
void ScannerSetBarcodeLength(Var stack[]) // (IIII)Z
{
   // values which are passed by the Scanner class
   #define ANY_LENGTH                                  0x00
   #define ONE_DISCRETE_LENGTH                         0x01
   #define TWO_DISCRETE_LENGTHS                        0x02
   #define LENGTH_WITHIN_RANGE                         0x03

   DECODER Decoder;
   DECODER_PARAMS decoder_params;
   DWORD dwMinLength, dwMaxLength, dwResult;

   // Allowable decode lengths are specified by dwMinLength and dwMaxLength as follows:

   // Variable length - dwMinLength = dwMaxLength = 0
   // Range (from a to b, including a and b) - dwMinLength = a, dwMaxLength = b
   // Two Discrete Lengths (either a or b, given a<b) - dwMinLength = b, dwMaxLength = a
   // One Discrete Length (only a) - dwMinLength = dwMaxLength = a

   switch (stack[1].intValue)
   {
      case ANY_LENGTH:
         dwMinLength = dwMaxLength = 0;
         break;
      case ONE_DISCRETE_LENGTH:
         dwMinLength = dwMaxLength = stack[2].intValue;
         break;
      case TWO_DISCRETE_LENGTHS:
         if (stack[3].intValue < stack[3].intValue)
         {
            dwMinLength = stack[3].intValue;
            dwMaxLength = stack[2].intValue;;
         }
         else
         {
            dwMinLength = stack[2].intValue;
            dwMaxLength = stack[3].intValue;;
         }
         break;
      case LENGTH_WITHIN_RANGE:
         dwMinLength = stack[2].intValue;
         dwMaxLength = stack[3].intValue;;
         break;
   }

   if (!TranslateDecoder(stack[0].intValue, (LPDECODER) &Decoder))
      RETURN32(0)

   // Initialize the structure to all zeroes
   memset(&decoder_params, 0, sizeof(decoder_params));

   // Mark entire buffer allocated and not used
   SI_INIT(&decoder_params);

   // Get current decoder parameters
   dwResult = SCAN_GetDecoderParams(hScanner, (LPDECODER) &Decoder, &decoder_params);

   // If there was an error getting the parameters
   if ( dwResult != E_SCN_SUCCESS )
      RETURN32(0)

   // Set all parameters into structure if changed
   SI_SET_IF_CHANGED(&decoder_params,dwMinLength,dwMinLength);
   SI_SET_IF_CHANGED(&decoder_params,dwMaxLength,dwMaxLength);

   // Return the result of setting parameters
   dwResult = SCAN_SetDecoderParams(hScanner, (LPDECODER) &Decoder, &decoder_params);
   if ( dwResult != E_SCN_SUCCESS )
      RETURN32(0)

   RETURN32(1)
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static boolean scannerCommitBarcodeParams()
//
// Commit is not relevant on WinCE, so just return true.
//----------------------------------------------------------------------------
void ScannerCommitBarcodeParams(Var stack[]) // ()Z
{
   RETURN32(1)
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static String scannerGetData()
//----------------------------------------------------------------------------
void ScannerGetData(Var stack[]) // ()Ljava/lang/String;
{
   LABELTYPE dwLabelType;
   int len;
   TCHAR *barcode;

   if ( SCNBUF_GETSTAT(lpScanBuffer) != E_SCN_SUCCESS )
      RETURN32(0)

   barcode = (TCHAR*)SCNBUF_GETDATA(lpScanBuffer);

   // handle the situation where the barcode returned by the scanner can
   // sometimes be repeated - ie barcode scanned is 123 but SCNBUF_GETDATA
   // returns 123123
   dwLabelType = SCNBUF_GETLBLTYP(lpScanBuffer);
   len = _tcslen(barcode);
   switch (dwLabelType)
   {
      case LABELTYPE_EAN13 :
         if (len > 13)
            _tcscpy(barcode, barcode + len - 13);
         break;
      case LABELTYPE_EAN8 :
         if (len > 8)
            _tcscpy(barcode, barcode + len - 8);
         break;
      default :
         if ( len%2 == 0 )    // possible double decode ? only look at even len strings
            if ( _tcsncmp( barcode, barcode+(len/2), len/2 ) == 0 )
                     barcode[ len/2 ] = 0;
   }
#ifdef UNICODE
   RETURN32((*vmGlobals->createStringFromUnicode)(barcode, _tcslen(barcode)))
#else
   RETURN32((*vmGlobals->createString)(barcode))
#endif
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static String scannerGetScanManagerVersion()
//----------------------------------------------------------------------------
void ScannerGetScanManagerVersion(Var stack[]) // ()Ljava/lang/String;
{
   SCAN_VERSION_INFO info;
   DWORD dwResult;

   SI_INIT(&info);
   dwResult = SCAN_GetVersionInfo(hScanner, &info);
   if (dwResult != E_SCN_SUCCESS)
      RETURN32(0)
   RETURN32((*vmGlobals->createString)((*vmGlobals->int2str)("",info.dwCAPIVersion)))
}

//----------------------------------------------------------------------------
// superwaba/ext/xplat/NativeMethods static String scannerGetScanPortDriverVersion()
//----------------------------------------------------------------------------
void ScannerGetScanPortDriverVersion(Var stack[]) // ()Ljava/lang/String;
{
   SCAN_VERSION_INFO info;
   DWORD dwResult;

   SI_INIT(&info);
   dwResult = SCAN_GetVersionInfo(hScanner, &info);
   if (dwResult != E_SCN_SUCCESS)
      RETURN32(0)
   RETURN32((*vmGlobals->createString)((*vmGlobals->int2str)("",info.dwPddVersion )))
}


//----------------------------------------------------------------------------
// Translate the decoder from the Palm format used by SuperWaba to the format
// used by the CE scanner API.  Symbol's CE & Palm API decoder values don't
// match.  Don't you just love consistent APIs
//----------------------------------------------------------------------------
static BOOL TranslateDecoder(int32 decoder, LPDECODER translated)
{
   switch (decoder)
   {
      case 0: strcpy(translated, DECODER_CODE39); break;
      case 1: strcpy(translated, DECODER_UPCA); break;
      case 2: strcpy(translated, DECODER_UPCE0); break;
      case 3: strcpy(translated, DECODER_EAN13); break;
      case 4: strcpy(translated, DECODER_EAN8); break;
      case 5: strcpy(translated, DECODER_D2OF5); break;
      case 6: strcpy(translated, DECODER_I2OF5); break;
      case 7: strcpy(translated, DECODER_CODABAR); break;
      case 8: strcpy(translated, DECODER_CODE128); break;
      case 9: strcpy(translated, DECODER_CODE93); break;
      case 11: strcpy(translated, DECODER_MSI); break;
      case 12: strcpy(translated, DECODER_UPCE1); break;
      case 13: strcpy(translated, DECODER_TRIOPTIC39); break;
   // the following decoders aren't available on PPT2700 & 2800
   // case 14: strcpy(translated, DECODER_EAN128); break;
   // case 83: strcpy(translated, DECODER_BROOKLAND); break;
   // case 84: strcpy(translated, DECODER_ISBT128); break;
   // case 85: strcpy(translated, DECODER_COUPON); break;
      default: return FALSE; // unknown decoder
   }
   return TRUE;
}

//----------------------------------------------------------------------------
// Start up a seperate thread to monitor the scanner
//----------------------------------------------------------------------------
static BOOL ScanStartMonitorThread()
{
   DWORD dwThreadId;

   hAbortEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
   hActivateScanner  = CreateEvent( NULL, TRUE, FALSE, NULL );
   hDeactivateScanner = CreateEvent( NULL, TRUE, FALSE, NULL );

   hThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE) ScanMonitorThread, NULL, 0,  &dwThreadId );
   if (hThread == NULL)
   {
      CloseHandle(hAbortEvent);
      CloseHandle(hActivateScanner);
      CloseHandle(hDeactivateScanner);
      hAbortEvent = NULL;
      hActivateScanner = NULL;
      hDeactivateScanner = NULL;
      return FALSE;
   }
   else
   {
      SetThreadPriority (hThread, THREAD_PRIORITY_ABOVE_NORMAL );
      return TRUE;
   }
}

//----------------------------------------------------------------------------
// Thread function which monitors the scanner.
//----------------------------------------------------------------------------
LRESULT ScanMonitorThread()
{
   HANDLE hEventPowerUp;
   HANDLE hScannerEvent;
   DWORD dwRequestId = 0;

   BOOL bAbort = FALSE;
   BOOL bIsScannerActive = FALSE;

   HANDLE h[ 5 ] ;
   int eventCount = 0;

   hScannerEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
   hEventPowerUp = CreatePowerUpEvent();

   //  Setup the events that we want to wait for
   h[ eventCount++ ] = hScannerEvent ;                 // WAIT_OBJECT_0
   h[ eventCount++ ] = hAbortEvent ;               // WAIT_OBJECT_0+1
   h[ eventCount++ ] = hDeactivateScanner ;               // WAIT_OBJECT_0+2
   h[ eventCount++ ] = hActivateScanner ;                // WAIT_OBJECT_0+3
   if ( hEventPowerUp )
      h[ eventCount++ ] = hEventPowerUp ;          // WAIT_OBJECT_0+4

   while( !bAbort )
   {
      DWORD dwResult = 0;
      ScanCancelReads();

      // If we are 'active' then put a read into the scanner otherwise we
      // are inactive, so do nothing
      if ( bIsScannerActive )
      {
         if ( hScannerEvent )
            CloseHandle( hScannerEvent );
         hScannerEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
         h[ 0 ] = hScannerEvent ;                 // WAIT_OBJECT_0
         dwResult = SCAN_ReadLabelEvent( hScanner, lpScanBuffer, hScannerEvent, 8460000, &dwRequestId );
         if ( dwResult != E_SCN_SUCCESS )
            PostErrorMessage(TEXT("SCAN_ReadLabelEvent"), dwResult);
      }

      dwResult = WaitForMultipleObjects( eventCount, h, FALSE, INFINITE );
      switch ( dwResult )
      {
         case WAIT_OBJECT_0 :           // scanner woke us up
         {
#if _WIN32_WCE >= 300
            // make sure the device does not go into suspended state.  The device does
            // not consider a scan to be a user activity, so may power off if we
            // don't call SystemIdleTimerReset.
            SystemIdleTimerReset();
#endif
            ResetEvent( hScannerEvent );
            if ( lpScanBuffer->dwStatus == E_SCN_SUCCESS )
            {
               bIsScannerActive = FALSE;
               // post a message back to the main thread that a barcode
               // has been scanned.  Since SuperWaba isn't multithreaded we
               // cannot post an event into the VM from this thread.
               PostMessage( vmGlobals->mainHWnd, UM_SCAN, 0, 0);
               continue;
             }
             else
             if ( lpScanBuffer->dwStatus == E_SCN_DEVICEFAILURE )
             {
                // problem with the device - try closing & re-opening
                ScanClose ();
                // if we can't reopen the scanner then abort
                if (!ScanOpen())
                   bAbort = TRUE;
             }
             break;
         }
      case (WAIT_OBJECT_0+1) :       // abort/exit event woke us up
         bAbort = TRUE ;
         continue;
      case (WAIT_OBJECT_0+2) :       // deactivate scanner - do not allow reads until enabled
         ResetEvent( hDeactivateScanner );
         bIsScannerActive = FALSE ;
         ResetEvent( hScannerEvent );
         break;
      case (WAIT_OBJECT_0+3) :       // activate scanner i.e. allow reads
         ResetEvent( hActivateScanner );
         bIsScannerActive = TRUE ;
         break;
      case (WAIT_OBJECT_0+4) :
         // power on event woke us up so wait to give time for the scanner & RF card to power up,
         // but still abort if abort event
         {
            const DWORD dwDelay = 7000;
            DWORD StartTime = GetTickCount();

            ScanCancelReads();

            do
            {
               DWORD dwResult;
               DWORD dwT = dwDelay - (GetTickCount()-StartTime) ;
               if ( dwT <= 0 || dwT > dwDelay )
                  break ;
               dwResult = WaitForMultipleObjects( eventCount, h, FALSE, dwT );
               switch ( dwResult )
               {
                  case WAIT_OBJECT_0 :           // scanner woke us up - ignore the read
                     ResetEvent( hScannerEvent );
                     break;
                  case (WAIT_OBJECT_0+1) :           // abort event
                     bAbort = TRUE;
                     StartTime = 0;
                     break;
                  case (WAIT_OBJECT_0+2) :           // deactivate scanner
                     ResetEvent( hDeactivateScanner );
                     ResetEvent( hScannerEvent );
                     bIsScannerActive = FALSE;
                     break;
                  case (WAIT_OBJECT_0+3) :           // activate scanner
                     ResetEvent( hActivateScanner );
                     if ( hScannerEvent )
                        ResetEvent( hScannerEvent );
                     bIsScannerActive = TRUE;
                     break;
                  case (WAIT_OBJECT_0+4) :       // power on event not reset so yield
                     Sleep(100);
                     break;
                  default :
                     break;
               }
            } while ( (GetTickCount()-StartTime) < dwDelay );
            break;
         }
         default :
            break;
      }
   }

   ScanClose();
   if ( hEventPowerUp )
      CloseHandle( hEventPowerUp );
   CloseHandle( hActivateScanner  );
   CloseHandle( hDeactivateScanner );
   CloseHandle( hScannerEvent );
   CloseHandle( hAbortEvent );
   return 0;
}

//----------------------------------------------------------------------------
// Return a handle to the power-up event, or null if the event cannot be
// found.  The power-up event must have been created by another application.
//----------------------------------------------------------------------------
static HANDLE CreatePowerUpEvent()
{
   HKEY hKey;
   TCHAR szEventName[ 256 ] = POWER_ON_EVENT; // default event name
   DWORD dwBufLen = 256;
   LONG rc=0;
   HANDLE hEventPowerUp;

   // look in the registry for the event name
   rc = RegOpenKeyEx( HKEY_CURRENT_USER, TEXT("Software\\SuperWaba\\Scanner"), 0, KEY_ALL_ACCESS, &hKey);
   if (rc == ERROR_SUCCESS )
   {
      DWORD dwType = REG_SZ;
      RegQueryValueEx( hKey, TEXT("pwrEvent"), 0, &dwType, (LPBYTE)szEventName, &dwBufLen );
       RegCloseKey( hKey );
   }

   hEventPowerUp = CreateEvent( NULL, TRUE, FALSE, szEventName );
   // If we the power-up Event doesn't already exist then we can't detect power on events
   if ( GetLastError() != ERROR_ALREADY_EXISTS )
   {
      CloseHandle( hEventPowerUp );
      hEventPowerUp = NULL;
   }
   return hEventPowerUp;
}

//----------------------------------------------------------------------------
// Open & enable the scanner, and create the scan buffer.
// Return TRUE if ok, FALSE otherwise
//----------------------------------------------------------------------------
static BOOL ScanOpen()
{
   DWORD dwResult;

   dwResult = SCAN_Open( TEXT("SCN1:"), &hScanner );
   if ( dwResult != E_SCN_SUCCESS )
   {
      hScanner = NULL;
      PostErrorMessage(TEXT("SCAN_Open"), dwResult);
        return FALSE;
   }

   dwResult = SCAN_Enable(hScanner) ;
   if ( dwResult != E_SCN_SUCCESS )
   {
      PostErrorMessage(TEXT("SCAN_Enable"), dwResult);
      SCAN_Close(hScanner);
      hScanner = NULL;
        return FALSE;
   }

   lpScanBuffer = SCAN_AllocateBuffer( TRUE, BUFFER_SIZE );
   if (lpScanBuffer == NULL)
   {
      PostErrorMessage(TEXT("SCAN_AllocateBuffer"), GetLastError());
      ScanClose();
      return FALSE;
   }

   if (!ScanSetup())
   {
      ScanClose();
      return FALSE;
   }

   return TRUE;
}

//----------------------------------------------------------------------------
// Set default parameters for the Scanner.
// Return TRUE if ok, FALSE otherwise
//----------------------------------------------------------------------------
static BOOL ScanSetup()
{
   DWORD dwResult;
   //
   //  Set the scanner parameters
   //
   {
      SCAN_PARAMS scan_params;
      memset(&scan_params,0, sizeof(scan_params));
      SI_ALLOC_ALL(&scan_params);
      scan_params.StructInfo.dwUsed = 0;
      dwResult = SCAN_GetScanParameters(hScanner,&scan_params);
      if ( dwResult != E_SCN_SUCCESS ) {
         PostErrorMessage(TEXT("SCAN_GetScanParameters"), dwResult);
         return FALSE;
      }

      scan_params.dwCodeIdType = CODE_ID_TYPE_NONE ;
      scan_params.dwScanType = SCAN_TYPE_FOREGROUND;
      scan_params.bLocalFeedback = TRUE;
      scan_params.dwDecodeBeepTime = 500;
      scan_params.dwDecodeBeepFrequency = 3000;
      scan_params.dwDecodeLedTime = 1000;
      memset( scan_params.szWaveFile, 0, sizeof(scan_params.szWaveFile) );
      dwResult = SCAN_SetScanParameters( hScanner, &scan_params ) ;
      if ( dwResult != E_SCN_SUCCESS ) {
         PostErrorMessage(TEXT("SCAN_SetScanParameters"), dwResult);
         return FALSE;
      }
   }
   //
   //  Set The reader paramers
   //
   {
      READER_PARAMS reader_params;
      memset(&reader_params,0, sizeof(reader_params));
      SI_ALLOC_ALL(&reader_params);
      reader_params.StructInfo.dwUsed = 0;
      dwResult = SCAN_GetReaderParams(hScanner,&reader_params);
      if ( dwResult != E_SCN_SUCCESS )
      {
         PostErrorMessage(TEXT("SCAN_GetReaderParams"), dwResult);
         return FALSE;
      }

      reader_params.dwReaderType = READER_TYPE_LASER ;
      reader_params.ReaderSpecific.laser_specific.dwAimType = AIM_TYPE_TRIGGER ;
      reader_params.ReaderSpecific.laser_specific.dwAimDuration = 0;
      reader_params.ReaderSpecific.laser_specific.dwAimMode = AIM_MODE_NONE;
      reader_params.ReaderSpecific.laser_specific.bNarrowBeam = TRUE;
      reader_params.ReaderSpecific.laser_specific.dwRasterMode = RASTER_MODE_NONE;
      reader_params.ReaderSpecific.laser_specific.dwBeamTimer = 1000;
      reader_params.ReaderSpecific.laser_specific.bControlScanLed = TRUE;
      reader_params.ReaderSpecific.laser_specific.bScanLedLogicLevel = TRUE;
      reader_params.ReaderSpecific.laser_specific.bKlasseEinsEnable = FALSE;
      reader_params.ReaderSpecific.laser_specific.bBidirRedundancy = TRUE;
      reader_params.ReaderSpecific.laser_specific.dwLinearSecurityLevel = SECURITY_ALL_THRICE;
      reader_params.ReaderSpecific.laser_specific.dwPointerTimer = 1000;
      dwResult = SCAN_SetReaderParams(hScanner,&reader_params);
      if ( dwResult != E_SCN_SUCCESS )
      {
         PostErrorMessage(TEXT("SCAN_SetReaderParams"), dwResult);
         return FALSE;
      }
   }

    return TRUE;
}

//----------------------------------------------------------------------------
// Cancel any outstanding reads on the scanner
//----------------------------------------------------------------------------
static void ScanCancelReads()
{
   if (lpScanBuffer != NULL)
   {
      if ( SCNBUF_GETSTAT(lpScanBuffer) == E_SCN_READPENDING )     // if an outstanding read then cancel read
      {
         //dwResult = SCAN_CancelRead( hScanner, SCNBUF_GETREQID(lpScanBuffer) );
         DWORD dwResult = SCAN_Flush( hScanner );
         if (dwResult == E_SCN_SUCCESS)
         {
            // the read may or may not have cancelled before flush returns, so
            // make sure the read is actually cancelled before continuing
            while ( SCNBUF_GETSTAT(lpScanBuffer) == E_SCN_READPENDING )
            {
               Sleep(100);
            }
         }
      }
   }
}

//----------------------------------------------------------------------------
// Close the Scanner, and cancel any reads pending
//----------------------------------------------------------------------------
static void ScanClose()
{
   if (hScanner)
   {
      ScanCancelReads();
      SCAN_Disable(hScanner);
      if (lpScanBuffer)
      {
         SCAN_DeallocateBuffer( lpScanBuffer );
         lpScanBuffer = NULL;
      }
      SCAN_Close(hScanner);
      hScanner = NULL;
   }
   return;
}

//----------------------------------------------------------------------------
// Send an error message to the main window - due to fatal error
//
// return true if success, false otherwise
//----------------------------------------------------------------------------
static BOOL PostErrorMessage(LPTSTR szInMsg, DWORD dwError )
{
   LPTSTR szMsg;

   // take a copy of the message - main thread must delete the memory
   szMsg = vmGlobals->xmalloc((_tcslen(szInMsg) + 20) * sizeof(TCHAR));
   if (!szMsg)
      return FALSE;

   wsprintf(szMsg, TEXT("%s error code %x"), szInMsg, dwError);
   // send the message to the main window - if we can't post the message then
   // make sure we free memory.
   if (!PostMessage( vmGlobals->mainHWnd, UM_SCAN_ERROR, 0, (LPARAM)szMsg))
   {
      xfree(szMsg);
      return FALSE;
   }
   return TRUE;
}
