/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_defines.c,v 1.8 2005/09/17 21:16:58 guich Exp $ */

#include "SW_defines.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

// from superwaba.c
extern int gc_();

#if defined(PALMOS)

// guich@200b4_99: put some big vars here to avoid exceeding the stack
DmSearchStateType state;

#if defined(__arm__)
uint32 ushr(uint32 number, int32 count) // can't use inline in win32 and wince
{
   return number >> count;
   //lsr.l d1,d0
}
#else
uint32 ushr(uint32 number, int32 count)
{
   asm __volatile__
   (
      "lsr.l %1,%0"
         : "=d" (number)
         : "0" (number), "d" (count)
   );
   return number;
}
#endif

#else
uint32 ushr(uint32 number, int32 count) // can't use inline in win32 and wince
{
   return number >> count;
   //lsr.l d1,d0
}

// Memory management for all non PalmOS systems

/*--------------------------------------------------------------------xmalloc-+
|                                                                             |
+----------------------------------------------------------------------------*/
void * xmalloc(uint32 size)
{
   void *p;
   size = ((size >> 2) << 2) + 4;
   p = dlmalloc(size);
   if (!p) {
      if (gc_()) // gc may free some memory blocks
      {
         p = dlmalloc(size);
      }
      if (!p) {
	      return 0;
      }
   }
   xmemzero(p, size); // guich@340_21: make sure we erase the allocated memory
   return p;
}

/*---------------------------------------------------------------------xfree_-+
|                                                                             |
+----------------------------------------------------------------------------*/
void xfree_(void *p)
{
   dlfree(p);
}

/*-------------------------------------------------------------------xrealloc-+
|                                                                             |
+----------------------------------------------------------------------------*/

// resizes the ptr. If newSize is smaller than the original size, always successful.
// Otherwise, may return null if no more memory. Note that the newSize may not
// be bigger than 64k-32 (PalmOS limit). Returns a ptr with the new size
// (maybe with different base address)
// note that this routine does not change where the pointer is allocated.
// If it was in storage heap, it remains in storage heap; if it was in
// dynamic heap, it remains in there.
// if growing, the new area allocated is zeroed out (keeping the old data intact).
// if p is null, memory is allocated using xmalloc.
void * xrealloc(void *p, uint32 *pNewSize)
#if defined WINCE && defined WINCE_USE_REGULAR_ALLOC
{
   uint32 newSize = *pNewSize;
   if (p == NULL)
      p = xmalloc(newSize);
   else
   {
      uint32 oldSize = _msize(p);
      p = realloc(p, newSize);
      if (!p)
      {
         if (gc_()) // guich@400_14, gc may free some memory blocks
            p = realloc(p, newSize);
      }
      else
      if (newSize > oldSize) // guich@340_21
         xmemzero((byte *)p+oldSize, newSize-oldSize);
   }
   if (!p) *pNewSize = 0;
   return p;
}
#else
{
   uint32 newSize = *pNewSize;
   if (p == NULL) 
   {
      p = xmalloc(newSize);
   }
   else 
   {
      #ifndef __ECOS
      uint32 oldSize;
      oldSize = dlmalloc_usable_size(p);
      p = dlrealloc(p, newSize);
      #else
      p = realloc_rtnoldsize(p, newSize, &oldSize);
      #endif
      if (!p && gc_()) // gc may free some memory blocks
      {
         #ifndef __ECOS
         p = dlrealloc(p, newSize);
         #else
         p = realloc_rtnoldsize(p, newSize, &oldSize);
         #endif
      }
      #ifndef __ECOS
      if (p && (newSize > oldSize)) 
         xmemzero((byte *)p+oldSize, newSize-oldSize);
      #endif
   }
   if (!p) *pNewSize = 0;
   return p;
}
#endif

/*---------------------------------------------------------------xFreeMemInfo-+
|                                                                             |
+----------------------------------------------------------------------------*/
uint32 xFreeMemInfo()
{
   return dlavailable();
}


_inline_ char toUp(char c)
{
   if (97 <= c && c <= 122) c -= 32;
   return c;
}

int16 xstrcaselesscmp(const char *s1, const char *s2)
{
   int16 i;
   for (i = 0; s1[i] && s2[i]; i++)
      if (toUp(s1[i]) != toUp(s2[i]))
         return 1;
   if (s1[i] || s2[i]) // different length?
      return 1;
   return 0;
}

#endif // PALMOS

/*===========================================================================*/
