/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_events.c,v 1.49 2005/11/17 22:12:20 guich Exp $ */

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "superwaba.h"
#include "SDL.h"
#include "SDL_syswm.h"

#define __DEFINE_FUNCTIONS
#include "SW_logging.h"
#include "SW_events.h"
#include "SW_String.h"
#include "SW_messages.h"
#include "SW_natives.h"
#include "SW_settings.h"
#include "SW_defines.h"


// from superwaba.c:
WClassMethod *getMethodByMapNum(WClass *wclass inA0, WClass **vclass inA1, uint16 mapNum inD0);
Var * executeMethod(WClass *wclass, WClassMethod *method, Var params[], uint32 numParams);
int16 getMethodMapNum(WClass *wclass, UtfString *name, UtfString *desc, byte searchType, WClassMethod **imethod, WClass **iclass);
WClassMethod *getMethod(WClass *wclass, UtfString name, UtfString desc, WClass **vclass);

#define SDL_USEREVENT_APPSTART SDL_USEREVENT
#define SDL_USEREVENT_REFRESH  SDL_USEREVENT+1

#define SK_ALL (0xFFFFFFFF & ~0x01000000)
#define SK_HARD1     0x00000008L
#define SK_HARD2     0x00000010L
#define SK_HARD3     0x00000020L
#define SK_HARD4     0x00000040L
#define SK_LAUNCH    0x01000000L
#define SK_FIND      0x02000000L
#define SK_CALC      0x04000000L
#define SK_ACTION    0x08000000L
#define SK_CONTRAST  0x10000000L // guich@401_31
#define SK_CLOCK     0x20000000L // guich@401_31
#define SK_SYNC      0x40000000L // guich@421_18
#define IS_KEY_INTERCEPTED(x) ((vmGlobals->interceptSystemKey & x) != 0)

#ifdef WIN32
typedef BOOL (__stdcall *UnregisterFunc1Proc)( UINT, UINT );
#endif

static WClass * runThreadsVclass;
static WClassMethod * runThreadsMethod;
static Var runThreadsParams[1];

static void checkMainWindowTimer();
static void executePostKeyEvent(int32 key, SDLMod mod);
static void executePostEvent(int32 type, int32 key, int32 x, int32 y,
                             int32 modifiers, int32 timeStamp);
static void drawMainWin();

int swExitCode=0; // guich@510_16

/*----------------------------------------------------filterOutIdenticalMoves-+
| Function called by SDL to discard unwanted events.
| Note: identical pen move positions have to be discarded on the PalmOS ARM port to prevent slow downs
| returns: 0 to filter out the event, 1 to proceed with normal event processing
+----------------------------------------------------------------------------*/
int sdlEventFiltering(const SDL_Event *event)
{
   static uint16 x, y = 0xFFFF;

#if defined(PALMOS)
   if (event->type == SDL_USEREVENT && event->user.code == 0x99) // fdie@ 0x99 is the PalmOS intercept checking code, see SDL_PalmEvents.c
   {
      SDL_keysym *keysym = (SDL_keysym *)(event->user.data1);

      int intercepted = 0;
      switch (keysym->sym)
      {
         // Unconditional intercepts
         case _SDLK_KEYBOARD_ALPHA:
         case _SDLK_KEYBOARD_NUM:
         case _SDLK_KEYBOARD:
            intercepted = 1;
            break;

         // Conditional intercepts
         case _SDLK_LAUNCH:      intercepted = IS_KEY_INTERCEPTED(SK_LAUNCH); break; // guich@200b4_5
         case _SDLK_FIND:        intercepted = IS_KEY_INTERCEPTED(SK_FIND);   break;
         case _SDLK_CALC:        intercepted = IS_KEY_INTERCEPTED(SK_CALC);   break;
         case _SDLK_HARD1:       intercepted = IS_KEY_INTERCEPTED(SK_HARD1);  break; // note: dont remove the == 0 or it wont work
         case _SDLK_HARD2:       intercepted = IS_KEY_INTERCEPTED(SK_HARD2);  break;
         case _SDLK_HARD3:       intercepted = IS_KEY_INTERCEPTED(SK_HARD3);  break;
         case _SDLK_HARD4:       intercepted = IS_KEY_INTERCEPTED(SK_HARD4);  break;
         case _SDLK_HARDCRADDLE: intercepted = IS_KEY_INTERCEPTED(SK_SYNC);   break; // guich@421_18
         case _SDLK_BRIGHTNESS: // guich@450_25
         case _SDLK_CONTRAST:    intercepted = IS_KEY_INTERCEPTED(SK_CONTRAST); break; // guich@402_2: fixed contrast
         case _SDLK_CLOCK:       intercepted = IS_KEY_INTERCEPTED(SK_CLOCK); break; // guich@402_2: fixed clock

         case menuCmdBarOpenEvent: 
         case 282: intercepted = 1; keysym->sym = _SDLK_COMMAND; break; // guich@552_31: intercept the COMMAND key.

//         default: // debug key press
//            vmGlobals->debug(vmGlobals->int2str("key",keysym->sym));
      }
      if (intercepted)
      {
         _LOG_DEBUG((LOG_CAT_EVENT, "scancode %x has been intercepted", keysym->scancode));
         return 0;
      }
   }
#endif

   if (event->type == SDL_MOUSEMOTION)
   {
      if (y != 0xFFFF && x == event->motion.x && y == event->motion.y)
      {
         return 0; // ignore this move event
      }
      // store new move position
      x = event->motion.x;
      y = event->motion.y;
   }
   return 1;
}

/*-----------------------------------------------------------------initEvents-+
|                                                                             |
+----------------------------------------------------------------------------*/
int32 initEvents()
{
   WClass *vclass;
   int32 runThreadsMethodMapNum;
   UtfString methodName, methodDesc;

   // cache method map numbers for commonly called methods
   vclass = WOBJ_class(vmGlobals->globalMainWin);
   methodName = createUtfString("_postEvent");
   methodDesc = createUtfString("(IIIIII)V");
   vmGlobals->postEventMethodMapNum = getMethodMapNum(vclass, &methodName, &methodDesc, CLASS_SEARCH_ALL, NULL, NULL);
   methodName = createUtfString("_onTimerTick");
   methodDesc = createUtfString("()V");
   vmGlobals->onTimerTickMethodMapNum = getMethodMapNum(vclass, &methodName, &methodDesc, CLASS_SEARCH_ALL, NULL, NULL);
   methodName = createUtfString("_runThreads");
   methodDesc = createUtfString("()V");
   runThreadsMethodMapNum = getMethodMapNum(vclass, &methodName, &methodDesc, CLASS_SEARCH_ALL, NULL, NULL);
   if ((vmGlobals->postEventMethodMapNum == -1) ||
       (vmGlobals->onTimerTickMethodMapNum == -1) ||
       (runThreadsMethodMapNum == -1))
   {
      return ERR_NoMainWindowObj;  // could be more explainatory
   }
   runThreadsVclass = vclass;
   runThreadsMethod = getMethodByMapNum(runThreadsVclass, &runThreadsVclass, (uint16)runThreadsMethodMapNum);
   runThreadsParams[0].obj = vmGlobals->globalMainWin;
   if (runThreadsMethod == NULL)
      return ERR_NoMainWindowObj;  // could be more explainatory
   vmGlobals->executePostEvent = executePostEvent;

   SDL_EnableUNICODE(1); // guich: enable unicode support (makes shift+a = A), if and only if SDL was compiled with GETKEYBOARD_TRANSLATED
   SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);  // Catch WM unhandled messages
   SDL_EventState(SDL_ACTIVEEVENT, SDL_ENABLE); // Tell me if I turn active
   SDL_SetEventFilter(sdlEventFiltering); // allow the interception of the system keys
   SDL_EnableKeyRepeat(750, 250); // allow key repetition

   return 0;  // success!
}

/*------------------------------------------------------------------nextEvent-+
| Poll and get the next event.  REFRESH event has the lowest priority.        |
| A mouse motion event is ignored if a newer one is waiting                   |
+----------------------------------------------------------------------------*/

static int nextEvent(SDL_Event * event)
{
    SDL_Event otherEvent;

    UNLOCK_MEMORY;
    while (SDL_PollEvent(event))    // poll and get next event
    {
       if (event->type == SDL_USEREVENT_REFRESH)
       {
          _LOG_DEBUG((LOG_CAT_EVENT, "SDL_USEREVENT_REFRESH"));
#if !defined(REFRESH_NODELAY)
          if (SDL_PeepEvents(&otherEvent, 1, SDL_GETEVENT, SDL_ALLEVENTS) != 0) // still other events, postpone refresh
          {
             _LOG_DEBUG((LOG_CAT_EVENT, "re-push refresh, higher prior. event=0x%x", otherEvent));
             SDL_PushEvent(event);  // put the refresh event back in the queue
             *event = otherEvent;   // take the other instead (higher priority)
          }
          else
#endif
          {
             LOCK_MEMORY;
             return 1;
          }
       }
       if ((event->type != SDL_MOUSEMOTION) ||
            SDL_PeepEvents(&otherEvent, 1, SDL_PEEKEVENT, SDL_MOUSEMOTIONMASK) == 0) // no other motion events
       {
          LOCK_MEMORY;
          return 1;
       }
    }
    LOCK_MEMORY;
    return 0;     // no events were found
}

#ifndef WM_NCACTIVATE
#define WM_NCACTIVATE 0x0086
#endif
#ifdef WIN32
static void appActivated(int activating, int focusGained)
{
   if (focusGained && activating)
   {
      // whenever the window is activated we must make sure we re-hide
      // the taskbar etc.
      SetForegroundWindow(vmGlobals->mainHWnd);
      #ifdef __AYGSHELL_H__
      if (vmGlobals->mainHWnd && vmGlobals->isFullScreen) SHFullScreen(vmGlobals->mainHWnd, SHFS_HIDESTARTICON | SHFS_HIDETASKBAR); // guich@504_12: only if fullScreen.
      #endif
   }
   #ifdef __AYGSHELL_H__
   if (!vmGlobals->useMinimizeButton) SHDoneButton(vmGlobals->mainHWnd, focusGained?SHDB_SHOW:SHDB_HIDE); // reenable the ok button - guich@504_12: show if showing, hide if hiding
   #endif
   SDL_WM_SetCaption(vmGlobals->appTitle, 0); // set the title again
   showSip(SIP_RESTORE_STATE,SIP_RESTORE_STATE);
}
#endif
/*-----------------------------------------------------------------pumpEvents-+
| Description:                                                                |
|   The 'wait' parameter allows the caller to specify a timelimit             |
|   for pumpEvents so that                                                    |
|   it will return if no events are available and the timelimit expires.      |
|   Note that if -1 is passed in for eventWait, then pumpEvents               |
|   will act exactly as it used to.  This is used by the main event loop for  |
|   efficiency.                                                               |
|   Note: it is guaranteed that at least one iteration through pending        |
|   threads will be made on every call to this function.                      |
+----------------------------------------------------------------------------*/
int32 pumpEvents(int32 wait)
{
   int32 i;
   int32 noEvent;
   SDL_Event event;

   //_LOG_DEBUG((LOG_CAT_EVENT, "pumpE(%d)", (int)wait));
   if (vmGlobals->vmStatus->errNum != 0)
   {
      sendError(vmGlobals->vmStatus);
      return vmGlobals->vmStatus->errNum;  // fatal error!
   }
   if (vmGlobals->threadCount > 0)
   {
      int32 t0 = SDL_GetTicks();
      do
      {
         // call at least once, even if we have a event pending
         SDL_Delay(2);  // let cpu take a breath
         executeMethod(runThreadsVclass, runThreadsMethod, runThreadsParams, 1);
         checkMainWindowTimer();
         noEvent = (nextEvent(&event) == 0);
      } while (noEvent && (wait < 0 || (SDL_GetTicks() - t0 < (unsigned int)wait)));
      if (noEvent) return 0;    // time out
   }
   else
   {
      checkMainWindowTimer();
      noEvent = (nextEvent(&event) == 0);
      if (noEvent) return 0;
   }

   _LOG_DEBUG((LOG_CAT_EVENT, "attached libs ?"));
#if defined(LINUX) || defined(__SYMBIAN32__) || defined(PALMOS)
   // search through the attached libraries so they can handle this event
   if (vmGlobals->nativeLibs)
   {
      for (i=0; vmGlobals->nativeLibs[i]; ++i)
      {
         NativeLibPtr lib = vmGlobals->nativeLibs[i];
         // fdie@ same SDL code now, the function header maybe changed now to pass a real SDL_Event...
         #if defined(PALMOS_NATIVES_SYSLIB)
         if (SWHandleEvent(lib, (EventPtr)&event) == SWErrEventHandled)
            return 0;
         #else
         if ((*lib->SWHandleEvent)(lib->ref,(EventPtr)&event) == SWErrEventHandled)
            return 0;
         #endif
      }
   }
#else
   // search through the attached libraries so they can handle this event
   if (vmGlobals->nativeLibs && event.type == SDL_SYSWMEVENT)
   {
      Event e;
      e.hwnd = event.syswm.msg->hwnd;
      e.wParam = event.syswm.msg->wParam;
      e.lParam = event.syswm.msg->lParam;
      e.message = event.syswm.msg->msg;
      for (i=0; vmGlobals->nativeLibs[i]; ++i)
      {
         // fdie@ same SDL code now
         if ((*vmGlobals->nativeLibs[i]->SWHandleEvent)(0,&e) == SWErrEventHandled)
            return 0;
      }
   }
#endif
   switch (event.type)
   {
      case SDL_USEREVENT_APPSTART:
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_USEREVENT_APPSTART"));
         drawMainWin();
         SDL_UpdateRect(SDL_GetVideoSurface(), 0, 0, 0, 0);
         break;

      case SDL_USEREVENT_REFRESH:
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_USEREVENT_REFRESH (%d,%d,%d,%d)",
                        (Sint16)((int32)event.user.data1), (Sint16)(((int32)event.user.data1) >> 16),
                        (Uint16)((int32)event.user.data2), (Uint16)(((int32)event.user.data2) >> 16)));
         SDL_UpdateRect(SDL_GetVideoSurface(),
                        (Sint16)((int32)event.user.data1), (Sint16)(((int32)event.user.data1) >> 16),
                        (Uint16)((int32)event.user.data2), (Uint16)(((int32)event.user.data2) >> 16));
         break;

      case SDL_MOUSEBUTTONDOWN:  /* 200: PenEvent.PEN_DOWN */
         if (event.button.button == SDL_BUTTON_LEFT)
         {
            _LOG_DEBUG((LOG_CAT_EVENT, "POST PEN_DOWN=(%d,%d)", (int)event.button.x, (int)event.button.y));
            executePostEvent(200, 0, event.button.x, event.button.y, 0, 0);
         }
         break;

      case SDL_MOUSEMOTION:      /* 201: PenEvent.PEN_MOVE */
         /*
         | SDL generates a mouse motion event before each mouse button events.
         | For Superwaba, we just want to report drag motions.
         | Also, in SuperWaba, only a MouseButtonDown changes the focus.
         | Reporting a MouseMotion before the MouseButtonDown that triggers it
         | would induce errors for the control that has just lost the focus.
         | Hence, mouse motions should only be reported for "drags", when
         | the control "captures" the mouse.
         */
         if ((event.motion.state & SDL_BUTTON(1)) &&        // was pressed
                (SDL_GetMouseState(0, 0) & SDL_BUTTON(1)))  // still pressed
         {                                                  // aka "drag"
            _LOG_DEBUG((LOG_CAT_EVENT, "POST PEN_DRAG=(%d,%d)", (int)event.motion.x, (int)event.motion.y));
            executePostEvent(201, 0, event.motion.x, event.motion.y, 0, 0);
         }
         else
         {
            _LOG_DEBUG((LOG_CAT_EVENT, "POST PEN_MOVE=(%d,%d)", (int)event.motion.x, (int)event.motion.y));
         }
         break;

      case SDL_MOUSEBUTTONUP:    /* 202: PenEvent.PEN_UP */
         if (event.button.button == SDL_BUTTON_LEFT)
         {
            _LOG_DEBUG((LOG_CAT_EVENT, "POST PEN_UP=(%d,%d)", (int)event.button.x, (int)event.button.y));
            executePostEvent(202, 0, event.button.x, event.button.y, 0, 0);
         }
         break;

      case SDL_ACTIVEEVENT:
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_ACTIVEEVENT"));
         #ifdef WIN32
         appActivated(event.active.state & SDL_APPACTIVE, event.active.gain);
         #endif
         break;

      case SDL_SYSWMEVENT:
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_SYSWMEVENT"));
         /* Pass on the first 256 WM_USER messages as SW keys in the 76500 to 76756 range. */
         #ifdef WIN32
         if ((event.syswm.msg->msg >= WM_USER) && (event.syswm.msg->msg < (WM_USER+256)))
         {
            executePostEvent(100, 76500 + event.syswm.msg->msg - WM_USER, 0, 0, 0, 0);
         }
         else
         {
            switch (event.syswm.msg->msg)
            {
            case WM_NCACTIVATE: // guich@550_41
               appActivated(1,event.syswm.msg->wParam);
               break;
            case WM_KILLFOCUS:
               _LOG_DEBUG((LOG_CAT_EVENT, "WM_KILLFOCUS"));
               if (vmGlobals->interceptSystemKey) {
                  unregisterHotkeys(); // unregister all hotkeys
               }
               #ifdef HAS_SIP
               if (SipStatus() == SIP_STATUS_AVAILABLE) // restore the original bounds of the sip
               {
                  CLSID Clsid;
                  SipSetDefaultRect(&vmGlobals->sipRect);
                  SipGetCurrentIM(&Clsid);
                  SipSetCurrentIM(&Clsid);
                  SipShowIM(SIPF_OFF);
                  #ifdef __AYGSHELL_H__
                  if (vmGlobals->mainHWnd) SHFullScreen(vmGlobals->mainHWnd, SHFS_HIDESIPBUTTON); // only supported in 3.0
                  #endif
               }
               #endif
               break;
            case WM_SETFOCUS:
               _LOG_DEBUG((LOG_CAT_EVENT, "WM_SETFOCUS"));
               if (vmGlobals->interceptSystemKey)
               {
                  registerHotkeys(); // register them back
               }
               showSip(SIP_RESTORE_STATE,SIP_RESTORE_STATE);
               break;

            case WM_COMMAND: // guich@320_48
               _LOG_DEBUG((LOG_CAT_EVENT, "WM_COMMAND=0x%x", (int)event.syswm.msg->wParam));
               if (event.syswm.msg->wParam == 0x10000001) // done button? (not sure why this parameter is this value)
               {
                  postStopEvent(0);
                  return 0;
               }
               break;

            case WM_HOTKEY:
               _LOG_DEBUG((LOG_CAT_EVENT, "WM_HOTKEY=0x%x", (int)LOWORD(event.syswm.msg->lParam)));
               if (0 == (LOWORD(event.syswm.msg->lParam) & 0x1000)) // key down
               {
                  int32 key;
                  SDLMod mod = SDL_GetModState();
                  switch (event.syswm.msg->wParam)
                  {
                     case 193:        // Calendar
                        key = 76003;
                        break;
                     case 194:        // Contacts
                        key = 76004;
                        break;
                     case 195:        // Inbox
                        key = 76005;
                        break;
                     case 196:        // iTask
                        key = 76006;
                        break;
                     case 197:        // Notes
                        key = 76007;
                        break;
                     default:
                        key = 76100 + event.syswm.msg->wParam - 193;
                        break;
                  }
                  executePostKeyEvent(key, mod);
               }
               break;
            }
         }
         // #elif LINUX...
         #endif
         break;

      case SDL_KEYDOWN:
      {
         int32 key = -1;
         SDLMod mod = event.key.keysym.mod;
         SDLKey sym = event.key.keysym.sym;
         Uint16 unicode = event.key.keysym.unicode;
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_KEYDOWN=(0x%x,0x%x,0x%x)", (int)mod, (int)sym, (int)unicode));

         switch (sym)
         {
            case SDLK_KP0:
            case SDLK_KP1:
            case SDLK_KP2:
            case SDLK_KP3:
            case SDLK_KP4:
            case SDLK_KP5:
            case SDLK_KP6:
            case SDLK_KP7:
            case SDLK_KP8:
            case SDLK_KP9: key = '0'+(sym-SDLK_KP0); break;
            case SDLK_KP_PERIOD:        key = '.';   break;
            case SDLK_KP_DIVIDE:        key = '/';   break;
            case SDLK_KP_MULTIPLY:      key = '*';   break;
            case SDLK_KP_MINUS:         key = '-';   break;
            case SDLK_KP_PLUS:          key = '+';   break;
            case SDLK_KP_ENTER:         key = 75009; break; // guich@502_13: fixed numeric keypad enter key code
            case SDLK_KP_EQUALS:        key = '=';   break; // guich@501_8: from here to up: fix support for numeric keypad
            case SDLK_PAGEUP:           key = 75000; break; // PAGE_UP
            case SDLK_PAGEDOWN:         key = 75001; break; // PAGE_DOWN
            case SDLK_HOME:             key = 75002; break; // HOME
            case SDLK_END:              key = 75003; break; // END
            case SDLK_UP:               key = 75004; break; // UP
            case SDLK_DOWN:             key = 75005; break; // DOWN
            case SDLK_LEFT:             key = 75006; break; // LEFT
            case SDLK_RIGHT:            key = 75007; break; // RIGHT
            case SDLK_INSERT:           key = 75008; break; // INSERT
#if !defined(SUPPORT_PUBLIC_SDL)
            case _SDLK_ACTION:          key = 75009; break; // ACTION
#endif            
            case SDLK_RETURN:           key = 75009; break; // ENTER
            case SDLK_TAB:              key = 75010; break; // TAB
            case SDLK_BACKSPACE:        key = 75011; break; // BACKSPACE
            case SDLK_ESCAPE:           key = 75012; break; // ESCAPE
            case SDLK_DELETE:           key = 75013; break; // DELETE
            case SDLK_MENU:             key = 75014; break; // MENU

#if defined(PALMOS)
            case _SDLK_COMMAND:         key = 75015; break; // COMMAND 
            case _SDLK_KEYBOARD_ALPHA:  key = 76000; break; // KEYBOARD_ABC
            case _SDLK_KEYBOARD_NUM:    key = 76001; break; // KEYBOARD_123
            case _SDLK_KEYBOARD:        key = 76002; break; // KEYBOARD
            case _SDLK_HARD1:           key = 76003; break; // HARD1
            case _SDLK_HARD2:           key = 76004; break; // HARD2
            case _SDLK_HARD3:           key = 76005; break; // HARD3
            case _SDLK_HARD4:           key = 76006; break; // HARD4
            case _SDLK_CALC:            key = 76007; break; // CALC
            case _SDLK_FIND:            key = 76008; break; // FIND
            case _SDLK_LAUNCH:          key = 76009; break; // LAUNCH
            case _SDLK_BRIGHTNESS:
            case _SDLK_CONTRAST:        key = 76011; break; // CONTRAST guich@450_25, guich@402_2: fixed contrast, guich@401_31
            case _SDLK_CLOCK:           key = 76012; break; // CLOCK guich@401_31, guich@402_2: fixed clock
            case _SDLK_HARDCRADDLE:     key = 76013; break; // SYNC guich@421_18
#endif

/* fdie@ not yet implemented PalmOS keys
   public static final int JOG_UP         = 76030;
   public static final int JOG_DOWN       = 76031;
   public static final int JOG_PUSH       = 76032;
   public static final int JOG_RELEASE    = 76033;
   public static final int JOG_PUSHREPEAT = 76034;
   public static final int JOG_PUSHUP     = 76035;
   public static final int JOG_PUSHDOWN   = 76036;
   public static final int JOG_BACK       = 76037;
*/

#ifdef WIN32 // !wince && win32 ?
            case SDLK_F1:  if (IS_KEY_INTERCEPTED(SK_HARD1))  key = 76003; break; // guich@502_10: changed VK_F1 to SDLK_F1 bc VK_F1 is a valid key (p) on win32
            case SDLK_F2:  if (IS_KEY_INTERCEPTED(SK_HARD2))  key = 76004; break;
            case SDLK_F3:  if (IS_KEY_INTERCEPTED(SK_HARD3))  key = 76005; break;
            case SDLK_F4:  if (mod & KMOD_LALT)               return -1;          // guich@501_7: fixed not closing on alt+f4
                           if (IS_KEY_INTERCEPTED(SK_HARD4))  key = 76006; break;
            case SDLK_F7:  if (IS_KEY_INTERCEPTED(SK_CALC))   key = 76007; break;
            case SDLK_F8:  if (IS_KEY_INTERCEPTED(SK_FIND))   key = 76008; break;
            case SDLK_F9:  if (IS_KEY_INTERCEPTED(SK_LAUNCH)) key = 76009; break;
            case SDLK_F12: if (IS_KEY_INTERCEPTED(SK_ACTION)) key = 76010; break;
            case SDLK_F10:         key = 76000; break;
            case SDLK_F11:         key = 76001; break;
            case SDLK_F6:          key = 75014; break;
            case SDLK_F5:          key = 75015; break;
#endif // WIN32
//#else
//            case 134:              key = 76010; break; // guich@501_6: IKeys.ACTION - debugging on sdl, this key turned from 134 to 13 (enter) - guich@504_11: fixed the value to 134
//#endif // WINCE
            default:             // ASCII

#if defined(PALMOS)
               if (sym >= 32)
                  key = sym;
#else
               if (sym >= 32)
                  key = unicode;
#endif
               break;
         }
         // post key only if the SuperWaba translation succeeded
         if (key >= 0) executePostKeyEvent(key, mod);

#if defined(PALMOS) //fdie@ TODO define new _SDLK_LAUNCH SDL syms or use custom like SDLK_F9
         // not intercepted launch key => quit
         if (sym == _SDLK_LAUNCH && (vmGlobals->interceptSystemKey & SK_LAUNCH) == 0)
         {
            _LOG_DEBUG((LOG_CAT_EVENT, "SDL_QUIT"));
            return -1;  // bye bye!
         }
#endif
         break;
      }

      case SDL_QUIT:
         _LOG_DEBUG((LOG_CAT_EVENT, "SDL_QUIT"));
         return -1;  // bye bye!
      default:
         _LOG_DEBUG((LOG_CAT_EVENT, "unhandled event=0x%x", (int)event.type));
         break;
   }
   return 0;
}

/*-------------------------------------------------------checkMainWindowTimer-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void checkMainWindowTimer()
{
   if (vmGlobals->globalTimerInterval > 0)
   {
      WObject mainWinObj = vmGlobals->globalMainWin;
      if (mainWinObj)
      {
         int32 now = getTimeStamp();
         int32 diff = now - vmGlobals->globalTimerStart;
         if (diff < 0) diff += (1L << 30); // max stamp is (1 << 30)
         if (diff >= vmGlobals->globalTimerInterval)
         {
            WClass * vclass = WOBJ_class(mainWinObj);
            WClassMethod * method = getMethodByMapNum(vclass, &vclass, (uint16)vmGlobals->onTimerTickMethodMapNum);
            vmGlobals->globalTimerStart = now;
            if (method)
            {
               Var params[1];
               params[0].obj = mainWinObj;
               executeMethod(vclass, method, params, 1);
            }
         }
      }
   }
}

/*--------------------------------------------------------executePostKeyEvent-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void executePostKeyEvent(int32 key, SDLMod mod)
{
   if (key)
   {
      int32 modifiers = 0;
      if (mod & KMOD_CTRL)  modifiers |= (1 << 1);
      if (mod & KMOD_SHIFT) modifiers |= (1 << 2);
      if (mod & KMOD_CAPS)  modifiers |= (1 << 3);
      if (mod & KMOD_ALT)   modifiers |= (1 << 5);
      /* 100: KeyEvent.KEY_PRESS */
      executePostEvent(100, key, 0, 0, modifiers, 0);
   }
}

/*-----------------------------------------------------------executePostEvent-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void executePostEvent(int32 type, int32 key, int32 x, int32 y,
                             int32 modifiers, int32 timeStamp)
{
   WObject mainWinObj;
   WClass * vclass;
   WClassMethod * method;
   Var params[7];

   mainWinObj = vmGlobals->globalMainWin;
   if (!mainWinObj)
      return;

   vclass = WOBJ_class(mainWinObj); // get runtime class
   method = getMethodByMapNum(vclass, &vclass, (uint16)vmGlobals->postEventMethodMapNum);
   if (method)
   {
      params[0].obj = mainWinObj;
      params[1].intValue = type;
      params[2].intValue = key;
      params[3].intValue = x;
      params[4].intValue = y;
      params[5].intValue = modifiers;
      params[6].intValue = timeStamp;
      _LOG_INFO((LOG_CAT_EVENT, "MainWindow._onEvent(): type=%d key=0x%x x=%d y=%d mod=0x%x",
               (int)type, (int)key, (int)x, (int)y, (int)modifiers));
      executeMethod(vclass, method, params, 7);
   }
}

/*----------------------------------------------------------------drawMainWin-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void drawMainWin()
{
   WObject mainWinObj;
   WClass *vclass;
   WClassMethod *method;
   Var params[5];

   mainWinObj = vmGlobals->globalMainWin;
   if (mainWinObj)
   {
      vclass = WOBJ_class(mainWinObj); // get runtime class
      method = getMethod(vclass, createUtfString("_doPaint"), createUtfString("(IIII)V"), &vclass);
      if (method == NULL) return;
      params[0].obj = mainWinObj;
      params[1].intValue = -5000; // -5000: ugly, i know. But necessary.
      params[2].intValue = 0; // y
      params[3].intValue = vmGlobals->windowWidth;
      params[4].intValue = vmGlobals->windowHeight;
      executeMethod(vclass, method, params, 5);
      _LOG_INFO((LOG_CAT_EVENT, "MainWindow._doPaint(): x=%d y=%d width=%d height=%d",
                  (int)params[1].intValue, (int)params[2].intValue, (int)params[3].intValue, (int)params[4].intValue));
   }
}

/*----------------------------------------------------------------atLoopStart-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void atLoopStart()
{
   SDL_Event appStartEvent;
   appStartEvent.type = SDL_USEREVENT_APPSTART;
   appStartEvent.user.code = 2;
   appStartEvent.user.data1 = (void*)0;
   appStartEvent.user.data2 = (void*)0;
   SDL_PushEvent(&appStartEvent);
}

/*--------------------------------------------------------------runEventsLoop-+
|                                                                             |
+----------------------------------------------------------------------------*/
int32 runEventsLoop()
{
   int32 rc;
   atLoopStart();
   while (!vmGlobals->quitRightNow && (rc=pumpEvents(-1)) == 0) // guich@552_1: added quitRightNow to make sure that the program will quit
   {
#ifdef WIN32
      Sleep(10);
#else
      SDL_Delay(1);  // let cpu take a breath
#endif
   }
   return rc;
}

/*-----------------------------------------------------------eventIsAvailable-+
|                                                                             |
+----------------------------------------------------------------------------*/
int32 eventIsAvailable()
{
   SDL_Event event;
#ifdef WIN32 // guich@552_38
      Sleep(10);
#else
      SDL_Delay(1);  // let cpu take a breath
#endif
   return SDL_PeepEvents(&event, 1, SDL_PEEKEVENT, SDL_ALLEVENTS);
}

/*--------------------------------------------------------------postStopEvent-+
|                                                                             |
+----------------------------------------------------------------------------*/
void postStopEvent(int32 exitCode)
{
   static byte once = 1;
#if defined(WINCE) && !defined(WIN32)
   if (exitCode == -1)
   {
      SDL_WM_IconifyWindow();  // on WINCE, the Window is hidden vs iconified - is this ever called????
   }
   else
#endif
   if (once)
   {
      SDL_Event ev;
      ev.type = SDL_QUIT;
      SDL_PushEvent(&ev);
      once = 0;
      swExitCode = exitCode;
      // PostQuitMessage(exitCode);
   }
}

/*---------------------------------------------------------------getTimeStamp-+
|                                                                             |
+----------------------------------------------------------------------------*/
int32 getTimeStamp()
{
   return (int32)(SDL_GetTicks() % ((uint32)1 << 30));
}

/*--------------------------------------------------------------millisToTicks-+
|                                                                             |
+----------------------------------------------------------------------------*/
static int32 tps;
uint32 millisToTicks(int32 millis)
{
#ifdef PALMOS
   if (tps == 0)                 // guich@551_4
      tps = SysTicksPerSecond(); // 1000 in new ARM devices
   return (tps==1000) ? millis : (millis * tps / 1000);
#else
   return millis;
#endif
}

/*------------------------------------------------------------registerHotkeys-+
|                                                                             |
+----------------------------------------------------------------------------*/
void registerHotkeys()
{
#ifdef WIN32
   if (!vmGlobals->mainHWnd) return;

   if (vmGlobals->interceptSystemKey == SK_ALL || vmGlobals->interceptSystemKey == 0xFFFFFFFFL)
   {
      int i;
      for (i=0xc1; i<=0xcf; i++)
         RegisterHotKey(vmGlobals->mainHWnd, i, MOD_WIN, i);
   }
   else
   {
      int32 which;
      if ((vmGlobals->interceptSystemKey & SK_HARD1)  != 0)
      {
         which = 193;
         RegisterHotKey(vmGlobals->mainHWnd, which, MOD_WIN, which);
      }
      if ((vmGlobals->interceptSystemKey & SK_HARD2)  != 0)
      {
         which = 194;
         RegisterHotKey(vmGlobals->mainHWnd, which, MOD_WIN, which);
      }
      if ((vmGlobals->interceptSystemKey & SK_HARD3)  != 0)
      {
         which = 195;
         RegisterHotKey(vmGlobals->mainHWnd, which, MOD_WIN, which);
      }
      if ((vmGlobals->interceptSystemKey & SK_HARD4)  != 0)
      {
         which = 196;
         RegisterHotKey(vmGlobals->mainHWnd, which, MOD_WIN, which);
      }
      if ((vmGlobals->interceptSystemKey & SK_CALC)  != 0)
      {
         which = 197;
         RegisterHotKey(vmGlobals->mainHWnd, which, MOD_WIN, which);
      }
   }
#endif
}

/*----------------------------------------------------------unregisterHotkeys-+
|                                                                             |
+----------------------------------------------------------------------------*/
void unregisterHotkeys()
{
#if defined(WINCE) // doesn't work under WIN32  dave@400_48
   HINSTANCE hCoreDll;
   UnregisterFunc1Proc procUnregisterFunc;

   if (!vmGlobals->mainHWnd) return;

   hCoreDll = LoadLibrary(_T("coredll.dll"));
   if (!hCoreDll)
   {
      (*vmGlobals->debug)("Could not register hotkeys");
      return;
   }

   procUnregisterFunc = (UnregisterFunc1Proc)GetProcAddress(hCoreDll, _T("UnregisterFunc1"));
   if (procUnregisterFunc)
   {
      int i;
      for (i=0xc1; i<=0xcf; i++)
      {
         procUnregisterFunc(MOD_WIN, i);
         UnregisterHotKey(vmGlobals->mainHWnd, i);
      }
   }
   FreeLibrary(hCoreDll);
#endif
}

/*===========================================================================*/
