/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_main.c,v 1.29 2005/10/31 19:35:50 guich Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"

/* Work-In-Progress:
| Except for superwaba.c, DEFINE_FUNCTIONS must be placed AFTER including
| superwaba.h.
|
| _DEFINE_FUNCTIONS defines the functions that can be directly called
| from the Kernel side, exclusively.  To access the same functions
| from the Native side, one must use the VMGlobals.  (*vmGlobals->fct)()
|
| Initially, these "private functions" were defined in SuperWaba.h.
| Since, some have been migrated to other H files (like SW_Class.h)
| but SuperWaba.h still has several of those.
|
| In order to make a clean inventory of what was needed, each kernel module
| lists the private functions it needs under the comments:
| "from superwaba.c".
|
| This is, of course, TEMPORARY.  The goal is to create a SuperWaba_Private.h,
| and to remove all the extra-prototypes.  We could also have a SW_kernel.c/h
| Or SW_functions, or VMGlobals...
|
| Being not yet fully designed and given that it takes a couple of days for
| implementing it (while we have time constraints) this work is not finished.
*/
#define __DEFINE_FUNCTIONS
#include "SDL.h"
#ifndef PALMOS
#include "SDL_net.h"
#endif
#include "SDL_syswm.h"
#include "SW_palmDb.h"
#include "SW_settings.h"  // for showSip
#include "SW_events.h"
#include "SW_natives.h"
#include "SW_messages.h"
#include "SW_utils.h"
#include "SW_String.h"
#include "SW_Class.h"
#include "SW_main.h"
#include "SW_logging.h"

#if defined(PALMOS)

#if defined(__arm__)

#if PALMOS_SDK_VERSION >= 0x0541
#define __palmos__ // required or AdnDebug instruction are discarded
#include <AdnDebugMgr.h>
#undef __palmos__
#else
#define AdnDebugEnableSet(x)
#define AdnDebugBreak()
#endif

#include <pealstub.h>
// this globals will be defined during m68k startup for later m68k callback
void *pealLoadLibrary68K, *pealGetProcAddress68K;
#endif //defined(__arm__)

#include "resources.h"
#include <palm_posix.h>
#endif //PALMOS

#ifndef __SYMBIAN32__
void showMsg(_unused_ int no, char * msg1, char * msg2)
{
   _LOG_INFO((LOG_CAT_NONE, "%s, %s", msg1, msg2));
}
void showError(int error)
{
   _LOG_ERROR((LOG_CAT_NONE, "error=%d", error));
}
#endif

// from superwaba.c
WObject allocObject(int32 size);
int32 arrayCount(char *array, int32 itemSize);
int32 arrayRangeCheck(WObject array, int32 start, int32 count);
int32 arraySize(int32 type, int32 len);
uint16 arrayType(char c);
byte callMethod(WObject classObj, char *methodName, char *methodDesc, Var *params, int32 paramCount);
int32 compatible(WClass *wclass inA0, WClass *target inA1);
int32 compatibleArray(WObject obj, UtfString arrayName);
Var constantToVar(WClass *wclass inA0, uint16 idx inD0);
void copyArray(Var stack[]);
int32 countMethodParams(UtfString desc);
_inline_ WObject createArrayObject(int32 type inD0, int32 len inD1);
WObject createMultiArray(int32 ndim, char *desc, Var *sizes);
_inline_ WObject createObject(WClass *wclass);
WObject double2str(double d, int32 places);
void doubleToStr(char dest[], double val, int32 totalLength, int32 decimal, int32 options);
void dumpStackTrace();
Var* executeMethod(WClass *wclass, WClassMethod *method,Var params[], uint32 numParams);
char *expandArray(char *array, int32 itemSize, int32 increment, int32 *insertIndex);
WObject float2str(float f);
void gc();
WClass * getClass(UtfString className, byte throwEx);
WClass *getClassByIndex(WClass *wclass inA0, uint16 classIndex inD0, byte throwEx inD1);
UtfString getClassName(WObject obj);
WClassField *getField(WClass *wclass, UtfString *name, UtfString *desc);
WClassMethod *getMethod(WClass *wclass, UtfString name, UtfString desc, WClass **vclass);
WClassMethod *getMethodByMapNum(WClass *wclass inA0, WClass **vclass inA1, uint16 mapNum inD0);
int16 getMethodMapNum(WClass *wclass, UtfString *name, UtfString *desc, byte searchType, WClassMethod **imethod, WClass **iclass);
WObject int642str(int64 i);
char *intToString(int32 i);
uchar *loadClassField(WClass *wclass, WClassField *field, uchar *p);
uchar *loadClassMethod(WClass *wclass, WClassMethod *method, uchar *p);
WObject popObject();
int32 pushObject(WObject obj);
double str2double(char *str);
int32 str2int32(char *str);
int64 str2int64(char *str);
void swQuickError(uint16 errNum, char *str);
WObject throwException(char *exceptionClass inA0, char *message inA1);
void tone(int32 freq, int32 dur);
void showCommand(int32 show); // guich  // ds@110 change Boolean to int to run on Win32
char *int2str(char *msg, int32 ii);

typedef struct
{
   uint32 creator;
   uint32 type;
   int32 count;
   DmOpenRef *classDbP;
} CountLibs_Vars;


static int32 initVMGlobals(int32 argc, char ** argv, char ** swPdbName);
static void destroyVMGlobals();
static WObject startApp(char * swPdbName);

static char defltSwPdbName[] = "SuperWaba.pdb";
static char const defltApp[] = "waba/ui/Welcome";

// from superwaba.c:
extern uint16 SWLibRefNum;
extern ErrorStatus vmStatus;
extern WObject firstException;

void swError(uint16 errNum, WClass *iclass, UtfString *desc1, UtfString *desc2);
void swInit();
WObject swStartApp(char *className);
void swStopApp(WObject mainWinObj);

//////////////////////////////////////////////////////////////////////////////
//                          Program Entry Point                             //
//////////////////////////////////////////////////////////////////////////////

#if defined(PALMOS)
//>>>PGR: This has to go on the Palm startup.c side.
#define MAX_LAUNCHCODES 12
typedef struct
{
   uint16 launchCode;
   uint16 refNum;
} Codes2Launch;
int32 nCodes2Launch;
Codes2Launch codes2Launch[MAX_LAUNCHCODES]; // i hope there's no more than 10 codes

#define APPLICATION_LAUNCHCODE   ((Word)40404)

#endif // PALMOS

#ifdef __ECOS
MTAB_ENTRY(romfs_mte1,"/","romfs","",(CYG_ADDRWORD)&filedata[0]);
#endif

#ifndef CS_BYTEALIGNCLIENT
   #define CS_BYTEALIGNCLIENT 0
#endif

extern int gc_(); // also calls the deleteUnusedBlocks

#undef ENDIAN_CHECK
#if defined(ENDIAN_CHECK)

/**
 * Do an endian check to see if 64bits integers and doubles stored
 * in LE (class bytecode format) are decoded correctly by the VM.
 */
void endianCheck()
{
   union
   {
      int64 i;
      double d;
      unsigned char uc[8];
   } u64;
   char BUFFER[1024];

   u64.i = 0xDDCCBBAA78563412L; // int64 0x12345678AABBCCDD in Little Endian
   u64.i = READ_BE_INT64(&u64.i);
   if (u64.i != 0x12345678AABBCCDDL)
   {
      _LOG_ERROR((LOG_CAT_MATH, "wrong int64 endian! 0x%08lx%08lx", (long)(u64.i>>32), (long)(u64.i & 0xFFFFFFFFL)));

      _LOG_DEBUG((LOG_CAT_MATH, "should be => 0x12345678AABBCCDD"));
      _LOG_DEBUG((LOG_CAT_MATH, "you got => 0x%08lx%08lx", (long)(u64.i>>32), (long)(u64.i & 0xFFFFFFFFL)));

      u64.i = 0x12345678AABBCCDDL;
      _LOG_DEBUG((LOG_CAT_MATH, "mem mapping of 0x12345678AABBCCDD"));
      _LOG_DEBUG((LOG_CAT_MATH, "%02xh,%02xh,%02xh,%02xh,%02xh,%02xh,%02xh,%02xh",
                                     u64.uc[0], u64.uc[1], u64.uc[2], u64.uc[3],
                                     u64.uc[4], u64.uc[5], u64.uc[6], u64.uc[7]));
   }

   u64.i = 0xD3957DCF37B6D23FL; // double 0.2923717047 in Little Endian
   u64.i = READ_BE_DOUBLE(&u64.i);
   if (u64.i != 0x3FD2B637CF7D95D3L)
   {
      _LOG_ERROR((LOG_CAT_MATH, "wrong doubles endian! 0x%08lx%08lx", (long)(u64.i>>32), (long)(u64.i & 0xFFFFFFFFL)));

      _LOG_DEBUG((LOG_CAT_MATH, "should be => 0x3FD2B637CF7D95D3"));
      _LOG_DEBUG((LOG_CAT_MATH, "you got => 0x%08lx%08lx", (long)(u64.i>>32), (long)(u64.i & 0xFFFFFFFFL)));

      u64.d = 0.2923717047;
      _LOG_DEBUG((LOG_CAT_MATH, "mem mapping of %lf", u64.d));
      _LOG_DEBUG((LOG_CAT_MATH, "%02xh,%02xh,%02xh,%02xh,%02xh,%02xh,%02xh,%02xh",
                                   u64.uc[0], u64.uc[1], u64.uc[2], u64.uc[3],
                                   u64.uc[4], u64.uc[5], u64.uc[6], u64.uc[7]));
      _LOG_DEBUG((LOG_CAT_MATH, "longbits(0x%08lx%08lx)", (long)(u64.i>>32), (long)(u64.i & 0xFFFFFFFFL)));
   }
}
#endif

/*-----------------------------------------------------------------------main-+
|                                                                             |
+----------------------------------------------------------------------------*/
int startup(StartupParams *params)
{
   uint32 rc;
   char *swPdbName;
#if defined(PALMOS)
   MemHandle posixCfgH;
   PalmPosixConfig *posixCfg;

   /* fdie@ warning: the alert resource have to be loaded in memory already or this
    * call will fail with the following error: MemoryMgr.c, Line: 3635, NULL handle.
    * Maybe due to the fact that the ALert is a m68k resource (big endian).
    */
   PEAL_INFO(VmInfoAlert, "ARM", "entered", NULL)

#if defined(__arm__)
   AdnDebugEnableSet(0xF);
   AdnDebugBreak();

   params->argc = Swap32(params->argc);
   params->argv = (char **)Swap32(params->argv);

   for (rc = 0; rc < params->argc; rc++)
   {
      params->argv[rc] = (char*)Swap32(params->argv[rc]);
   }
#endif
   PEAL_INFO(VmInfoAlert, "ARM", "params swapped", NULL);

   // NOTE: We need VoidHand to fall on a byte boundry for our xmalloc()
   // routine to work correctly (see xmalloc())
   if ((sizeof(VoidHand) & 3) != 0)
   {
      swQuickError(ERR_SanityCheckFailed,NULL);
      return 0;
   }

   PEAL_INFO(VmInfoAlert, "ARM", "1st MemHandleNew", NULL);
   posixCfgH = MemHandleNew(sizeof(PalmPosixConfig)); // dynalloc guaranties ARM correct alignment
   if (!posixCfgH)
   {
      return -1;
   }
   posixCfg = (PalmPosixConfig*) MemHandleLock(posixCfgH);

   memset(posixCfg, 0, sizeof(*posixCfg));
   posixCfg->appl                 = "Superwaba";
   posixCfg->creatorID            = SUPERWABA_CREATOR_ID;
   posixCfg->useStorageHeap       = true;
   posixCfg->releaseFunc          = releaseAndClose;
   posixCfg->garbageCollectFunc   = gc_;
   palmPosixInitialize(posixCfg);

   UNLOCK_MEMORY;
   PEAL_INFO(VmInfoAlert, "ARM", "palmPosixInitialize.", NULL);
   LOCK_MEMORY;

#endif // PALMOS

   // create the VMGlobals structure
   SWLibRefNum = 0xFF; // can't be 0 - required by initVMGlobals
   rc = initVMGlobals(params->argc, params->argv, &swPdbName);
   UNLOCK_MEMORY;
   PEAL_INFO(VmInfoAlert, "ARM", "initVMGlobals", 0);
   LOCK_MEMORY;
   if (rc != 0)
   {
      if (vmGlobals)
         destroyVMGlobals();
      return rc;
   }

   UNLOCK_MEMORY;
   PEAL_INFO(VmInfoAlert, "ARM", "SDL_Init", 0);
   LOCK_MEMORY;
   _LOG_INFO((LOG_CAT_INIT, "SDL_Init"));

#if defined(ENDIAN_CHECK)
   endianCheck();
#endif

   /* perform SDL initialization  (Audio initialization performed on the fly) */
   if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE) < 0) {
      destroyVMGlobals();
      return ERR_SdlInitFailed;
   }

#if !defined linux && !defined WIN32 // in win32 we want to show the cursor
   SDL_ShowCursor(SDL_DISABLE);
#endif

   /* if no width and height were set, default to appropriate values */
   if (!vmGlobals->screenWidth || !vmGlobals->screenHeight)
   {
      int width;
      int height;
      SDL_Rect ** rects = SDL_ListModes(
         0,
         vmGlobals->isFullScreen? SDL_FULLSCREEN : 0
      );
      if (rects == 0) // no size fits
      {
         destroyVMGlobals();
         return ERR_SdlInitFailed;
      }
      else if (rects == (SDL_Rect **)-1) // any size fits
      {
         #ifdef WINCE
         width = GetSystemMetrics(SM_CXSCREEN);
         height = GetSystemMetrics(SM_CYSCREEN);
         #else
         width = 240;
         height = 320;
         #endif
      }
      else
      {
         width = rects[0]->w;
         height = rects[0]->h;
      }
      vmGlobals->screenWidth = width;
      vmGlobals->screenHeight = height;
   }
   _LOG_INFO((LOG_CAT_INIT, "screenSize=%dx%d", vmGlobals->screenWidth, vmGlobals->screenHeight));

/*{  // guich@550_46: sanity check for xfree,xmalloc and xrealloc - non of these should fail
   int i;
   byte *p;
   uint32 ns = 20;
   byte *b = xmalloc(10);
   for (p=b,i=0; i < 10; i++)
      *p++ = i;
   xfree(b);
   b = xmalloc(10);
   if (!b)
   {
      debug("xmalloc failed!");
      destroyVMGlobals();
      return 0;
   }
   for (p=b,i=0; i < 10; i++)
      *p++ = i;
   b = xrealloc(b,&ns);
   if (!b)
   {
      debug("xrealloc failed!");
      destroyVMGlobals();
      return 0;
   }
   for (i=10; i < 20; i++)
      *p++ = i;
   for (p=b,i=0; i < 20; i++)
      if (*p++ != i)
      {
         debug("xrealloc test failed!");
         destroyVMGlobals();
         return 0;
      }
   xfree(b);
} */

   #if !defined(PALMOS)
   SDL_WM_SetCaption(vmGlobals->appTitle, 0);
   #endif

   #ifdef WIN32
   {
      SDL_SysWMinfo winfo;
      SDL_VERSION(&(winfo.version));
      if (SDL_GetWMInfo(&winfo) != 1)
         vmGlobals->mainHWnd = NULL;
      else
         vmGlobals->mainHWnd = winfo.window;
      BringWindowToTop(vmGlobals->mainHWnd); // trying to make the window appear on top.
   }
   #endif

   /* Initialize SuperWaba */
   _LOG_INFO((LOG_CAT_INIT, "SWLibOpenStatic"));
   SWLibOpenStatic(SWLibRefNum, vmGlobals);

   _LOG_DEBUG((LOG_CAT_INIT, "PUBLIC_GraphicsInit"));
   if (SWCallPublicStatic(0,PUBLIC_GraphicsInit, 0, 0) != 0)  // guich@200b3
      swQuickError(ERR_NoFontsFound,NULL);
   else
      vmGlobals->globalMainWin = startApp(swPdbName);

   if (vmGlobals->globalMainWin == 0)
      rc = (vmStatus.errNum>0) ? vmStatus.errNum : ERR_NoMainWindowObj; // guich@504_8: use the errNum if assigned.
   else
   {
      _LOG_DEBUG((LOG_CAT_INIT, "initEvents"));
      rc = initEvents();
   }

   if ((rc != 0) || (vmStatus.errNum > 0))
   {
      _LOG_DEBUG((LOG_CAT_INIT, "firstException?"));
      if (firstException)
      {
         UtfString utf;
         // guich@360_70: if an exception was thrown in the initializer, show it.
         WClass *wclass;
         WObject obj = firstException[1].obj;
         wclass = WOBJ_class(firstException);  // get the class name
         if (obj)    // get the message
            utf = stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE);
         else
            utf.len = 0;
         swError(ERR_UncaughtException, wclass, &utf, NULL);
      }
      else
      {
         swError(ERR_CantInitializeApp, NULL, NULL, NULL);
      }
      if (vmStatus.errNum > 0)
         sendError(&vmStatus);
   }

   _LOG_INFO((LOG_CAT_INIT, "Start main loop"));
   vmGlobals->globalNetState = NET_READY_FOR_OPEN;

   if (rc == 0)
   {
      int i;
      // if (!vmGlobals->quitRightNow) // guich@504_2: check if the application asked to quit before the main event loop has started - guich@552_1: this test is now made in runEventsLoop
      {
         rc = runEventsLoop();
      }
      _LOG_DEBUG((LOG_CAT_INIT, "swStopApp"));
      swStopApp(vmGlobals->globalMainWin);

      if (vmGlobals->interceptSystemKey)
      {
         unregisterHotkeys();
         vmGlobals->interceptSystemKey = 0;
      }
      _LOG_DEBUG((LOG_CAT_INIT, "SWLibClose"));
      #if defined(PALMOS_SWNATIVES_SYSLIB)
      SWLibClose(SWLibRefNum);
      SysLibRemove(SWLibRefNum);
      #else
      SWLibCloseStatic();
      #endif

      if (vmGlobals->nativeLibs)
      {
         for (i = 0; vmGlobals->nativeLibs[i]; ++i)
         {
			NativeLibPtr lib;
            _LOG_DEBUG((LOG_CAT_INIT, "Close native lib %d", i));
         	lib = vmGlobals->nativeLibs[i];
            #if defined(PALMOS_NATIVES_SYSLIB)
            SWLibClose(lib);
            SysLibRemove(lib);
            #else
            (*lib->SWLibClose)(lib->ref);
            #endif
         }
      }

      systemOutClose();

      i = vmGlobals->reboot; // save before destroying!
      destroyVMGlobals();
      // guich@401_42: reboot the device if requested
      if (i)   // reboot?
      {
#if defined(PALMOS)
         SysReset();
#elif defined WINCE
         KernelIoControl(IOCTL_HAL_REBOOT, NULL, 0, NULL, 0, NULL);
#elif defined WIN32
         ExitWindowsEx(EWX_REBOOT,0);
#endif
      }
   }

   /* Clean up SDL */
#ifndef PALMOS
   SDLNet_Quit();  /* in case it was open */
#endif

   SDL_Quit();

#if defined(PALMOS)
   palmPosixUninitialize();
   MemHandleUnlock(posixCfgH);
   MemHandleFree(posixCfgH);
#endif

   return rc;
}

/*--------------------------------------------------------------initVMGlobals-+
|                                                                             |
+----------------------------------------------------------------------------*/
void setPalmDBProcs(struct VMGlobals * vmGlobals);
static int32 initVMGlobals(int32 argc, char ** argv, char ** swPdbName)
{
   char * arg;
   char const * swPath = 0;
#if !defined PALMOS
   char const * appPath = 0;
#endif
   int32 romVersion = 0;

   // guich@200b4: alloc space for the globals
   if ((vmGlobals = xmalloc(sizeof(struct VMGlobals))) == NULL)
   {
      // guich@200b4_78: most critical error, fail directly
      swQuickError(ERR_CantAllocateMemory, NULL); // guich@200b4_78
      return ERR_CantAllocateMemory;
   }
   xmemzero(vmGlobals, sizeof(struct VMGlobals));
   UNLOCK_MEMORY;
   PEAL_INFO(VmInfoAlert, "ARM", "vmGlobals allocated", 0);
   LOCK_MEMORY;

#if defined(PALMOS)
   // Check for a required version of PalmOS - currently PalmOS 2.0
   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

#if defined(PALMOS5)
   if (romVersion < 0x04000000) // note: PalmOS5 also functions also supports Tungsten W and other 16bpp devices
#else
   if (romVersion < 0x02000000)
#endif
   {
      UNLOCK_MEMORY;
      PEAL_INFO(VmInfoAlert, "ARM", "Incompatible device!", 0);
      LOCK_MEMORY;
      WinDrawChars("Incompatible device!",20,10,10); // guich@340_61
      SysTaskDelay(millisToTicks(2000));
      return sysErrRomIncompatible;
   }
   vmGlobals->romVersion = romVersion;
#endif //PALMOS

   UNLOCK_MEMORY;
   PEAL_INFO(VmInfoAlert, "ARM", "got romVersion", 0);
   LOCK_MEMORY;

   vmGlobals->SWLibRefNum = SWLibRefNum;
   *swPdbName = defltSwPdbName;

   /* initialize the functions table right now,
      to be able to use the Superwaba core library */

   // nmwin32_defines.h
   vmGlobals->xmalloc = xmalloc;
   vmGlobals->xrealloc = xrealloc;
   vmGlobals->xfree_ = xfree_;
   // nmwin32_bridge.c
   vmGlobals->getTimeStamp = getTimeStamp;
   vmGlobals->millisToTicks = millisToTicks;
   vmGlobals->postStopEvent = postStopEvent;
   vmGlobals->int2str = int2str;
   vmGlobals->debug = debug;
   vmGlobals->showCommand = showCommand;
   vmGlobals->tone = tone;
   vmGlobals->lockWarpRec = lockWarpRec;
   vmGlobals->unlockWarpRec = unlockWarpRec;
#ifndef DEPRECATED_PALMOS
   vmGlobals->registerHotkeys = registerHotkeys;
   vmGlobals->unregisterHotkeys = unregisterHotkeys;
   vmGlobals->ascii2unicode = ascii2unicode;
   vmGlobals->unicode2ascii = unicode2ascii;
   vmGlobals->createStringFromUnicode = createStringFromUnicode;
#endif
   vmGlobals->showSip = showSip;
   vmGlobals->eventIsAvailable = eventIsAvailable;
   vmGlobals->pumpEvents = pumpEvents;
   // xplat_javalang.c
   vmGlobals->toLowerCase = toLowerCase;
   vmGlobals->toUpperCase = toUpperCase;
   vmGlobals->unsigned2hex = unsigned2hex;
   vmGlobals->StringToLowerCase = StringToLowerCase;
   vmGlobals->StringToUpperCase = StringToUpperCase;
   vmGlobals->StringBufferEnsureCapacity = StringBufferEnsureCapacity;
   vmGlobals->StringBufferSetLength = StringBufferSetLength;
   vmGlobals->appendString = appendString;
   vmGlobals->appendCharArray = appendCharArray;
   vmGlobals->appendCharArrayOffLen = appendCharArrayOffLen;
   vmGlobals->appendChar = appendChar;
   vmGlobals->appendInt = appendInt;
   vmGlobals->appendFloat = appendFloat;
   vmGlobals->appendDouble = appendDouble;
   vmGlobals->valueOfDouble = valueOfDouble;
   vmGlobals->StringCopyChars = StringCopyChars;
   vmGlobals->StringEquals = StringEquals;
   vmGlobals->StringCompareTo = StringCompareTo;
   vmGlobals->indexOfChar = indexOfChar;
   vmGlobals->indexOfCharStart = indexOfCharStart;
   vmGlobals->StringIndexOf = StringIndexOf;
   vmGlobals->StringHashCode = StringHashCode;
   vmGlobals->ObjectToStringNative = ObjectToStringNative;
   vmGlobals->ObjectGetClass = ObjectGetClass;
   vmGlobals->ThrowablePrintStackTraceNative = ThrowablePrintStackTraceNative;
   vmGlobals->ClassForName = ClassForName;
   vmGlobals->ClassNewInstance = ClassNewInstance;
   vmGlobals->ClassIsInstance = ClassIsInstance;
   // superwaba.c
   vmGlobals->doubleToStr = doubleToStr;
   vmGlobals->float2str = float2str;
   vmGlobals->int642str = int642str;
   vmGlobals->double2str = double2str;
   vmGlobals->str2int32 = str2int32;
   vmGlobals->str2double = str2double;
   vmGlobals->str2int64 = str2int64;
   vmGlobals->intToString = intToString;
   vmGlobals->stringToUtf16 = stringToUtf16;
   vmGlobals->dumpStackTrace = dumpStackTrace;
   vmGlobals->callMethod = callMethod;
   vmGlobals->createString = createString;
   vmGlobals->stringToUtf = stringToUtf;
   vmGlobals->gc = gc;
   vmGlobals->createArrayObject = createArrayObject;
   vmGlobals->pushObject = pushObject;
   vmGlobals->popObject = popObject;
   vmGlobals->arrayRangeCheck = arrayRangeCheck;
   vmGlobals->VmQuickError = swQuickError;
   vmGlobals->throwException = throwException;
   vmGlobals->createStringFromCharPtr = createStringFromCharPtr;
   vmGlobals->createStringObject = createStringObject;
   vmGlobals->getClassName = getClassName;
   vmGlobals->getClass = getClass;
   vmGlobals->compatible = compatible;
   vmGlobals->executeMethod = executeMethod;
   vmGlobals->getMethodByMapNum = getMethodByMapNum;
   vmGlobals->getMethodMapNum = getMethodMapNum;
   vmGlobals->printToBuf = printToBuf;
   vmGlobals->nativeLoadClass = nativeLoadClass;
   vmGlobals->loadClassField = loadClassField;
   vmGlobals->constantToVar = constantToVar;
   vmGlobals->loadClassMethod = loadClassMethod;
   vmGlobals->createUtfString = createUtfString;
   vmGlobals->getUtfString = getUtfString;
   vmGlobals->createObject = createObject;
   vmGlobals->arraySize = arraySize;
   vmGlobals->arrayType = arrayType;
   vmGlobals->createMultiArray = createMultiArray;
   vmGlobals->createStringFromUtf = createStringFromUtf;
   vmGlobals->copyArray = copyArray;
   vmGlobals->getField = getField;
   vmGlobals->getClassByIndex = getClassByIndex;
   vmGlobals->getMethod = getMethod;
   vmGlobals->countMethodParams = countMethodParams;
   vmGlobals->compatibleArray = compatibleArray;
   vmGlobals->allocObject = allocObject;
   vmGlobals->expandArray = expandArray;
   vmGlobals->arrayCount = arrayCount;
   vmGlobals->vmStatus = &vmStatus;
   vmGlobals->logDebug = log_debug;
   vmGlobals->logInfo = log_info;
   vmGlobals->logError = log_error;

   // used in Native Libraries
   setPalmDBProcs(vmGlobals);
   _LOG_DEBUG((LOG_CAT_INIT, "initVM done"));

#if HAVE_GETOPT_LONG

   int option_index = 0;
   int c;
   int iValue;

   while ((c = getopt_long(argc, argv, "Fw:h:x:y:vp:n:t:", long_options, &option_index)) != -1)
   {
      switch (c)
      {
         case 0:
            showOption(long_options[option_index].name, optarg);
            break;

         case 'F':
            vmGlobals->isFullScreen = 1;
            break;
         case 'm': // guich@420_19: now /m specify if we want to replace Ok by x in Pocket PC,
            vmGlobals->useMinimizeButton = 1;
            break;
         case 'h':
            iValue = atoi(optarg);
            if (iValue <= 0)
            {
               debug("invalid height");
               usage();
               return ERR_NoRunRequired;
            }
            vmGlobals->screenHeight = iValue;
            break;
         case 'w':
            iValue = atoi(optarg);
            if (iValue <= 0)
            {
               debug("invalid width");
               usage();
               return ERR_NoRunRequired;
            }
            vmGlobals->screenWidth = iValue;
            break;
         case 'x':
            vmGlobals->desiredX = atoi(optarg);
            break;
         case 'y':
            vmGlobals->desiredY = atoi(optarg);
            break;
         case 'p':  // path to the PDB
            swPath = optarg;
            break;
         case 'n':
            *swPdbName = optarg;  // petrus: for my tests (check that it is not transient)
            break;
         case 't':
#ifndef PALMOS
            vmGlobals->appTitle = optarg;
#endif
            break;

         case 'v':
            version();
            return ERR_NoRunRequired;

         case 'H':
         case '?':
         case ':':
            usage();
            return ERR_NoRunRequired;

         default:
            // printf ("?? getopt returned character code 0%o ??\n", c);
            return ERR_BadOption;
      }
   }

   // adjust args for the non-options common argument processing
   argv += optind;
   argc -= optind;
   arg = *argv;

#else   // !HAVE_GETOPT_LONG

   /*
   | apply the options from reading the command line
   | most of these options are not really "optional":
   | w, h, and p are required b/c they are system dependent.
   | It's up to the caller to guess and set the appropriate values
   | e.g.: setSuperWabaPath() is outside this code.
   */
   while ((--argc > 0) && (*(arg = *(++argv)) == '/'))
   {
      char c = arg[1];
      _LOG_DEBUG((LOG_CAT_INIT, "arg='%c'", c));

      switch (c)
      {
      #ifndef PALMOS
         case 'F':
            vmGlobals->isFullScreen = 1;
            break;
         case 'm': // guich@420_19: now /m specify if we want to replace Ok by x in Pocket PC,
            vmGlobals->useMinimizeButton = 1;
            break;
      #endif
			case '1': // guich@421_50: preparse the command line to see if we want 16bpp screens
	      	if (arg[2] != '6') return ERR_BadOption;
			   vmGlobals->wants16bpp = 1;
            break;
         case 'v':
            version();
            return ERR_NoRunRequired;
         case '?':
            usage();
            return ERR_NoRunRequired;
         default:
            if (--argc <= 0)
               return ERR_BadOption;
            arg = *++argv;
            switch (c)
            {
               case 'w':
                  vmGlobals->screenWidth = my_atoi(arg);
                  break;
               case 'h':
                  vmGlobals->screenHeight = my_atoi(arg);
                  break;
            #ifndef PALMOS
               case 'x':
                  vmGlobals->desiredX = my_atoi(arg);
                  break;
               case 'y':
                  vmGlobals->desiredY = my_atoi(arg);
                  break;
               case 't':
                  vmGlobals->appTitle = arg;
                  break;
            #endif
               case 'p':  // path to the PDB. ex: "\Program Files\SuperWaba\"
                  swPath = arg;
                  break;
               case 'n':
                  *swPdbName = arg;  // petrus: for my tests
                  break;
               default:
                  return ERR_BadOption;
               }
            }
      }
#endif

   if (argc > 0)  // unkeyed options...
   {
      // 1) class name
      xstrcpy(vmGlobals->className = xmalloc(1+xstrlen(arg)), arg);
#ifdef PALMOS
      // 2) the creator id
      if (--argc > 0)
      {
         arg = *++argv;
         vmGlobals->appCreatorId = READ_BE_UINT32((UInt8 *)arg);
         _LOG_DEBUG((LOG_CAT_INIT, "appCreatorId=0x%x", vmGlobals->appCreatorId));
      }
#else
      // 2) the program directory
      if (--argc > 0)
      {
         arg = *++argv;
         appPath = arg;
         // 3) the creator id
         if (--argc > 0)
         {
            arg = *++argv;
            vmGlobals->appCreatorId = READ_BE_UINT32(arg);
         }
      }
#endif
   }

   /* if anything was missing, keep on running the default app */
   if (argc <= 0)
   {
      xstrcpy(vmGlobals->className = xmalloc(sizeof(defltApp)), defltApp);
      vmGlobals->appCreatorId = SUPERWABA_CREATOR_ID;
      _LOG_DEBUG((LOG_CAT_INIT, "defaultApp=%s", vmGlobals->className));
   }
   else // guich@504_1: otherwise, the rest may be an optional commandline to be passed to the application
   {
      // get the total length
      int i,len=0;
      argc--; argv++; // skip the creator id
      for (i = 0; i < argc; i++)
         len += xstrlen(argv[i])+1;
      if (len == 0)
         vmGlobals->commandLine = NULL;
      else
      {
         // concat all strings
         arg = vmGlobals->commandLine = xmalloc(len);
         for (i = 0; i < argc; i++)
         {
            len = xstrlen(argv[i]);
            xstrncpy(arg, argv[i],len);
            arg += len;
            *arg++ = ' ';
         }
         *(arg-1) = 0; // replace last space by the \0
      }
   }
   if (!vmGlobals->appTitle)           // if appTitle wasn't defined
   {
      vmGlobals->appTitle = vmGlobals->className;
   }
   _LOG_DEBUG((LOG_CAT_INIT, "appTitle=%s", vmGlobals->appTitle));

#ifndef PALMOS

   /*
   | set the paths, all must end with a dirSep.
   | - swPath argument might already end with a dirSep
   | - appPath argument generally does not not end with a dirSep
   */
   {
      #if HAVE_GETENV
        char buf[MAX_PATH+1];
      #endif
      TCHAR tchBuf[MAX_PATH+1];
      int swPathLen;
      char dirSep;
      int dirSepValid;
      byte isAbsolute;
      int appPathLen;
      char appPathEndingChar;
      int swPathUcLen;

      if (!swPath) {
         #ifdef __SYMBIAN32__
            #ifdef __WINS__
               swPath = "./system/Apps/SuperWaba/";
            #else
               swPath = ".\\";
            #endif
         #elif HAVE_GETENV
            char * cur = getenv("HOME");
            if (!cur) {
               swPath = "./";
            }
            else
            {
               char * max = buf + sizeof buf;
               cur = boundedCopy(buf, max, cur);
               if ((cur[-1] != '\\') && (cur[-1] != '/'))
               {
                  cur = boundedCat(cur, max, "/");
               }
               cur = boundedCat(cur, max, "SuperWaba/");
               swPath = buf;
            }
         #else
            swPath = "./";
         #endif
      }

      swPathLen = xstrlen(swPath);
      dirSep = swPath[swPathLen-1];
      dirSepValid = ((dirSep == '\\') || (dirSep == '/')); // better check!
      swPathUcLen = ascii2unicode(tchBuf, swPath, (sizeof tchBuf / sizeof tchBuf[0]));

      if (!dirSepValid)
      {
         dirSep = '/';               // default directory separator
      }
      // fix the superwaba path, with a mandatory ending dirSep (add 1 extra ucs)
      vmGlobals->swPath = (TCHAR *)xmalloc((sizeof(TCHAR)) * (swPathUcLen+2));
      lstrcpy(vmGlobals->swPath, tchBuf);
      if (!dirSepValid)
      {
         vmGlobals->swPath[swPathUcLen] = (TCHAR)dirSep;
         vmGlobals->swPath[swPathUcLen+1] = 0;
         ++swPathLen;  // the multibyte length
      }

      // pre-compute the final length of the appPath
      if (!appPath)
      {
         appPath = swPath;
         appPathLen = swPathLen;
         isAbsolute = 1;
      }
      else
      {
         appPathLen = strlen(appPath);
         /*
         | if \... or /... or c:...? this is a full path
         | else we must add the SwPath in front
         | and possibly a dirsep in the between
         */
         isAbsolute = (
            (*appPath == '\\') || (*appPath == '/') ||
            (*appPath && (appPath[1] == ':'))
         );
         if (!isAbsolute) {
            appPathLen += swPathLen;
         }
      }

      // 1 extra byte for possible ending dirsep
      vmGlobals->appPath = (char *)xmalloc(appPathLen+2);
      //>>PGR check this malloc!!!

      if (isAbsolute)
      {
         xstrcpy(vmGlobals->appPath, appPath);
      }
      else
      {
         xstrncpy(vmGlobals->appPath, swPath, swPathLen);
         if (!dirSepValid) vmGlobals->appPath[swPathLen-1] = dirSep;
         xstrcpy(vmGlobals->appPath+swPathLen, appPath);
      }

      appPathEndingChar = vmGlobals->appPath[appPathLen - 1];
      if ((appPathEndingChar != '\\') && (appPathEndingChar != '/'))
      {
         vmGlobals->appPath[appPathLen] = dirSep;
         vmGlobals->appPath[appPathLen+1] = '\0';
      }
   }
#endif

   _LOG_DEBUG((LOG_CAT_INIT, "initVM end"));

   return 0;   // everything is OK!
}

/*-----------------------------------------------------------destroyVMGlobals-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void destroyVMGlobals()
{
   if (vmGlobals->className)  xfree(vmGlobals->className);
   if (vmGlobals->classHooks) xfree(vmGlobals->classHooks);
   if (vmGlobals->appPath) xfree(vmGlobals->appPath);
   if (vmGlobals->swPath) xfree(vmGlobals->swPath);
   if (vmGlobals->nativeLibs)
   {
      #if !(defined(PALMOS) && !defined(__arm__))
         int i;
         for (i=0; vmGlobals->nativeLibs[i]; ++i)
         {
            xfree(vmGlobals->nativeLibs[i]);
         }
      #endif
      xfree(vmGlobals->nativeLibs);
   }
   xfree(vmGlobals);
}

/*------------------------------------------------------------------countLibs-+
|                                                                             |
+----------------------------------------------------------------------------*/
#ifndef PALMOS
static byte countLibs(TCHAR *fileName, void *userVars)
{
   uint32 type, creator;
   CountLibs_Vars *vars = (CountLibs_Vars *)userVars;

   if (DmGetDatabaseTypeCreatorCount(fileName, &type, &creator, NULL) == PALM_SUCCESS
        && vars->type == type && vars->creator == creator)
   {
      vars->count++;
      if (vars->classDbP)
      {
         DmOpenRef dm = DmOpenDatabaseReadOnly(fileName, creator, type);
         if (dm)
            *vars->classDbP++ = dm;
      }
   }
   return 0;
}
#endif

/*-------------------------------------------------------------------startApp-+
|                                                                             |
+----------------------------------------------------------------------------*/
static WObject startApp(char * swPdbName)
{
   int i;
   int32 extG=0;
   int32 extL=0;
   DmOpenRef * classDbP;
#ifdef PALMOS
   Boolean first;
   _unused_ UInt cardNo;
   LocalID dbID;
#else
   CountLibs_Vars clv;
   TCHAR tchFullSwPdbName[MAX_PATH+1];
#endif
   int runningApp = (vmGlobals->appCreatorId != SUPERWABA_CREATOR_ID);

#if !defined(PALMOS)
   xmemzero(&clv,sizeof(clv));
#endif

   _LOG_INFO((LOG_CAT_INIT, "startApp(%s)", swPdbName));

   if (runningApp)
   {
      // guich@200b2: search for extended libraries SWAX
      // count the local extension libraries (only at the app's directory)

#if defined(PALMOS)
         //>>>PGR: FIXME!   (see nmplam_bridge.c, around line 777)
         // We should have mimiced for POSIX DmGetNextDatabaseByTypeCreator...
         // or, better, mimic a POSIX filesystem on Dm (I have this code)
         // in anycase, these ifdefs should not be there.
         first = true;
         while (dmErrCantFind != DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT,
         												vmGlobals->appCreatorId, false, PALM_CARD(&cardNo) &dbID))
         {
            ++extL;
            first = false;
         }
#else
         clv.count = 0;
         clv.type = SUPERWABA_TYPE_EXT;
         clv.creator = vmGlobals->appCreatorId;
         DmListDatabasesIn(vmGlobals->appPath, &countLibs, &clv, 0);
         extL = clv.count;
#endif
   }

   // count the global extension libraries (only at SW directory)
#if defined(PALMOS)
   first = true;

   while (dmErrCantFind != DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT,
   															SUPERWABA_CREATOR_ID, false, PALM_CARD(&cardNo) &dbID))
   {
      ++extG;
      first = false;
   }
#else
   clv.count = 0;
   clv.type = SUPERWABA_TYPE_EXT;
   clv.creator = SUPERWABA_CREATOR_ID;
   DmListDatabases(&countLibs, &clv, 0);
   extG = clv.count;
#endif

   vmGlobals->classDbList = xmalloc(sizeof(DmOpenRef) * ((runningApp?1:0)+extG+extL+2));
   if (vmGlobals->classDbList == 0)
   {
      swQuickError(ERR_CantAllocateMemory,NULL);
      return 0;
   }
   xmemzero(vmGlobals->classDbList, sizeof(DmOpenRef) * ((runningApp?1:0)+extG+extL+2));
   classDbP = (DmOpenRef *)vmGlobals->classDbList;

   // Load the SuperWaba.pdb
#if defined(PALMOS)
   *classDbP = DmOpenDatabaseByTypeCreator(SUPERWABA_TYPE_ID, SUPERWABA_CREATOR_ID,  dmModeReadOnly);
#else
   // dave@400_48 - allow to run from any dir
   i = lstrlen(vmGlobals->swPath);
   lstrcpy(tchFullSwPdbName, vmGlobals->swPath);
   ascii2unicode(tchFullSwPdbName + i, swPdbName, (sizeof tchFullSwPdbName / sizeof tchFullSwPdbName[0])-i);
   *classDbP = DmOpenDatabaseReadOnly(tchFullSwPdbName, SUPERWABA_CREATOR_ID, SUPERWABA_TYPE_ID);
   clv.classDbP = classDbP+1; // this line instructs the countLibs function to open and store the filenames
#endif

   if (*classDbP == 0)
   {
      swQuickError(ERR_CantAccessCoreClasses, NULL);
      return 0;
   }

   // Load the Application pdb
   if (runningApp)
   {
#if defined(PALMOS)
      *++classDbP = DmOpenDatabaseByTypeCreator(SUPERWABA_TYPE_ID, vmGlobals->appCreatorId, dmModeReadOnly);
      if (*classDbP == 0)
      {
         UtfString utf;
         utf.str = vmGlobals->className;
         utf.len = xstrlen(vmGlobals->className);
         swError(ERR_CantAccessAppClasses, NULL, &utf, NULL);
         return 0;
      }
#else
      clv.count = 0;
      clv.type = SUPERWABA_TYPE_ID;
      clv.creator = vmGlobals->appCreatorId;
      DmListDatabasesIn(vmGlobals->appPath, &countLibs, &clv, 0);
      if (clv.count == 0)
      {
         UtfString utf;
         utf.str = vmGlobals->appPath;
         utf.len = xstrlen(utf.str);
         swError(ERR_CantAccessAppClasses, NULL, &utf, NULL);
         return 0;
      }
#endif
   }

   // now get the extended libraries

   // 1. get the global extension libraries
   _LOG_DEBUG((LOG_CAT_INIT, "get the %d global extension libraries", (int)extG));

#if defined(PALMOS)
   first = true;
   for (i = 0; i < extG; i++)
   {
      DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT, SUPERWABA_CREATOR_ID,
      									false, PALM_CARD(&cardNo) &dbID);
      *++classDbP = DmOpenDatabase(PALM_CARD(cardNo) dbID, dmModeReadOnly);
      first = false;
   }
#else
   clv.count = 0;
   if (extG > 0)
   {
      clv.type = SUPERWABA_TYPE_EXT;
      clv.creator = SUPERWABA_CREATOR_ID;
      DmListDatabases(&countLibs, &clv, 0);
   }
#endif

   // 2. get the local extension libraries
   _LOG_DEBUG((LOG_CAT_INIT, "get the %d local extension libraries", (int)extL));

#if defined(PALMOS)
   first = true;
   for (i = 0; i < extL; i++)
   {
      DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT,
      								vmGlobals->appCreatorId, false, PALM_CARD(&cardNo) &dbID);
      *++classDbP = DmOpenDatabase(PALM_CARD(cardNo) dbID, dmModeReadOnly);
      first = false;
   }
   *++classDbP = 0;
#else
   if (extL > 0)
   {
      clv.type = SUPERWABA_TYPE_EXT;
      clv.creator = vmGlobals->appCreatorId;
      DmListDatabasesIn(vmGlobals->appPath, &countLibs, &clv, 0);
   }
   *clv.classDbP = 0;
#endif

   swInit();
   return swStartApp(vmGlobals->className);
}

/*===========================================================================*/
