/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id$ */

#if UNDER_CE >= 300 && !defined(WIN32_PLATFORM_HPC2000)
#include <stddef.h>
#endif
#include <stdio.h>
#include <string.h>
#include <tchar.h>
#include "win32/dirent.h"

#include "superwaba.h"
#include "SDL_endian.h"
#include "SW_palmDb.h"
#include "SW_palmDbPrivate.h"

// wince is 4.5x faster when using the system io, due to the FILE_FLAG_RANDOM_ACCESS flag
// this is why we have a special version not using POSIX

HANDLE _myCreateFile(TCHAR *fullPath, byte createIt, byte readOnly)
{
   HANDLE h;
   TCHAR *fp = replaceDataPath(fullPath);
   int32 err=0;

   h = CreateFile(fp,
                  readOnly?GENERIC_READ:(GENERIC_READ|GENERIC_WRITE), // font files must be open in readonly, or two instances will not be able to run
                  FILE_SHARE_READ, // important!
                  NULL,
                  createIt?OPEN_ALWAYS:OPEN_EXISTING,
                  FILE_ATTRIBUTE_NORMAL|FILE_FLAG_RANDOM_ACCESS|FILE_FLAG_WRITE_THROUGH,
                  NULL);
   if (h == INVALID_HANDLE_VALUE)
      err = GetLastError();
   return (h == INVALID_HANDLE_VALUE) ? NULL : h;
}

void _myCloseFile(HANDLE h)
{
   CloseHandle(h);
}

void _myRename(TCHAR *oldName, TCHAR *newName)
{
   MoveFile(oldName, newName);
}

void _myRemove(TCHAR *name)
{
   DeleteFile(name);
}

byte _myRead(HANDLE h, void *buf, uint32 size)
{
   DWORD read;
   return ReadFile(h, buf, size, &read, NULL) ? PALM_SUCCESS : PALM_ERROR;
}
byte _myReadAt(HANDLE h, void *buf, Uint32 size, uint32 startingAt)
{
   if (SetFilePointer(h, startingAt, NULL, FILE_BEGIN) != 0xFFFFFFFFL)
      return _myRead(h, buf, size);
   return PALM_ERROR;
}
byte _myWrite(HANDLE h, void *buf, uint32 size)
{
   DWORD written;
   return WriteFile(h, buf, size, &written, NULL) ? PALM_SUCCESS : PALM_ERROR;
}
byte _myWriteAt(HANDLE h, void *buf, Uint32 size, uint32 startingAt)
{
   if (SetFilePointer(h, startingAt, NULL, FILE_BEGIN) != 0xFFFFFFFFL)
      return _myWrite(h, buf, size);
   return PALM_ERROR;
}
uint32 _myGetFileSize(HANDLE h)
{
   uint32 size = GetFileSize(h, NULL); // should not return error
   return (size != 0xFFFFFFFFL) ? size : 0;
}
byte _myGrowFileSize(HANDLE h, int32 oldSize, int32 growSize)
{
   int32 newSize = oldSize + growSize;
   return ((SetFilePointer(h, (LONG)newSize, NULL, FILE_BEGIN) != 0xFFFFFFFFL)
       && (SetEndOfFile(h) > 0)) ? PALM_SUCCESS : PALM_ERROR;
}

void _myListDatabasesIn(TCHAR* path, HandlePDBSearchProcType proc, void *userVars, byte recursive)
{
   TCHAR fullPath[MAX_PATH];
   int const len = lstrlen(path);
   struct dirent * entry;
   DIR * targetDirectory = opendir(path);

   if (!targetDirectory) return;

   lstrcpy(fullPath, path);
   while (!ldinStopSearch && (entry = readdir(targetDirectory)))
   {
      TCHAR const * name = entry->d_name;
      int isDir = (entry->d_type & DT_DIR);
      if (isDir)
      {
         if (recursive &&          /* warning: order matters! */
            ((name[0] != '.') || ((name[1] != '\0') && ((name[1] != '.') || (name[2] != '\0')))))
         {
            lstrcpy(fullPath+len, entry->d_name);
            lstrcat(fullPath+len,TEXT("\\"));
            _myListDatabasesIn(fullPath, proc, userVars, 1);
         }
      }
      else
      { // not a dir
         if (endsWithPDB((TCHAR*)entry->d_name))
         {
            ++ldinPdbCount;
            if (proc)
            {
               lstrcpy(fullPath+len, entry->d_name);
               ldinStopSearch = (*proc)(fullPath, userVars);
            }
         }
      }
   }
   closedir(targetDirectory);
}

// date and time functions

static byte leapYear(int y)
{
   if (y % 400 == 0) return 1;
   if (y % 100 == 0) return 0;
   if (y % 4 == 0)   return 1;
   return 0;
}

static uint32 daysAccrued(int m, int y)
{
   static int daysAc[] = {0,31,59,90,120,151,181,212,243,273,304,334,365}; // dias acumulados no ano
   if (leapYear(y) && m >=2) return daysAc[m]+1;
   return daysAc[m];
}

static uint32 daysInYear(int y)
{
   return 365 + (leapYear(y)?1:0);
}

static uint32 date2days(int y, int m, int d)
{
   uint32 total= d; // start the count with the current day
   int i;
   if (m > 1) total += daysAccrued(m-1,y); // not necessary count current month neither january
   // count since last year until base year how many days have gone
   for (i = y-1; i >= 1900; i--)   // was 1904, but that's MacOS, isn't it?
      total += daysInYear(i);
   return total-1;
}

uint32 _myGetNow()
{
   // Mac date is based in # of seconds since 1904
   SYSTEMTIME st;
   uint32 timeX;

   GetLocalTime(&st);
   timeX = date2days(st.wYear, st.wMonth, st.wDay);
   timeX *= 24*60*60; // get the number of seconds
   timeX += st.wHour * 60 * 60 + st.wMinute * 60 + st.wSecond; // add the number of seconds in today
   return timeX;
}

char* getAppPath()
{
	return vmGlobals->appPath;
}

TCHAR* getSWPath()
{
	return vmGlobals->swPath;
}
