/* $Id: SW_settings.c,v 1.2 2005/09/25 12:56:34 guich Exp $ */

#include <locale.h>
#include "superwaba.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#define __DEFINE_FUNCTIONS
#include "SW_String.h"
#include "SW_utils.h"
#include "SW_settings.h"

#if HAVE_STRING_H
#include <string.h>
#endif

#ifdef WIN32
  #ifndef WINCE
    #include <time.h>
    #include <sys/timeb.h>
  #endif
#elif !defined PALMOS        // POSIX
  #include <time.h>
  #include <sys/time.h>
  #include <unistd.h>
  #include <errno.h>
  #include <sys/fcntl.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

// from superwaba.c:
WClass * getClass(UtfString className, byte throwEx);

static WClass * prefClass; // guich@230_22


#ifdef PALMOS // NOTE: BETTER TO PUT THESE DIFFERENCES IN A FILE!

static byte isHanderaSilk()
{
#ifdef SUPPORT_HANDERA_HIGHRES
   UInt32 version;
   if (FtrGet(TRGSysFtrID, TRGSilkFtrNum, &version) == 0)
      if (sysGetROMVerMajor(version) >= 1)
      {
         return 1;
      }
#endif
   return 0;
}

static void showSip(int16 on, int16 ontop)
{
   if (isHanderaSilk())
   {
      if (!on)
      {
         SilkMinimizeWindow();
         //drawMainWin(1);
      }
      else
      {
/* does not work!
         RectangleType originalAlpha, originalNum;
         SilkRestoreDefaultTemplates();
         if (ontop)
         {
            SilkGetAreas(&originalAlpha, &originalNum);
            originalAlpha.topLeft.y = 0;
            originalNum.topLeft.y = 0;
            SilkSetAreas(&originalAlpha, &originalNum);
         }
         else
         {
            //originalAlpha.topLeft.y = vmGlobals->screenHeight+16 - originalAlpha.extent.y;
            //originalNum.topLeft.y   = vmGlobals->screenHeight+16 - originalAlpha.extent.y; // same extent
         }
         */
         SilkMaximizeWindow();
      }
   }
}

static Var* getSettingsField(char *name, char *desc)
{
	UtfString fieldName = createUtfString(name);
	UtfString fieldDesc = createUtfString(desc);
	WClassField *field = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
	return &field->var.staticVar;
}

static void storeSettings(char *name, int32 crtr)
{
   WObject obj;
	uint16 len;
	UtfString s;

   obj = getSettingsField(name,"Ljava/lang/String;")->obj;
	if (obj == 0 || WOBJ_StringCharArrayObj(obj) == 0) // if string null, delete it - guich@240_3: first condition added.
	{
	   PrefSetAppPreferences (crtr, 0, 2, 0, 0, true);  // guich@241_20: now storing in saved referrences database
	   PrefSetAppPreferences (crtr, 1, 2, 0, 0, true);
	}
   else
   {
      s = vmGlobals->stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   	len = s.len;
   	PrefSetAppPreferences (crtr, 0, 2, &len, 2, true);
   	PrefSetAppPreferences (crtr, 1, 2, s.str, len, true);
   }
}

static uint32 getSecretKeyCreator(uint32 crtr)
{
   uint32 c0 = ((crtr >> 24) & 0xFF)+64;
   uint32 c1 = ((crtr >> 16) & 0xFF)+64;
   uint32 c2 = ((crtr >> 8 ) & 0xFF)+64;
   uint32 c3 = ((crtr      ) & 0xFF)+64;
   return (c0 << 24) | (c1 << 16) | (c2 << 8) | c3;
}

static void VmStoreSettings() // guich@230_22
{
   storeSettings("appSettings",vmGlobals->appCreatorId);
   storeSettings("appSecretKey",getSecretKeyCreator(vmGlobals->appCreatorId)); // guich@330_47
}

// guich@200b3: to minimize trouble, these functions are implemented here and not in nmpalm_c.c
static void VmGetSettings() // guich
{
   uint32 pref;
   int date;
   char dateSep;
   int weekStart;
   int time24h;
   char timeSep;
   char ts; // thousandsSeparator;
   char ds; // decimalSeparator;
   WObject userName;
   int32 maxColors;
   int isColor;
   WObject platform;
   int screenWidth,screenHeight;
   int32 romVersion;
   WObject appSettings;
   uint16  appSettingsLen;
   WObject appSecretKey;
   uint16 len;
   char crid[5];
   WObject appCreatorId;
   int haveVirtualKeyboard;
   uint32 secretCrtr;
   byte daylightSavings=0;
   int16 timeZone=0;
   WObject romSerialNumber=0;
   char* bufP;
   uint16 bufLen;
   uint16 showPrivateRecs; // guich@421_35

	char name[dlkUserNameBufSize];

   haveVirtualKeyboard = isHanderaSilk();
   if (haveVirtualKeyboard) showSip(0,0);

   romVersion   = vmGlobals->romVersion;
   maxColors    = vmGlobals->maxColors;
	screenWidth  = vmGlobals->screenWidth;
	screenHeight = vmGlobals->screenHeight;
	isColor      = maxColors > 16; // isColor guich@200b3: now returns the real status. In PalmOS, if <= 16 colors, then its grayScale
	platform     = (*vmGlobals->createString)("PalmOS"); // getPlatform
	userName = DlkGetSyncInfo(NULL, NULL, NULL, name, NULL, NULL)?0:(*vmGlobals->createString)(name);

   // guich@230_25: get the preferences of the Welcome app and set the accordingly flags
   len = 2;
   if (PrefGetAppPreferences('SWAB', 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences ('SWAB', 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      vmGlobals->audibleGC = buf[0] == '1';
      vmGlobals->dumpMemStats = buf[1] == '1';
      xfree(buf);
   }

   // get the preferences
   pref = PrefGetPreference(prefDateFormat);

   switch (pref)
   {
   	case dfDMYWithSlashes: date = 2; dateSep = '/'; break;  // 31/12/95
   	case dfDMYWithDots:	  date = 2; dateSep = '.'; break;  // 31.12.95
   	case dfDMYWithDashes:  date = 2; dateSep = '-'; break;  // 31-12-95
   	case dfYMDWithSlashes: date = 3; dateSep = '/'; break;  // 95/12/31
   	case dfYMDWithDots:    date = 3; dateSep = '.'; break;  // 95.12.31
   	case dfYMDWithDashes:  date = 3; dateSep = '-'; break;  // 95-12-31
   	case dfMDYWithSlashes:                                  // 12/31/95
   	default:               date = 1; dateSep = '/';  break;
   }
	weekStart = PrefGetPreference(prefWeekStartDay);
	pref = PrefGetPreference(prefTimeFormat);
	timeSep = TimeSeparator(pref);
	time24h = Use24HourFormat(pref);
	pref = PrefGetPreference(prefNumberFormat);
	switch (pref)
	{
      case nfCommaPeriod      : ts=',';  ds='.'; break;
    	case nfPeriodComma      : ts='.';  ds=','; break;
     	case nfSpaceComma       : ts=' ';  ds=','; break;
     	case nfApostrophePeriod : ts='\''; ds='.'; break;
     	case nfApostropheComma  : ts='\''; ds=','; break;
   }
   // guich@230_22: get the size then the app preferrences, if any
   len = 2;
   if (PrefGetAppPreferences(vmGlobals->appCreatorId, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (vmGlobals->appCreatorId, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSettings = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSettings = 0;
   // guich@330_47: get the size then the app secret key, if any
   secretCrtr = getSecretKeyCreator(vmGlobals->appCreatorId);
   len = 2;
   if (PrefGetAppPreferences(secretCrtr, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (secretCrtr, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSecretKey = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSecretKey = 0;

   // guich@230_22: get the creator id
   crid[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   crid[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   crid[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   crid[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   crid[4] = 0;
   appCreatorId = (*vmGlobals->createString)(crid);

   // store the preferences in the waba.sys.Settings class
   vmGlobals->prefClass = getClass(createUtfString("waba/sys/Settings"),1);

   if (vmGlobals->romVersion >= 0x04000000) // guich@340_33
   {
      daylightSavings = PrefGetPreference(prefDaylightSavingAdjustment) > 0;
      timeZone = PrefGetPreference(prefTimeZone);
      timeZone /= 60; // in hours
   }

   // guich@421_16: get the serial number
   if (vmGlobals->romVersion >= 0x03000000 && SysGetROMToken(0, sysROMTokenSnum, (UInt8**) &bufP, &bufLen) == 0 && bufP != NULL && (byte)*bufP != 0xFF)
      romSerialNumber = (*vmGlobals->createStringFromCharPtr)(bufP,bufLen);

   // guich@421_35: determine if the user has choosen to hide the private records
   showPrivateRecs = PrefGetPreference(prefShowPrivateRecords) == showPrivateRecords;

   // guich@421_17: don't call the constructor; just set directly the values
   getSettingsField("dateFormat","B")->intValue = date;
   getSettingsField("dateSeparator","C")->intValue = dateSep;
   getSettingsField("weekStart","B")->intValue = weekStart;
   getSettingsField("is24Hour","Z")->intValue = time24h;
   getSettingsField("timeSeparator","C")->intValue = timeSep;
   getSettingsField("thousandsSeparator","C")->intValue = ts;
   getSettingsField("decimalSeparator","C")->intValue = ds;
   getSettingsField("screenWidth","I")->intValue = screenWidth;
   getSettingsField("screenHeight","I")->intValue = screenHeight;
   getSettingsField("isColor","Z")->intValue = isColor;
   getSettingsField("maxColors","I")->intValue = maxColors;
   getSettingsField("onDevice","Z")->intValue = 1;
   getSettingsField("romVersion","I")->intValue = romVersion;
   getSettingsField("platform","Ljava/lang/String;")->obj = platform;
   getSettingsField("userName","Ljava/lang/String;")->obj = userName;
   getSettingsField("appSettings","Ljava/lang/String;")->obj = appSettings;
   getSettingsField("appCreatorId","Ljava/lang/String;")->obj = appCreatorId;
   getSettingsField("virtualKeyboard","Z")->intValue = haveVirtualKeyboard;
   getSettingsField("appSecretKey","Ljava/lang/String;")->obj = appSecretKey;
   getSettingsField("daylightSavings","Z")->intValue = daylightSavings;
   getSettingsField("timeZone","I")->intValue = timeZone;
   getSettingsField("romSerialNumber","Ljava/lang/String;")->obj = romSerialNumber;
   getSettingsField("showSecrets","Z")->intValue = showPrivateRecs; // guich@421_35
   getSettingsField("isHighColor","Z")->intValue = vmGlobals->screenBPP > 8; // guich@421_35

   GsiInitialize();
   GsiSetLocation(2800,2800); // guich: move to an invisible position
#ifdef PALMOS5
   if (!vmGlobals->isSony) // guich@350_28: sony palm os 5 doesnt like this command.
#endif
   GsiEnable(true);
}


#else

#define BLK_SZ					128
#define DEFAULT_DIR_PERMS	0775
#define DEFAULT_FILE_PERMS	0664

// the crid replaces the last 4 characters of the given string
// @pre buf contains a string of at least 4 chars
static TCHAR *createRegistryKey(TCHAR *buf, uint32 crid)
{
   TCHAR *cur = buf + lstrlen(buf) - 4;

   // set the key name to the creator id of the application
   *cur++ = (TCHAR)      (crid >> 24);
   *cur++ = (TCHAR)LOWER8(crid >> 16);
   *cur++ = (TCHAR)LOWER8(crid >> 8);
   *cur++ = (TCHAR)LOWER8(crid);

   return buf;
}

// frank@500_01: class member checking
static WClassField * getSettingsField(char * name, char * desc)
{
   UtfString fieldName = createUtfString(name);
   UtfString fieldDesc = createUtfString(desc);
   return vmGlobals->getField(prefClass, &fieldName, &fieldDesc);
}

static void setSettingsIntField(char * name, char * desc, int32 value)
{
   WClassField *field = getSettingsField(name, desc);
   if (field)
      field->var.staticVar.intValue = value;
}
static void setSettingsObjField(char * name, char * desc, WObject obj)
{
   WClassField *field = getSettingsField(name, desc);
   if (field)
      field->var.staticVar.obj = obj;
}

static void storeSettingsEntry(char *name, uint32 crtr)
{
   WObject obj = getSettingsField(name,"Ljava/lang/String;")->obj;
   if (obj == NULL || WOBJ_StringCharArrayObj(obj) == NULL) // if string null, delete it - guich@240_3: first condition added.
      deleteAppPreferences(crtr);
   else
      setAppPreferences(crtr, stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE));
}

////////////////////////////////////////////////////////////////////////////
// retrieve the superwaba user's preferences path
// @arg \c buf    [in] destination buffer
// @arg \c buflen [in] destination buffer length
// @return superwaba user's preferences path or 0 if failed
static char *getSuperwabaPath(char *buf, int buflen)
{
#if HAVE_GETENV
	char *max = buf + buflen;
	char *cur = boundedCopy(buf, max, getenv("HOME"));
   if (*cur != '/')
	   cur = boundedCat(cur, max, "/");
   cur = boundedCat(cur, max, ".superwaba/");
   return buf;
#else
  	return 0;
#endif
}

void deleteAppPreferences(uint32 crid)
{
   // for POSIX, we cannot use the registry.
   // we need to use a basic file system

   char buf[MAX_PATH];
	char *home = getSuperwabaPath(buf, MAX_PATH);
   if (home != 0)
   {
      char *cur = boundedCat(home, buf + MAX_PATH, "appSettings/");
      boundedCat(cur, buf+MAX_PATH, "1234");
      createRegistryKey(cur, crid);
      unlink(buf);
   }
}

static void setAppPreferences(uint32 crid, UtfString utf)
{
   // for POSIX, we cannot use the registry.
   // we need to use a basic file system

   char buf[MAX_PATH];
	char *home = getSuperwabaPath(buf, MAX_PATH);
   if (home &&
       ((access(home, W_OK) == 0) || ((errno == ENOENT) && (mkdir(home, DEFAULT_DIR_PERMS) == 0))))
   {
	  char *cur = boundedCat(home, buf + MAX_PATH, "appSettings/");
            if (
         (access(buf, W_OK) != 0) ||
         ((errno == ENOENT) && (mkdir(buf, DEFAULT_DIR_PERMS) == 0))
      ) {
         int f;
	boundedCat(cur, buf + MAX_PATH, "1234");
	createRegistryKey(cur, crid);
         f = open(buf, O_WRONLY|O_CREAT, DEFAULT_FILE_PERMS);
	if (f != -1)
	{
		write(f, utf.str, utf.len);
		close(f);
	}
}
   }
}

static char * getAppPreferences(uint32 crid)
{
   // for POSIX, we cannot use the registry.
   // we need to use a basic file system

   char *pref = 0;
   char buf[MAX_PATH];
	char *home = getSuperwabaPath(buf, MAX_PATH);
   if (home)
   {
      int f;
	char *cur = boundedCat(home, buf + MAX_PATH, "appSettings/");
	boundedCat(cur, buf + MAX_PATH, "1234");
	createRegistryKey(cur, crid);
      f = open(buf, O_RDONLY);
	if (f != -1)
	{
      char temp[BLK_SZ];
      uint32 total = 0;
      int c;
      while ((c = read(f, temp, BLK_SZ)) > 0)
      {
         uint32 t;
      	total += c;
      	t = total;
			pref = xrealloc(pref, &t);
			memcpy(pref + total - c, temp, c);
      }
		close(f);
	}
   return pref;
}

void getSettings()
{
   char tmpbuf[10];

   char ts;                 // decimal separator
   char ds;                 // thousands separator
   char dateSep = '/';      // datefmt separator
   char timeSep = ':';      // time separator
   int weekStart = 0;       // week start
   int time24h;             // time 0-24h
   int datefmt;             // date format
   int gmtBias;             // gmt+0
   int daylightSavings;     // 1 when DST is on
   int haveVirtualKeyboard;
   int32 romVersion = 0;

   WObject userName;
   WObject platform;
   WObject appSettings = 0;
   WObject appCreatorId = 0;
   WObject appSecretKey = 0;
   WObject romSerialNumber = 0;

   time_t t;
   struct tm * tm;
   struct lconv * locale;

// #ifndef NDEBUG
//   vmGlobals->audibleGC = 1;
//   vmGlobals->dumpMemStats = 1;
// #endif

#if HAVE_GETENV
   {
      char * user = getenv("USER");
      userName = createString(user ? user:"noname");
   }
#else
      userName = createString("noname");
#endif

#if defined WINCE
      platform = createString("WindowsCE");
      haveVirtualKeyboard = 1;
#elif defined WIN32
      platform = createString("Win32");
      haveVirtualKeyboard = 0;
#else
      platform = createString("POSIX");  // String platform
      haveVirtualKeyboard = 0;
#endif

   // obtain the time locale data
   locale = localeconv();
   ts = locale->thousands_sep[0];
   ds = locale->decimal_point[0];
   if (ts == '\0') ts = ',';  // so poor microsoft!
   time(&t);
   daylightSavings = localtime(&t)->tm_isdst;
   t = 30326400;   // 12/17/70 00:00:00 - GMT
   tm = localtime(&t);
   gmtBias = (tm->tm_hour * 60) + tm->tm_min;
   if ((gmtBias >= 720) && (tm->tm_mday == 17))
      gmtBias -= 1440;
   strftime(tmpbuf, sizeof tmpbuf, "%x", tm);
   if (strlen(tmpbuf) == 8)
   {
      dateSep = tmpbuf[2];
      if (tmpbuf[1] == '0')
         datefmt = 3;                // 70/12/17
      else
      if (tmpbuf[1] == '7')
         datefmt = 2;                // 17/12/70
      else
         datefmt = 1;                // 12/17/70
   }
   tm->tm_hour = 22;
   strftime(tmpbuf, sizeof tmpbuf, "%X", tm);
   if (strlen(tmpbuf) == 8)
      timeSep = tmpbuf[2];
   time24h = (tmpbuf[0] > '1');

   // guich@230_22: get the creator id
   tmpbuf[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   tmpbuf[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   tmpbuf[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   tmpbuf[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   tmpbuf[4] = 0;
   appCreatorId = (*vmGlobals->createString)(tmpbuf);

   // store the preferences in the waba.sys.Settings class
   prefClass = getClass(createUtfString("waba/sys/Settings"), 1);

   // guich@421_17: don't call the constructor; just set directly the values
   // frank@500_01: don't crash if the member doesn't exist
   setSettingsIntField("dateFormat", "B", datefmt);
   setSettingsIntField("dateSeparator", "C", dateSep);
   setSettingsIntField("weekStart", "B", weekStart);
   setSettingsIntField("is24Hour", "Z", time24h);
   setSettingsIntField("timeSeparator", "C", timeSep);
   setSettingsIntField("thousandsSeparator", "C", ts);
   setSettingsIntField("decimalSeparator", "C", ds);
   setSettingsIntField("screenWidth", "I", vmGlobals->screenWidth);
   setSettingsIntField("screenHeight", "I", vmGlobals->screenHeight);
   setSettingsIntField("isColor", "Z", (vmGlobals->maxColors == 0) || (vmGlobals->maxColors > 16));
   setSettingsIntField("maxColors", "I", vmGlobals->maxColors);
   setSettingsIntField("onDevice", "Z", 1);
   setSettingsIntField("romVersion", "I", romVersion);
   setSettingsObjField("platform", "Ljava/lang/String;", platform);
   setSettingsObjField("userName", "Ljava/lang/String;", userName);
   setSettingsObjField("appSettings", "Ljava/lang/String;", appSettings);
   setSettingsObjField("appCreatorId", "Ljava/lang/String;", appCreatorId);
   setSettingsIntField("virtualKeyboard", "Z", haveVirtualKeyboard);
   setSettingsObjField("appSecretKey", "Ljava/lang/String;", appSecretKey);
   setSettingsIntField("daylightSavings", "Z", daylightSavings);
   setSettingsIntField("timeZone", "I", gmtBias);
   setSettingsObjField("romSerialNumber", "Ljava/lang/String;", romSerialNumber);
   setSettingsIntField("isHighColor", "Z", (vmGlobals->maxColors > 256) || (vmGlobals->screenBPP > 8)); // guich@421_35
}

void storeSettings() // guich@230_22
{
   storeSettingsEntry("appSettings", vmGlobals->appCreatorId);
   storeSettingsEntry("appSecretKey", getSecretKeyCreator(vmGlobals->appCreatorId)); // guich@330_47
}

#endif
/*===========================================================================*/
