/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_settings_palm.c,v 1.13 2005/11/05 08:34:57 guich Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "superwaba.h"

#ifndef PALMOS
#include <locale.h>
#endif

#define __DEFINE_FUNCTIONS
#include "SW_String.h"
#include "SW_utils.h"
#include "SW_settings.h"

// from superwaba.c:
WClass * getClass(UtfString className, byte throwEx);

//static void deleteAppPreferences(uint32 crid);
//static void setAppPreferences(uint32 crid, UtfString utf);

static Var* getSettingsField(char *name, char *desc)
{
   UtfString fieldName = createUtfString(name);
   UtfString fieldDesc = createUtfString(desc);
   WClassField *field = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
   return &field->var.staticVar;
}

static void storeSettingsEntry(char *name, int32 crtr)
{
   WObject obj;
   uint16 len;
   UtfString s;

   obj = getSettingsField(name,"Ljava/lang/String;")->obj;
   if (obj == 0 || WOBJ_StringCharArrayObj(obj) == 0) // if string null, delete it - guich@240_3: first condition added.
   {
      PrefSetAppPreferences (crtr, 0, 2, 0, 0, true);  // guich@241_20: now storing in saved referrences database
      PrefSetAppPreferences (crtr, 1, 2, 0, 0, true);
   }
   else
   {
      s = vmGlobals->stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE);
      len = s.len;
      PrefSetAppPreferences (crtr, 0, 2, &len, 2, true);
      PrefSetAppPreferences (crtr, 1, 2, s.str, len, true);
   }
}

static uint32 getSecretKeyCreator(uint32 crtr)
{
   uint32 c0 = ((crtr >> 24) & 0xFF)+64;
   uint32 c1 = ((crtr >> 16) & 0xFF)+64;
   uint32 c2 = ((crtr >> 8 ) & 0xFF)+64;
   uint32 c3 = ((crtr      ) & 0xFF)+64;
   return (c0 << 24) | (c1 << 16) | (c2 << 8) | c3;
}

/*--------------------------------------------------------------storeSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
void storeSettings() // guich@230_22
{
   storeSettingsEntry("appSettings",vmGlobals->appCreatorId);
   storeSettingsEntry("appSecretKey",getSecretKeyCreator(vmGlobals->appCreatorId)); // guich@330_47
}

extern int use8bppEmulation;

#if 0
// guich@552_15: check if its a Treo device, to return the correct information on the rom serial number
// Treo 600 related code names
#define TreoComID 'hspr'
#define Treo600DevID 'H101'
#define Treo600SimID 'H201'
#define Treo650DevID 'H102'
#define Treo650SimID 'H202'

static int isTreo6xx()
{
   UInt32 companyID;
   UInt32 deviceID;

   FtrGet(sysFtrCreator, sysFtrNumOEMCompanyID, &companyID);
   FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);

   return (companyID == TreoComID) &&
          (deviceID == Treo600DevID || deviceID == Treo650DevID || deviceID == Treo600SimID || deviceID == Treo650SimID);
}

Err HsGetVersionString (UInt16 /*HsVerStrEnum*/selector, Char* outStrP, UInt16* sizeP) SYS_TRAP(entryNumHsGetVersionString/*hsSelGetVersionString*/);
#endif

/*----------------------------------------------------------------getSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
// guich@200b3: to minimize trouble, these functions are implemented here.
void getSettings() // guich
{
   uint32 pref;
   int date;
   char dateSep;
   int weekStart;
   int time24h;
   char timeSep;
   char ts; // thousandsSeparator;
   char ds; // decimalSeparator;
   WObject userName;
   int32 maxColors;
   int isColor;
   WObject platform;
   int screenWidth,screenHeight;
   int32 romVersion;
   WObject appSettings;
   uint16  appSettingsLen;
   WObject appSecretKey;
   uint16 len;
   char crid[5];
   WObject appCreatorId;
   int haveVirtualKeyboard=0;
   uint32 secretCrtr;
   byte daylightSavings=0;
   int16 timeZone=0;
   WObject romSerialNumber=0;
   char* bufP;
   uint16 bufLen;
   uint16 showPrivateRecs; // guich@421_35

   char name[dlkUserNameBufSize];

   romVersion   = vmGlobals->romVersion;
   maxColors    = vmGlobals->maxColors;
   if (use8bppEmulation) maxColors = 16;
   screenWidth  = vmGlobals->screenWidth;
   screenHeight = vmGlobals->screenHeight;
   isColor      = maxColors > 16; // isColor guich@200b3: now returns the real status. In PalmOS, if <= 16 colors, then its grayScale
   platform     = (*vmGlobals->createString)("PalmOS/SDL"); // getPlatform
   userName     = DlkGetSyncInfo(NULL, NULL, NULL, name, NULL, NULL) ? 0 : (*vmGlobals->createString)(name);

   // guich@230_25: get the preferences of the Welcome app and set the accordingly flags
   len = 2;
   if (PrefGetAppPreferences('SWAB', 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences ('SWAB', 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      vmGlobals->audibleGC = buf[0] == '1';
      vmGlobals->dumpMemStats = buf[1] == '1';
      xfree(buf);
   }

   // get the preferences
   pref = PrefGetPreference(prefDateFormat);

   switch (pref)
   {
      case dfDMYWithSlashes: date = 2; dateSep = '/'; break;  // 31/12/95
      case dfDMYWithDots:    date = 2; dateSep = '.'; break;  // 31.12.95
      case dfDMYWithDashes:  date = 2; dateSep = '-'; break;  // 31-12-95
      case dfYMDWithSlashes: date = 3; dateSep = '/'; break;  // 95/12/31
      case dfYMDWithDots:    date = 3; dateSep = '.'; break;  // 95.12.31
      case dfYMDWithDashes:  date = 3; dateSep = '-'; break;  // 95-12-31
      case dfMDYWithSlashes:                                  // 12/31/95
      default:               date = 1; dateSep = '/';  break;
   }
   weekStart = PrefGetPreference(prefWeekStartDay);
   pref = PrefGetPreference(prefTimeFormat);
   timeSep = TimeSeparator(pref);
   time24h = Use24HourFormat(pref);
   pref = PrefGetPreference(prefNumberFormat);
   switch (pref)
   {
      case nfCommaPeriod      : ts=',';  ds='.'; break;
      case nfPeriodComma      : ts='.';  ds=','; break;
      case nfSpaceComma       : ts=' ';  ds=','; break;
      case nfApostrophePeriod : ts='\''; ds='.'; break;
      case nfApostropheComma  : ts='\''; ds=','; break;
      default                 : ts='.';  ds=','; break;
   }
   // guich@230_22: get the size then the app preferrences, if any
   len = 2;
   if (PrefGetAppPreferences(vmGlobals->appCreatorId, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (vmGlobals->appCreatorId, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSettings = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSettings = 0;
   // guich@330_47: get the size then the app secret key, if any
   secretCrtr = getSecretKeyCreator(vmGlobals->appCreatorId);
   len = 2;
   if (PrefGetAppPreferences(secretCrtr, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (secretCrtr, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSecretKey = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSecretKey = 0;

   // guich@230_22: get the creator id
   crid[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   crid[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   crid[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   crid[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   crid[4] = 0;
   appCreatorId = (*vmGlobals->createString)(crid);

   // store the preferences in the waba.sys.Settings class
   vmGlobals->prefClass = getClass(createUtfString("waba/sys/Settings"),1);

   if (vmGlobals->romVersion >= 0x04000000) // guich@340_33
   {
      daylightSavings = PrefGetPreference(prefDaylightSavingAdjustment) > 0;
      timeZone = PrefGetPreference(prefTimeZone);
      timeZone /= 60; // in hours
   }

   // guich@552_15: added specific code to get treo's rom serial number
#if 0
   if(isTreo6xx())
   {
      char bufP2[32];
      bufP2[0] = 0;
      HsGetVersionString(1/*hsVerStrSerialNo*/, bufP2, &bufLen); // NOT YET HANDLED BY PALMOS5RE
      romSerialNumber = (*vmGlobals->createStringFromCharPtr)(bufP2,bufLen);
   }
   else
#endif
   // guich@421_16: get the serial number
   if (vmGlobals->romVersion >= 0x03000000 && SysGetROMToken(PALM_CARD(0) sysROMTokenSnum, (UInt8**) &bufP, &bufLen) == 0 && bufP != NULL && (byte)*bufP != 0xFF)
      romSerialNumber = (*vmGlobals->createStringFromCharPtr)(bufP,bufLen);

   // guich@421_35: determine if the user has choosen to hide the private records
   showPrivateRecs = PrefGetPreference(prefShowPrivateRecords) == showPrivateRecords;

   // guich@421_17: don't call the constructor; just set directly the values
   getSettingsField("dateFormat","B")->intValue = date;
   getSettingsField("dateSeparator","C")->intValue = dateSep;
   getSettingsField("weekStart","B")->intValue = weekStart;
   getSettingsField("is24Hour","Z")->intValue = time24h;
   getSettingsField("timeSeparator","C")->intValue = timeSep;
   getSettingsField("thousandsSeparator","C")->intValue = ts;
   getSettingsField("decimalSeparator","C")->intValue = ds;
   getSettingsField("screenWidth","I")->intValue = screenWidth;
   getSettingsField("screenHeight","I")->intValue = screenHeight;
   getSettingsField("isColor","Z")->intValue = isColor;
   getSettingsField("maxColors","I")->intValue = maxColors;
   getSettingsField("onDevice","Z")->intValue = 1;
   getSettingsField("romVersion","I")->intValue = romVersion;
   getSettingsField("platform","Ljava/lang/String;")->obj = platform;
   getSettingsField("userName","Ljava/lang/String;")->obj = userName;
   getSettingsField("appSettings","Ljava/lang/String;")->obj = appSettings;
   getSettingsField("appCreatorId","Ljava/lang/String;")->obj = appCreatorId;
   getSettingsField("virtualKeyboard","Z")->intValue = haveVirtualKeyboard;
   getSettingsField("appSecretKey","Ljava/lang/String;")->obj = appSecretKey;
   getSettingsField("daylightSavings","Z")->intValue = daylightSavings;
   getSettingsField("timeZone","I")->intValue = timeZone;
   getSettingsField("romSerialNumber","Ljava/lang/String;")->obj = romSerialNumber;
   getSettingsField("showSecrets","Z")->intValue = showPrivateRecs; // guich@421_35
   getSettingsField("isHighColor","Z")->intValue = vmGlobals->screenBPP > 8; // guich@421_35

#ifdef DEPRECATED_PALMOS // have to ask what this is for ?
   GsiInitialize();
   GsiSetLocation(2800,2800); // guich: move to an invisible position
	#ifdef PALMOS5
   if (!vmGlobals->isSony) // guich@350_28: sony palm os 5 doesnt like this command.
	#endif
      GsiEnable(true);
#endif // DEPRECATED_PALMOS
}

/*--------------------------------------------------------------------showSip-+
| show, hide or move the sip                                                  |
+----------------------------------------------------------------------------*/
void showSip(int16 on, int16 ontop)
{
}

/*=============================================================================*/
