/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id$ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "SW_main.h"
#include "SDL_main.h"

#if !defined(PALMOS)

/*-----------------------------------------------------------------------main-+
|                                                                             |
+----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
   StartupParams start_params;
   start_params.argc = argc;
   start_params.argv = argv;
   return startup(&start_params);
}
#else //PALMOS

#include <PalmOS.h>
#include "resources.h"

#if defined(ARM_DEBUG)
#define DISPLAY_EXIT_CODE
#endif

#ifdef USE_PEAL
#include "peal.h"
#include "dbtypes.h"

uint32_t pealLoadLibrary(const char *name, DmResType resType, DmResID resID)
{
   DmSearchStateType state;
   LocalID dbID;
   UInt16 cardNo;
   char dbName[32];
   UInt32 creatorId;
   DmOpenRef dbRef;

   Boolean first = true;
   PealModule *module = NULL;

   if (name == NULL)
   {
      module = PealLoadFromResources(NULL, resType, resID);
      ErrFatalDisplayIf(module == 0, "Failed to load ARMC/1000 resource");
      return (uint32_t)module;
   }

   while (1)
   {
      Err err = DmGetNextDatabaseByTypeCreator(first, &state, SWLibTypeID, 0, false, &cardNo, &dbID);
      if (err == dmErrCantFind)
      {
         ErrFatalDisplay("cannot find ARMC database");
         break;
      }      
      
      if (DmDatabaseInfo(cardNo, dbID, dbName, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &creatorId) == 0
                        && strncmp(dbName,name,strlen(name)) == 0)
      {
         // ok, we found the library. try to load it
 
         dbRef = DmOpenDatabase (cardNo, dbID, dmModeReadWrite);
         if (dbRef == NULL)
         {
            ErrFatalDisplay("Failed to open ARMC database");
            break;
         }

         module = PealLoadFromResources(dbRef, resType, resID);
         if (module == 0)
         {
            ErrFatalDisplay("Failed to load ARMC resource");
         }
         else
         {
            break;
         }
      }
      first = false;
   }  
   ErrFatalDisplayIf(module == 0, "Cannot load library");
   return (uint32_t)module;
}

uint32_t pealGetProcAddress(PealModule *module, char *func)
{
   return (uint32_t)PealLookupSymbol(module, func);
}

#endif//USE_PEAL

#define SAMPLE_NONE        0
#define SAMPLE_BENCH       1
#define SAMPLE_PING        2
#define SAMPLE_CONTAINER   3
#define SAMPLE_HELLOWORLD  4
#define SAMPLE_ZLIB        5
#define SAMPLE_ALLTESTS    6

#define RUN_SAMPLE         SAMPLE_NONE

#if defined(USE_PEAL)
static const char *cpuID(UInt32 processorType)
{
   const char *cpu = "unknown";
   switch(processorType)
   {
      case sysFtrNumProcessor328:        cpu = "Motorola 68328 (Dragonball)"; break;
      case sysFtrNumProcessorARM710A:    cpu = "ARM710A"; break;
      case sysFtrNumProcessorARM720T:    cpu = "ARM 720T"; break;
      case sysFtrNumProcessorARM7TDMI:   cpu = "ARM7TDMI"; break;
      case sysFtrNumProcessorARM920T:    cpu = "ARM920T"; break;
      case sysFtrNumProcessorARM922T:    cpu = "ARM922T"; break;
      case sysFtrNumProcessorARM925:     cpu = "ARM925"; break;
      case sysFtrNumProcessorEZ:         cpu = "Motorola 68EZ328 (Dragonball EZ)"; break;
      case sysFtrNumProcessorStrongARM:  cpu = "Strong ARM"; break;
      case sysFtrNumProcessorSuperVZ:    cpu = "Motorola 68SZ328 (Dragonball SuperVZ)"; break;
      case sysFtrNumProcessorVZ:         cpu = "Motorola 68VZ328 (Dragonball VZ)"; break;
      case sysFtrNumProcessorx86:        cpu = "The Palm OS ARM Simulator"; break;
      case sysFtrNumProcessorXscale:     cpu = "X-scale"; break;    
   }
   return cpu;
}
#endif

/*-----------------------------------------------------------------------main-+
|                                                                             |
+----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
   UInt32 rc = 0;
   StartupParams *start_params = NULL;
#if defined(USE_PEAL)
   PealModule *pealmod = NULL;
   void *arm_startup;
#endif
#if RUN_SAMPLE != SAMPLE_NONE
   char **ARGV = NULL;
#endif
   
#if defined(USE_PEAL)
   UInt32 processorType;

   // get the processor type
   FtrGet(sysFileCSystem, sysFtrNumProcessorID, &processorType);
    
   if (!sysFtrNumProcessorIsARM(processorType))
   {
      FrmCustomAlert(WrongCpuAlert, "\nunsupported cpu : ", cpuID(processorType), 0);
      return -1;
   }
   else
   {
      PEAL_INFO(CpuInfoAlert, cpuID(processorType), 0, 0);
   }
#endif

   start_params = MemPtrNew(sizeof(StartupParams));
   if (start_params == NULL) rc = -2;
   if (rc == 0)
   {
#if RUN_SAMPLE != SAMPLE_NONE
       ARGV = MemPtrNew(4 * sizeof(char*));
       if (ARGV == NULL) rc = -3;
       if (rc == 0)
       {
           ARGV[0] = "superwaba";           
           #if   RUN_SAMPLE == SAMPLE_BENCH
           ARGV[1] = "superwaba/samples/app/bench/Bench";
           ARGV[2] = "MGQi";
           #elif RUN_SAMPLE == SAMPLE_PING
           ARGV[1] = "superwaba/samples/ext/game/ping/Ping";
           ARGV[2] = "PiNg";
           #elif RUN_SAMPLE == SAMPLE_CONTAINER
           ARGV[1] = "superwaba/samples/ui/containerswitch/Main";
           ARGV[2] = "zMTw";
           #elif RUN_SAMPLE == SAMPLE_HELLOWORLD
           ARGV[1] = "superwaba/samples/app/helloworld/HelloWorld";
           ARGV[2] = "HeWd";
           #elif RUN_SAMPLE == SAMPLE_ZLIB
           ARGV[1] = "superwaba/samples/ext/zlib/test/ZLibTest";
           ARGV[2] = "fcOU";
           #elif RUN_SAMPLE == SAMPLE_ALLTESTS
           ARGV[1] = "superwaba/samples/ext/sql/pdbdriver/testcases/AllTests";
           ARGV[2] = "gqVX";
           #endif
           ARGV[3] = NULL;   
           start_params->argc = 3;
           start_params->argv = ARGV;
       }
#else
       start_params->argc = argc;
       start_params->argv = argv;
#endif   
   }

   if (rc == 0)
   {
      uint32_t *pealLoadLibrary68K, *pealGetProcAddress68K;
      
#if defined(USE_PEAL)
      PEAL_INFO(VmInfoAlert, "pealLoadLibrary", "ARMC", 0);
      pealmod = (PealModule*)pealLoadLibrary(NULL, 'ARMC', SWVM_CODE_RESID);
      if (pealmod == NULL) rc = -4;
      if (rc == 0)
      {
         PEAL_INFO(VmInfoAlert, "Call an ARM function that uses global data",
                                            "pealGetProcAddress", "startup");
	     arm_startup = (void *)pealGetProcAddress(pealmod, "startup");
         if (arm_startup == NULL) rc = -5;
      }
      if (rc == 0)
      {
         PEAL_INFO(VmInfoAlert, "Set the pealLoadLibrary m68k callback", 
                                "PealLookupSymbol", "pealLoadLibrary68K");
         pealLoadLibrary68K = (void *)PealLookupSymbol(pealmod, "pealLoadLibrary68K");
         if (pealLoadLibrary68K == NULL) rc = -6;
         else *pealLoadLibrary68K = Swap32(&pealLoadLibrary);
      }
      if (rc == 0)
      {    
         PEAL_INFO(VmInfoAlert, "Set the pealGetProcAddress m68k callback", 
                                "PealLookupSymbol", "pealGetProcAddress68K");
         pealGetProcAddress68K = (void *)PealLookupSymbol(pealmod, "pealGetProcAddress68K");
         if (pealGetProcAddress68K == NULL) rc = -7;
         else *pealGetProcAddress68K = Swap32(&pealGetProcAddress);
      }
      if (rc == 0)
      {
#if defined(DEBUG_PEAL)
         int i;
         char vm[256];
         char args[512]; args[0] = '\0';
         sprintf(vm, "pealCall arm_startup()\n%s with %d args", argv[0] ? argv[0]:"?", (int)argc);
         for (i = 0; i < argc; i++)
         {
             sprintf(args + strlen(args), "arg%d:%s\n", i, argv[i] ? argv[i] : "none");
         }
         PEAL_INFO(VmInfoAlert, vm, args, 0);
#endif
    	 rc = PealCall(pealmod, arm_startup, start_params);
      }
      if (pealmod != NULL) PealUnload(pealmod);
#else
      rc = startup(&start_params);
#endif
   }

   if (start_params != NULL) MemPtrFree(start_params);
#if RUN_SAMPLE != SAMPLE_NONE
   if (ARGV != NULL) MemPtrFree(ARGV);
#endif

   #if defined(DISPLAY_EXIT_CODE)
   if (rc != -1)
   {
      char buf[9 + 1];
      int i;
       
      for (i = 0; i < 8; i++)
      {
         int d = (rc >> ((8 - i - 1) * 4)) & 0x0F;
         if (d > 9) d += ('A' - 10);
         else d += '0';
         buf[i] = d;       
      }
      buf[8] = '\n';
      buf[9] = 0;
      FrmCustomAlert(ExitAlert, buf, NULL, NULL);
   }
   #endif
   
   return rc;
}

#endif//PALMOS

/*===========================================================================*/
