/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_types.h,v 1.20 2005/10/16 23:34:53 fdie Exp $ */

#ifndef SW_TYPES_H_INCLUDED
#define SW_TYPES_H_INCLUDED

#if !defined(UNDER_CE) || (UNDER_CE >= 300 && !defined(WIN32_PLATFORM_HPC2000))
#include <stddef.h>
#endif
#include "sysdefs.h"
#ifndef WIN32
#include <string.h>  // for memmove
#endif
#if defined(PALMOS)
#if defined(__arm__)
#include <PalmOSARM.h>
#include <PalmCompatibility.h>
#include <SerialMgrOld.h> // Palm OS 3.5 Support
#define PALM_CARD(x)	
#else
#define PALM_CARD(x)	x,
#define ALLOW_ACCESS_TO_INTERNALS_OF_CLIPBOARDS
#define ALLOW_ACCESS_TO_INTERNALS_OF_CONTROLS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FIELDS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FINDPARAMS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FORMS
#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_MENUS
#define ALLOW_ACCESS_TO_INTERNALS_OF_PROGRESS
#define ALLOW_ACCESS_TO_INTERNALS_OF_SCROLLBARS
#define ALLOW_ACCESS_TO_INTERNALS_OF_TABLES
#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FONTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_WINDOWS
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <SysEvtMgr.h>
#include <SerialMgr.h>
#include <SerialMgrOld.h> // Palm OS 3.5 Support
#include <DataMgr.h>
#include <NetMgr.h>
#include <FloatMgr.h>
#include <DLServer.h>
#include <VFSMgr.h>
#endif
#endif

#if defined __ECOS
 // Doug allocator is available on eCos
  #include <stdlib.h>
  #define dlmalloc malloc
  #define dlfree free
  #define dlrealloc realloc
  #define dlavailable() mallinfo().fordblks
  #define dlmalloc_usable_size malloc_usable_size
  extern void * realloc_rtnoldsize(unsigned char *,int,int *);
#else
  #undef malloc
  #define malloc  deprecated_malloc  // make sure we don't use these fellows
  #undef free
  #define free    deprecated_free
  #undef realloc
  #define realloc deprecated_realloc
  
  #if defined __SYMBIAN32__
    #include "epoc_alloc.h"
    #define dlmalloc epoc_malloc
    #define dlfree epoc_free
    #define dlrealloc epoc_realloc
    #define dlavailable epoc_alloc_available
    #define dlmalloc_usable_size epoc_malloc_usable_size
  #elif !defined PALMOS
    //>>>PGR: Why?  doug allocator works for PalmOS as well
    // fdie@ indeed, but dlmalloc do not use the palmOS useful storage heap
    #include "dlmalloc.h"      // Doug allocator
    #define dlavailable() dlmallinfo().fordblks
  #endif
#endif

#ifdef WIN32
  #include <windows.h>
  #include <tchar.h>  // dave@400_48
  #include <math.h> // dave@400_48
  #define __utimbuf _utimbuf
  #define __utime(x,y) _utime(x,y)
#else
  #define __utimbuf utimbuf
  #define __utime(x,y) utime(x,y)
  #define MAX_PATH 256
  #ifdef UNICODE
    #define TEXT(t) L ## t
    #define TCHAR wchar_t
    #define lstrlen wcslen
    #define lstrcmp wscmp
    #define lstrcpy wscpy
    #define lstrcat wscat
    #define lstrstr wsstr  //>>>PGR: i don't think this exists
  #else
    #define TEXT(t) t
    #define TCHAR char
    #define lstrlen strlen
    #define lstrcmp strcmp
    #define lstrcpy strcpy
    #define lstrcat strcat
    #define lstrstr strstr
  #endif
#endif

#define GR_ENTER
#define GR_EXIT

#ifndef max
  #define max(x,y) (((x)>=(y))?(x):(y))
  #define min(x,y) (((x)>=(y))?(y):(x))
#endif

#if defined(__GNUC__)
#define _inline_ inline
#define _inline_asm_ inline volatile 
#define _unused_  __attribute__((unused))
//#elif defined __SYMBIAN32__  // fdie@ TODO Symbian seems to remove inlined function if not used localy
//#define _inline_
#else
#define _inline_ __inline
#define _inline_asm_ __inline asm
#define _unused_ 
#endif

//
// basic types
//

#if defined WIN32 || (defined __SYMBIAN32__  && defined __WINS__)
   typedef __int64 int64;
#else
   typedef long long int64;
#endif

#if defined(PALMOS)
   typedef Int32 int32;
   typedef UInt32 uint32;
#else
   typedef int int32;
   typedef unsigned int uint32;
   #define kCoordinatesStandard  72
#endif

typedef float float32;
typedef short int16;
typedef unsigned short uint16;
typedef unsigned char uint8;
typedef uint16 char16;
typedef unsigned char byte;
typedef unsigned char uchar;

#ifndef VIDEO_BITSPERPIXEL
  #if defined PALMOS
    #define VIDEO_BITSPERPIXEL 8
  #else
    #define VIDEO_BITSPERPIXEL 32
  #endif
#endif

#if VIDEO_BITSPERPIXEL == 8
  typedef uchar Pixel;
#elif VIDEO_BITSPERPIXEL == 16
  typedef uint16 Pixel;
#else
  typedef uint32 Pixel;
#endif

#ifndef __cplusplus
#ifndef bool // bool is defined when running on win32 with c++
	typedef int bool;
#endif
#ifndef true
    #define true 1
#endif

#ifndef false
    #define false 0
#endif
#endif

#ifdef WIN32
	#define PATH_SEPARATOR	'\\'
#else
	#define PATH_SEPARATOR	'/'
#endif

//
// x portability functions
//

// NOTE: The str, mem and malloc routines aren't documented in the reference
// manuals for CE, however, the methods do exist in the CE library and have
// prototypes, etc.

#if defined(PALMOS)
  #define xstrncmp(s1, s2, n) StrNCompareAscii(s1, s2, n)
  #define xstrncpy(dst, src, n) StrNCopy(dst, src, (uint32)n)
  #define xstrcpy StrCopy
  #define xstrlen(s) StrLen(s)
  #define xstrcat(dst, src) StrCat(dst, src)
  #define xmemmove(dst, src, size) MemMove(dst, src, size)
  #undef xmemzero
  #define xmemzero(mem, len) MemSet(mem, len, (Byte)0)
  #define xmemset(mem, what, len) MemSet(mem, len, (Byte)what)
  #define xstrprintf StrPrintF
  #define xstrcaselesscmp StrCaselessCompare
  #define xstrstr StrStr
  #define xmemcmp MemCmp
  #if !defined(PALMOS) || defined(PALMOS5) || defined(__arm__) // In PalmOS 5, those functions are ARM (30x faster)
     #define xfer32(from, to, count) xmemmove(to,from,((count) << 2)) // total of size*4 bytes transfered
     #define xzero32(mem,len) xmemzero(mem,((len) << 2))
  #else
     // guich@200b4_86: ultrafast data transfer. note that the total bytes transfered are count*4
     #if __GNUC__
        static _inline_ void xfer32(byte *from, byte *to, uint16 count)
        {
         __asm__ volatile( 
            "  subq.l   #1,%2         \n"
            "0:                       \n"
            "  move.l   (%0)+,(%1)+   \n"
            "  dbf.w    %2,0b         \n" 
               : 
               : "a"(from), "a"(to), "d"(count) 
               : "memory");
        }
        static _inline_ void xzero32(byte *start, uint16 count)
        {
         __asm__ volatile( 
            "  lea.l    (%0), %%a0      \n"
            "  subq.l   #1,%1         \n"
            "1:                       \n"
            "  clr.l    (%%a0)+       \n"
            "  dbf.w    %1,1b         \n" 
               : 
               : "a"(start), "d"(count) 
               : "memory");
        }
     #else
        asm static _inline_ void xfer32(byte *from:__A0, byte *to:__A1, uint16 count: __D0)
        {
            subq.l    #1,d0
         loop:
            move.l    (a0)+,(a1)+
            dbf.w     d0,loop
        }
        asm static _inline_ void xzero32(byte *start:__A0, uint16 count: __D0)
        {
            subq.l    #1,d0
         loop:
            clr.l     (a0)+
            dbf.w     d0,loop
        }
     #endif
  #endif
#else
  #define xstrncmp strncmp
  #define xstrncpy strncpy
  #define xstrcpy strcpy
  #define xstrlen strlen
  #define xstrcat strcat
  #define xmemmove memmove
  #define xmemzero(mem, len) memset(mem, 0, len)
  #define xmemset memset
  #define xstrprintf sprintf
  // #define xstrcaselesscmp (see SW_defines.h)
  #define xstrstr strstr
  #define xmemcmp memcmp
  #define xfer32(from, to, count) xmemmove(to,from,((count) << 2)) // total of size*4 bytes transfered
  #define xzero32(mem,len) xmemzero(mem,((len) << 2))
#endif

/*
| Endiannesses
|
| On ARM, while each word of a double is stored little-endian, the most
| significant word (the word containing the sign bit) is at the lower address.
| This is neither pure big-endian (as implemented in 68k processors) nor pure
| little-endian (as implemented in x86 processors).
| Litterature describes it as "mix" or "cross" endianess.
|
| The MARM layout comes from the design of the ARM floating point coprocessor,
| which specifies that doubles are stored in memory as two 32-bit words, with
| the first word containing the top 32 bits, the second word the low 32 bits.
| EPOC uses little-endian storage for words, giving rise to the rather strange
| mixture of big-endian and little-endian layout.
*/

/* The two types of endianness */
#define SW_LIL_ENDIAN	1234
#define SW_BIG_ENDIAN	4321

/* Pardon the mess, I'm trying to determine the endianness of this host.
   I'm doing it by preprocessor defines rather than some sort of configure
   script so that application code can use this too.  The "right" way would
   be to dynamically generate this file on install, but that's a lot of work.
 */
#if  defined(__i386__) || defined(__ia64__) || defined(WIN32) || \
    (defined(__alpha__) || defined(__alpha)) || \
     defined(__arm__) || \
    (defined(__mips__) && defined(__MIPSEL__)) || \
     defined(__SYMBIAN32__) || \
     defined(__x86_64__) || \
     defined(__LITTLE_ENDIAN__)
#define SW_BYTEORDER	SW_LIL_ENDIAN
#else
#define SW_BYTEORDER	SW_BIG_ENDIAN
#endif

#define SW_WORDORDER SW_BYTEORDER // Always pure (for integers) until someone could prove the opposite.

// fdie@550 Pierre, we can't acknowledge general mixed mode on any device. I bet, you wanted to handle 
// ARM's double mixed endianness described in the little paragraph above. Well, this is handled by our 
// READ_DOUBLE macro, all other macros apply to integer values only and don't suffer mixed mode.

//#if (defined(__SYMBIAN32__) && !defined(__WINS__))
//  #define SW_WORDORDER SW_BIG_ENDIAN  // Mixed!
//#else
//  #define SW_WORDORDER SW_BYTEORDER   // Pure
//#endif

#if (defined(PALMOS) && defined(__arm__)) || \
    (defined(__SYMBIAN32__) && defined(__MARM__))
#define SW_DOUBLES_MIXED  1   // Mixed endian for doubles only
#else
#define SW_DOUBLES_MIXED  0
#endif

// align same byte endianness
static __inline uint16 sw_align16(byte * b)
{
   uint16 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   return tgt;
}

// align diff byte endianness
static __inline uint16 sw_align16_r(byte * b)
{
   uint16 tgt;
   *((byte *)&tgt+0) = b[1];
   *((byte *)&tgt+1) = b[0];
   return tgt;
}

// align same byte endianness
static __inline uint32 sw_align32(byte * b)
{
   uint32 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   *((byte *)&tgt+2) = b[2];
   *((byte *)&tgt+3) = b[3];
   return tgt;
}

// align diff byte endianness
static __inline uint32 sw_align32_r(byte * b)
{
   uint32 tgt;
   *((byte *)&tgt+0) = b[3];
   *((byte *)&tgt+1) = b[2];
   *((byte *)&tgt+2) = b[1];
   *((byte *)&tgt+3) = b[0];
   return tgt;
}

// align same byte endianness, same word-endianness (pure endiannesses)
static __inline int64 sw_align64(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[0];
   *((byte *)&tgt+1) = b[1];
   *((byte *)&tgt+2) = b[2];
   *((byte *)&tgt+3) = b[3];
   *((byte *)&tgt+4) = b[4];
   *((byte *)&tgt+5) = b[5];
   *((byte *)&tgt+6) = b[6];
   *((byte *)&tgt+7) = b[7];
   return tgt;
}

// align diff byte endianness, diff word-endianness (pure endiannesses)
static __inline int64 sw_align64_r(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[7];
   *((byte *)&tgt+1) = b[6];
   *((byte *)&tgt+2) = b[5];
   *((byte *)&tgt+3) = b[4];
   *((byte *)&tgt+4) = b[3];
   *((byte *)&tgt+5) = b[2];
   *((byte *)&tgt+6) = b[1];
   *((byte *)&tgt+7) = b[0];
   return tgt;
}

// align same byte endianness, diff word-endianness (mix endiannesses)
static __inline int64 sw_align64_w(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[4];
   *((byte *)&tgt+1) = b[5];
   *((byte *)&tgt+2) = b[6];
   *((byte *)&tgt+3) = b[7];
   *((byte *)&tgt+4) = b[0];
   *((byte *)&tgt+5) = b[1];
   *((byte *)&tgt+6) = b[2];
   *((byte *)&tgt+7) = b[3];
   return tgt;
}

// align diff byte endianness, same word-endianness (mix endiannesses)
static __inline int64 sw_align64_b(byte * b)
{
   int64 tgt;
   *((byte *)&tgt+0) = b[3];
   *((byte *)&tgt+1) = b[2];
   *((byte *)&tgt+2) = b[1];
   *((byte *)&tgt+3) = b[0];
   *((byte *)&tgt+4) = b[7];
   *((byte *)&tgt+5) = b[6];
   *((byte *)&tgt+6) = b[5];
   *((byte *)&tgt+7) = b[4];
   return tgt;
}

/*
| Read a numeric with the specified endianness
*/
#if SW_BYTEORDER == SW_LIL_ENDIAN
  #define READ_LE16(b)       sw_align16     ((byte*)b)
  #define READ_BE16(b)       sw_align16_r   ((byte*)b)
  #define READ_LE32(b)       sw_align32     ((byte*)b)
  #define READ_BE32(b)       sw_align32_r   ((byte*)b)
  #if SW_WORDORDER == SW_LIL_ENDIAN
    #define READ_LE64(b)     sw_align64     ((byte*)b)
    #define READ_LEMix64(b)  sw_align64_w   ((byte*)b)
    #define READ_BE64(b)     sw_align64_r   ((byte*)b)
    #define READ_BEMix64(b)  sw_align64_b   ((byte*)b)
  #else
    #define READ_LE64(b)     sw_align64_w   ((byte*)b)
    #define READ_LEMix64(b)  sw_align64     ((byte*)b)
    #define READ_BE64(b)     sw_align64_b   ((byte*)b)
    #define READ_BEMix64(b)  sw_align64_r   ((byte*)b)
  #endif
#else
  #define READ_BE16(b)       sw_align16     ((byte*)b)
  #define READ_LE16(b)       sw_align16_r   ((byte*)b)
  #define READ_BE32(b)       sw_align32     ((byte*)b)
  #define READ_LE32(b)       sw_align32_r   ((byte*)b)
  #if SW_WORDORDER != SW_LIL_ENDIAN
    #define READ_BE64(b)     sw_align64     ((byte*)b)
    #define READ_BEMix64(b)  sw_align64_w   ((byte*)b)
    #define READ_LE64(b)     sw_align64_r   ((byte*)b)
    #define READ_LEMix64(b)  sw_align64_b   ((byte*)b)
  #else
    #define READ_BE64(b)     sw_align64_w   ((byte*)b)
    #define READ_BEMix64(b)  sw_align64     ((byte*)b)
    #define READ_LE64(b)     sw_align64_b   ((byte*)b)
    #define READ_LEMix64(b)  sw_align64_r   ((byte*)b)
  #endif
#endif

#if SW_DOUBLES_MIXED // Mixed!
    #define READ_DOUBLE(b)  READ_BEMix64(b)
#else    
    #define READ_DOUBLE(b)  READ_BE64(b)
#endif

#define READ_LE_UINT16(b)              READ_LE16(b)
#define READ_BE_UINT16(b)              READ_BE16(b)
#define READ_LE_UINT32(b)              READ_LE32(b)
#define READ_BE_UINT32(b)              READ_BE32(b)
#define READ_LE_UINT64(b)      (uint64)READ_LE64(b)
#define READ_LEMix_UINT64(b)   (uint64)READ_LEMix64(b)
#define READ_BE_UINT64(b)      (uint64)READ_BE64(b)
#define READ_BEMix_UINT64(b)   (uint64)READ_BEMix64(b)

#define READ_LE_INT16(b)        (int16)READ_LE16(b)
#define READ_BE_INT16(b)        (int16)READ_BE16(b)
#define READ_LE_INT32(b)        (int32)READ_LE32(b)
#define READ_BE_INT32(b)        (int32)READ_BE32(b)
#define READ_LE_INT64(b)               READ_LE64(b)
#define READ_LEMix_INT64(b)            READ_LEMix64(b)
#define READ_BE_INT64(b)               READ_BE64(b)
#define READ_BE_DOUBLE(b)              READ_DOUBLE(b)
#define READ_BEMix_INT64(b)            READ_BEMix64(b)

#define Swap16(n) (((((unsigned int) n) << 8) & 0xFF00) | \
                   ((((unsigned int) n) >> 8) & 0x00FF))

#define Swap32(n) (((((unsigned long) n) << 24) & 0xFF000000) |   \
                   ((((unsigned long) n) <<  8) & 0x00FF0000) |   \
                   ((((unsigned long) n) >>  8) & 0x0000FF00) |   \
                   ((((unsigned long) n) >> 24) & 0x000000FF))

#endif
/*===========================================================================*/
