/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

// Before including superwaba.h, define BUILDING_CUSTOM_LIB so that when
// we make calls to the library's routines from this file, we generate
// the correct dll import/export symbols.

/********************************************************************
 * Private variables
 ********************************************************************/

// This array is used to map a hash value to a corresponding native function.
// It must remain sorted by hash value because a binary search is performed
// to find a method by its hash value. There is a small chance of collision
// if two hashes match and if one occurs, the function name should be changed
// to avoid collision. To prevent users from creating invalid native methods
// that hash to a valid value, native methods could be restricted to a specific
// set of classes (this is probably not necessary since any verifier probably
// wouldn't allow native methods to get by anyway).
// guich: since i introduced a hash table, this table dont has anymore to stay sorted. :-))

// guich@102 - implements a hash table, whose key is a int32 and value can be anything.
// ps: HASHSIZE is computed in the NativeHashes program.

// IMPORTATN!!! THIS IS THE WINCE SWLIB.C

struct htElem
{
   uint32 key;
   NativeFunc value;
   struct htElem *next;
};

static struct htElem *hashTable[HASHSIZE];         /* The storage array for the methods */

struct VMGlobals *vmGlobals; // setted by the SWInvokeNative

#ifdef WIN32
#define appErrorClass 0
#endif
/********************************************************************
 * Private functions
 ********************************************************************/
#define htHash(key) (unsigned)((key) % HASHSIZE)

/* return the htElem found, else NULL if key doesn't exist */
static struct htElem *htGet(uint32 key)
{
   struct htElem *np;
   for (np = hashTable[htHash(key)]; np; np = np->next)
      if (np->key == key)
         return np;
   return NULL;
}
/* return struct containing value or NULL if unable to store */
static struct htElem *htPut(uint32 key, NativeFunc value)
{
   struct htElem *np;
   unsigned hashval;
   if ((np = htGet(key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElem *) (*vmGlobals->xmalloc) (sizeof(struct htElem));
      if (np != NULL)
      {
          hashval = htHash(key);
          /* point to last w/ this hash */
          np->next = hashTable[hashval];
          np->key = key;
          np->value = value;
          /* attach to hashtab array */
          hashTable[hashval] = np;
      }
   }
   else
      (*vmGlobals->int2str)("**** duplicated key in hashtable ****",key);
   return np;
}
/*************************************************************/
static void disposeHashtable()
{
   // dispose the hashtable
   int i;
   for (i = 0; i < HASHSIZE; i++)
   {
      struct htElem *np;
      np = hashTable[i];
      while (np)
      {
         struct htElem *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      hashTable[i] = NULL;
   }
}
/*************************************************************/
Err SWLibOpen(uint16 inRefnum, void *inGlobalsPtr)
{
   ClassHook *libClassHooks;
   int i;
   // 1. associate the vmGlobals struct with the given one
   vmGlobals = (struct VMGlobals *)inGlobalsPtr;
   // 2. install our hooks
   libClassHooks = getClassHooks();
   if (libClassHooks)
   {
      int hookLibCount = (*vmGlobals->arrayCount)((char *)libClassHooks,sizeof(ClassHook));
      int insertIndex;
      vmGlobals->classHooks = (ClassHook *)(*vmGlobals->expandArray)((char *)vmGlobals->classHooks,sizeof(ClassHook),hookLibCount,&insertIndex);
      if (insertIndex == -1) return SWErrMemory;
      for (i =0; i < hookLibCount; i++)
         libClassHooks[i].libRefnum = inRefnum;

      xmemmove (vmGlobals->classHooks+insertIndex,libClassHooks, hookLibCount*sizeof(ClassHook));
   }
   buildNativeFuncArray();
   for (i=0; nativeHashes[i]; i++)
      htPut(nativeHashes[i], nativeFuncs[i]);
	return 0;
}
/*************************************************************/
Err SWLibClose(uint16 inRefnum)
{
   onClose();
   disposeHashtable();
   return 0;
}
/*************************************************************/
Err SWGetNativeFunc(uint16 inRefnum, int32 hash, NativeFunc *func)
{
   Err err = 0;
   struct htElem *elem = htGet(hash);
   if (elem && elem->value)
      *func = elem->value;
   else
      err = SWErrMethodNotFound;
   return err;
}
/*************************************************************/
Err SWInvokeNativeFunc(uint16 inRefnum, NativeFunc func, Var stack[])
{
   func(stack);
   return 0;
}
/*************************************************************/
Err SWInvokeNative(uint16 inRefnum, int32 hash, Var stack[])
{
   Err err = 0;
   struct htElem *elem = htGet(hash);
   if (elem && elem->value)
      elem->value(stack);
   else
   {
   	(*vmGlobals->debug)((*vmGlobals->int2str)("SWInvokeNative coud'nt find method with hash ",hash));
      err = SWErrMethodNotFound;
   }
   return err;
}
/*************************************************************/
Err SWCallPublic(uint16 inRefnum, int index, uint32 *ret, uint32 *args)
{
   return callPublic(index, ret, args);
}
/*************************************************************/
Err SWHandleEvent(uint16 inRefnum, EventPtr eventP)
{
   return handleEvent(eventP);
}
