/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 1998 Ken Krugler                                               *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#ifndef __SW_LIB_H__
#define __SW_LIB_H__


#define EMULATION_LEVEL		EMULATION_NONE		// Force this to no emulation:

// If this header is being used by the library code itself, then
// BUILDING_SW_LIB will be defined, and thus our routine declarations
// will be just that, as opposed to both a declaration and a definition
// (the SYS_TRAP macro generates inline code for trap dispatching).

#ifdef BUILDING_SW_LIB
	#define SW_LIB_TRAP(trapNum)
#else
	#define SW_LIB_TRAP(trapNum) SYS_TRAP(trapNum)
#endif

/********************************************************************
 * Library Specific Constants
 ********************************************************************/

#define		SWLibTypeID		'libr'		// Standard library database type

#define PUBLIC_subLaunched 987  // Calls this library when a sublaunch occured
                                // with the subscribed launch code (use
                                // (*vmGlobals->registerLaunchCode) to register and
                                // (*vmGlobals->unregisterLaunchCode) to unregister.
                                // This is exclusive for Palm OS.

#if defined(BUILDING_SW_LIB)
#include "CustomLib.h"
#endif

// Custom routine traps defined for SW Library. Each library call,
// including the standard set, gets a trap number. The constant sysLibTrapCustom
// is the first available library trap number after the system's set of four
// predefined library routines (Open, Close, Sleep, Wake).

#if defined(PALMOS) && defined( __GNUC__)
/**
 * The palmdev sdk SYS_CALL macro does not support enums for trap numbers.
 * If enums are used, you will get "bad callseq" errors.
 * see http://www.mail-archive.com/palm-dev-forum@3com.com/msg13273.html
 */
   #define  SWLibTrapSWLibGetLibAPIVersion   (sysLibTrapCustom+0)
   #define  SWLibTrapSWInvokeNative          (sysLibTrapCustom+1)
   #define  SWLibTrapSWCallPublic            (sysLibTrapCustom+2)
   #define  SWLibTrapSWHandleEvent           (sysLibTrapCustom+3)
   #define  SWLibTrapSWGetNativeFunc         (sysLibTrapCustom+4)
   #define  SWLibTrapSWInvokeNativeFunc      (sysLibTrapCustom+5)
   #define  SWLibTrapLast                    (sysLibTrapCustom+6)
#else
typedef enum
{
	SWLibTrapSWLibGetLibAPIVersion = sysLibTrapCustom,
	SWLibTrapSWInvokeNative,
	SWLibTrapSWCallPublic,
	SWLibTrapSWHandleEvent,
	SWLibTrapSWGetNativeFunc,
	SWLibTrapSWInvokeNativeFunc,
	SWLibTrapLast
} SWLibTrapNumbers;
#endif

// Our library reference number
extern uint16 libRefNum;

/********************************************************************
 * Public functions
 ********************************************************************/

// Standard library open, close, sleep and wake functions

Err SWLibOpen(uint16 inRefnum, void *inGlobalsPtr)    SW_LIB_TRAP(sysLibTrapOpen);
Err SWLibClose(uint16 inRefnum)                       SW_LIB_TRAP(sysLibTrapClose);
Err SWLibSleep(uint16 inRefnum)                       SW_LIB_TRAP(sysLibTrapSleep);
Err SWLibWake(uint16 inRefnum)                        SW_LIB_TRAP(sysLibTrapWake);

// Custom library functions

// Get our library API version
Err SWLibGetLibAPIVersion(uint16 inRefnum, DWordPtr outVersion)            SW_LIB_TRAP(SWLibTrapSWLibGetLibAPIVersion);
// Invoke the native method
Err SWInvokeNative (uint16 inRefnum, int32 hash, Var stack[])              SW_LIB_TRAP(SWLibTrapSWInvokeNative);
// Invoke a public method
Err SWCallPublic (uint16 inRefnum, int index, uint32 *ret, uint32 *args)   SW_LIB_TRAP(SWLibTrapSWCallPublic);
// Let the native lib handle events if it want so
Err SWHandleEvent (uint16 inRefnum, EventPtr eventP)                       SW_LIB_TRAP(SWLibTrapSWHandleEvent);
// Gets the NativeFunc given the hash value
Err SWGetNativeFunc(uint16 uRefNum, int32 hash, NativeFunc *func)          SW_LIB_TRAP(SWLibTrapSWGetNativeFunc);
// Invoke the native method given the native func (faster)
Err SWInvokeNativeFunc(UInt uRefNum, NativeFunc func, Var stack[])       SW_LIB_TRAP(SWLibTrapSWInvokeNativeFunc);

#endif	// __SW_LIB_H__
