/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: dirent.c,v 1.3 2004/12/18 22:21:18 fdie Exp $ */

/* dirent functions for windoze and CE */

#include <string.h>
#include "dirent.h"

#ifdef WINCE
 #define FIND_DATA WIN32_FIND_DATA
 #define FIND_DATA_HANDLE void *
 #define FIND_DATA_INVALID_HANDLE_VALUE (void *)-1
 #define NAME_IN_FDATA(fdata) (fdata).cFileName
 #define _tfindfirst FindFirstFile
 #define _tfindnext FindNextFile
 #define _findclose FindClose
 #define dlmalloc malloc
 #define dlfree free
#else
 #define FIND_DATA struct _tfinddata_t
 #define FIND_DATA_HANDLE long
 #define FIND_DATA_INVALID_HANDLE_VALUE -1
 #define NAME_IN_FDATA(fdata) (fdata).name
 #include <io.h>
 #include <direct.h>
 #include "../dlmalloc.h"   // Doug allocator
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define SUFFIX "*"
#define SLASH  "\\"

struct __dirstream           // typedef'ed as DIR
{
   FIND_DATA fdata;
   struct dirent dir;
   FIND_DATA_HANDLE handle;
   short no;
   TCHAR name[1];
};

/*--------------------------------------------------------------------opendir-+
|                                                                             |
+----------------------------------------------------------------------------*/
DIR * opendir(TCHAR const * path)
{
   DIR * dirp;
   int len;
   unsigned int rc;

   if ((!path) || (path[0] == 0) || (rc = GetFileAttributes(path), (rc == -1)) ||
       ((rc & FILE_ATTRIBUTE_DIRECTORY) == 0) ||
       ((dirp = (DIR *)dlmalloc(sizeof(DIR) + (sizeof(TCHAR)*(lstrlen(path) + 2)))) == 0)) 
   {
      return (DIR *)0;
   }
   lstrcpy(dirp->name, path);
   len = lstrlen(path) - 1;
   if ((dirp->name[0] != 0) && (dirp->name[len] != '\\') && (dirp->name[len] != '/')) 
      dirp->name[++len] = '\\';
   dirp->name[++len] = '*';
   dirp->name[++len] = 0;
   dirp->handle = FIND_DATA_INVALID_HANDLE_VALUE;
   dirp->no = 0;
   dirp->dir.d_ino = 0;
   dirp->dir.d_reclen = 0;
   dirp->dir.d_namlen = 0;
   dirp->dir.d_name = NAME_IN_FDATA(dirp->fdata);
   return dirp;
}

/*--------------------------------------------------------------------readdir-+
|                                                                             |
+----------------------------------------------------------------------------*/
struct dirent * readdir(DIR * dirp)
{
   if ((dirp) && (dirp->dir.d_name == NAME_IN_FDATA(dirp->fdata)) && (dirp->no >= 0)) 
   {
      if (dirp->no == 0) 
      {
         dirp->handle = _tfindfirst(dirp->name, &(dirp->fdata));
         if (dirp->handle == FIND_DATA_INVALID_HANDLE_VALUE) 
            dirp->no = -1;
         else 
            dirp->no = 1;
      }
      else 
      {
         long err;
#ifdef WINCE         
   		if ((err=_tfindnext(dirp->handle, &(dirp->fdata))) == 0)  // guich: without == 0, wince doesnt works!
#else				   
			if ((err=_tfindnext(dirp->handle, &(dirp->fdata)))) 
#endif				   
         {
            _findclose(dirp->handle);
            dirp->handle = FIND_DATA_INVALID_HANDLE_VALUE;
            dirp->no = -1;
         }
         else 
         {
            ++dirp->no;
         }
      }
      if (dirp->no > 0) 
      {
         dirp->dir.d_namlen = lstrlen(dirp->dir.d_name);
#ifdef WINCE
         if ((dirp->fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
#else
         if ((dirp->fdata.attrib & FILE_ATTRIBUTE_DIRECTORY) != 0)
#endif
            dirp->dir.d_type = DT_DIR;
         else  
            dirp->dir.d_type = DT_REG;
         return &dirp->dir;
      }
   }
   return (struct dirent *)0;
}

/*-------------------------------------------------------------------closedir-+
|                                                                             |
+----------------------------------------------------------------------------*/
int closedir(DIR * dirp)
{
   if (!dirp) 
      return -1;

   if (dirp->handle != FIND_DATA_INVALID_HANDLE_VALUE) 
      _findclose(dirp->handle);
   dlfree(dirp);
   return 0;
}

/*===========================================================================*/
