/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#define BUILDING_SW_STATIC
#include "superwaba.h"

#ifndef max
#define max(x,y) (((x)>=(y))?(x):(y))
#define min(x,y) (((x)>=(y))?(y):(x))
#endif
#define abs(x) (((x)>0)?(x):-(x))
#define sgn(x) (((x)<0)?-1:(((x)>0)?1:0))

////////////  Primitives that must be implemented in the ports, if possible in Assembler ///////////
typedef int32 (*GetColor)       (int r, int g, int b);
typedef void  (*GetBitmapRow)   (GrHandle h, int x, int y, int w, Pixel *row);
typedef void  (*DrawBitmap)     (GrHandle h, int x, int y, int width, int height, int dstX, int dstY, int drawOp, int32 backColor, int32 foreColor, int doClip);
typedef void  (*DrawScreen)     (GrHandle src, GrHandle dst, int ySrc, int yDst, int height);

// translate and clip. eg: x += TRANSX; y += TRANSY; if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= y && y < CLIPY2) ...
typedef void  (*DrawVLine)      (int x, int y, int h, int32 color);
typedef void  (*DrawHLine)      (int x, int y, int w, int32 color);
typedef void  (*SetPixel_)       (int x, int y, int32 color);
typedef void  (*SetPixelNOCT)   (int x, int y, int32 color);
typedef int32 (*GetPixel_)       (int x, int y);
// NOCT: without Translation and Clipping check
#ifndef WIN32
typedef void  (*InvPixelNOCT)   (int x, int y);
typedef void  (*InvVLineNOCT)   (int x, int y, int h);
typedef void  (*EraseVLineNOCT) (int x, int y, int h, int32 cfrom, int32 cto);
typedef void  (*FillRectNOCT)   (int x, int y, int w, int h, int32 color);
#endif

int scaleImage(GrHandle srcG, uint16 width, uint16 height, GrHandle dstG, uint16 newWidth, uint16 newHeight); // guich@420_14
void drawHourGlass(int step);
void initMainGrHandle(); // used in WinCE
void graphicsSwap(int on);
void graphicsStop();
void graphicsStart();
byte graphicsInit(); // called by the vm to init the graphics
void graphicsSetup();
byte loadAllFonts();
void releaseAllFonts();
void graphicsReset();

void freeGrResources(); // guich@200b4
byte intersects(RectangleType *r1, RectangleType *r2);
#define makeRect(xx, yy, ww, hh, r) {r.topLeft.x = xx; r.topLeft.y = yy; r.extent.x = ww; r.extent.y = hh;}
////////////////////////////////////////////////////////////////////////////
void intersectWith(RectangleType *r1, RectangleType *r2);
byte translateAndClip(int *x, int *y, int *width, int *height);
GrHandle CreateOffscreenWindow(WObject obj, int width, int height, int allocWinMem, uint16 *error);
void SetDrawWindow(GrHandle h);
void drawRow(int xStart, int xEnd, int y, unsigned char * rgbOrIndex, uint32 * colorMap);
void graphicsStartImage(uint32 * ret, uint32 * args);
void graphicsEndImage(GrHandle oldG);

WObject lastGraphics=NULL;
GrHandle lastGrHandle=NULL;
GrHandle oldGrHandle; // guich@402_18
byte *activeScr;
int32 transXY; // saves the translation coords x in the upper 16 bits and the y in the lower 16 bits
int32 clipX,clipY; // saves the clip coords. the lower 16 bits have x1/y1 and the upper 16 bits have x2/y2
int maxX,maxY;
int widthInBPP;
uint32 displayBPP; // bits per pixel for the display
GrHandle mainGrHandle,activeGrHandle;
int32 currentForeColor;
int32 currentBackColor;
struct UserFont *currentUserFont;
int MaxScreenX;
int MaxScreenY;
int32 *palette; // guich@421_8: alloc from heap from now on to shrink prc.
Pixel imgBackColor; // guich@402_32

int isColor; // true if the display supports color
uint32 maxColors; // ps: 0 means 2^32

DrawVLine drawVLine;
DrawHLine drawHLine;
SetPixel_ setPixel;
GetPixel_ getPixel;
SetPixel_ setPixelXM;
SetPixelNOCT setPixelNOCT;
#ifndef WIN32
InvPixelNOCT invPixelNOCT;
InvVLineNOCT invVLineNOCT;
EraseVLineNOCT eraseVLineNOCT;
FillRectNOCT fillRectNOCT;
#endif
DrawHLine drawHLineNOCT;
GetColor getColor;
GetBitmapRow getBitmapRow;
DrawBitmap drawBitmap;
DrawScreen drawScreen;

int32 *systemPalette; // websafe palette - guich@421_20: created by algorithm

//////////////////////////// INCLUDES
#ifdef PALMOS5
 #include "nmpalm5_graphics.c"
#endif
#if defined(PALMOS) && !defined(PALMOS5)
 #include "nmpalm_graphics.c"
#endif
#ifdef WIN32
 #include "ce\nmwin32_graphics.c"
#endif

#include "native_methods.h"
/////////////////////////////////////

//////  prototypes  //////
void GraphicsDestroy(WObject obj);
static void preArcPiePointDrawAndFill(int xc, int yc, int rx, int ry, float startAngle, float endAngle, int32 c, int32 c2, byte fill, byte pie);

static int32* createWebSafePalette() // guich@421_20
{
   int32 palAux[] =
   {
      0x111111,0x222222,0x444444,0x555555,0x777777,0x888888,0xAAAAAA,0xBBBBBB,0xDDDDDD,0xEEEEEE,
      0xC0C0C0,0x800000,0x800080,0x008000,0x008080
   };
   int32 i=0,r,g,b;
   int32 *newpal = vmGlobals->xmalloc(1024);
   int32 samples[] = {0x00, 0x33, 0x66, 0x99, 0xcc, 0xff};
   int32 samples8[] = {0x0000, 0x3300, 0x6600, 0x9900, 0xcc00, 0xff00};
   int32 samples16[] = {0x000000, 0x330000, 0x660000, 0x990000, 0xcc0000, 0xff0000};
   int32 steps[] = {120,114,108,12,6,0};
   for (r = 5; r >=0; r--,i += 12)
      for (g = 5; g >=0; g--,i++)
         for (b = 5; b >= 0; b--)
            newpal[i+steps[b]] = samples16[r] | samples8[g] | samples[b];
   xmemmove(&newpal[215],palAux,15*4);
   xmemzero(&newpal[230],25*4);
   newpal[255] = 0x060003;
   return newpal;
}

////  Methods that must be implemented by the Platforms ////

byte graphicsInit() // called by the vm to init the graphics
{
   systemPalette = createWebSafePalette();
   transXY=MAKE32(0,0); // saves the translation coords x in the upper 16 bits and the y in the lower 16 bits
   graphicsSetup(); // call the port routines
   registerPrimitives(NULL);
   if (!loadAllFonts())
      return 0;
   graphicsStart();
   return 1;
}

static int SurfaceGetType(WObject surface)
{
   static WClass *imageClass = 0;
	// cache class pointers for performance
	if (!imageClass)
		imageClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/fx/Image"),1);
	return (surface != NULL && (*vmGlobals->compatible)(WOBJ_class(surface), imageClass)) ? SURF_IMAGE : SURF_CONTROL;
}

////////////////////////////////////////////////////////////////////////////
// ellipse, arc, pie and circle functions
#define drawEllipse(x,y,rx,ry,c) ellipseDrawAndFill(x,y,rx,ry,c,0)
#define fillEllipse(x,y,rx,ry,c) ellipseDrawAndFill(x,y,rx,ry,c,1)
#define drawCircle(xc, yc,r, c)  ellipseDrawAndFill(xc,yc,r,r,c,0)
#define fillCircle(xc, yc,r, c)  ellipseDrawAndFill(xc,yc,r,r,c,1)

#define drawArc(xc,yc,r,startAngle,endAngle,c) preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,c,c,0,0)
#define drawPie(xc,yc,r,startAngle,endAngle,c) preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,c,c,0,1)
#define fillPie(xc,yc,r,startAngle,endAngle,c,c2) preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,c,c2,1,1)
#define drawEllipticalArc(xc,yc,rx,ry,startAngle,endAngle,c) preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,c,0,0,0)
#define drawEllipticalPie(xc,yc,rx,ry,startAngle,endAngle,c) preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,c,0,0,1) // guich@402_57
#define fillEllipticalPie(xc,yc,rx,ry,startAngle,endAngle,c,c2) preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,c,c2,1,1) // guich@402_57
////////////////////////////////////////////////////////////////////////////
// Bresenham algorithm to draw lines (modified by guich to improve performance in vertical and horizontal lines)
static void drawLine(int Ax, int Ay, int Bx, int By, int32 c)
{
    // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE SLOPE OR DIRECTION OF THE LINE
    int dX = abs(Bx-Ax);    // store the change in X and Y of the line endpoints
    int dY = abs(By-Ay);

    int CurrentX = Ax;      // store the starting point (just point A)
    int CurrentY = Ay;

    // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
    int Xincr, Yincr;

    // guich: if its a pixel, draw only the pixel
    if (!dX && !dY)
       (*setPixel)(Ax,Ay,c); else
    // guich: if they are vertical or horizontal lines, use specific methods
    if (!dY) // horizontal line?
       (*drawHLine)(min(Ax,Bx),min(Ay,By),dX+1,c); else
    if (!dX) // vertical line?
       (*drawVLine)(min(Ax,Bx),min(Ay,By),dY+1,c);
    else // diagonal
    {
        if (Ax > Bx) { Xincr=-1; } else { Xincr=1; }    // which direction in X?
        if (Ay > By) { Yincr=-1; } else { Yincr=1; }    // which direction in Y?

        // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
        // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
        // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
        if (dX >= dY)   // if X is the independent variable
        {
            int dPr     = dY<<1;                        // amount to increment decision if right is chosen (always)
            int dPru    = dPr - (dX<<1);                // amount to increment decision if up is chosen
            int P       = dPr - dX;                     // decision variable start value

            for (; dX>=0; dX--)                         // process each point in the line one at a time (just use dX)
            {
                (*setPixel)(CurrentX, CurrentY, c);        // plot the pixel
                CurrentX+=Xincr;                        // increment independent variable
                if (P > 0)                              // is the pixel going right AND up?
                {
                    CurrentY+=Yincr;                    // increment dependent variable
                    P+=dPru;                            // increment decision (for up)
                }
                else                                    // is the pixel just going right?
                    P+=dPr;                             // increment decision (for right)
            }
        }
        else            // if Y is the independent variable
        {
            int dPr     = dX<<1;                        // amount to increment decision if right is chosen (always)
            int dPru    = dPr - (dY<<1);                    // amount to increment decision if up is chosen
            int P       = dPr - dY;                     // decision variable start value

            for (; dY>=0; dY--)                         // process each point in the line one at a time (just use dY)
            {
                (*setPixel)(CurrentX, CurrentY, c);        // plot the pixel
                CurrentY+=Yincr;                        // increment independent variable
                if (P > 0)                              // is the pixel going up AND right?
                {
                    CurrentX+=Xincr;                    // increment dependent variable
                    P+=dPru;                            // increment decision (for up)
                }
                else                                    // is the pixel just going up?
                    P+=dPr;                             // increment decision (for right)
            }
        }
    }
}
////////////////////////////////////////////////////////////////////////////
// Bresenham algorithm to draw lines. draws one pixel with color c1 and one pixel with color c2
static void drawDottedLine(int Ax, int Ay, int Bx, int By, int32 c1, int32 c2)
{
    byte on=1;
    // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE SLOPE OR DIRECTION OF THE LINE
    int dX = abs(Bx-Ax);    // store the change in X and Y of the line endpoints
    int dY = abs(By-Ay);

    int CurrentX = Ax;      // store the starting point (just point A)
    int CurrentY = Ay;

    // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
    int Xincr, Yincr;

    if (Ax > Bx) { Xincr=-1; } else { Xincr=1; }    // which direction in X?
    if (Ay > By) { Yincr=-1; } else { Yincr=1; }    // which direction in Y?

    // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
    // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
    // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
    if (dX >= dY)   // if X is the independent variable
    {
        int dPr     = dY<<1;                        // amount to increment decision if right is chosen (always)
        int dPru    = dPr - (dX<<1);                // amount to increment decision if up is chosen
        int P       = dPr - dX;                     // decision variable start value

        for (; dX>=0; dX--)                         // process each point in the line one at a time (just use dX)
        {
            (*setPixel)(CurrentX, CurrentY, on?c1:c2); // plot the pixel
            on = !on;
            CurrentX+=Xincr;                        // increment independent variable
            if (P > 0)                              // is the pixel going right AND up?
            {
                CurrentY+=Yincr;                    // increment dependent variable
                P+=dPru;                            // increment decision (for up)
            }
            else                                    // is the pixel just going right?
                P+=dPr;                             // increment decision (for right)
        }
    }
    else            // if Y is the independent variable
    {
        int dPr     = dX<<1;                        // amount to increment decision if right is chosen (always)
        int dPru    = dPr - (dY<<1);                    // amount to increment decision if up is chosen
        int P       = dPr - dY;                     // decision variable start value

        for (; dY>=0; dY--)                         // process each point in the line one at a time (just use dY)
        {
            (*setPixel)(CurrentX, CurrentY, on?c1:c2); // plot the pixel
            on = !on;
            CurrentY+=Yincr;                        // increment independent variable
            if (P > 0)                              // is the pixel going up AND right?
            {
                CurrentX+=Xincr;                    // increment dependent variable
                P+=dPru;                            // increment decision (for up)
            }
            else                                    // is the pixel just going up?
                P+=dPr;                             // increment decision (for right)
        }
    }
}
////////////////////////////////////////////////////////////////////////////
static void _inline_ quadPixel(int xc, int yc, int x, int y, int32 c)
{
    // draw 4 points using symmetry
    (*setPixel)(xc + x, yc + y, c);
    (*setPixel)(xc + x, yc - y, c);
    (*setPixel)(xc - x, yc + y, c);
    (*setPixel)(xc - x, yc - y, c);
}
static void _inline_ quadLine(int xc, int yc, int x, int y, int32 c)
{
    int w = x+x+1; // plus 1 for the drawHLine (draws to width-1)
    // draw 2 lines using symmetry
    (*drawHLine)(xc - x, yc - y, w, c);
    (*drawHLine)(xc - x, yc + y, w, c);
}
// draws an ellipse incrementally
static void ellipseDrawAndFill (int xc, int yc, int rx, int ry, int32 c, byte fill)
{
    // intermediate terms to speed up loop
    long t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
    long t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
    long t7 = rx*t5, t8 = t7<<1, t9 = 0L;
    long d1 = t2 - t7 + (t4>>1);    // error terms
    long d2 = (t1>>1) - t8 + t5;

    register int x = rx, y = 0; // ellipse points
    if (rx < 0 || ry < 0) // guich@501_13
       return;

    while (d2 < 0)          // til slope = -1
    {
        if (fill)
            quadLine(xc,yc,x,y,c);
        else
            quadPixel(xc,yc,x,y,c);
        y++;        // always move up here
        t9 += t3;
        if (d1 < 0) // move straight up
        {
            d1 += t9 + t2;
            d2 += t9;
        }
        else        // move up and left
        {
            x--;
            t8 -= t6;
            d1 += t9 + t2 - t8;
            d2 += t9 + t5 - t8;
        }
    }

    do              // rest of top right quadrant
    {
        // draw 4 points using symmetry
        if (fill)
            quadLine(xc,yc,x,y,c);
        else
            quadPixel(xc,yc,x,y,c);
        x--;        // always move left here
        t8 -= t6;
        if (d2 < 0) // move up and left
        {
            y++;
            t9 += t3;
            d2 += t9 + t5 - t8;
        }
        else        // move straight left
            d2 += t5 - t8;
    } while (x >= 0);
}
////////////////////////////////////////////////////////////////////////////
static void drawDottedRect(int x, int y, int w, int h, int32 c1, int32 c2)
{
   int x2 = x+w-1;
   int y2 = y+h-1;
   drawDottedLine(x, y, x2,y, c1,c2);
   drawDottedLine(x, y, x ,y2,c1,c2);
   drawDottedLine(x2,y, x2,y2,c1,c2);
   drawDottedLine(x, y2,x2,y2,c1,c2);
}
////////////////////////////////////////////////////////////////////////////
// Generalized Polygon Fill * Implementation by Curt McDowell,
struct edge
{
    struct edge *next;
    int yTop, yBot;
    int xNowWhole, xNowNum, xNowDen, xNowDir;
    int xNowNumStep;
};
// UpdateActive first removes any edges which curY has entirely passed by. The removed edges are freed. It then removes any edges from the edge table at curY and places them on the active list.
static struct edge *UpdateActive(struct edge *active, struct edge *edgeTable[], int curY)
{
    struct edge *e, **ep;
    for (ep = &active, e = *ep; e != NULL; e = *ep)
        if (e->yBot < curY)
        {
            *ep = e->next;
            xfree(e);
        }
        else ep = &e->next;
    *ep = edgeTable[curY];
    return active;
}

static void fillPolygon(int32 *xPoints, int32 *yPoints, int nPoints, int32 c)
{
    // @todo: important: if memory allocation fails, this routine does not free all memory. this needs to be fixed.
    struct edge **edgeTable;
    struct edge *active;
    int curY;
    int *xCoords; // guich@400_33
    int i, j;
    struct edge *e;
    int32 *p = yPoints;
    int maxy = *p, miny=*p; // guich@421_37: calculates also miny to fix problem when the polygon has negative y values
    for (p++,i = nPoints; --i > 0; p++) // guich@350_26: now we compute the real polygon height, and also add 2 to the table size - guich@502_1: increment p in the first iteration
    {
       if (*p > maxy) maxy = *p;
       if (*p < miny) miny = *p;
    }
    if (maxy == miny) return; // guich@501_13: nothing to be drawn.
    xCoords = (int *)((*vmGlobals->xmalloc)(vmGlobals->screenHeight/3*sizeof(int))); // guich@500_10: depends on the screen height from now on
    if (!xCoords) return; // guich@501_13
    edgeTable = (*vmGlobals->xmalloc)(((maxy-miny)+2) * sizeof(struct edge *));
    if (!edgeTable)
    {
       xfree(xCoords); // guich@402_25
       (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In native fillPolygon 1"); // guich@200b4_78
       return;
    }
    edgeTable -= miny; // allow the edgetable to store negative values
    for (i = maxy+2; --i >= miny;)
        edgeTable[i] = NULL;

    for (i = 0; i < nPoints; i++)
    {
        int p1x,p1y,p2x,p2y,p3y;
        p1y = yPoints[i];
        p2y = yPoints[(i + 1) % nPoints];
        if (p1y == p2y)
            continue;   // Skip horiz. edges
        p1x = xPoints[i];
        p2x = xPoints[(i + 1) % nPoints];

        // Find next vertex not level with p2
        for (j = (i + 2) % nPoints; ; j = (j + 1) % nPoints)
        {
            p3y = yPoints[j];
            if (p2y != p3y)
                break;
        }
        e = (*vmGlobals->xmalloc)(sizeof(struct edge));
        if (!e)
        {
           (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In native fillPolygon 2"); // guich@200b4_78 - note that we will not free some chunks
           goto end;
        }
        e->xNowNumStep = abs(p1x - p2x);
        if (p2y > p1y)
        {
            e->yTop = p1y;
            e->yBot = p2y;
            e->xNowWhole = p1x;
            e->xNowDir = sgn(p2x - p1x);
            e->xNowDen = e->yBot - e->yTop;
            e->xNowNum = (e->xNowDen >> 1);
            if (p3y > p2y)
                e->yBot--;
        }
        else
        {
            e->yTop = p2y;
            e->yBot = p1y;
            e->xNowWhole = p2x;
            e->xNowDir = sgn(p1x - p2x);
            e->xNowDen = e->yBot - e->yTop;
            e->xNowNum = (e->xNowDen >> 1);
            if (p3y < p2y)
            {
                e->yTop++;
                e->xNowNum += e->xNowNumStep;
                while (e->xNowNum >= e->xNowDen)
                {
                    e->xNowWhole += e->xNowDir;
                    e->xNowNum -= e->xNowDen;
                }
            }
        }
        e->next = edgeTable[e->yTop];
        edgeTable[e->yTop] = e;
    }
    for (curY = miny; edgeTable[curY] == NULL; curY++)
        if (curY == maxy - 1)
           goto end;
    // draw runs
    for (active = NULL; (active = UpdateActive(active, edgeTable, curY)) != NULL; curY++)
    {
        /* DrawRuns first uses an insertion sort to order the X
         * coordinates of each active edge.  It updates the X coordinates
         * for each edge as it does this.
         * Then it draws a run between each pair of coordinates,
         * using the specified fill pattern.
         */
        int numCoords = 0;
        for (e = active; e != NULL; e = e->next)
        {
            for (i = numCoords; i > 0 && xCoords[i - 1] > e->xNowWhole; i--)
               xCoords[i] = xCoords[i - 1];
            xCoords[i] = e->xNowWhole;
            if (++numCoords >= 50)
            {
               (*vmGlobals->debug)("too much points in fill polygon. please inform support@superwaba.com.br");
               goto end; // too much points! maybe a fatal error?
            }
            e->xNowNum += e->xNowNumStep;
            while (e->xNowNum >= e->xNowDen)
            {
                e->xNowWhole += e->xNowDir;
                e->xNowNum -= e->xNowDen;
            }
        }
        if (numCoords & 1)  /* Protect from degenerate polygons */
        {
            xCoords[numCoords] = xCoords[numCoords - 1], /** !!! */
            numCoords++;
        }
        for (i = 0; i < numCoords; i += 2)
        {
            int w = xCoords[i+1]-xCoords[i]+1;
            if (w > 0)
               (*drawHLine)(xCoords[i],(int)curY,w,c);
        }
    }
end:
    edgeTable += miny;
    xfree(xCoords);
    xfree(edgeTable);
}

////////////////////////////////////////////////////////////////////////////
// draws a polygon. if the polygon is not closed, close it
static void drawPolygon(int32 *xPoints, int32 *yPoints, int nPoints, int32 c, byte close)
{
   int i;
   if (nPoints < 2) return;
   for (i=1; i < nPoints; i++)
      drawLine(xPoints[i-1],yPoints[i-1],xPoints[i],yPoints[i],c);
   if (close)
   {
	   nPoints--;
	   if (xPoints[0] != xPoints[nPoints] || yPoints[0] != yPoints[nPoints])
	      drawLine(xPoints[0],yPoints[0],xPoints[nPoints],yPoints[nPoints],c);
	}
}
////////////////////////////////////////////////////////////////////////////
// draw an elliptical arc from startAngle to endAngle. c is the fill color and c2 is the outline color (if in fill mode - otherwise, c = outline color)
static void arcPiePointDrawAndFill(int xc, int yc, int rx, int ry, float startAngle, float endAngle, int32 c, int32 c2, byte fill, byte pie, int32 *startAnglePoint)
{
   static int32 *xPoints=0,*yPoints=0,nPoints=0;
   static int lastRX=-1,lastRY=-1,lastXC = -1, lastYC = -1,lastSize=0;
   int limitx1 = CLIPX1-TRANSX; // guich@501_13: store these for faster computation
   int limitx2 = CLIPX2-TRANSX;
   int limity1 = CLIPY1-TRANSY;
   int limity2 = CLIPY2-TRANSY;
   static float lastPPD=0;
   // this algorithm was created by Guilherme Campos Hazan
   float ppd;
   int startIndex,endIndex,index,i;
   int nq,size=0;
   int oldX=0,oldY=0;
   byte checkClipX = (byte)((xc+rx) > limitx2 || (xc-rx) < limitx1); // guich@340_3 - guich@401_2: added TRANSX/Y
   byte checkClipY = (byte)((yc+ry) > limity2 || (yc-ry) < limity1);
   if (rx < 0 || ry < 0) // guich@501_13
      return;
   // step 0: correct angle values
   if (startAngle < 0.1 && endAngle > 359.9) // full circle? use the fastest routine instead
   {
      if (fill)
         fillEllipse(xc,yc,rx,ry,c);
      drawEllipse(xc,yc,rx,ry,fill?c2:c);
      return;
   }
   // step 0: if possible, use cached results
   if (xc != lastXC || yc != lastYC || rx != lastRX || ry != lastRY)
   {
       // step 1: computes how many points the circle has (computes only 45 degrees and mirrors the rest)
       // intermediate terms to speed up loop
       int32 t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
       int32 t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
       int32 t7 = rx*t5, t8 = t7<<1, t9 = 0L;
       int32 d1 = t2 - t7 + (t4>>1);    // error terms
       int32 d2 = (t1>>1) - t8 + t5;

       register int x = rx, y = 0; // ellipse points

       while (d2 < 0)          // til slope = -1
       {
          y++;        // always move up here
          t9 += t3;
          if (d1 < 0) // move straight up
          {
             d1 += t9 + t2;
             d2 += t9;
          }
          else        // move up and left
          {
             x--;
             t8 -= t6;
             d1 += t9 + t2 - t8;
             d2 += t9 + t5 - t8;
          }
          size++;
       }

       do              // rest of top right quadrant
       {
          x--;        // always move left here
          t8 -= t6;
          if (d2 < 0) // move up and left
          {
             y++;
             t9 += t3;
             d2 += t9 + t5 - t8;
          }
          else        // move straight left
             d2 += t5 - t8;
        size++;
       } while (x >= 0);
       nq = size;
       size *= 4;
       // step 2: computes how many points per degree
       ppd = (float)size / 360.0f;
       // step 3: create space in the buffer so it can save all the circle
       size+=2;
       if (nPoints < size)
       {
          uint32 temp = sizeof(int32)*size;
          xPoints = (*vmGlobals->xrealloc)(xPoints, &temp);
          if (!xPoints)
          {
             (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "in ArcPiePointDrawAndFill 1"); // guich@200b4_78
             return;
          }
          yPoints = (*vmGlobals->xrealloc)(yPoints, &temp);
          if (!yPoints)
          {
             xfree(xPoints);
             (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "in ArcPiePointDrawAndFill 2"); // guich@200b4_78
             return;
          }
       }
       nPoints = size;
       // step 4: stores all the circle in the array. the odd arcs are drawn in reverse order
        // intermediate terms to speed up loop
        t2 = t1<<1; t3 = t2<<1;
        t8 = t7<<1; t9 = 0L;
        d1 = t2 - t7 + (t4>>1); // error terms
        d2 = (t1>>1) - t8 + t5;

        x = rx;
        y = 0;  // ellipse points

        i=0;
        while (d2 < 0)          // til slope = -1
        {
            // save 4 points using symmetry
            // guich@340_3: added clipping
            #define doClipX(x) ( ((x) < limitx1) ? limitx1 : ((x) >= limitx2) ? limitx2 : (x) ) // guich@401_2: added TRANSX/Y
            #define doClipY(y) ( ((y) < limity1) ? limity1 : ((y) >= limity2) ? limity2 : (y) )

            index = nq*0+i;      // 0/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2-i-1;    // 1/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2+i;      // 2/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            index = nq*4-i-1;    // 3/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            i++;
            y++;        // always move up here
            t9 += t3;
            if (d1 < 0) // move straight up
            {
                d1 += t9 + t2;
                d2 += t9;
            }
            else        // move up and left
            {
                x--;
                t8 -= t6;
                d1 += t9 + t2 - t8;
                d2 += t9 + t5 - t8;
            }
        }

        do              // rest of top right quadrant
        {
            // save 4 points using symmetry
            // guich@340_3: added clipping
            index = nq*0+i;    // 0/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2-i-1;  // 1/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2+i;    // 2/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            index = nq*4-i-1;  // 3/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            i++;
            x--;        // always move left here
            t8 -= t6;
            if (d2 < 0) // move up and left
            {
                y++;
                t9 += t3;
                d2 += t9 + t5 - t8;
            }
            else        // move straight left
                d2 += t5 - t8;
        } while (x >= 0);
       // save last arguments
       lastXC = xc;
       lastYC = yc;
       lastRX = rx;
       lastRY = ry;
       lastPPD = ppd;
       lastSize = size;
   }
   else ppd = lastPPD;
   // step 5: computes the start and end indexes that will become part of the arc
   startIndex = (int)(ppd * startAngle);
   endIndex = (int)(ppd * endAngle);
   // step 5 1/2: if only computing the point, return it
   if (startAnglePoint)
   {
      startAnglePoint[0] = xPoints[startIndex];
      startAnglePoint[1] = yPoints[startIndex];
      return;
   }
   if (endIndex == (lastSize-2)) // 360?
      endIndex--;
   // step 6: fill or draw the polygons
   endIndex++;
   if (pie)
   {
      oldX = xPoints[endIndex];
      oldY = yPoints[endIndex];
      xPoints[endIndex] = xc;
      yPoints[endIndex] = yc;
      endIndex++;
   }
   if (fill)
      fillPolygon(xPoints+startIndex,yPoints+startIndex,endIndex-startIndex,c2); // guich@401_3
   drawPolygon(xPoints+startIndex,yPoints+startIndex,endIndex-startIndex,/*fill?c2:*/c,pie); // guich@402_57 and guich@402_58: see waba.fx.Graphics.java
   if (pie) // restore saved points
   {
      xPoints[endIndex-1] = oldX;
      yPoints[endIndex-1] = oldY;
   }

}
////////////////////////////////////////////////////////////////////////////
// calls arcPiePointDrawAndFill twice if the angles cross the 0
static void preArcPiePointDrawAndFill(int xc, int yc, int rx, int ry, float startAngle, float endAngle, int32 c, int32 c2, byte fill, byte pie)
{
   // make sure the values are -359 <= x <= 359
   while (startAngle <= -360) startAngle += 360;
   while (endAngle   <= -360) endAngle   += 360;
   while (startAngle >   360) startAngle -= 360;
   while (endAngle   >   360) endAngle   -= 360;

   if (startAngle == endAngle) // guich@501_13
      return;
   if (startAngle > endAngle) // eg 235 to 45
      startAngle -= 360; // set to -45 to 45 so we can handle it correctly
   if (startAngle >= 0 && endAngle <= 0) // eg 135 to -135
      endAngle += 360; // set to 135 to 225

   if (startAngle >= 0 && endAngle >= 0)
      arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, endAngle, c, c2, fill, pie, 0);
   else
   if (startAngle <= 0 && endAngle >= 0) // eg -45 to 45
   {
      startAngle += 360;
      arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, 360, c, c2, fill, pie, 0);
      arcPiePointDrawAndFill(xc, yc, rx, ry, 0, endAngle, c, c2, fill, pie, 0);
   } else (*vmGlobals->debug)("arc/pie/point could not be filled with given angles");
}
////////////////////////////////////////////////////////////////////////////
// for a angle, computes its point. useful if you want to place a label in the pie and dont want to use sin/cos. index 0 is x ad index 1 is y
static void getAnglePoint(int xc, int yc, int rx, int ry, float angle, int32* twoInts)
{
   while (angle >= 360) angle -= 360;
   while (angle < 0) angle += 360;
   arcPiePointDrawAndFill(xc,yc,rx,ry,angle,angle,0,0,0,0,twoInts);
}
////////////////////////////////////////////////////////////////////////////
static void drawRoundRect(int x, int y, int width, int height, int r, int32 c)
{
   int w,h,x1,y1,x2,y2,dec,xx,yy;
   if (r > (width/2) || r > (height/2)) r = min(width/2,height/2); // guich@200b4_6: correct bug that crashed the device.
   w = width - 2*r;
   h = height - 2*r;
   x1 = x+r;
   y1=y+r;
   x2=x+width-r-1;
   y2=y+height-r-1;
   dec = 3-2*r;

   (*drawHLine)(x+r,y,w,c); // top
   (*drawHLine)(x+r,y+height-1,w,c); // bottom
   (*drawVLine)(x,y+r,h,c); // left
   (*drawVLine)(x+width-1,y+r,h,c); // right

   for (xx = 0, yy = r; xx <= yy; xx++)
   {
      (*setPixel)(x2+xx,y2+yy,c);
      (*setPixel)(x2+xx,y1-yy,c);
      (*setPixel)(x1-xx,y2+yy,c);
      (*setPixel)(x1-xx,y1-yy,c);

      (*setPixel)(x2+yy,y2+xx,c);
      (*setPixel)(x2+yy,y1-xx,c);
      (*setPixel)(x1-yy,y2+xx,c);
      (*setPixel)(x1-yy,y1-xx,c);
      if (dec >= 0)
         dec += -4*(yy--)+4;
      dec += 4*xx+6;
   }
}
////////////////////////////////////////////////////////////////////////////
static void fillRoundRect(int x, int y, int width, int height, int r, int32 c)
{
   int x1,y1,x2,y2,xx,yy,dec;
   if (r > (width/2) || r > (height/2)) r = min(width/2,height/2); // guich@200b4_6: correct bug that crashed the device.

   x1 = x+r;
   y1=y+r;
   x2=x+width-r-1;
   y2=y+height-r-1;
   dec = 3-2*r;

   height -= 2*r;
   y += r;
   while (height--)
      (*drawHLine)(x,y++,width,c);

   for (xx = 0, yy = r; xx <= yy; xx++)
   {
      drawLine(x1-xx,y1-yy,x2+xx,y1-yy,c);
      drawLine(x1-xx,y2+yy,x2+xx,y2+yy,c);

      drawLine(x1-yy,y1-xx,x2+yy,y1-xx,c);
      drawLine(x1-yy,y2+xx,x2+yy,y2+xx,c);

      if (dec >= 0)
         dec += -4*(yy--)+4;
      dec += 4*xx+6;
   }
}
////////////////////////////////////////////////////////////////////////////
static void drawHatchedRect(int x, int y, int width, int height, byte top, byte bottom, int32 c)
{
   int x2 = x+width-1;
   int y2 = y+height-1;
   if (top && bottom)
   {
      drawLine(x,y+2,x,y2-2,c); // left
      drawLine(x+2,y2,x2-2,y2,c); // bottom
      drawLine(x2,y+2,x2,y2-2,c); // right
      drawLine(x+2,y,x2-2,y,c); // top
      setPixel(x+1,y+1,c); // top left
      setPixel(x2-1,y+1,c); // top right
      setPixel(x+1,y2-1,c); // bottom left
      setPixel(x2-1,y2-1,c); // bottom right
   }
   else
   if (top && !bottom)
   {
      drawLine(x,y+2,x,y2,c); // left
      drawLine(x,y2,x2,y2,c); // bottom
      drawLine(x2,y+2,x2,y2,c); // right
      drawLine(x+2,y,x2-2,y,c); // top
      setPixel(x+1,y+1,c); // top left
      setPixel(x2-1,y+1,c); // top right
   }
   else
   if (!top && bottom)
   {
      drawLine(x,y,x,y2-2,c); // left
      drawLine(x+2,y2,x2-2,y2,c); // bottom
      drawLine(x2,y,x2,y2-2,c); // right
      drawLine(x,y,x2,y,c); // top
      setPixel(x+1,y2-1,c); // bottom left
      setPixel(x2-1,y2-1,c); // bottom right
   }
}
////////////////////////////////////////////////////////////////////////////
static void fillHatchedRect(int x, int y, int width, int height, byte top, byte bottom, int32 c)
{
   int x2 = x+width-1;
   int y2 = y+height-1;
   if (top && bottom)
   {
      fillRect(x,y+2,width,height-4,c); // middle
      drawLine(x+2,y,x2-2,y,c); // 1st line
      drawLine(x+1,y+1,x2-1,y+1,c); // 2nd line
      drawLine(x+1,y2-1,x2-1,y2-1,c); // last-1 line
      drawLine(x+2,y2,x2-2,y2,c); // last line
   }
   else
   if (top && !bottom)
   {
      drawLine(x+2,y,x2-2,y,c); // 1st line
      drawLine(x+1,y+1,x2-1,y+1,c); // 2nd line
      fillRect(x,y+2,width,height-2,c); // middle
   }
   else
   if (!top && bottom)
   {
      fillRect(x,y,width,height-2,c); // middle
      drawLine(x+1,y2-1,x2-1,y2-1,c); // last-1 line
      drawLine(x+2,y2,x2-2,y2,c); // last line
   }
}
static void drawHighLightFrame(int x, int y, int w, int h, WObject topLeft, WObject bottomRight, int yMirror)
{
   int x2 = x+w-1;
   int y2 = y+h-1;
   if (topLeft)
   {
   	Var *cObj = topLeft; // class object
      int32 c = (*getColor)(WOBJ_ColorRedP(cObj),WOBJ_ColorGreenP(cObj),WOBJ_ColorBlueP(cObj));
      drawLine(x,y,x,y2,c);
      if (!yMirror)
         drawLine(x,y,x2,y,c);
      else
         drawLine(x,y2,x2,y2,c);
   }
   if (bottomRight)
   {
   	Var *cObj = bottomRight; // class object
      int32 c = (*getColor)(WOBJ_ColorRedP(cObj),WOBJ_ColorGreenP(cObj),WOBJ_ColorBlueP(cObj));
      drawLine(x2,y2,x2,y,c);
      if (!yMirror)
         drawLine(x2,y2,x,y2,c);
      else
         drawLine(x2,y,x,y,c);
   }
}

////////////////////////////////////////////////////////////////////////////
// sets the clip rect to (x,y,x+w-1,y+h-1), translated to the current translated origin
static void setClip(int x, int y, int w, int h)
{
   int clipX1 = x+TRANSX;
   int clipY1 = y+TRANSY;
   int clipX2 = clipX1+w;
   int clipY2 = clipY1+h;

   if (clipX1 <     0) clipX1 = 0;
   if (clipY1 <     0) clipY1 = 0;
   if (clipX1 > maxX) clipX1 = maxX;
   if (clipY1 > maxY) clipY1 = maxY;

   if (clipX2 <     0) clipX2 = 0;
   if (clipY2 <     0) clipY2 = 0;
   if (clipX2 > maxX) clipX2 = maxX;
   if (clipY2 > maxY) clipY2 = maxY;

   GRHANDLE_ClipX1(activeGrHandle) = clipX1;
   GRHANDLE_ClipY1(activeGrHandle) = clipY1;
   GRHANDLE_ClipX2(activeGrHandle) = clipX2;
   GRHANDLE_ClipY2(activeGrHandle) = clipY2;

   clipX = MAKE32(clipX1,clipX2);
   clipY = MAKE32(clipY1,clipY2);
}
////////////////////////////////////////////////////////////////////////////
// translates the origin of the drawing area by dx and dy. the final translation coords cannot be lesser than 0.
static void translate(int dx, int dy)
{
   int transX = TRANSX;
   int transY = TRANSY;

   transX += dx;
   transY += dy;

   GRHANDLE_TransX(activeGrHandle) = transX;
   GRHANDLE_TransY(activeGrHandle) = transY;
   transXY = MAKE32(transX,transY);

/*{char buf[80];
xstrprintf(buf,"transxy: %lx = (%d,%d)",(int32)transXY,(int)TRANSX,(int)TRANSY);
DbgMessage(buf);
}*/

}
////////////////////////////////////////////////////////////////////////////
static SetPixel_ setRGB32;

static void convertToGray(int x, int y, int32 rgb); // y not used
uint32 *colorTable=0;
static Pixel *indexes;

static void readRLE(int height, byte *in, byte rle8)
{
    byte val;
    byte len,esc;
    byte colors0,colors1;
    int x,y;
    Pixel color;
    Pixel back = imgBackColor; // cache it

    x = 0;
    y = height-1;

    while (1)
    {
        esc = *in++;
        // encoded mode starts with a run length, and then a byte with two colour indexes to alternate between for the run
        if (esc)
        {
            if (rle8)
		      {
		         Pixel color = indexes[*in++];
		         if (color != back)
   		         while (esc--)
	   		         (*setRGB32)(x++,y,color);
	   		   else x += esc;
		      }
		      else
		      {
	            val = *in++;
               colors0 = (val>>4) & 0x0F;
               colors1 =  val     & 0x0F;
               for (len = 1; len <= esc; len++,x++)
	            {
	               color = indexes[(len & 1)?colors0:colors1];
	               if (color != back)
   	               (*setRGB32)(x, y, color); // odd count, low nybble
   	         }
	         }
        }
        else // A leading zero is an escape; it may signal the end of the bitmap, a cursor move, or some absolute data.
        {    // zero tag may be absolute mode or an escape
            esc = *in++;
            switch (esc)
            {
               case 0:                             // end of line
                  x = 0; y--;
                  break;
               case 1:                             // end of bitmap
                  return;
               case 2:                             // delta
                  x += *in++;
                  y -= *in++;
                  break;
               default:        // no compression
                  if (rle8)
                  {
			            len = esc;
			            while (len--)
			            {
			               color = indexes[*in++];
			               if (color != back)
   			               (*setRGB32)(x,y,color);
   			            x++;
   			         }
			            if (esc & 1) // Must have even nunber of bytes
			               in++;
                  }
                  else
                  {
                     for (len = 1; len <= esc; len++,x++)
                     {
                        if ((len & 1) == 1)
                        {
                           val = *in++;
                           colors0 = (val>>4) & 0x0F;
                           colors1 =  val     & 0x0F;
                           if (colors0 != back) (*setRGB32)(x, y, colors0);
                        }
                        else if (colors1 != back) (*setRGB32)(x, y, colors1); // odd count, low nybble
                     }
                     if ((((esc+1)>>1) & 1) != 0) // Must have even nunber of words
                        in++;
	               }
                  break;
            }
        }
    }
}

// readRGB reads in pixels values that are stored uncompressed.
// The bits represent indices into the color table.
static void readRGB(int32 width, int32 height, int bpp, uchar *in)
{
    // How many pixels can be stored in a byte?
    int pixelsPerByte = 8 / bpp;
    // A bit mask containing the number of bits in a pixel
    int bitMask = (1 << bpp) - 1;
    int bitShifts[8];
    int i,h,w;
    int whichBit = 0;
    int currByte;
    int div = 32 / bpp;
    Pixel color;
    Pixel back = imgBackColor; // cache it

    width = ((width+div-1) / div)*div; // make sure are in a 4 byte boundary - those extra pixels will be stripped off by the current clip

    // Start at the bottom of the pixel array and work up
    if (pixelsPerByte == 1) // guich@200b3: if 8bpp, use a faster routine
    {
#if defined(WINCE) || defined(PALMOS)
       if (displayBPP < 8 || !read8bppImage(width,height,in)) // guich@420_15: try the faster way if we have enough memory
#endif
	    for (h=height-1; h >= 0; h--)
	        for (w=0; w < width; w++)
	        {
	           color = indexes[*in++];
	           if (color != back)
                 (*setRGB32)(w,h,color);
           }
    }
    else
    {
       // The shift values that will move each pixel to the far right
       for (i=0; i < pixelsPerByte; i++)
           bitShifts[i] = 8 - ((i+1) * bpp);

	    // Read in the first byte
	    currByte = *in++;

	    for (h=height-1; h >= 0; h--)
	        for (w=0; w < width; w++)
	        {
	            color = indexes[(currByte >> bitShifts[whichBit]) & bitMask];
	            if (color != back)
   	            (*setRGB32)(w,h,color); // Get the next pixel from the current byte
	            // If the current bit position is past the number of pixels in a byte, you advance to the next byte
	            if (++whichBit >= pixelsPerByte)
	            {
	                whichBit = 0;
	                currByte = *in++;
	            }
	        }
	 }
}
////////////////////////////////////////////////////////////////////////////
/* Floyd/Steinberg error diffusion dithering algorithm in color.  The array
** line[][] contains the RGB values for the current line being processed;
** line[0][x] = red, line[1][x] = green, line[2][x] = blue.
*/
// ps: this algorithm was heavily modified and optimized by guich
// note: dithering an 160x160x256 color bitmap takes about 5 minutes in a palm iiic, so it will only be done in gray scale up to 16 shades.
static int bmpW,bmpH, scale,maxColorIndex;
static Pixel *scannedLine;

static void convertToGray(int x, int y, int32 index) // y not used
{
   int r,g,b;
   int ind;
   byte *rgb;
   int med;

   if (x > bmpW)
   	return;

   rgb = (byte*)&colorTable[index];
   r = *(rgb+1);//GET_RED(rgb);
   g = *(rgb+2);//GET_GREEN(rgb);
   b = *(rgb+3);//GET_BLUE(rgb);

   med = (r+g+b)/3;
   ind = med >> scale;

   scannedLine[x] = maxColorIndex-ind;
   if (x == bmpW) // full row computed
      drawScanLine(0,y,bmpW+1,scannedLine);
}

// Intel-architecture getUInt32 - these are not #defines to make the executable smaller
static uint32 inGetUInt32(uchar *b)
{
	return (uint32)( (uint32)b[3]<<24 | (uint32)b[2]<<16 | (uint32)b[1]<<8 | (uint32)b[0] );
}

static uint32 inGetUInt16(uchar *b)
{
	return (uint32)( (uint16)b[1]<<8 | (uint16)b[0] );
}

// created by guich to handle all types of modern bitmaps,
#define _BI_RGB 0
#define _BI_RLE8 1
#define _BI_RLE4 2
static void ImageLoadBMPCompressed(WObject image, uchar *p)
{
   int32 bitmapOffset, infoSize, width, height;
   uint32 bmpBPP;
   int32 compression, imageSize,usedColors,color;
   int i,j,baseIndex;
   uint16 error;
   GrHandle imgHandle;
   byte doDither,storePalette,confirmDithering, highColor;
   //displayBPP = 4; maxColors = 16; debug
   colorTable = NULL;
   indexes = NULL;
   oldGrHandle = NULL;

   // header (54 bytes)
   // 0-1   magic chars 'BM'
   // 2-5   uint32 filesize (not reliable)
   // 6-7   uint16 0
   // 8-9   uint16 0
   // 10-13 uint32 bitmapOffset
   // 14-17 uint32 info size
   // 18-21 int32  width
   // 22-25 int32  height
   // 26-27 uint16 nplanes
   // 28-29 uint16 bits per pixel
   // 30-33 uint32 compression flag
   // 34-37 uint32 image size in bytes
   // 38-41 int32  biXPixelsPerMeter
   // 42-45 int32  biYPixelsPerMeter
   // 46-49 uint32 colors used
   // 50-53 uint32 important color count
   // 54-   uchar  bitmap bytes depending to type
   // Each scan line of image data is padded to the next four byte boundary

   if (p[0] != 'B' || p[1] != 'M')
      return; // not a BMP file
   bitmapOffset = inGetUInt32(p+10);
   infoSize = inGetUInt32(p+14);
   if (infoSize != 40)
      return; // old-style BMP
   width = inGetUInt32(p+18);
   height = inGetUInt32(p+22);

   if (width > 65535 || height > 65535)
   {
      (*vmGlobals->debug)("bad bitmap dimension");
      return; // bad width/height
   }
   bmpBPP = inGetUInt16(p+28);
   if (bmpBPP == 24)
   {
      (*vmGlobals->debug)("24 bpp bitmaps not supported");
      return; // support for 24 bits bitmap was removed because the pics gets too large, and palm will take some more years to have support for 24 bits displays
   }
   compression = inGetUInt32(p+30);
   imageSize = inGetUInt32(p+34);

   usedColors = inGetUInt32(p+46);
   if (usedColors == 0) usedColors = (int32)1 << bmpBPP;

   // Create space for the pixels
   imgHandle = CreateOffscreenWindow(image, (int16)width, (int16)height, 1, &error);
   if (error) return;
   bmpW = width-1;
   bmpH = height-1;

   baseIndex    = (displayBPP >= 8 && usedColors == 16) ? 231 : 0; // if in color device and the image has 16 colors, store the colors in the user defined colors region
   doDither     =  (byte)(displayBPP <  8 && bmpBPP > displayBPP); // this is the normal situation.
   confirmDithering = (byte)(displayBPP < 8); // But if we have a 16 colors bitmap in a 16 gray device, we also must do dithering. So, we must confirm that this bmp has a gray palette.
#if _WIN32_WCE == 212
   highColor    = 1; // for the intermec 600 grayscale version, we need to convert the palette
   storePalette = 0;
#else
   highColor    = (byte)(displayBPP > 8);
   storePalette = (byte)(displayBPP >= 8 && !doDither && usedColors > 2); // guich@300: changed "displayBPP >= 8" to "== 8" to avoid storing the palette in high bpp devices
#endif
   colorTable   = (uint32*) (*vmGlobals->xmalloc)(4*(storePalette ? 256 : ((int32)1 << bmpBPP/*usedColors*/))); // guich@340_59: now always use 1<<bpp bc some bitmaps can bring wrong information for colorsUsed
   if (!colorTable)
      goto memErr;
   indexes      = (Pixel *) (*vmGlobals->xmalloc)(usedColors*sizeof(Pixel));
   if (!indexes)
      goto memErr;
   if (storePalette && usedColors != 256)
      xmemmove(colorTable,systemPalette,1024); // copy the system palette

   // Read the bitmap's color table
   for (i=0,j=54; i < usedColors; i++,j+=4)
   {
      colorTable[i+baseIndex] = color = (inGetUInt32(p+j) & 0x00FFFFFF);
      indexes[i] = i+baseIndex;
      if (confirmDithering && !doDither && (GET_RED(color) != GET_GREEN(color) || GET_RED(color) != GET_BLUE(color)))
         doDither = 1;
   }

   if (!isColor || doDither)
      if (maxColors == 16)
         {scale = 4; maxColorIndex = 15;}
      else
         {scale = 6; maxColorIndex =  3;}

   // prepares the color table for this screen bpp.
   if (!isColor)
      indexes[usedColors-1] = maxColors-1; // case 1 bit bpp on a gray device - make the black realy black
   else
   if (usedColors == 2) indexes[1] = 230; // monochrome bitmap in a color device must map the black color to the correct index
   if (bmpBPP == 1 && colorTable[1] == 0xFFFFFF) // the bitmap is monochrome, are the colors iverted?
   {
      Pixel temp = indexes[1];
      indexes[1] = indexes[0];
      indexes[0] = temp;
   }
   if (highColor) // guich@300: if using direct color devices, store the color directly in the indexes
   {
      for (i = 0; i < usedColors; i++)
      {
         color = colorTable[i+baseIndex];
         indexes[i] = (*getColor)(GET_RED(color), GET_GREEN(color), GET_BLUE(color));
      }
   }
   if (storePalette)
   {
      GRHANDLE_PaletteSize(imgHandle) = 256;
      GRHANDLE_Palette(imgHandle) = colorTable;
   }
   // set the appropriated algorithm to put the pixel
   oldGrHandle = activeGrHandle;
   SetDrawWindow(imgHandle);

   scannedLine = NULL;
   if (doDither)
   {
	   scannedLine = (*vmGlobals->xmalloc)(sizeof(Pixel)*(bmpW+1));
      if (!scannedLine)
         goto memErr;
      setRGB32 = &convertToGray;
      imgBackColor=-1; // when dithering, we must compute all pixels
   }
   else
   {
      setRGB32 = setPixelXM;
      fillRect(0,0,width,height,imgBackColor=(*getColor)(255,255,255)); // guich@402_32
   }

   WOBJ_ControlImageGraphicsGrHandle(image) = imgHandle;
   WOBJ_ImageWidth(image)    = width;
   WOBJ_ImageHeight(image)   = height;

   // Read the pixels from the stream based on the compression type directly into the selected offscreen image
   if (compression == _BI_RGB)
      readRGB(width, height, bmpBPP, p+bitmapOffset);
   else
      readRLE((int)height, p+bitmapOffset, (byte) (compression == _BI_RLE8));

   // set the image palette
   if (isColor)
   {
      int len = storePalette ? 256 : usedColors;  // guich@310_29: fixed a problem with the image had less colors
	   WObject intArray = (*vmGlobals->createArrayObject)(10, len);
      if (intArray)
      {
         int32 *ints = (int32 *)WOBJ_arrayStart(intArray);
         xmemmove(ints,colorTable,len*4); // guich@310_28: changed usedColors to 256, since in an 16 color image, the palette stored is in reality with 256 colors
         WOBJ_ImagePalette(image) = intArray;
      }
   }
   xfree(scannedLine);
   xfree(indexes);
   if (!storePalette) xfree(colorTable);
   if (oldGrHandle != NULL) SetDrawWindow(oldGrHandle); // guich@310_2
   return;
memErr:
   xfree(colorTable); // note: xfree already test if xxx is null
   xfree(scannedLine);
   xfree(indexes);
   if (oldGrHandle != NULL) SetDrawWindow(oldGrHandle); // guich@402_18: forgot to add this
   (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In imageLoadBmpCompressed"); // guich@200b4_78
}
////////////////////////////////////////////////////////////////////////////
// returns 1 if the coords intersects the clip rectangle
static byte intersectsClipRect(int x1,int y1,int x2,int y2)
{
   return (byte)!((x2 <= CLIPX1) || (y2 <= CLIPY1) || (x1 >= CLIPX2) || (y1 >= CLIPY2));
}
////////////////////////////////////////////////////////////////////////////
byte intersects(RectangleType *r1, RectangleType *r2)
{
   return (byte)!((r1->topLeft.x + r1->extent.x <= r2->topLeft.x) || (r1->topLeft.y + r1->extent.y <= r2->topLeft.y) || (r1->topLeft.x >= r2->topLeft.x + r2->extent.x) || (r1->topLeft.y >= r2->topLeft.y + r2->extent.y));
}
////////////////////////////////////////////////////////////////////////////
void intersectWith(RectangleType *r1, RectangleType *r2)
{
   int x1 = max(r1->topLeft.x, r2->topLeft.x);
   int x2 = min(r1->topLeft.x + r1->extent.x, r2->topLeft.x + r2->extent.x);
   int y1 = max(r1->topLeft.y, r2->topLeft.y);
   int y2 = min(r1->topLeft.y + r1->extent.y, r2->topLeft.y + r2->extent.y);
   r1->topLeft.x = x1;
   r1->topLeft.y = y1;
   r1->extent.x = x2 - x1;
   r1->extent.y = y2 - y1;
}
////////////////////////////////////////////////////////////////////////////
static Pixel *cldRow1=NULL;
static Pixel *cldRow2=NULL;
void freeGrResources() // guich@200b4
{
   if (cldRow1) xfree(cldRow1);
   if (cldRow2) xfree(cldRow2);
}

// returns true if the two bitmaps collids, ie, if there are any poits != backcolor intersected
static byte collisionDetect(GrHandle bmp1, RectangleType rect1, GrHandle bmp2, RectangleType rect2, int backcolor)
{
   RectangleType rect;
   int x1,y1,x2,y2;
   int i,j,w,h;
   static uint32 lastW=0;
   uint32 size;
   // first and simplier test: if rects dont intersect, no colision could occur.
   if (!intersects(&rect1,&rect2)) return 0;
   RctGetIntersection(&rect1,&rect2,&rect);
   w = rect.extent.x;
   h = rect.extent.y;
   x1 = rect.topLeft.x - rect1.topLeft.x;
   y1 = rect.topLeft.y - rect1.topLeft.y;
   x2 = rect.topLeft.x - rect2.topLeft.x;
   y2 = rect.topLeft.y - rect2.topLeft.y;
   if ((uint32)w > lastW)
   {
      lastW = ((w/20)+1)*20; // lets grow a little to make memory less fragmented
      size = lastW * sizeof(cldRow1[0]);
      cldRow1 = (*vmGlobals->xrealloc)(cldRow1,&size);
      if (!cldRow1)
      {
         (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In collisionDetect 1"); // guich@200b4_78
         return 0;
      }
      cldRow2 = (*vmGlobals->xrealloc)(cldRow2,&size);
      if (!cldRow2)
      {
         xfree(cldRow1);
         (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In collisionDetect 2"); // guich@200b4_78
         return 0;
      }
   }
   // now scans for pixels != color in both images
   for (j = 0; j < h; j++)
   {
      (*getBitmapRow)(bmp1,x1,y1+j,w,cldRow1);
      (*getBitmapRow)(bmp2,x2,y2+j,w,cldRow2);
      for (i = 0; i < w; i++)
         if ((cldRow1[i] != backcolor) && (cldRow2[i] != backcolor))
            return 1;
   }
   return 0;
}
////////////////////////////////////////////////////////////////////////////
// transltates the given coords and returns the intersection between the clip rect and the coords passed. returns 0 if the coords are outside the clip rect
byte translateAndClip(int *x, int *y, int *width, int *height)
{
   *x += TRANSX;
   *y += TRANSY;
   if (!intersectsClipRect(*x,*y,*x+*width,*y+*height)) return 0;
   if (*x < CLIPX1) {*width  -= CLIPX1-*x; *x = CLIPX1;}
   if (*y < CLIPY1) {*height -= CLIPY1-*y; *y = CLIPY1;}
   if ((*x+*width)  > CLIPX2) *width  = CLIPX2 - *x;
   if ((*y+*height) > CLIPY2) *height = CLIPY2 - *y;
   return 1;
}
////////////////////////////////////////////////////////////////////////////
//         INTERFACE OF THE GRAPHICS CLASS METHODS WITH THE C CODE        //
////////////////////////////////////////////////////////////////////////////
static void FreeGrHandle(GrHandle h)
{
   if (h && (h != mainGrHandle))
   {
      if (h == activeGrHandle) // guich@200b3: is this the active GrHandle?
      {
         if (mainGrHandle)
            SetDrawWindow(mainGrHandle); // switch to other handle
         lastGraphics = NULL;
         lastGrHandle = NULL;
      }
	   xfree(h);
	}
}
static void SaveState(GrHandle h)
{
   GRHANDLE_TransX(h) = TRANSX;
   GRHANDLE_TransY(h) = TRANSY;
   GRHANDLE_ClipX1(h) = CLIPX1;
   GRHANDLE_ClipX2(h) = CLIPX2;
   GRHANDLE_ClipY1(h) = CLIPY1;
   GRHANDLE_ClipY2(h) = CLIPY2;
}
//////////////////////////////////////////////////////////////////////////
#define SELECT_GRAPHICS(GR) if (!lastGraphics || (lastGrHandle != WOBJ_ControlImageGraphicsGrHandle(GR))) selectGraphicsObj(GR) // the objects may differ, but the GrHandle wont; dont try to optimize this by changing the order or the logic will fail!

static void selectGraphicsObj(WObject gr)
{
   WObject lastG1 = lastGraphics;
   GrHandle lastGH = lastGrHandle;
   GrHandle lastSto = lastG1?WOBJ_ControlImageGraphicsGrHandle(lastGraphics):NULL;

   if (lastG1 != NULL && lastSto != NULL && lastGH != NULL)
//   if (lastGraphics != NULL && WOBJ_ControlImageGraphicsGrHandle(lastGraphics) != NULL) // guich@200b2: corrected a bug after a Graphics object was freed and here was called. guich@200b3: removed comment from 2nd part of if
   {
      // saves back the values in the object
      SaveState(WOBJ_ControlImageGraphicsGrHandle(lastGraphics));
      WOBJ_GraphicsForeRGB(lastGraphics) = currentForeColor;
      WOBJ_GraphicsBackRGB(lastGraphics) = currentBackColor;
      WOBJ_GraphicsFont(lastGraphics) = currentUserFont;
   }
   lastGraphics = gr;
   lastGrHandle = WOBJ_ControlImageGraphicsGrHandle(lastGraphics);
   SetDrawWindow(WOBJ_ControlImageGraphicsGrHandle(lastGraphics));
   currentForeColor = WOBJ_GraphicsForeRGB(lastGraphics);
   currentBackColor = WOBJ_GraphicsBackRGB(lastGraphics);
   currentUserFont = WOBJ_GraphicsFont(lastGraphics);
}
void drawHourGlass(int step)
{
   if (graphicsActive)
   {
      int32 c1,c2,c3,c4,c5;
      int x = (MaxScreenX - 160) / 2;
      int y = (MaxScreenY - 160) / 2;
      if (maxColors >= 256)
      {
	      c1 = (*getColor)(0,0,0);
	      c2 = (*getColor)(0,0,255);
	      c3 = (*getColor)(0,255,0);
	      c4 = (*getColor)(255,0,0);
	   }
	   else
      if (maxColors == 16)
      {
        c1 = 0xF;
        c2 = 0x2;
        c3 = 0x6;
        c4 = 0xA;
      }
      else
      {
	      c1 = 3;
	      c2 = 1;
	      c3 = 2;
	      c4 = 3;
	   }
      c5 = (*getColor)(255,255,255);
      if (step == 0)
      {
         // draw hourglass
         fillRect(x+76,y+73,9,14,c1);
         // top line
         drawLine(x+77, y+74, x+83, y+74,c5);
         // left indent
         drawLine(x+76, y+77, x+76, y+82,c5);
         drawLine(x+77, y+78, x+77, y+81,c5);
         drawLine(x+78, y+79, x+78, y+80,c5);
         // right indent
         drawLine(x+82, y+79, x+82, y+80,c5);
         drawLine(x+83, y+78, x+83, y+81,c5);
         drawLine(x+84, y+77, x+84, y+82,c5);
         // dots
         drawLine(x+80, y+80, x+80, y+80,c5);
         drawLine(x+80, y+82, x+80, y+82,c5);
         drawLine(x+79, y+83, x+79, y+83,c5);
         drawLine(x+81, y+83, x+81, y+83,c5);
         // top line
         drawLine(x+77, y+85, x+83, y+85,c5);
      }
      // draw rgb colors
      if (step == 0)
      {
	      fillRect(x+70,y+90,5,8,c4);
	      fillRect(x+75,y+90,5,8,c3);
	      fillRect(x+80,y+90,5,8,c2);
	      drawRect(x+70,y+90,20,8,c1);
	   }
	   else fillRect(x+70,y+90,20,8,c5);
   }
}
//////////////////////////////////////////////////////////////////////////
//              Functions called by the Java routines                   //
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void create()
void GraphicsCreate(Var stack[])
{
	WObject gr, surface;
	uint16 error=0;
	GrHandle h;
	int stype;
	GR_ENTER

	gr = stack[0].obj;
	surface = WOBJ_GraphicsSurface(gr);
   stype = SurfaceGetType(surface);

   if (stype == SURF_IMAGE)
  	   h = WOBJ_ControlImageGraphicsGrHandle(surface);
  	else
      h = CreateOffscreenWindow(gr, vmGlobals->screenWidth,vmGlobals->screenHeight,0,&error); // same bitmap of MainWin, but with its own clip and translate - guich@400_62: instead of MaxScreenX/Y, use screenW/H
  	if (!error)
  	{
      WOBJ_GraphicsSurfType(gr) = stype;
      WOBJ_ControlImageGraphicsGrHandle(gr) = h;
		WOBJ_GraphicsFont(gr) = vmGlobals->allFonts+0;
	   WOBJ_ControlImageGraphicsGrHandle(surface) = h; // guich@402_28
   }
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
void GraphicsDestroy(WObject obj)
{
	if (obj != NULL && WOBJ_ControlImageGraphicsGrHandle(obj) != NULL) // guich@210_7: fixed when an Image Graphics was the last used and it was freed by the GC.
	{
	   GrHandle gr = WOBJ_ControlImageGraphicsGrHandle(obj);
	   if (oldGrHandle == gr) // guich@402_18: are we freeing the current stored grhandle? (used in imageLoadBmpCompressed)
	      oldGrHandle = mainGrHandle;
	   if (WOBJ_GraphicsSurfType(obj) != SURF_IMAGE) // note: images cannot have their Graphics handle deleted!
   	{
   		WOBJ_ControlImageGraphicsGrHandle(obj) = NULL;
   		if (WOBJ_GraphicsSurfType(obj) != SURF_IMAGE) // guich@200b4_165: avoid freeing graphics pertinent to Images
   		   FreeGrHandle(gr);
   	}
   	else
      if (gr == activeGrHandle) // guich@200b3: is this the active GrHandle?
      {
         if (mainGrHandle)
            SetDrawWindow(mainGrHandle); // switch to other handle
         lastGraphics = NULL;
         lastGrHandle = NULL;
      }
   }
}
// waba/fx/Graphics void free()
void GraphicsFree(Var stack[]) // ()V
{
	GR_ENTER
	GraphicsDestroy(stack[0].obj); // class object
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics static void nativeApplyPalette(int []palrgb, int startIndex, int endIndex)
void Graphics_applyPalette(Var stack[]) // ([III)V
{
	WObject intArray;
	int32 *ints;
	int startIndex,endIndex;
	GR_ENTER
	intArray = stack[0].obj;
	startIndex = stack[1].intValue;
	endIndex = stack[2].intValue;
   ints = (int32 *)WOBJ_arrayStart(intArray);
   setPalette(ints,startIndex,endIndex);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void eraseRect(int x, int y, int w, int h)
void GraphicsEraseRect(Var stack[]) // (IIII)V
{
   GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   eraseRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipse(int xc, int yc, int rx, int ry)
void GraphicsDrawEllipse(Var stack[]) // (IIII)V
{
   GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawEllipse(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillEllipse(int xc, int yc, int rx, int ry)
void GraphicsFillEllipse(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   fillEllipse(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentBackColor); // guich@340_41: fixed bug that was using currentForeColor instead of currentBackColor
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawArc(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsDrawArc(Var stack[]) // (IIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawArc(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].floatValue,stack[5].floatValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPie(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsDrawPie(Var stack[]) // (IIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawPie(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].floatValue,stack[5].floatValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillPie(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsFillPie(Var stack[]) // (IIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   fillPie(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].floatValue,stack[5].floatValue,currentForeColor,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipticalArc(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsDrawEllipticalArc(Var stack[]) // (IIIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawEllipticalArc(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].floatValue,stack[6].floatValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsDrawEllipticalPie(Var stack[]) // (IIIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawEllipticalPie(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].floatValue,stack[6].floatValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsFillEllipticalPie(Var stack[]) // (IIIIFF)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   fillEllipticalPie(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].floatValue,stack[6].floatValue,currentForeColor,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCircle(int xc, int yc, int r)
void GraphicsDrawCircle(Var stack[]) // (III)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawCircle(stack[1].intValue,stack[2].intValue,stack[3].intValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillCircle(int xc, int yc, int r)
void GraphicsFillCircle(Var stack[]) // (III)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   fillCircle(stack[1].intValue,stack[2].intValue,stack[3].intValue,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Coord getAnglePoint(int xc, int yc, int rx, int ry, float angle)
void GraphicsGetAnglePoint(Var stack[]) // (IIIIF)waba.fx.Coord
{
	WObject coord;
	int32 ints[2];
	GR_ENTER

   SELECT_GRAPHICS(stack[0].obj);

   coord = (*vmGlobals->createObject)((*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/fx/Coord"),1));
   if (!coord)
   {
      GR_EXIT
      RETURN32(0)
   }
   if ((*vmGlobals->pushObject)(coord) == -1)  // make sure it doesn't get GC'd
   {
      GR_EXIT
      RETURN32(0)
   }
   getAnglePoint(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].floatValue, ints);
   WOBJ_CoordX(coord) = ints[0];
   WOBJ_CoordY(coord) = ints[1];
	(*vmGlobals->popObject)(); // coord
	GR_EXIT
	RETURN32(coord)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics int getPixel (int x, int y)
void GraphicsGetPixel(Var stack[]) // (II)I
{
	int32 v;
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   v = (*getPixel)(stack[1].intValue,stack[2].intValue);
   GR_EXIT
	RETURN32(v)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setPixel (int x, int y)
void GraphicsSetPixel(Var stack[]) // (II)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   (*setPixel)(stack[1].intValue,stack[2].intValue, currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawLine(int Ax, int Ay, int Bx, int By)
void GraphicsDrawLine(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawLine(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void clearScreen()
void GraphicsClearScreen(Var stack[]) // ()V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   clearScreen(currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDots(int Ax, int Ay, int Bx, int By)
void GraphicsDrawDots(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawDottedLine(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCursor(int x, int y, int w, int h)
void GraphicsDrawCursor(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawCursor(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCursorOutline(int x, int y, int w, int h)
void GraphicsDrawCursorOutline(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawCursorOutline(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDottedCursor(int x, int y, int w, int h)
void GraphicsDrawDottedCursor(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawDottedCursor(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawRect(int x, int y, int w, int h)
void GraphicsDrawRect(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillRect(int x, int y, int w, int h)
void GraphicsFillRect(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   if (stack[3].intValue == maxX && stack[4].intValue == maxY && transXY == 0 && stack[1].intValue == 0 && stack[2].intValue == 0 && CLIPX1 == 0 && CLIPY1 == 0 && CLIPX2 == maxX && CLIPY2 == maxY) // guich@210_6: forgot to test agains clip boundaries.
      clearScreen(currentBackColor); // guich@200b4_10 - fix wrong window paint
   else
      fillRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDottedRect(int x, int y, int w, int h)
void GraphicsDrawDottedRect(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawDottedRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,currentForeColor,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillPolygon(int []xPoints, int []yPoints, int nPoints)
void GraphicsFillPolygon(Var stack[]) // ([I[II)V
{
	WObject xPoints,yPoints;
	int32 *xp;
	int32 *yp;
	int32 nPoints;
	GR_ENTER

   SELECT_GRAPHICS(stack[0].obj);
   xPoints = stack[1].obj;
   yPoints = stack[2].obj;
	nPoints = stack[3].intValue;
	xp = (int32 *)WOBJ_arrayStart(xPoints);
	yp = (int32 *)WOBJ_arrayStart(yPoints);

   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) || !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange,"In fillPolygon");
   else
      fillPolygon(xp,yp,nPoints,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPolygon(int []xPoints, int []yPoints, int nPoints)
void GraphicsDrawPolygon(Var stack[]) // ([I[II)V
{
	WObject xPoints,yPoints;
	int32 *xp;
	int32 *yp;
	int32 nPoints;
	GR_ENTER

   SELECT_GRAPHICS(stack[0].obj);
   xPoints = stack[1].obj;
   yPoints = stack[2].obj;
	nPoints = stack[3].intValue;
	xp = (int32 *)WOBJ_arrayStart(xPoints);
	yp = (int32 *)WOBJ_arrayStart(yPoints);

   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) || !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange, "In drawPolygon");
   else
      drawPolygon(xp,yp,nPoints,currentForeColor,1);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPolyline(int []xPoints, int []yPoints, int nPoints)
void GraphicsDrawPolyline(Var stack[]) // ([I[II)V
{
	WObject xPoints,yPoints;
	int32 *xp;
	int32 *yp;
	int32 nPoints;
	GR_ENTER

   SELECT_GRAPHICS(stack[0].obj);
   xPoints = stack[1].obj;
   yPoints = stack[2].obj;
	nPoints = stack[3].intValue;
	xp = (int32 *)WOBJ_arrayStart(xPoints);
	yp = (int32 *)WOBJ_arrayStart(yPoints);

   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) || !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange, "In drawPolyline");
   else
      drawPolygon(xp,yp,nPoints,currentForeColor,0);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawRoundRect(int x, int y, int width, int height, int r)
void GraphicsDrawRoundRect(Var stack[]) // (IIIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   drawRoundRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].intValue,currentForeColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillRoundRect(int x, int y, int width, int height, int r)
void GraphicsFillRoundRect(Var stack[]) // (IIIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   fillRoundRect(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].intValue,currentBackColor);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawText(String text, int x, int y)
void GraphicsDrawText(Var stack[]) // (Ljava/lang/String;II)V
{
   WObject charArray;
   uint16 *charStart;
   uint32 len;

   GR_ENTER
   if (stack[1].obj == 0) // kdarabi@400_84: null Strings lead to reset
      return;
	// guich@200b4_91: use the uint16* charArray directly
   charArray = WOBJ_StringCharArrayObj(stack[1].obj);
   if (charArray == 0)
      return;
   charStart = (uint16 *)WOBJ_arrayStart(charArray);
   len = WOBJ_arrayLen(charArray);

   SELECT_GRAPHICS(stack[0].obj);
	drawText(charStart,len,stack[2].intValue,stack[3].intValue,currentForeColor);
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics static void drawText(char []chars, int start, int count, int x, int y)
void GraphicsDrawTextChars(Var stack[]) // ([CIIII)V  - guich@402_29
{
   WObject charArray;
   uint16 *charStart;
   uint32 len,start,strLen;

   GR_ENTER
   if (stack[1].obj == 0) // kdarabi@400_84: null Strings lead to reset
      return;
	// guich@200b4_91: use the uint16* charArray directly
   charArray = stack[1].obj;
   charStart = (uint16 *)WOBJ_arrayStart(charArray);
   strLen = WOBJ_arrayLen(charArray);
   start = stack[2].intValue;
   len = stack[3].intValue;
   if (start < 0 || len < 0 || (start+len) > strLen)
      return;

   SELECT_GRAPHICS(stack[0].obj);
	drawText(&charStart[start],len,stack[4].intValue,stack[5].intValue,currentForeColor);
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setClip(int x, int y, int w, int h)
void GraphicsSetClip(Var stack[]) // (IIII)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   setClip(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Rect getClip(waba.fx.Rect r)
void GraphicsGetClip(Var stack[]) // (Lwaba/fx/Rect;)Lwaba/fx/Rect;
{
	WObject rect;

   GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
	rect = stack[1].obj;
	if (rect == 0)
	{
	   GR_EXIT
		RETURN32(0)
	}

	WOBJ_RectX(rect) = CLIPX1-TRANSX;
	WOBJ_RectY(rect) = CLIPY1-TRANSY;
	WOBJ_RectWidth(rect)  = CLIPX2-CLIPX1; // guich@300_36: removed +1 here and line below
	WOBJ_RectHeight(rect) = CLIPY2-CLIPY1;
	GR_EXIT
	RETURN32(rect)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void clearClip()
void GraphicsClearClip(Var stack[]) // ()V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   setClip(0,0,GRHANDLE_Width(activeGrHandle),GRHANDLE_Height(activeGrHandle));
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics boolean clip(waba.fx.Rect r)
void GraphicsClip(Var stack[]) // (Lwaba/fx/Rect;)Z
{
	int32 v;
	WObject rect;
	int x,y,w,h;
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
	rect = stack[1].obj;
	if (rect == 0)
	{
	   GR_EXIT
		RETURN32(0)
   }
	x = WOBJ_RectX(rect);
	y = WOBJ_RectY(rect);
	w = WOBJ_RectWidth(rect);
	h = WOBJ_RectHeight(rect);
   v = translateAndClip(&x,&y,&w,&h);
   if (v) // intersect?
   {
		WOBJ_RectX(rect) = x-TRANSX;
		WOBJ_RectY(rect) = y-TRANSY;
		WOBJ_RectWidth(rect) = w;
		WOBJ_RectHeight(rect) = h;
   }
   GR_EXIT
	RETURN32(v)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void translate(int dx, int dy)
void GraphicsTranslate(Var stack[]) // (II)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   if (stack[1].intValue || stack[2].intValue)
      translate(stack[1].intValue,stack[2].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// used by the drawImage method
// waba/fx/Graphics void copyRect(waba.fx.ISurface surface, int x, int y, int width, int height, int dstX, int dstY)
void GraphicsCopyRect(Var stack[]) // (Lwaba/fx/ISurface;IIIIII)V
{
	WObject surfOrig,surfDest;
	GrHandle hOrig,hDest;
	surfOrig = stack[1].obj;
	surfDest = stack[0].obj;
	if (!surfOrig || !surfDest) return;
	hOrig = WOBJ_ControlImageGraphicsGrHandle(surfOrig);
   if (!hOrig) return;
   hDest = WOBJ_ControlImageGraphicsGrHandle(surfDest);
   if (!hDest) return;
   SELECT_GRAPHICS(surfDest);
   GR_ENTER
   (*drawBitmap)(hOrig, stack[2].intValue, stack[3].intValue, stack[4].intValue, stack[5].intValue, stack[6].intValue, stack[7].intValue, WOBJ_GraphicsDrawOp(surfDest), currentBackColor, currentForeColor,1);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void copyScreen(waba.fx.ISurface src, int ySrc, int yDst, int height)
void GraphicsCopyScreen(Var stack[]) // (Lwaba/fx/ISurface;III)V
{
	WObject surfOrig,surfDest;
	GrHandle hOrig,hDest;

	surfOrig = stack[1].obj;
	surfDest = stack[0].obj;
	if (!surfOrig || !surfDest) return;
	hOrig = WOBJ_ControlImageGraphicsGrHandle(surfOrig);
   if (!hOrig) return;
   hDest = WOBJ_ControlImageGraphicsGrHandle(surfDest);
   if (!hDest) return;
   GR_ENTER
#ifdef WIN32
	SELECT_GRAPHICS(surfDest);
#endif
   (*drawScreen)(hOrig, hDest, stack[2].intValue, stack[3].intValue, stack[4].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
WObject lastColor;
int32 lastColorValue,lastTranspPixelDest,lastTranspPixelOrig=-1;
// waba/fx/Graphics void drawImage(waba.fx.Image image, int x, int y, int drawOp, waba.fx.Color backColor, boolean doClip)
void GraphicsDrawImageSprite(Var stack[]) // (Lwaba/fx/Image;IIILwaba/fx/Color;Z)V
{
	WObject surfOrig,surfDest;
	GrHandle hOrig,hDest;
   int32 transpPixel;

	surfOrig = stack[1].obj;   // DESCOBRIR QUANDO QUE O 0 QUER DIZER TRANSPARENTE PRETO OU COR BRANCA (PALETTE INDEX)
	surfDest = stack[0].obj;
	if (!surfOrig || !surfDest) return;
   hOrig = WOBJ_ControlImageGraphicsGrHandle(surfOrig);
   if (!hOrig) return;
   hDest = WOBJ_ControlImageGraphicsGrHandle(surfDest);
   if (!hDest) return;
   // petrus@402_01
   transpPixel = WOBJ_ImageTranspPixel(surfOrig);
#ifdef PALMOS
   if (displayBPP > 8) // guich@421a_65: if in 16bpp, convert the color to the index
   {
      if (transpPixel == lastTranspPixelOrig)
         transpPixel = lastTranspPixelDest;
      else
      {
         lastTranspPixelOrig = transpPixel;
         transpPixel = (*getColor)(GET_RED(transpPixel),GET_GREEN(transpPixel),GET_BLUE(transpPixel));
         lastTranspPixelDest = transpPixel;
      }
   }
#endif
   GR_ENTER
   SELECT_GRAPHICS(surfDest);
   if ((stack[5].obj) && (stack[4].intValue != 0) && ((transpPixel != lastColorValue) || (stack[5].obj != lastColor)))
   {
      lastColor = stack[5].obj;
      transpPixel = lastColorValue = (*getColor)(WOBJ_ColorRedP(lastColor),WOBJ_ColorGreenP(lastColor),WOBJ_ColorBlueP(lastColor));
   }
   if (transpPixel < 0)               // defense! - guich: really necessary?
      transpPixel = lastColorValue;

   (*drawBitmap)(hOrig, 0, 0, WOBJ_ImageWidth(surfOrig), WOBJ_ImageHeight(surfOrig), stack[2].intValue, stack[3].intValue, stack[4].intValue, transpPixel, currentForeColor, stack[6].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void copyImageRect(waba.fx.Image image, int x, int y, int width, int height, int drawOp, waba.fx.Color backColor, boolean doClip)
void GraphicsCopyImageRect(Var stack[]) // (Lwaba/fx/Image;IIIIILwaba/fx/Color;Z)V
{
	WObject surfOrig,surfDest;
	GrHandle hOrig,hDest;
   int32 transpPixel;

	surfOrig = stack[1].obj;
	surfDest = stack[0].obj;
	if (!surfOrig || !surfDest) return;
   hOrig = WOBJ_ControlImageGraphicsGrHandle(surfOrig);
   if (!hOrig) return;
   hDest = WOBJ_ControlImageGraphicsGrHandle(surfDest);
   if (!hDest) return;
   // petrus@402_01
   transpPixel = WOBJ_ImageTranspPixel(surfOrig);
   GR_ENTER
   SELECT_GRAPHICS(surfDest);
#ifdef PALMOS
   if (displayBPP > 8) // guich@421a_65: if in 16bpp, convert the color to the index
   {
      if (transpPixel == lastTranspPixelOrig)
         transpPixel = lastTranspPixelDest;
      else
      {
         lastTranspPixelOrig = transpPixel;
         transpPixel = (*getColor)(GET_RED(transpPixel),GET_GREEN(transpPixel),GET_BLUE(transpPixel));
         lastTranspPixelDest = transpPixel;
      }
   }
#endif
   if ((stack[7].obj) && (stack[6].intValue != 0) && ((transpPixel != lastColorValue) || (stack[7].obj != lastColor)))
   {
      lastColor = stack[7].obj;
      transpPixel = lastColorValue = (*getColor)(WOBJ_ColorRedP(lastColor),WOBJ_ColorGreenP(lastColor),WOBJ_ColorBlueP(lastColor));
   }
   if (transpPixel < 0)               // defense!
      transpPixel = lastColorValue;
   (*drawBitmap)(hOrig, stack[2].intValue, stack[3].intValue, stack[4].intValue, stack[5].intValue,
       0, 0, stack[6].intValue, transpPixel, currentForeColor, stack[8].intValue);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsSetForeColorC(Graphics obj, waba.fx.Color c)
void GraphicsSetForeColor(Var stack[]) // (Lwaba/fx/Graphics;Lwaba/fx/Color;)V
{
	Var *c;
	c = stack[1].obj; // class object
   GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   currentForeColor = (*getColor)(WOBJ_ColorRedP(c),WOBJ_ColorGreenP(c),WOBJ_ColorBlueP(c));
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsSetBackColorC(Graphics obj, waba.fx.Color c)
void GraphicsSetBackColor(Var stack[]) // (Lwaba/fx/Graphics;Lwaba/fx/Color;)V
{
	Var *c;
	GR_ENTER
	c = stack[1].obj; // class object
   SELECT_GRAPHICS(stack[0].obj);
   currentBackColor = (*getColor)(WOBJ_ColorRedP(c),WOBJ_ColorGreenP(c),WOBJ_ColorBlueP(c));
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setFont(waba.fx.Font font)
void GraphicsSetFont(Var stack[]) // (Lwaba/fx/Font;)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
   currentUserFont = WOBJ_FontUserFont(stack[1].obj);
   WOBJ_GraphicsLastFont(stack[0].obj) = stack[1].obj;
   WOBJ_GraphicsFont(stack[0].obj) = currentUserFont;
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics boolean collisionDetect(waba.fx.ISurface surface, int x1, int y1, waba.fx.Image img, int x2, int y2)
void GraphicsCollisionDetect(Var stack[]) // (Lwaba/fx/ISurface;IILwaba/fx/Image;II)Z
{
	WObject surface;
	GrHandle h1,h2;
	RectangleType r1,r2;
	GR_ENTER
	surface = stack[1].obj;
   SELECT_GRAPHICS(stack[0].obj);
	r1.topLeft.x = stack[2].intValue;
	r1.topLeft.y = stack[3].intValue;
	r2.topLeft.x = stack[5].intValue;
	r2.topLeft.y = stack[6].intValue;
	if (!surface)
	   h1 = mainGrHandle;
	else
   if (SurfaceGetType(surface) == SURF_IMAGE)
   {
	   h1 = WOBJ_ControlImageGraphicsGrHandle(surface);
	   r1.extent.x = WOBJ_ImageWidth(surface);
	   r1.extent.y = WOBJ_ImageHeight(surface);
	}
	else
	{
      h1 = WOBJ_ControlImageGraphicsGrHandle(surface);
      r1.extent.x = GRHANDLE_Width(h1) - r1.topLeft.x; // guich@200b2: sub the width
      r1.extent.y = GRHANDLE_Height(h1) - r1.topLeft.y; // guich@200b2: sub the height
   }
   surface = stack[4].obj;
   h2 = WOBJ_ControlImageGraphicsGrHandle(surface);
   r2.extent.x = WOBJ_ImageWidth(surface);
   r2.extent.y = WOBJ_ImageHeight(surface);

	GR_EXIT
	RETURN32(collisionDetect(h1,r1,h2,r2,currentBackColor))
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Coord getTranslation()
void GraphicsGetTranslation(Var stack[]) // ()Lwaba/fx/Coord;
{
	WObject coord;
	GR_ENTER

   SELECT_GRAPHICS(stack[0].obj);
   coord = (*vmGlobals->createObject)((*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/fx/Coord"),1));
   if (!coord)
      RETURN32(0)
   WOBJ_CoordX(coord) = TRANSX;
   WOBJ_CoordY(coord) = TRANSY;
	GR_EXIT
	RETURN32(coord)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Palette int []getCurrentPalette()
void PaletteGetCurrentPalette(Var stack[]) // ()[I
{
	WObject intArray;
	int32 *ints;

	intArray = (*vmGlobals->createArrayObject)(10, 256);
	if (intArray)
   {
	   ints = (int32 *)WOBJ_arrayStart(intArray);
	   xmemmove(ints,palette,256*4);
		RETURN32(intArray)
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsDrawHatchedRect(Graphics obj, int x, int y, int width, int height, boolean top, boolean bottom)
void GraphicsDrawHatchedRect(Var stack[]) // (Lwaba/fx/Graphics;IIIIZZ)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
	drawHatchedRect((int)stack[1].intValue,(int)stack[2].intValue,(int)stack[3].intValue,(int)stack[4].intValue,(byte)stack[5].intValue,(byte)stack[6].intValue,(int32)currentForeColor);
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsFillHatchedRect(Graphics obj, int x, int y, int width, int height, boolean top, boolean bottom)
void GraphicsFillHatchedRect(Var stack[]) // (Lwaba/fx/Graphics;IIIIZZ)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
	fillHatchedRect((int)stack[1].intValue,(int)stack[2].intValue,(int)stack[3].intValue,(int)stack[4].intValue,(byte)stack[5].intValue,(byte)stack[6].intValue,(int32)currentBackColor);
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsDrawHighLightFrame(Graphics obj, int x, int y, int w, int h, waba.fx.Color topLeft, waba.fx.Color bottomRight, boolean yMirror)
void GraphicsDrawHighLightFrame(Var stack[]) // (Lwaba/fx/Graphics;IIIILwaba/fx/Color;Lwaba/fx/Color;Z)V
{
	GR_ENTER
   SELECT_GRAPHICS(stack[0].obj);
	drawHighLightFrame(stack[1].intValue,stack[2].intValue,stack[3].intValue,stack[4].intValue,stack[5].obj,stack[6].obj,stack[7].intValue);
	GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
void ImageDestroy(WObject image)
{
	GrHandle h;

	h = WOBJ_ControlImageGraphicsGrHandle(image);
	if (h == NULL)
		return;
	DestroyOffscreenWindow(h);
	WOBJ_ControlImageGraphicsGrHandle(image) = NULL;
}


void ImageFree(Var stack[])
{
	WObject image;
	image = stack[0].obj;
	ImageDestroy(image);
	WOBJ_ImageWidth(image) = 0;
	WOBJ_ImageHeight(image) = 0;
}

void ImageLoad(Var stack[])
{
	WObject image;
	UtfString path;
	WObject pathString;
	uint16 pathLen;
	uchar *p;

   GR_ENTER
	image = stack[0].obj;
	// NOTE: we don't have to free an existing bitmap because this is only called
	// from an image constructor
	WOBJ_ControlImageGraphicsGrHandle(image) = NULL;
	WOBJ_ImageWidth(image) = 0;
	WOBJ_ImageHeight(image) = 0;
	pathString = stack[1].obj;
	path = (*vmGlobals->stringToUtf)(pathString, STU_USE_STATIC);
	if (path.len == 0)
		return;

	pathLen = (uint16)path.len;
	p = (*vmGlobals->lockWarpRec)(path.str, (uint16)path.len, NULL);
	if (p == NULL)
		return;
	ImageLoadBMPCompressed(image, p);
	(*vmGlobals->unlockWarpRec)(p, pathLen);
	GR_EXIT
}
// waba/fx/NativeMethods void imageParse(Image obj, byte []fullBmpDescription)
void ImageParse(Var stack[]) // (Lwaba/fx/Image;[B)V
{
	WObject image = stack[0].obj; // class object
	GR_ENTER
	ImageLoadBMPCompressed(image, (byte *)WOBJ_arrayStart(stack[1].obj));
	GR_EXIT
}
void ImageCreate(Var stack[])
{
	WObject image;
	int16 width, height;
	GrHandle imgHandle;
	uint16 error;

   GR_ENTER
	image = stack[0].obj;
	width = (int16)WOBJ_ImageWidth(image);
	height = (int16)WOBJ_ImageHeight(image);
	if (width > 0 && height > 0)
	{
		imgHandle = CreateOffscreenWindow(image, width, height,1, &error);
		if (error != 0)
			imgHandle = NULL;
	}
	else
		imgHandle = NULL;
	WOBJ_ControlImageGraphicsGrHandle(image) = imgHandle;
	GR_EXIT
}

// waba/fx/Image void nativeChangeColors(int fromR, int fromG, int fromB, int toR, int toG, int toB)
void ImageChangeColors(Var stack[]) // (IIIIII)V
{
	WObject image = stack[0].obj; // class object
	int width = (int)WOBJ_ImageWidth(image);
	int height = (int)WOBJ_ImageHeight(image);
	GrHandle h = WOBJ_ControlImageGraphicsGrHandle(image);
	int32 colorFrom = (*getColor)(stack[1].intValue,stack[2].intValue,stack[3].intValue);
	int32 colorTo = (*getColor)(stack[4].intValue,stack[5].intValue,stack[6].intValue);
	if (!h) return;
	GR_ENTER
   imageChangeColors(h, width, height, colorFrom, colorTo);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
int scaleImage(GrHandle srcG, uint16 width, uint16 height, GrHandle dstG, uint16 newWidth, uint16 newHeight);

// waba/fx/NativeMethods waba.fx.Image imageGetScaledInstance(Image obj, int newWidth, int newHeight)
void ImageGetScaledInstance(Var stack[]) // (Lwaba/fx/Image;II)Lwaba/fx/Image;
{
   int32 v;
	WObject imageSrc,imageDst;
	int16 width, height, newWidth, newHeight;
	GrHandle srcG,dstG;
	uint16 error;

   // get the source data
	imageSrc = stack[0].obj;
	newWidth = (int16)stack[1].intValue;
	newHeight = (int16)stack[2].intValue;
	width = (int16)WOBJ_ImageWidth(imageSrc);
	height = (int16)WOBJ_ImageHeight(imageSrc);
	if (width > 0 && height > 0)
	{
	   static WClass *imgClass = NULL;
	   if (imgClass == NULL)
	   {
	      imgClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/fx/Image"),1);
	      if (imgClass == NULL)
	         RETURN32(0)
	   }
	   // create the new image object with the given sizes
      imageDst = (*vmGlobals->createObject)(imgClass);
      if (imageDst == NULL || (*vmGlobals->pushObject)(imageDst) == -1) // make sure it doesn't get GC'd
         RETURN32(0)
      GR_ENTER
		dstG = CreateOffscreenWindow(imageDst, newWidth, newHeight,1, &error);
   	GR_EXIT
		if (error != 0) // no more memory?
		   goto end;
		WOBJ_ImageWidth(imageDst) = newWidth;
		WOBJ_ImageHeight(imageDst) = newHeight;
   	WOBJ_ControlImageGraphicsGrHandle(imageDst) = dstG;
      WOBJ_ImageTranspPixel(imageDst) = WOBJ_ImageTranspPixel(imageSrc); // guich@402_8: forgot to copy those attributes - petrus@402_01
      WOBJ_ImagePalette(imageDst) = WOBJ_ImagePalette(imageSrc);

   	srcG = WOBJ_ControlImageGraphicsGrHandle(imageSrc);
   	if (!scaleImage(srcG,width,height, dstG,newWidth,newHeight))
   	{
   		WOBJ_ImageWidth(imageDst) = WOBJ_ImageHeight(imageDst) = 0; // guich@401_30
   	}
   	v = (int32)imageDst;
	}
end:
   (*vmGlobals->popObject)(); // guich@402_24: moved to here, bc the scaleImage could trigger the GC and clean up this object
	RETURN32(v)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Image void getPixelRow(byte []fillIn, int y)
void ImageGetPixelRow(Var stack[]) // ([BI)V
{
	WObject image = stack[0].obj; // class object
	WObject fillIn;
	int y;
	byte *fi;

   fillIn = stack[1].obj;
   y = stack[2].intValue;
	fi = (byte *)WOBJ_arrayStart(fillIn);

   GR_ENTER
   fillInRow(WOBJ_ControlImageGraphicsGrHandle(image),fi,y);
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// guich@401_38: added by petrus to give more flexibility in implementing image decoders

void drawRow(int xStart, int xEnd, int y, unsigned char * rgbOrIndex, uint32 * colorMap)
{
#ifdef PALMOS_AUTO_RESET // guich@420_13
   static int called;
   if (called-- == 0)
   {
      SemaphoreRelease(1);
      SemaphoreReserve(1);
      called = 30;
   }
#endif
   if (colorMap == 0)
   {
      // if colorMap is null, then rgbOrIndex is a 24 bits RGB.
      #if defined(PALMOS5) || defined(PALMOS_AUTO_RESET) // petrus@421_1
      // guich: at simulator, no speed gain, but at device, about 8% gained with windrawline - well, it really depends on the picture
/*      RGBColorType rgbct;
      rgbct.index = 0;
      while (xStart < xEnd)
      {
         rgbct.r = rgbOrIndex[0];
         rgbct.g = rgbOrIndex[1];
         rgbct.b = rgbOrIndex[2];
         WinSetForeColorRGB(&rgbct, NULL);
         WinDrawPixel(xStart++, y);
         rgbOrIndex += 3;
      }*/
      // slower in photos taken from zire71
      typedef union
      {
         RGBColorType rgbct;
         int32 i;
      } RGBi;
      RGBi rgbi,last;
      int saved =0;
      int lastX=0;
      last.i = -1;
      rgbi.rgbct.index = 0;
      while (xStart < xEnd)
      {
         rgbi.rgbct.r = rgbOrIndex[0];
         rgbi.rgbct.g = rgbOrIndex[1];
         rgbi.rgbct.b = rgbOrIndex[2];
         if (last.i != rgbi.i)
         {
            WinSetForeColorRGB(&last.rgbct, 0);
            WinDrawLine(lastX,y,xStart,y);
            lastX = xStart;
            last.i = rgbi.i;
         }
         xStart++;
         rgbOrIndex += 3;
      }
      WinSetForeColorRGB(&last.rgbct, 0);
      WinDrawLine(lastX,y,xStart,y);
      #else
      while (xStart < xEnd)
      {
         #ifdef PALMOS // petrus@420_2
         uint32 k = 0xFFFFFF - (
            (((uint32)(rgbOrIndex[2])) << 16) |
            (((uint32)(rgbOrIndex[1])) << 8) |
            rgbOrIndex[0]
         );
         #else
         uint32 k = (
            (((uint32)(rgbOrIndex[2])) << 16) |
            (((uint32)(rgbOrIndex[1])) << 8) |
            rgbOrIndex[0]
         );
         #endif
         (*setPixelXM)(xStart++, y, k);
         rgbOrIndex += 3;
      }
      #endif
   }
   else
   {
      // if colorMap is not null, then rgbOrIndex is an index into the colormap.
      // guich: changed the old algorithm to use a scan-line approach
      byte lastColor = *rgbOrIndex;
      int dstX = xStart;
      int width = xEnd-xStart;
      int dstY = y;
      int lastI=0, i=1;
      for (; i < width; i++)
         if (rgbOrIndex[i] != lastColor)
         {
            (*drawHLineNOCT)(dstX+lastI,dstY,i-lastI,(int)lastColor);
            lastI = i;
            lastColor = rgbOrIndex[i];
         }
      (*drawHLineNOCT)(dstX+lastI,dstY,i-lastI,(int)lastColor);

      //while (xStart < xEnd)
        // setPixel(xStart++, y, *rgbOrIndex++);
   }
}

void graphicsStartImage(uint32 * ret, uint32 * args)     /* PUBLIC_GraphicsStartImage */
{
   // NOTE: it is assumed that the image is brand new, and was created using a 'NULL' constructor, so we don't have to free an existing bitmap.
   WObject image = (WObject)args[0];
   int width = WOBJ_ImageWidth(image);
   int height = WOBJ_ImageHeight(image);

   *ret = 0;
   GR_ENTER
   if (width > 0 && height > 0)
   {
      uint16 error;
      GrHandle oldG, newG;

      DrawRowProcType *pDrawRow = (DrawRowProcType *)args[2];
      newG = CreateOffscreenWindow(image, width, height, 2, &error);
      if (error != 0) return; // no more memory?
      WOBJ_ControlImageGraphicsGrHandle(image) = newG;
      oldG = activeGrHandle;
      SetDrawWindow(newG);   // this does "registerPrimitives"
      *pDrawRow = drawRow;
      *(GrHandle*)ret = oldG;
      if (args[1] != 0 && vmGlobals->screenBPP <= 8)       // redoTransparentPixel - only if not highcolor!
      {
         /*
         | redoTransparentPixel means that
         | 1) a transparent Pixel has been defined for this image; and,
         | 2) its value is a RGB 888 value (not an index in the colormap).
         |
         | Windoze (incl. CE) implementation takes the color as BGR 888.
         | We assume that other implemntation takes the transparent Pixel
         | as RGB 565  (FIXME?)
         */
         uint32 tp = WOBJ_ImageTranspPixel(image);
         #if (defined WINCE || defined WIN32)
         tp = (((tp & 0xFF) << 16) + (tp & 0xFF00) + (tp >> 16));
         #else
         tp = (((tp >> 8) & 0xF800) + ((tp >> 5) & 0x7E0) + ((tp >> 3) & 0x1F));
         #endif
         WOBJ_ImageTranspPixel(image) = tp;
      }
   }
}

void graphicsEndImage(GrHandle oldG)    /* PUBLIC_GraphicsEndImage */
{
   GR_EXIT
   if (oldG) SetDrawWindow(oldG);
}
// waba/fx/NativeMethods waba.fx.Image imageGetRotatedScaledInstance(waba.fx.Image obj, int percScale, int angle, waba.fx.Color fill)
void ImageGetRotatedScaledInstance(Var stack[]) // (Lwaba/fx/Image;IILwaba/fx/Color;)Lwaba/fx/Image;
{
   RETURN32(stack[0].intValue)
#if 0
   WObject imageIn = stack[0].obj;
   Sint32 perScale = stack[1].intValue;
   Sint32 angle = stack[2].intValue;
   WObject color = stack[3].obj;

   SDL_Surface * surfIn = setCompatibleSurface(imageIn);
   Pixel * pixelsIn = (Pixel *)surfIn->pixels;
   int pitchIn = surfIn->pitch;
   int hIn = surfIn->h;
   int wIn = surfIn->w;

   Uint32 backColor = ((!color) ? WOBJ_ImageTranspPixel(imageIn)
                                : SDL_MapRGB(surfIn->format, (Uint8)WOBJ_ColorRedP(color),
                                             (Uint8)WOBJ_ColorGreenP(color), (Uint8)WOBJ_ColorBlueP(color)));
   // if (perScale <= 0) return; // defense! (done on the java side)

   /* xplying by 0x10000 allow integer math, while not loosing much prec. */

   Sint32 rawSine;
   Sint32 rawCosine;
   Sint32 sine;
   Sint32 cosine;

   WObject imageOut;
   SDL_Surface * surfOut;
   int pitchOut;
   Pixel * pixelsOut;
   int hOut, wOut;
   Sint32 x0, y0;
   int i;

   angle = angle % 360;
   if ((angle % 90) == 0)
   {
      if (angle < 0) angle += 360;
      switch (angle)
      {
         case 0:
            rawSine = sine = 0;
            rawCosine = 0x10000;
            cosine = 0x640000 / perScale;
            break;
         case 90:
            rawSine = 0x10000;
            rawCosine = cosine = 0;
            sine = 0x640000 / perScale;
            break;
         case 180:
            rawSine = sine = 0;
            rawCosine = -0x10000;
            cosine = -0x640000 / perScale;
            break;
         default: // case 270:
            rawSine = -0x10000;
            rawCosine = cosine = 0;
            sine = -0x640000 / perScale;
            break;
      }
   }
   else
   {
      double rad = angle * 0.0174532925;
      rawSine = (Sint32)(sin(rad) * 0x10000);
      rawCosine = (Sint32)(cos(rad) * 0x10000);
      sine = (rawSine * 100) / perScale;
      cosine = (rawCosine * 100) / perScale;
   }

   /* create imageOut */
   {
      GrHandle grhOut;
      int x[3], y[3];
      int xMin = 0;
      int yMin = 0;
      int xMax = 0;
      int yMax = 0;
      x[0] = (wIn * rawCosine) >> 16;
      y[0] = (wIn * rawSine) >> 16;
      x[2] = (-hIn * rawSine) >> 16;
      y[2] = (hIn * rawCosine) >> 16;
      x[1] = x[0] + x[2];
      y[1] = y[0] + y[2];

      i = sizeof x / sizeof x[0];
      while (--i >= 0)
      {
         if (x[i] < xMin)
            xMin = x[i];
         else
         if (x[i] > xMax)
            xMax = x[i];

         if (y[i] < yMin)
            yMin = y[i];
         else
         if (y[i] > yMax)
            yMax = y[i];
      }
      wOut = ((xMax - xMin) * perScale) / 100;
      hOut = ((yMax - yMin) * perScale) / 100;
      imageOut = createSimilarImage(imageIn, wOut, hOut);
      if (!imageOut)
         RETURN32(0);
      /* center */
      x0 = (((uint32)wIn<<16) - (((xMax-xMin) * rawCosine) - ((yMax-yMin) * rawSine)) - 1) / 2;
      y0 = (((uint32)hIn<<16) - (((xMax-xMin) * rawSine) + ((yMax-yMin) * rawCosine)) - 1) / 2;

      grhOut = WOBJ_ControlImageGraphicsGrHandle(imageOut);
      surfOut = (SDL_Surface *)GRHANDLE_WinHandle(grhOut);
      pixelsOut = (Pixel *)surfOut->pixels;
      pitchOut = surfOut->pitch;
   }

   /* and draw! */
   SDL_LockSurface(surfIn);
   SDL_LockSurface(surfOut);
   while (--hOut >= 0)
   {
      Sint32 x = x0;
      Sint32 y = y0;
      Pixel * out = pixelsOut;
      for (i=wOut; --i >= 0; x += cosine, y += sine)
      {
         Sint32 u = x>>16;
         Sint32 v = y>>16;
         if (0 <= u && u < wIn && 0 <= v && v < hIn)
            *out++ = *(Pixel *)(((char *)pixelsIn + ((v * pitchIn) + (u * sizeof(Pixel)))));
         else
            *out++ = backColor;
      }
      x0 -= sine;
      y0 += cosine;
      pixelsOut = (Pixel *)((char *)pixelsOut + pitchOut);
   }
   SDL_UnlockSurface(surfIn);
   SDL_UnlockSurface(surfOut);
   RETURN32(imageOut);  // success
#endif
}

/*-------------------------------------------------------computeContrastTable-+
| Generates 128 points from a kind of symmetrical                             |
| yn = sum{1 to inf}(e-xt/tn)/dt,                                             |
| n being proportional to 'level'.                                            |
| The contrast level must be in the range -127 ... + 127 excluded             |
+----------------------------------------------------------------------------*/
static void computeContrastTable(byte * table, char level)
{
#if 0
POW IS UNDEFINED
   int i = 0;
   double factor;
   byte * tbend = table+256;

   if (level < 0)
   {
      level += 127;
      if (level <= 0) level = 1;  // defense!
      factor = (double)level / 127.0;
   }
   else
   {
      level = 127-level;
      if (level <= 0) level = 1;  // defense!
      factor = 127.0 / (double)level;
   }
   do
   {
      *--tbend = 255 - (*table++ = (int)(127.0 * pow((double)i/127.0, factor)) & 0xff);
   } while (++i < 128);
#endif
}

/*--------------------------------------------------ImageGetTouchedUpInstance-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/fx/NativeMethods waba.fx.Image imageGetTouchedUpInstance(waba.fx.Image obj, byte brightness, byte contrast)
void ImageGetTouchedUpInstance(Var stack[]) // (Lwaba/fx/Image;BB)Lwaba/fx/Image;
{
   RETURN32(stack[0].intValue)
#if 0
   #define NO_TOUCHUP         0
   #define BRITE_TOUCHUP      1
   #define CTRST_TOUCHUP      2
   #define CTRSTBRITE_TOUCHUP 3  // OR'ed

   WObject imageIn = stack[0].obj;
   Sint8 iBrightness = stack[1].intValue;
   Sint8 iContrast = stack[2].intValue;
   Sint8 touchup = NO_TOUCHUP;
   SDL_Surface * surfIn = setCompatibleSurface(imageIn);

   Pixel * pixelsIn = (Pixel *)surfIn->pixels;
   SDL_PixelFormat * fmt = surfIn->format;
   int pitch = surfIn->pitch;
   int w = surfIn->w;
   int h = surfIn->h;

   WObject imageOut = createSimilarImage(imageIn, w, h);
   GrHandle grhOut;
   SDL_Surface * surfOut;
   Pixel * pixelsOut;
   Uint8 table[256];
   int m, k;
   int32 max;

   if (!imageOut)
      RETURN32(0);
   grhOut = WOBJ_ControlImageGraphicsGrHandle(imageOut);
   surfOut = (SDL_Surface *)GRHANDLE_WinHandle(grhOut);
   pixelsOut = surfOut->pixels;

   if (iContrast)
   {
      touchup |= CTRST_TOUCHUP;
      computeContrastTable(table, iContrast);
      /*
      | Table[0] must be more or less positive, and table[255]
      | more or less lesser than 255 when the contrast level is negative.
      */
      if (iContrast < 0)
      {
         table[255] = table[254];
         table[0] = table[1];
      }
   }
   if (iBrightness)
   {
      double brightness = ((float)iBrightness+128.0)/128.0;  // [0.0 ... 2.0]
      touchup |= BRITE_TOUCHUP;
      if (brightness <= 1.0)
      {
         m = (int)(sqrt(brightness) * 0x10000);
         k = 0;
         max = 0xFFFFFF;
      }
      else
      {
         double f;
         max = (int)(0xFF / brightness);
         f = brightness - 1.0;
         f = f * f;
         k = (int)(f * 0xFF0000);
         m = (int)((1.0-f) * brightness * 0x10000);
      }
   }
   #ifdef __GNUC__
   else
   {
      m = k = max = 0;  // remove a gcc warning
   }
   #endif
   SDL_LockSurface(surfIn);
   SDL_LockSurface(surfOut);
   if (fmt->palette)          // #if VIDEO_BITSPERPIXEL == 8
   {
      SDL_Color * colors = fmt->palette->colors;
      while (--h >= 0)
      {
         int i = w;
         Pixel * in = pixelsIn;
         Pixel * out = pixelsOut;
         switch (touchup)
         {
            case NO_TOUCHUP:
               while (--i >= 0)
                  *out++ = *in++;
               break;
            case BRITE_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  int r = color.r;
                  int g = color.g;
                  int b = color.b;
                  *out++ = SDL_MapRGB(fmt, (Uint8)((r > max)? 255 : ((((uint32)m*r) + k) >> 16)),
                                           (Uint8)((g > max)? 255 : ((((uint32)m*g) + k) >> 16)),
                                           (Uint8)((b > max)? 255 : ((((uint32)m*b) + k) >> 16)));
               }
               break;
            case CTRST_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  *out++ = SDL_MapRGB(fmt, table[color.r], table[color.g], table[color.b]);
               }
               break;
            default: // case CTRSTBRITE_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  int r = table[color.r];
                  int g = table[color.g];
                  int b = table[color.b];
                  *out++ = SDL_MapRGB(fmt, (Uint8)((r > max)? 255 : ((((uint32)m*r) + k) >> 16)),
                                           (Uint8)((g > max)? 255 : ((((uint32)m*g) + k) >> 16)),
                                           (Uint8)((b > max)? 255 : ((((uint32)m*b) + k) >> 16)));
               }
               break;
         }
         pixelsIn  = (Pixel *)((char *)pixelsIn  + pitch);
         pixelsOut = (Pixel *)((char *)pixelsOut + pitch);
      }
   }
   else                  // #else // VIDEO_BITSPERPIXEL != 8
   {
      Uint8 rloss  = fmt->Rloss;
      Uint8 gloss  = fmt->Gloss;
      Uint8 bloss  = fmt->Bloss;
      Uint8 rloss2 = rloss + 16;
      Uint8 gloss2 = gloss + 16;
      Uint8 bloss2 = bloss + 16;
      Uint8 rshift = fmt->Rshift;
      Uint8 gshift = fmt->Gshift;
      Uint8 bshift = fmt->Bshift;
      Uint32 rmask = fmt->Rmask;
      Uint32 gmask = fmt->Gmask;
      Uint32 bmask = fmt->Bmask;
      Uint32 amask = fmt->Amask;

      while (--h >= 0)
      {
         int i = w;
         Pixel * in = pixelsIn;
         Pixel * out = pixelsOut;
         switch (touchup)
         {
            case NO_TOUCHUP:
               while (--i >= 0)
                  *out++ = *in++;
               break;
            case BRITE_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  int r = ((px & rmask) >> rshift) << rloss;
                  int g = ((px & gmask) >> gshift) << gloss;
                  int b = ((px & bmask) >> bshift) << bloss;
                  *out++ = (((((r > max)? 0xFF0000 : ((m*r) + k)) >> rloss2) << rshift) |
                            ((((g > max)? 0xFF0000 : ((m*g) + k)) >> gloss2) << gshift) |
                            ((((b > max)? 0xFF0000 : ((m*b) + k)) >> bloss2) << bshift) |
                            (amask & px));
               }
               break;
            case CTRST_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  *out++ = (((table[((px & rmask) >> rshift) << rloss] >> rloss) << rshift) |
                            ((table[((px & gmask) >> gshift) << gloss] >> gloss) << gshift) |
                            ((table[((px & bmask) >> bshift) << bloss] >> bloss) << bshift) |
                            (amask & px));
               }
               break;
            default: // case CTRSTBRITE_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  int r = table[((px & rmask) >> rshift) << rloss];
                  int g = table[((px & gmask) >> gshift) << gloss];
                  int b = table[((px & bmask) >> bshift) << bloss];
                  *out++ = (((((r > max)? 0xFF0000 : ((m*r) + k)) >> rloss2) << rshift) |
                            ((((g > max)? 0xFF0000 : ((m*g) + k)) >> gloss2) << gshift) |
                            ((((b > max)? 0xFF0000 : ((m*b) + k)) >> bloss2) << bshift) |
                            (amask & px));
               }
               break;
         }
         pixelsIn  = (Pixel *)((char *)pixelsIn  + pitch);
         pixelsOut = (Pixel *)((char *)pixelsOut + pitch);
      }
   }                           // #endif // VIDEO_BITSPERPIXEL
   SDL_UnlockSurface(surfIn);
   SDL_UnlockSurface(surfOut);
   RETURN32(imageOut);
   #endif
}
