/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#ifndef __SWNATIVE_CUSTOM_LIB_H // note! this special __SWNATIVE_CUSTOM_LIB_H directive can only be defined by SWNatives\CustomLib.h. If you create your own lib, use the __CUSTOM_LIB_H directive instead.
#define __SWNATIVE_CUSTOM_LIB_H

/********************************************************************
 * Note: you must define the following constants to create your lib *
 ********************************************************************/

// The library version scheme follows the system versioning scheme.
// See sysMakeROMVersion and friends in SystemMgr.h.
//
// For reference:
//
// 0xMMmfsbbb, where MM is major version, m is minor version
// f is bug fix, s is stage: 3-release,2-beta,1-alpha,0-development,
// bbb is build number for non-releases
// V1.12b3   would be: 0x01122003
// V2.00a2   would be: 0x02001002
// V1.01     would be: 0x01013000
#define prvSWLibVersion sysMakeROMVersion(2, 0, 0, 2, 1)

// Library database creator ID
#define         SWLibCreatorID  'SWNL'

// Name of the library which apps can pass to SysLibFind()
#define         SWLibName                       "SWNatives"

/***********************************************************************
 * You can add custom error codes, but don't change those 6 basic ones *
 ***********************************************************************/

// SW Library result codes. Note that appErrorClass is reserved for
// 3rd party apps/libraries, and thus prevents error number collisions
// between PalmOS and libraries.

#define SWErrParam            (appErrorClass | 0)  // invalid parameter
#define SWErrNotOpen          (appErrorClass | 1)  // library is not open
#define SWErrMethodNotFound   (appErrorClass | 2)  // returned from SWInvokeNative if the method is not found in the lib hashtable
#define SWErrMemory           (appErrorClass | 3)  // memory error occurred
#define SWErrEventHandled     (appErrorClass | 4)  // an event was handled by SWHandleEvent
#define SWErrEventNotHandled  (appErrorClass | 5)  // an event was NOT handled by SWHandleEvent

/********************************************************************
 * The following constant define the size of the hashtable. It may  *
 * be a prime number and be larger than the number of functions     *
 * defined in nativeHashes to minimize collisions. The best number  *
 * is computed by the MethodsHash program; just place the output    *
 * here                                                             *
 ********************************************************************/

#define HASHSIZE 502 // (20 collisions)

/********************************************************************
 * Indexes to be used in the callPublic function       .            *
 ********************************************************************/

// MUST NOT START FROM 0 !
#define PUBLIC_GraphicsStop        1
#define PUBLIC_GraphicsInit        2
#define PUBLIC_GraphicsSwap        3
#define PUBLIC_FreeGrResources     4
#define PUBLIC_DrawHourGlass       5
#define PUBLIC_GraphicsReset       6    // In WinCE is PUBLIC_InitMainGrHandle
#define PUBLIC_GraphicsStartImage  13
#define PUBLIC_GraphicsEndImage    14

// Object destroyers
#define PUBLIC_GraphicsDestroy     7
#define PUBLIC_ImageDestroy        8
#define PUBLIC_CatalogDestroy      9
#define PUBLIC_SocketDestroy       10
#define PUBLIC_SerialPortDestroy   11
#define PUBLIC_FileDestroy         12

/*********************************************************************
 * These array and functions must be implemented by your custom Lib. *
 * See the SuperwabaSDK/vm/palm/SWSymbol for an example.             *
 ********************************************************************/

extern uint32 nativeHashes[];
extern NativeFunc nativeFuncs[];
void buildNativeFuncArray();

// Called by the Lib when a public method is invoked by the vm or someone else. This defines a interface that lets you add custom methods to interface with the vm without modifying the basic architecture of SW Libs
Err callPublic(int index, uint32 *ret, uint32 *args);
// Called by the vm to let the native libraries handle particular events. Must return SWErrEventHandled or SWErrEventNotHandled.
Err handleEvent(EventPtr eventP);
// Called by the system to get the ClassHooks in this custom lib. Must return NULL if no hooks are defined
ClassHook *getClassHooks();
// Called by the system when this lib is about to be closed. The library must free any allocated resources.
void onClose();

#endif
