/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#include <PalmOS.h>											// Standard Palm stuff
#include <PalmCompatibility.h>								// use new headers
#include <SysEvtMgr.h>
#include "stub_res.h"

//#define appFileCreator 'XYXY'
//#define appVersionNum 0x01
//#define appPrefID 0x00
//#define appPrefVersionNum 0x01

#define version20 0x02000000

static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags)
{
   DWord romVersion;

   // See if we're on in minimum required version of the ROM or later.
   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
   if (romVersion < requiredVersion)
   {
      if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) == (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
      {
         // PALMOS 1.0 will continuously relaunch this app unless we switch to
         // another safe one.
         if (romVersion < version20)
            AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
      }
      return (sysErrRomIncompatible);
   }
   return (0);
}

Ptr userCmdLine;

// ATTENTION: do not modify those strings!
char launchString[] =
// guich@200: expanded from 80 to 200 chars. codewarrior puts strange chars in string lengths > 127, so we need to be careful
"L2345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";
//        1         2         3         4         5         6         7         8         9         10        11        12        13        14        15        16        17        18        19        20
static short launchWabaApp()
{
   unsigned short err;
   UInt cardNo;
   LocalID dbID;
   DmSearchStateType searchState;
   int len1,len0;
   CharPtr s;

   // guich@200b2: corrected the param line problem.

   // guich@200 compute the total commandline length
   len0 = StrLen(launchString);
   len1 = userCmdLine?(StrLen(userCmdLine)+1):0; // plus 1 for the space

   s = MemPtrNew(len0+len1 + 1);
   StrCopy(s, launchString);
   if (len1)
   {
      s[len0] = ' ';
      StrCopy(s+len0+1,userCmdLine);
   }
   s[len0+len1] = 0;
   MemPtrSetOwner(s, 0);

   // now launch the vm
   err = DmGetNextDatabaseByTypeCreator(true, &searchState, 'appl', 'SWAB', true, &cardNo, &dbID);
   if (err)
      return err;
   SysUIAppSwitch(cardNo, dbID, 40404, s);
   return 0;
}

static Boolean FormHandleEvent(EventPtr eventP)
{
   switch (eventP->eType)
   {
      case frmUpdateEvent:
      case frmOpenEvent:
      {
         FrmDrawForm(FrmGetActiveForm());
         return true;
      }
   }
   return false;
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
   EventType event;
   int formId;
   FormPtr formPtr;
   Err error;
   unsigned short err;
   short done;

   if (cmd != sysAppLaunchCmdNormalLaunch)
      return 0;
   error = RomVersionCompatible (version20, launchFlags);
   if (error)
      return (error);

   userCmdLine = cmdPBP;

   err = launchWabaApp();
   if (err)
      FrmGotoForm(MissingForm);
   done = 0;
   while(!done)
   {
      EvtGetEvent(&event, 100);
      if (SysHandleEvent(&event))
         continue;
      if (MenuHandleEvent((void *)0, &event, &err))
         continue;
      switch (event.eType)
      {
         case ctlSelectEvent:
            if (event.data.ctlSelect.controlID == MissingCloseButton)
            done = 1;
            FrmDispatchEvent(&event);
            break;
         case frmLoadEvent:
            formId = event.data.frmLoad.formID;
            formPtr = FrmInitForm(formId);
            FrmSetActiveForm(formPtr);
            FrmSetEventHandler(formPtr, FormHandleEvent);
            break;
         case appStopEvent:
            done = 1;
            break;
         default:
            FrmDispatchEvent(&event);
            break;
      }
   }
   return 0;
}

