/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

/*
    This file contains functions that directly interface with the VM:
    . The file functions to make the classloader work
    . The event translation from OS-specific to the SW ones
    . The main program entry point
    . Some string convertion routines
    . The debug support (like System.out.println - used in Vm.debug).
    . The presentation of the fatal error window
    . Application startup and stop
    . Settings and preferrences
    . Some auxiliary OS-specific routines
    . Initialization of the VmGlobals
*/

// Defines used to launch an application during debugging
//#define DEBUG_APP "/m H UIGadgets DoHD"
//#define DEBUG_APP "versapalm/finance/VPFinance AnGD" // put app creator id at end. to get the crtr, open the application pdb and see offset 64
//#define DEBUG_APP "Bench MGQi"
//#define DEBUG_APP "Tour kFCb"
//#define DEBUG_APP "TestSearch ANTW"
//#define DEBUG_APP "GuiBuilder QIRV"
//#define DEBUG_APP "Aktien/Aktien Nee4"
//#define DEBUG_APP "SocketTest twyz"
//#define DEBUG_APP "TestMain zmTW"
//#define DEBUG_APP "UIBench OhdB"

#include "BmpGlue.h"

byte callLauncher = 0; // set to 1 if we need to call launcher when exiting the vm.
// guich@200b4_99: put some big vars here to avoid exceeding the stack
DmSearchStateType state;
Var params[7]; // 7 is the biggest number used

static byte checkSony();
static byte isHanderaSilk();

//////////////////////////////////////////////////////////////////////////////
//                              String convertions                          //
//////////////////////////////////////////////////////////////////////////////
static int isostrncmp(uchar *s1, uchar *s2, uint16 n)
{
	uint16 i;

	for (i = 0; i < n; i++)
	{
		if (s1[i] == s2[i])
			;
		else if (s1[i] < s2[i])
			return -1;
		else
			return 1;
	}
	return 0;
}
//////////////////////////////////////////////////////////////////////////////
//                                 Error Messages                           //
//////////////////////////////////////////////////////////////////////////////
// guich@200b4: simplify the drawErrorWin function
static void showErr(char *msg, int x, int *y)
{
   if (msg && *msg)
   {
   	WinDrawChars(msg, xstrlen(msg), x, *y);
   	debug(msg);
   	*y += FntLineHeight();
   }
}
static void drawErrorWin()
{
#ifndef PALMOS5
	RectangleType rect;
	FontID savedFont;
#endif
	int y;
	int h = FntLineHeight();

   if (vmGlobals->errorWinDisplayed) return;

   vmGlobals->errorWinDisplayed = 1;

   SWCallPublic(SWLibRefNum, PUBLIC_GraphicsStop, 0,0);

	WinEraseWindow();

	debug("SuperWaba VM Fatal Error");
	// draw error message
	y = 5+h;
	showErr("An error was found in the program", 5, &y);
	showErr("being run by the SuperWaba VM.", 5, &y);
   showErr("Error:", 5, &y);
	showErr(errorMessages[vmStatus.errNum - 1], 10, &y);
	showErr(vmStatus.arg1, 10, &y);
	showErr(vmStatus.arg2, 10, &y);
	y += 4;
	if (vmStatus.className[0] || vmStatus.methodName[0])
	{
		showErr("In:", 5, &y);
		showErr(vmStatus.className, 10, &y);
		showErr(vmStatus.methodName, 10, &y);
   	y += 4;
	}

	if (vmStatus.errNum == ERR_CantAccessCoreClasses)
	{
	   y += 4;
		showErr("The SuperWaba VM is not fully", 5, &y);
		showErr("installed. Please make sure the",5, &y);
		showErr("files SWNatives.prc, SuperWaba.prc",5, &y);
		showErr("and SuperWaba.pdb are in the device.",5, &y);
   }
	else
	if (vmStatus.errNum == ERR_UncaughtException && xstrncmp(vmStatus.arg2,"Can't find class",16) == NULL) // guich@350_31: give a better error message. note: the error can be a Uncaught exception
	{
	   showErr("Check if the application's .PDB", 5, &y);
	   showErr("and/or a required library (E.G.", 5, &y);
	   showErr("XPlat*.pdb or other)", 5, &y);
	   showErr("are correctly installed.", 5, &y);
	   y += 4;
		showErr("See debug memo for more details.", 5, &y);
	}
	else
	{
		showErr("Please notify the program's author.", 5, &y);
		showErr("See debug memo for more details.", 5, &y);
   }

#ifndef PALMOS5
	// draw line across top
	rect.topLeft.x = 0;
	rect.topLeft.y = 12;
	rect.extent.x = 160;
	rect.extent.y = 2;
	WinDrawRectangle(&rect, 0);

	// draw rounded rect at top for title
	rect.topLeft.x = 0;
	rect.topLeft.y = 0;
	rect.extent.x = 92;
	rect.extent.y = 13;
	WinDrawRectangle(&rect, 3);

	// draw title
	savedFont = FntSetFont(boldFont);
	WinEraseChars("SuperWaba VM Notice", 13, 5, 2);
	FntSetFont(savedFont);
#else
   WinDrawChars("SuperWaba VM Notice", 13, 5, 2);
#endif
#ifdef PALMOS5
   SemaphoreRelease(1); // guich@400_61: in PalmOS, if we don't release the semaphore, SysTaskDelay does not work.
   SysTaskDelay(millisToTicks(5000)); // guich: added this because sometimes a click is buffered and there is no time to see the screen.
   SemaphoreReserve(1);
#else
   SysTaskDelay(millisToTicks(5000)); // guich: added this because sometimes a click is buffered and there is no time to see the screen.
#endif
#ifndef PALMOS5
	if (vmStatus.errNum != ERR_IncompatibleDevice)
	{
		// draw Stop Program button
		rect.topLeft.x = 2;
		rect.topLeft.y = 148;
		rect.extent.x = 72;
		rect.extent.y = 11;
		WinDrawRectangleFrame(roundFrame, &rect);
		WinDrawChars("Stop Program", 12, 10, 148);
	}
#endif
}

//////////////////////////////////////////////////////////////////////////////
//                         Settings and Preferrences                        //
//////////////////////////////////////////////////////////////////////////////

static Var* getSettingsField(char *name, char *desc)
{
	UtfString fieldName = createUtfString(name);
	UtfString fieldDesc = createUtfString(desc);
	WClassField *field = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
	return &field->var.staticVar;
}

static void storeSettings(char *name, int32 crtr)
{
   WObject obj;
	uint16 len;
	UtfString s;

   obj = getSettingsField(name,"Ljava/lang/String;")->obj;
	if (obj == 0 || WOBJ_StringCharArrayObj(obj) == 0) // if string null, delete it - guich@240_3: first condition added.
	{
	   PrefSetAppPreferences (crtr, 0, 2, 0, 0, true);  // guich@241_20: now storing in saved referrences database
	   PrefSetAppPreferences (crtr, 1, 2, 0, 0, true);
	}
   else
   {
      s = vmGlobals->stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   	len = s.len;
   	PrefSetAppPreferences (crtr, 0, 2, &len, 2, true);
   	PrefSetAppPreferences (crtr, 1, 2, s.str, len, true);
   }
}

static uint32 getSecretKeyCreator(uint32 crtr)
{
   uint32 c0 = ((crtr >> 24) & 0xFF)+64;
   uint32 c1 = ((crtr >> 16) & 0xFF)+64;
   uint32 c2 = ((crtr >> 8 ) & 0xFF)+64;
   uint32 c3 = ((crtr      ) & 0xFF)+64;
   return (c0 << 24) | (c1 << 16) | (c2 << 8) | c3;
}

static void VmStoreSettings() // guich@230_22
{
   storeSettings("appSettings",vmGlobals->appCreatorId);
   storeSettings("appSecretKey",getSecretKeyCreator(vmGlobals->appCreatorId)); // guich@330_47
}

// guich@200b3: to minimize trouble, these functions are implemented here and not in nmpalm_c.c
static void VmGetSettings() // guich
{
   uint32 pref;
   int date;
   char dateSep;
   int weekStart;
   int time24h;
   char timeSep;
   char ts; // thousandsSeparator;
   char ds; // decimalSeparator;
   WObject userName;
   int32 maxColors;
   int isColor;
   WObject platform;
   int screenWidth,screenHeight;
   int32 romVersion;
   WObject appSettings;
   uint16  appSettingsLen;
   WObject appSecretKey;
   uint16 len;
   char crid[5];
   WObject appCreatorId;
   int haveVirtualKeyboard;
   uint32 secretCrtr;
   byte daylightSavings=0;
   int16 timeZone=0;
   WObject romSerialNumber=0;
   char* bufP;
   uint16 bufLen;
   uint16 showPrivateRecs; // guich@421_35

	char name[dlkUserNameBufSize];

   haveVirtualKeyboard = isHanderaSilk();
   if (haveVirtualKeyboard) showSip(0,0);

   romVersion   = vmGlobals->romVersion;
   maxColors    = vmGlobals->maxColors;
	screenWidth  = vmGlobals->screenWidth;
	screenHeight = vmGlobals->screenHeight;
	isColor      = maxColors > 16; // isColor guich@200b3: now returns the real status. In PalmOS, if <= 16 colors, then its grayScale
	platform     = (*vmGlobals->createString)("PalmOS"); // getPlatform
	userName = DlkGetSyncInfo(NULL, NULL, NULL, name, NULL, NULL)?0:(*vmGlobals->createString)(name);

   // guich@230_25: get the preferences of the Welcome app and set the accordingly flags
   len = 2;
   if (PrefGetAppPreferences('SWAB', 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences ('SWAB', 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      vmGlobals->audibleGC = buf[0] == '1';
      vmGlobals->dumpMemStats = buf[1] == '1';
      xfree(buf);
   }

   // get the preferences
   pref = PrefGetPreference(prefDateFormat);

   switch (pref)
   {
   	case dfDMYWithSlashes: date = 2; dateSep = '/'; break;  // 31/12/95
   	case dfDMYWithDots:	  date = 2; dateSep = '.'; break;  // 31.12.95
   	case dfDMYWithDashes:  date = 2; dateSep = '-'; break;  // 31-12-95
   	case dfYMDWithSlashes: date = 3; dateSep = '/'; break;  // 95/12/31
   	case dfYMDWithDots:    date = 3; dateSep = '.'; break;  // 95.12.31
   	case dfYMDWithDashes:  date = 3; dateSep = '-'; break;  // 95-12-31
   	case dfMDYWithSlashes:                                  // 12/31/95
   	default:               date = 1; dateSep = '/';  break;
   }
	weekStart = PrefGetPreference(prefWeekStartDay);
	pref = PrefGetPreference(prefTimeFormat);
	timeSep = TimeSeparator(pref);
	time24h = Use24HourFormat(pref);
	pref = PrefGetPreference(prefNumberFormat);
	switch (pref)
	{
      case nfCommaPeriod      : ts=',';  ds='.'; break;
    	case nfPeriodComma      : ts='.';  ds=','; break;
     	case nfSpaceComma       : ts=' ';  ds=','; break;
     	case nfApostrophePeriod : ts='\''; ds='.'; break;
     	case nfApostropheComma  : ts='\''; ds=','; break;
   }
   // guich@230_22: get the size then the app preferrences, if any
   len = 2;
   if (PrefGetAppPreferences(vmGlobals->appCreatorId, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (vmGlobals->appCreatorId, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSettings = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSettings = 0;
   // guich@330_47: get the size then the app secret key, if any
   secretCrtr = getSecretKeyCreator(vmGlobals->appCreatorId);
   len = 2;
   if (PrefGetAppPreferences(secretCrtr, 0, &appSettingsLen, &len, true) != noPreferenceFound && appSettingsLen > 0) // guich@241_20: now storing in saved referrences database
   {
      char *buf;
      buf = xmalloc(appSettingsLen);
      PrefGetAppPreferences (secretCrtr, 1, buf, &appSettingsLen, true); // guich@241_20: now storing in saved referrences database
      appSecretKey = vmGlobals->createStringFromCharPtr(buf,appSettingsLen);
      xfree(buf);
   }
   else appSecretKey = 0;

   // guich@230_22: get the creator id
   crid[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   crid[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   crid[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   crid[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   crid[4] = 0;
   appCreatorId = (*vmGlobals->createString)(crid);

   // store the preferences in the waba.sys.Settings class
   vmGlobals->prefClass = getClass(createUtfString("waba/sys/Settings"),1);

   if (vmGlobals->romVersion >= 0x04000000) // guich@340_33
   {
      daylightSavings = PrefGetPreference(prefDaylightSavingAdjustment) > 0;
      timeZone = PrefGetPreference(prefTimeZone);
      timeZone /= 60; // in hours
   }

   // guich@421_16: get the serial number
   if (vmGlobals->romVersion >= 0x03000000 && SysGetROMToken(0, sysROMTokenSnum, (UInt8**) &bufP, &bufLen) == 0 && bufP != NULL && (byte)*bufP != 0xFF)
      romSerialNumber = (*vmGlobals->createStringFromCharPtr)(bufP,bufLen);

   // guich@421_35: determine if the user has choosen to hide the private records
   showPrivateRecs = PrefGetPreference(prefShowPrivateRecords) == showPrivateRecords;

   // guich@421_17: don't call the constructor; just set directly the values
   getSettingsField("dateFormat","B")->intValue = date;
   getSettingsField("dateSeparator","C")->intValue = dateSep;
   getSettingsField("weekStart","B")->intValue = weekStart;
   getSettingsField("is24Hour","Z")->intValue = time24h;
   getSettingsField("timeSeparator","C")->intValue = timeSep;
   getSettingsField("thousandsSeparator","C")->intValue = ts;
   getSettingsField("decimalSeparator","C")->intValue = ds;
   getSettingsField("screenWidth","I")->intValue = screenWidth;
   getSettingsField("screenHeight","I")->intValue = screenHeight;
   getSettingsField("isColor","Z")->intValue = isColor;
   getSettingsField("maxColors","I")->intValue = maxColors;
   getSettingsField("onDevice","Z")->intValue = 1;
   getSettingsField("romVersion","I")->intValue = romVersion;
   getSettingsField("platform","Ljava/lang/String;")->obj = platform;
   getSettingsField("userName","Ljava/lang/String;")->obj = userName;
   getSettingsField("appSettings","Ljava/lang/String;")->obj = appSettings;
   getSettingsField("appCreatorId","Ljava/lang/String;")->obj = appCreatorId;
   getSettingsField("virtualKeyboard","Z")->intValue = haveVirtualKeyboard;
   getSettingsField("appSecretKey","Ljava/lang/String;")->obj = appSecretKey;
   getSettingsField("daylightSavings","Z")->intValue = daylightSavings;
   getSettingsField("timeZone","I")->intValue = timeZone;
   getSettingsField("romSerialNumber","Ljava/lang/String;")->obj = romSerialNumber;
   getSettingsField("showSecrets","Z")->intValue = showPrivateRecs; // guich@421_35
   getSettingsField("isHighColor","Z")->intValue = vmGlobals->screenBPP > 8; // guich@421_35

   GsiInitialize();
   GsiSetLocation(2800,2800); // guich: move to an invisible position
#ifdef PALMOS5
   if (!vmGlobals->isSony) // guich@350_28: sony palm os 5 doesnt like this command.
#endif
   GsiEnable(true);
}

//////////////////////////////////////////////////////////////////////////////
//                                  Events                                  //
//////////////////////////////////////////////////////////////////////////////
static void drawMainWin(byte allWindows)
{
	WObject mainWinObj;
	WClass *vclass;
	WClassMethod *method;

	mainWinObj = vmGlobals->globalMainWin;
	if (!mainWinObj)
		return;
	// WinEraseWindow(); guich@200b3: in color it paints with a ugly pink! lets get rid of it!
	vclass = WOBJ_class(mainWinObj); // get runtime class
	method = getMethod(vclass, createUtfString("_doPaint"),
		createUtfString("(IIII)V"), &vclass);
	if (method == NULL)
		return;
	params[0].obj = mainWinObj;
	params[1].intValue = allWindows?-5000:0; // -5000: ugly, i know. But necessary.
	params[2].intValue = 0; // y
	params[3].intValue = vmGlobals->screenWidth;
	params[4].intValue = vmGlobals->screenHeight;
	executeMethod(vclass, method, params, 5);
}
static void timerCheck()
{
	WObject mainWinObj;
	int32 now, diff;
	WClass *vclass;
	WClassMethod *method;
	Var params[1];

	mainWinObj = vmGlobals->globalMainWin;
	if (!mainWinObj)
		return;
	now = getTimeStamp();
	diff = now - vmGlobals->globalTimerStart;
	if (diff < 0)
		diff += (1L << 30); // max stamp is (1 << 30)
	if (diff < vmGlobals->globalTimerInterval)
		return;
	vclass = WOBJ_class(mainWinObj); // get runtime class
	method = getMethodByMapNum(vclass, &vclass, (uint16)vmGlobals->onTimerTickMethodMapNum);
	if (method != NULL)
	{
		params[0].obj = mainWinObj;
		executeMethod(vclass, method, params, 1);
	}
	vmGlobals->globalTimerStart = now;
}

void executePostEvent(int32 type, int32 key, int32 x, int32 y, int32 modifiers, int32 timeStamp)
{
	WObject mainWinObj;
	WClass *vclass;
	WClassMethod *method;

	mainWinObj = vmGlobals->globalMainWin;
	if (!mainWinObj)
		return;

	vclass = WOBJ_class(mainWinObj); // get runtime class
	method = getMethodByMapNum(vclass, &vclass, (uint16)vmGlobals->postEventMethodMapNum);
	if (method != NULL)
	{
		params[0].obj = mainWinObj;
		params[1].intValue = type;
		params[2].intValue = key;
		params[3].intValue = x;
		params[4].intValue = y;
		params[5].intValue = modifiers;
		params[6].intValue = timeStamp;
		executeMethod(vclass, method, params, 7);
	}
}
char last1,last2,last3;
static int winExitEventCount=0;

static void handleMainWinEvent(EventPtr eventP)
{
	static Boolean isCommand;

	if (!vmGlobals->globalMainWin)
		return;

	switch (eventP->eType)
	{
		case penUpEvent:
		case penDownEvent:
		case penMoveEvent:
		{
			int32 type, x, y;

			if (eventP->eType == penDownEvent)
				type = 200; // PenEvent.PEN_DOWN
			else if (eventP->eType == penUpEvent)
				type = 202; // PenEvent.PEN_UP
			else
				type = 201; // PenEvent.PEN_MOVE
			x = eventP->screenX;
			y = eventP->screenY;
#ifndef PALMOS5
			if (vmGlobals->isSony) // Unbelievable, unbelievable. Sony's highres 320x320 mode make PenEvents act as 160x160 mode. So, to keep things in sync, we need to multiply coords by 2.
#else
         if (vmGlobals->isHighDensity && vmGlobals->screenWidth != 160)
#endif
			{
			   x <<= 1;
			   y <<= 1;
			}
			executePostEvent(type, 0, x,y,0,0);
			break;
		}
      case menuCmdBarOpenEvent: // guich@230_6 - guich@320_33
         eventP->data.keyDown.chr = commandChr;
		case keyDownEvent:
		{
			int32 type, key;
			Word chr;
			Boolean capsLockP=0;
         Boolean numLockP=0;
         UInt16 tempShiftP=0;
         Boolean autoShiftedP=0;
         int32 modifiers = 0;
			type = 100; // KeyEvent.KEY_PRESS
			chr = eventP->data.keyDown.chr;
			key = 0;

			last1 = last2;
			last2 = last3;
			last3 = chr;
			if (last1 == '@' && last2 == '?' && last3 == '@')
			   debugMemory();

			switch (chr)
			{
			   case vchrNavChange: // 5-way
			      // each key
               if (NavKeyPressed(eventP, Select))
                  key = 75009; else
               if (NavKeyPressed(eventP, Left))
                 key = 75006; else
               if (NavKeyPressed(eventP, Right))
                  key = 75007;
               // simultaneous key
               vmGlobals->fiveWay = 0xFFFFFFFFL; // guich@340_55: will be anded with the KeyCurrentState when the method Vm.getSystemKeysPressed is called
               if (!NavKeyPressed(eventP, Select)) vmGlobals->fiveWay &= ~0x08000000;
               if (!NavKeyPressed(eventP, Left))   vmGlobals->fiveWay &= ~0x00001000;
               if (!NavKeyPressed(eventP, Right))  vmGlobals->fiveWay &= ~0x00002000;
               if (!NavKeyPressed(eventP, Up))	   vmGlobals->fiveWay &= ~0x00000002;
               if (!NavKeyPressed(eventP, Down))	vmGlobals->fiveWay &= ~0x00000004;
	   	      break;
			   case vchrJogUp:    // sony
			   case vchrTrgJogUp: // handera
                                     key = vmGlobals->isSony?76030:76012; break; // guich@320_11 - guich@450_23: on M505, the clock is also vchrJogUp
			   case vchrJogDown:
			   case vchrTrgJogDown:
			                            key = vmGlobals->isSony?76031:76012; break; // guich@450_23: on M505, the clock is also vchrJogUp
            case vchrJogPush:
            case vchrTrgJogPush:
                                     key = 76032; break;
            case vchrJogRelease:     key = 76033; break;
            case vchrJogPushRepeat:  key = 76034; break;
            case vchrJogPushedUp:    key = 76035; break;
            case vchrJogPushedDown:  key = 76036; break;
            case vchrJogBack:        key = 76037; break; // guich@450_28
				// NOTE: these should go somewhere:
				//		nextFieldChr
				//		prevFieldChr
				//		linefeedChr
				case pageUpChr:    key = 75000; break; // PAGE_UP
				case pageDownChr:  key = 75001; break; // PAGE_DOWN
//				              :    key = 75002; break; // HOME
//				case :             key = 75003; break; // END
				case upArrowChr:   key = 75004; break; // UP
				case downArrowChr: key = 75005; break; // DOWN
				case leftArrowChr: key = 75006; break; // LEFT
				case rightArrowChr:key = 75007; break; // RIGHT
//				case :             key = 75008; break; // INSERT
                case 10:                                          // guich@200
				case returnChr:    key = 75009; break; // ENTER
				case tabChr:       key = 75010; break; // TAB
				case backspaceChr: key = 75011; break; // BACKSPACE
				case escapeChr:    key = 75012; break; // ESCAPE
//				case :             key = 75013; break; // DELETE
				case menuChr:      key = 75014; break; // MENU
				case commandChr:   key = 75015; isCommand = !isCommand; showCommand(isCommand); break; // COMMAND
				// lines below added by guich
            case keyboardAlphaChr: key = 76000; break;
            case keyboardNumericChr: key = 76001; break;
            case keyboardChr: key = 76002; break;
            case hard1Chr: key = 76003; break;
            case hard2Chr: key = 76004; break;
            case hard3Chr: key = 76005; break;
            case hard4Chr: key = 76006; break;
            case calcChr: key = 76007; break;
            case findChr: key = 76008; break;
				case launchChr:    key = 76009; break; // HOME guich@200b4_5
				case brightnessChr: // guich@450_25
            case contrastChr:
            case vchrPalmMin + 2: // guich@402_2: fixed contrast
               key = 76011; break; // guich@401_31
            case 0x20E:
            case vchrPalmMin + 0: // guich@402_2: fixed clock
               key = 76012; break; // guich@401_31
            case vchrHardCradle: // guich@421_18
               key = 76013; break;
			}
			if (!key)
			{
				if (chr > 255)
					break;
				key = chr;
			}
		   if (key != 75015 && isCommand) // guich: different key and command?
		   {
		      isCommand = false;
		      showCommand(false);
		   }

         GrfGetState(&capsLockP,&numLockP,&tempShiftP,&autoShiftedP);
   		if (eventP->data.keyDown.modifiers & controlKeyMask || chr == commandChr) modifiers |= (1 << 1);
   		if (eventP->data.keyDown.modifiers & shiftKeyMask || tempShiftP || autoShiftedP)   modifiers |= (1 << 2);
   		if (eventP->data.keyDown.modifiers & capsLockMask || capsLockP)   modifiers |= (1 << 3);
   		if (eventP->data.keyDown.modifiers & commandKeyMask) modifiers |= (1 << 4);
   		if (eventP->data.keyDown.modifiers & optionKeyMask)  modifiers |= (1 << 5);
		   executePostEvent(type, key, 0,0, modifiers,0);
			break;
		}
		default:
			break;
	}
}

static Long calcEventTimeout()
{
	Long timeout;

	if (vmGlobals->globalTimerInterval <= 0)
		timeout = -1;
	else
	{
		int32 now, diff;

		// NOTE: calculate the ticks until the next interval
		now = getTimeStamp();
		diff = now - vmGlobals->globalTimerStart;
		if (diff < 0)
			diff += (1L << 30); // max stamp is (1 << 30)
		timeout = vmGlobals->globalTimerInterval - diff;
		if (timeout <= 0)
			timeout = 0;
		timeout = millisToTicks(timeout);
		if (timeout <= 0)
			timeout = 1;
	}
	return timeout;
}
//////////////////////////////////////////////////////////////////////////////
//                               Class Loader                               //
//////////////////////////////////////////////////////////////////////////////
uchar *lockWarpRec(char *path, uint16 pathLen, uint32 *size)
{
	uchar *p;
	VoidHand recHandle;
	VoidPtr recPtr;
	uint32 top, bot, mid, numRecs;
	uint16 i, nameLen, minLen;
	int cmp;

	for (i = 0; vmGlobals->classDbList[i]; i++)
	{
		// binary search to find class
		numRecs = DmNumRecords(vmGlobals->classDbList[i]);
		if (numRecs == 0)
			continue;
		top = 0;
		bot = numRecs;
		while (1)
		{
			mid = (bot + top) / 2;
			recHandle = DmQueryRecord(vmGlobals->classDbList[i], mid);
			recPtr = MemHandleLock(recHandle);
			p = (uchar *)recPtr;
			nameLen = getUInt16(p);
			p += 2;
			if (pathLen > nameLen)
				minLen = nameLen;
			else
				minLen = pathLen;
			cmp = isostrncmp((uchar *)path, p, minLen);
			if (!cmp)
			{
				if (pathLen == nameLen)
				{
					if (size != NULL)
						*size = MemHandleSize(recHandle) - nameLen - 2;
					p += pathLen;
					return p;
				}
				if (pathLen > nameLen)
					cmp = 1;
				else
					cmp = -1;
			}
			MemHandleUnlock(recHandle);
			if (mid == top)
				break; // not found
			if (cmp < 0)
				bot = mid;
			else
				top = mid;
		}
	}
	return NULL;
}

void unlockWarpRec(uchar *ptr, uint16 pathLen)
{
	// NOTE: The pointer given is "inside" the record. To unlock the record
	// we need to find the original pointer locked which is at the front
	// of the path.
	MemPtrUnlock(ptr - pathLen - 2);
}

uchar *nativeLoadClass(UtfString className, uint32 *size)
{
	uchar *p;
   char *path;
	uint16 pathLen;

	// make full path by appending ".class"
	pathLen = className.len + 6;
	path = xmalloc(pathLen+1); // guich@350_29: now xmalloc the buffer
	xstrncpy(path, className.str, className.len);
	xstrncpy(&path[className.len], ".class", 6);
	path[pathLen] = 0;
	// NOTE: We unlock the record locked here in stopApp()
	p = lockWarpRec(path, pathLen, size);
	if (p != NULL && getUInt32(p) != (uint32)0xCAFEBABE)
	{
		VmError(ERR_BadClass, NULL, &className, NULL);
		unlockWarpRec(p, pathLen);
		p = NULL;
	}
	xfree(path);
	return p;
}

static void releaseClass(UtfString className, uchar *p)
{
   unlockWarpRec(p, className.len+6);
}

static void releaseAndClose(DmOpenRef dmRef, byte removeRecord)
{
	UInt16 attrP;
   int32 count = DmNumRecords(dmRef);
   while (--count >= 0)
   {
      // is the record locked?
      DmRecordInfo(dmRef, count, &attrP, NULL, NULL); // guich@300_32: changed 0 always to count
      if (attrP & dmRecAttrBusy)
      {
		   MemHandleUnlock(DmQueryRecord(dmRef,count));
	      DmReleaseRecord(dmRef,count,false); // guich@300_46: moved to inside this if
	   }
	   if (removeRecord)
         DmRemoveRecord(dmRef,count); // guich@200b4_103: fix "records left locked in closed unprotected database"
   }
   // delete the database
   DmCloseDatabase(dmRef); // guich@200b2: removed from before the last block to correct the "address error" from pose
}

//////////////////////////////////////////////////////////////////////////////
//                           App Start and Stop                             //
//////////////////////////////////////////////////////////////////////////////
void postStopEvent(int exitCode)
{
	EventType event;

   callLauncher = 1; // guich@200b4_74

	// NOTE: Here we launch the "launcher". If we simply
	// queued an appStopEvent to stop, it would return
	// to the calling application which was the stub that
	// executed the vm, causing it to rerun the same program
	// Note2: we must enqueue something to make the system
	// know that we want to stop the application
	event.eType = keyDownEvent;
	event.data.keyDown.chr = launchChr;
	event.data.keyDown.modifiers = commandKeyMask;
	EvtAddEventToQueue(&event);
}
////////////////// SuperWaba 2.0 beta 3 new memory routines ////////////////////
// note: now, to allocate memory in the heap db, we use xmalloc. xmalloc now is smart enough to know where to get the memory from.
DmOpenRef heapDmRef = 0;
static void destroyHeapDb()
{
#ifdef USE_STORAGE_HEAP
	UInt cardNo=0;
	LocalID dbID=0;
   Err err;

   err = DmGetNextDatabaseByTypeCreator(true, &state, 'Heap', SUPERWABA_CREATOR_ID, true, &cardNo, &dbID);
   if (err == dmErrCantFind)
      return;
   if (!heapDmRef && !err)
	   heapDmRef = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);

	if (heapDmRef)
   {
      releaseAndClose(heapDmRef,1);
	   // delete the database
		DmDeleteDatabase(cardNo, dbID);
	}
	heapDmRef = NULL;
#endif
}

static void createHeapDb()
{
#ifdef USE_STORAGE_HEAP
   if (!heapDmRef)
   {
	   destroyHeapDb();
		// create a db (ObjHeap) for the object heap
		if (DmCreateDatabase(0, "SuperWabaHeap", SUPERWABA_CREATOR_ID, 'Heap', false) == 0)
			heapDmRef = DmOpenDatabaseByTypeCreator('Heap', SUPERWABA_CREATOR_ID, dmModeReadWrite);
   }
#endif
}

static WObject startApp(int runApp, CharPtr cmdLine)
{
	int i,j, n,extG=0,extL=0;
	UInt cardNo;
	LocalID dbID;
	Err err;
	Boolean first;
	DmOpenRef dmRef;
	char *cmd;
	uint32 dummy = 0;
	DmOpenRef *classDbP;

#ifndef WIN32 // in win32 we do not have a valid screen until the MainWindow construction
   SWCallPublic(SWLibRefNum, PUBLIC_DrawHourGlass,0,&dummy);
#endif

	// NOTE: We need VoidHand to fall on a byte boundry for our xmalloc()
	// routine to work correctly (see xmalloc())
	if (sizeof(VoidHand) & 3 != 0)
	{
		VmQuickError(ERR_SanityCheckFailed, NULL);
		return 0;
	}

   n = 2;
	if (!runApp)
	{
#ifndef DEBUG_APP
	   n = 1;
		cmdLine = "waba/ui/Welcome"; // run "welcome" app (no app class specified)
#else
      // debug
		cmdLine = DEBUG_APP;
		runApp = 1;
#endif
		////////
   }

	// parse options
	cmd = cmdLine;
	while (*cmd == '/')
	{
		if (!*++cmd)
			break;
		if (*++cmd != ' ')
			break;
		if (*(cmd-1) == 'm') // guich@420_19: now /m specify if we want to replace Ok by x in Pocket PC,
		   cmd++;
	   cmd++;
	   while (*cmd && *cmd != ' ')  // skip any other characters
	      cmd++;
	   // skip spaces
		while (*cmd == ' ')
		   cmd++;
	}

	// parse class name
	for (i=0; cmd[i] && cmd[i] != ' '; i++) ; // find the length
	vmGlobals->className = xmalloc(i+1); // guich@340_44: now the size of the class name is dynamically allocated
	j = 0;
	while (*cmd && *cmd != ' ')
		vmGlobals->className[j++] = *cmd++;
	vmGlobals->className[j] = 0;

   // guich@200b2: search for extended libraries SWAX
   if (runApp) // local extension only exists if were running a app.
   {
		cmd++; // skip space
		vmGlobals->appCreatorId = getUInt32((uchar *)cmd);
		cmd += 4;
      // count the local extension libraries
   	first = true;
		while (1)
		{
			err = DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT, vmGlobals->appCreatorId, false, &cardNo, &dbID);
			if (err == dmErrCantFind)
				break;
		   extL++;
			first = false;
		}
   }
   else vmGlobals->appCreatorId = SUPERWABA_CREATOR_ID; // guich@230_22: store the crid, even if just running the welcome app
   // count the global extension libraries
	first = true;
	while (1)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT, SUPERWABA_CREATOR_ID, false, &cardNo, &dbID);
		if (err == dmErrCantFind)
			break;
	   extG++;
		first = false;
	}
	vmGlobals->classDbList = xmalloc(sizeof(DmOpenRef) * (n+extG+extL+1));
	if (vmGlobals->classDbList == 0)
	{
		VmQuickError(ERR_CantAllocateMemory, NULL);
		return 0;
	}
	xmemzero(vmGlobals->classDbList, sizeof(DmOpenRef) * (n+extG+extL+1));
	classDbP = vmGlobals->classDbList;
	*classDbP++ = DmOpenDatabaseByTypeCreator(SUPERWABA_TYPE_ID, SUPERWABA_CREATOR_ID,  dmModeReadOnly);
	if (vmGlobals->classDbList[0] == 0)
	{
		VmQuickError(ERR_CantAccessCoreClasses, NULL);
		return 0;
	}

	if (runApp) // get the database of the app
	{
		dmRef = DmOpenDatabaseByTypeCreator(SUPERWABA_TYPE_ID, vmGlobals->appCreatorId, dmModeReadOnly);
		if (dmRef == 0)
		{
		   UtfString utf;
		   utf.str = cmdLine;
		   utf.len = cmdLine?xstrlen(cmdLine):0;
			VmError(ERR_CantAccessAppClasses, NULL,&utf,NULL); // guich@330_15: add the cmdline to the error message
			return 0;
		}
		*classDbP++ = dmRef;
	}
	// store the command line
	if (*cmd)
	{
	   int len = StrLen(cmd+1);	    // store a copy
  	   vmGlobals->commandLine = MemPtrNew(len+1);
  	   StrCopy(vmGlobals->commandLine, cmd+1);
  	   vmGlobals->commandLine[len] = 0;
 	}
	// now get the extended libraries
   // 1. get the global extension libraries
	first = true;
	for (i = 0; i < extG; i++)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT, SUPERWABA_CREATOR_ID, false, &cardNo, &dbID);
		*classDbP++ = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
		first = false;
	}
	// 2. get the local extension libraries
	first = true;
	for (i = 0; i < extL; i++)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, SUPERWABA_TYPE_EXT, vmGlobals->appCreatorId, false, &cardNo, &dbID);
		*classDbP++ = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
		first = false;
	}
	*classDbP = 0;

   SndGetDefaultVolume (NULL, &vmGlobals->defSoundAmp, NULL);
	VmInit(); // no more memory sizes!
// guich@220_41	if (getDeviceFreeMem() > 40000)
//   	heapGap = xmalloc(getDeviceFreeMem()>80000?50000:20000); // guich@200b4: alloc the gap so we can warranty that the user wont steal all the memory and SW can't start again.
	return VmStartApp(vmGlobals->className);
}
static void stopApp(WObject mainWinObj)
{
	WClass *wclass, *nextClass;
	uint32 i;

   if (vmStatus.errNum == 0) // guich@200b4_25: dont call onExit if VM crashed
   {
   	VmStopApp(mainWinObj);
      SWCallPublic(SWLibRefNum, PUBLIC_GraphicsStop, 0,0); // guich@200: bye bye gray scale
   }

	if (vmInitialized) // set by VmInit()
	{
	   FreeObjectHeap();
		// unlock all record pointers
		for (i = 0; i <= CLASS_HASH_SIZE; i++)
		{
			wclass = classHashList[i];
			while (wclass != NULL)
			{
				UtfString className;

				// NOTE: The pointer to the record is before the byteRep
				// pointer because the class' absolute location comes first
				// in the record. The first bytes in the record are the
				// class name length (2) plus the class name plus the .class
				// extension of the class name.
				className = wclass->className;
				nextClass = wclass->nextClass;
				unlockWarpRec(wclass->byteRep, className.len + 6); // + 6 for ".class"
				wclass = nextClass;
			}
		}
		VmFree();
	}
	for (i = 0; vmGlobals->classDbList[i]; i++)
	   if (vmGlobals->classDbList[i]) // guich@200b4_99: fixed bug when no app class was found
   		releaseAndClose(vmGlobals->classDbList[i],0); // guich@241_7: now we make sure all records are released
	if (vmGlobals->classDbList != 0)
		xfree(vmGlobals->classDbList);

   //guich@220_41 if (heapGap) xfree(heapGap); // guich@200b4: free the gap so we can warranty that the user wont steal all the memory and SW can't start again.
   SWCallPublic(SWLibRefNum, PUBLIC_FreeGrResources, 0,0); // guich@200b4: free some possible allocated resources, like some arrays used in collision detect.

	systemOutClose(); // guich@200
   // guich@200b3: put the sound back to its default
	SndSetDefaultVolume (NULL, &vmGlobals->defSoundAmp, NULL);
}

//////////////////////////////////////////////////////////////////////////////
//                          Program Entry Point                             //
//////////////////////////////////////////////////////////////////////////////
#define DEFINE_varInCodeSegment_FUNC(funcName) \
static asm long funcName(int doStore, long p) \
{															 \
   lea bag,a0;											 \
   cmp.w #1,4(a7);									 \
   beq store;											 \
retrieve:   											 \
   move.l (a0),d0;									 \
   rts;													 \
store:													 \
   move.l 6(a7),(a0);								 \
   rts;													 \
bag:														 \
   dc.l 0;												 \
}

// from here to bottom: implemented Gennadi Pergament modifications to correct the fatal errors when poping up some palm dialogs. this bug is since waba 1.0.
static Boolean formHandler (EventPtr /*eventP*/)
{
   return false;
}

#ifndef PALMOS5

DEFINE_varInCodeSegment_FUNC(varInCodeSegmentVMGlobals)

typedef void (*sysTrapWinSetActiveWindowType) (WinHandle winHandle); // SYS_TRAP(sysTrapWinSetActiveWindow);
typedef void (*sysTrapFrmCustomAlertType) (UInt16 alertId, const Char *s1, const Char *s2, const Char *s3); // guich@330_13
typedef void (*EvtGetPenType) (SWord *pScreenX, SWord *pScreenY, Boolean *pPenDown);

static void trappedEvtPenType (SWord *pScreenX, SWord *pScreenY, Boolean *pPenDown)
{
   // at first call the original Palm OS code ...
   struct VMGlobals *globals = (struct VMGlobals *)varInCodeSegmentVMGlobals(0,0);
   ((EvtGetPenType)globals->evtGetPenTrapPtr) (pScreenX,pScreenY,pPenDown);

   // and now brake it a little bit ...
   SysTaskDelay(1);
}
// end of GP modifications
////////////////////////////////////////////////////////////////////////////////////////
// trap the winSetActiveWindow to turn off grayscale when a normal form is displayed  //
////////////////////////////////////////////////////////////////////////////////////////
static void trappedSysTrapFrmCustomAlertType (UInt16 alertId, const Char *s1, const Char *s2, const Char *s3)
{
   struct VMGlobals *globals = (struct VMGlobals *)varInCodeSegmentVMGlobals(0,0);
   uint32 enableGraphics = 0;
   SWCallPublic(globals->SWLibRefNum, PUBLIC_GraphicsSwap, 0, &enableGraphics);
   SWCallPublic(globals->SWLibRefNum, PUBLIC_GraphicsReset, 0, 0L);
   // call the original Palm OS code ...
   ((sysTrapFrmCustomAlertType)globals->sysTrapFrmCustomAlertPtr) (alertId, s1, s2, s3);
}
static void trappedSysTrapWinSetActiveWindowType (WinHandle winHandle)
{
   struct VMGlobals *globals = (struct VMGlobals *)varInCodeSegmentVMGlobals(0,0);
   if (winHandle || globals->romVersion >= 0x03500000) // guich@200b3: sometimes this winHandle is passed null. guich@300_49: PalmOS 3.5 must have this parameter passed or it will HALT!
   {
      uint32 enableGraphics = globals->wabaFrmP && (!winHandle || &globals->wabaFrmP->window == winHandle);
      if (winHandle || enableGraphics)
      {
         SWCallPublic(globals->SWLibRefNum, PUBLIC_GraphicsSwap, 0, &enableGraphics);
   //      if (enableGraphics) - guich@421_77: with this if, when a call was received, the kyocera was freezing
   //         SWCallPublic(globals->SWLibRefNum, PUBLIC_GraphicsReset, 0, 0L);
      }
      else
      if (!winHandle && !enableGraphics) // the active form has changed?
         SWCallPublic(globals->SWLibRefNum, PUBLIC_GraphicsReset, 0, 0L);
      // call the original Palm OS code ...
      ((sysTrapWinSetActiveWindowType)globals->sysTrapWinSetActiveWindowPtr) (winHandle);
   }
}

#endif // PALMOS5

#define MAX_LAUNCHCODES 12
typedef struct
{
   uint16 launchCode;
   uint16 refNum;
} Codes2Launch;
int nCodes2Launch;
Codes2Launch codes2Launch[MAX_LAUNCHCODES]; // i hope there's no more than 10 codes

/// guich@350_3: added support for custom launchcode handling ///
void registerLaunchCode(uint16 launchCode, uint16 refNum)
{
   // add the given launchCode to the array.
   codes2Launch[nCodes2Launch].launchCode = launchCode;
   codes2Launch[nCodes2Launch].refNum = refNum;
   nCodes2Launch++;
}
void unregisterLaunchCode(uint16 launchCode, uint16 refNum)
{
   int i;
   // remove the given launchCode from the array
   for (i = 0; i < nCodes2Launch; i++)
      if (codes2Launch[i].launchCode == launchCode && codes2Launch[i].refNum == refNum)
      {
         // move the rest of the array into the current position
         if (i < (MAX_LAUNCHCODES-1) && i < nCodes2Launch)
            xmemmove(&codes2Launch[i], &codes2Launch[i+1],sizeof(codes2Launch[0])*(nCodes2Launch-i-1));
         nCodes2Launch--;
         break;
      }
}
/////////
/********************************************************/
/* Purpose..: The Startup routine for Palm Applications */
/********************************************************/
// Our internal startup code
UInt32 __Startup__(void);
UInt32 __Startup__(void)
{
	UInt32			result=0;
	void *			prevGlobalsP;
	void *			globalsP;
	Err				err;
	SysAppInfoPtr	appInfoP;

	// Call the Standary system code for allocating and initializing globals and
	//  setting up A5, and getting the command line arguments
	err = SysAppStartup(&appInfoP, &prevGlobalsP, &globalsP);
	if (err)
	{
		ErrDisplay("Error launching application");
		return 0;
	}

   if ((appInfoP->cmd == sysAppLaunchCmdNormalLaunch) || ((Word)appInfoP->cmd == (Word)40404) ||
      ((appInfoP->launchFlags & sysAppLaunchFlagSubCall) && nCodes2Launch > 0)) // guich@400_28: if there are registered launchcodes by libraries, then allow them to process the codes
   	// Call the PilotMain routine and get the result
   	result = PilotMain(appInfoP->cmd, appInfoP->cmdPBP, appInfoP->launchFlags);

	// Call SysAppExit to dispose of application allocated memory chunks
	//  and delete the AMX task if this app was on it's own thread
	SysAppExit(appInfoP, prevGlobalsP, globalsP);

	return result;
}

static void SizeForm(FormType *frmP)
{
   RectangleType rect;

   rect.topLeft.x = 0;
   rect.topLeft.y = 0;

   WinGetDisplayExtent(&rect.extent.x, &rect.extent.y);
   WinSetWindowBounds(FrmGetWindowHandle(frmP), &rect);
}

static WClass *vclass, *runThreadsVclass;
static WClassMethod *runThreadsMethod;
static Var runThreadsParams[1];
static byte mustSwap;
static DWord romVersion;
#define SK_HARD1     0x00000008L
#define SK_HARD2     0x00000010L
#define SK_HARD3     0x00000020L
#define SK_HARD4     0x00000040L
#define SK_LAUNCH    0x01000000L
#define SK_FIND      0x02000000L
#define SK_CALC      0x04000000L
#define SK_CONTRAST  0x10000000L // guich@401_31
#define SK_CLOCK     0x20000000L // guich@401_31
#define SK_SYNC      0x40000000L // guich@421_18

// andy@330_25
// Fixed bug where Event.isAvailable would return 'true' but Window.pumpEvents would block.
// The 'eventWait' parameter allows the caller to specify a timelimit for pumpEvents so that
// it will return if no events are available and the timelimit expires.  Note that if evtWaitForever
// is passed in for eventWait, then pumpEvents will act exactly as it used to.  This is used
// by the main event loop for effeciency.
// Note: it is guaranteed that at least one iteration through pending threads will be made on
// every call to this function.
int pumpEvents(int eventWait)
{
   byte cont;
   int interceptKey;
   Long timeout;
   EventType event;
   int i;
   Err err;

    // guich@200b4_84: implement a simple thread model.
    if (vmStatus.errNum == 0 && vmGlobals->threadCount > 0)
    {
       byte dontStop;
       do // call at least once, even if we have a event pending
       {
          SysTaskDelay(1); // let cpu take a breath
          executeMethod(runThreadsVclass, runThreadsMethod, runThreadsParams, 1); // guich@200b4_83: run the threads
          if (calcEventTimeout() != -1)
             timerCheck();
#ifdef PALMOS5
          if (vmGlobals->isHighDensity) SemaphoreRelease(1);
#endif
          dontStop = eventWait == evtWaitForever && !EvtEventAvail();
#ifdef PALMOS5
          if (vmGlobals->isHighDensity) SemaphoreReserve(1);
#endif
       } while (dontStop); // 11/12/02 Andy: added check for 'eventWait' so that only one iteration will be performed if a timelimit was given.
       timeout = 1;
    }
    else
    {
       timeout = calcEventTimeout();
       if (calcEventTimeout() != -1)
        timerCheck();
    }
    unlockHeap();

    // andy@330_25
    // This 'if' statement added to fix a bug where EvtEventAvail returns true, but
    // EvtGetEvent doesn't see an event (and so blocks).  If someone calls
    // pumpEvents with 'evtWaitForever' as the parameter, then pumpEvents() will act
    // exactly as it used to.  This change also allows the Window.pumpEvents() method
    // to pass in a harmless number (such as '1') and thus not block when there
    // really is no event available (why oh why does EvtEventAvail lie?).
    if(timeout == evtWaitForever || (eventWait != evtWaitForever && timeout > eventWait))
      timeout = eventWait;

    EvtGetEvent(&event, timeout);
    lockHeap();

    if (callLauncher) // guich@200b4_77: called MainWindow.exit()?
       return -1;

    if (vmStatus.errNum != 0) return -2; // guich@200b4_90

    if (event.eType == winExitEvent) // in the app start, there's an event thrown
    {
       WinSetDrawWindow(&vmGlobals->wabaFrmP->window); // guich@200b3: after a popup window is called and closed (like the contrast button), it sets the DrawWindow to NULL, so we set it back to our window.
       if (mustSwap) // guich@300_49: fixed bug when popping up a window like find or socket dialing. This also fix warning "proscribed function call" (only in PalmOS >= 3.5)
       {
          uint32 dummy = 1;
          SWCallPublic(SWLibRefNum, PUBLIC_GraphicsSwap, 0, &dummy); // make sure we are on again...
       }
       drawMainWin(1); // draw again all windows
       return 1;   // anodos: return 1
    }
    // search through the attached libraries so they can handle this event
    cont = 1;
    if (vmGlobals->nativeLibs)
       for (i = 0; vmGlobals->nativeLibs[i].ref; i++)
          if (SWHandleEvent(vmGlobals->nativeLibs[i].ref,&event) == SWErrEventHandled)
          {
             cont = 0;
             break; // break this loop
          }
    if (!cont)
       return 1; // guich@200b4_115: fixes device freezing when scanner get low batteries.

    // modified by guich: let the user intercept these keys (and do what they want!)
    interceptKey = 0;
    if (event.eType == keyDownEvent)
    {
       //debug(int2str("key",event.data.keyDown.chr));
       switch (event.data.keyDown.chr) // guich@200b4: changed to a switch to improve performance
       {
          // Inconditional intercepts
          case 5641: // hsChrPostProcess: guich@450_34: treo ALT key
             event.data.keyDown.chr = keyboardAlphaChr;
          case keyboardAlphaChr:
          case keyboardNumericChr:
          case keyboardChr:
             interceptKey = 1;
             break;
          // Conditional intercepts
          case hard1Chr   : interceptKey = (vmGlobals->interceptSystemKey & SK_HARD1)    != 0; break; // note: dont remove the != 0 or it wont work
          case hard2Chr   : interceptKey = (vmGlobals->interceptSystemKey & SK_HARD2)    != 0; break;
          case hard3Chr   : interceptKey = (vmGlobals->interceptSystemKey & SK_HARD3)    != 0; break;
          case hard4Chr   : interceptKey = (vmGlobals->interceptSystemKey & SK_HARD4)    != 0; break;
          case findChr    : interceptKey = (vmGlobals->interceptSystemKey & SK_FIND)     != 0; break;
          case calcChr    : interceptKey = (vmGlobals->interceptSystemKey & SK_CALC)     != 0; break;
          case vchrHardCradle: interceptKey = (vmGlobals->interceptSystemKey & SK_SYNC)  != 0; break; // guich@421_18
          case brightnessChr: // guich@450_25
          case contrastChr:
          case vchrPalmMin + 2: // guich@402_2: fixed contrast
             interceptKey = (vmGlobals->interceptSystemKey & SK_CONTRAST) != 0; break;
          case 0x20E      :
          case 5888:
          case 5889: // guich@450_23: M505 also see this as CLOCK, but it intercepts with vchrJogUp/vchrJogDown
            if (vmGlobals->isSony) break;
          case vchrPalmMin + 0: // guich@402_2: fixed clock
            interceptKey = (vmGlobals->interceptSystemKey & SK_CLOCK)     != 0; break;
          case launchChr  : interceptKey = (vmGlobals->interceptSystemKey & SK_LAUNCH)   != 0; break; // guich@200b4_5
          //default: debug(int2str("key",event.data.keyDown.chr)); // find out new codes
       }
    }
    if (!interceptKey)
    {
       if (event.eType == keyDownEvent && event.data.keyDown.chr == launchChr) // if we are not intercepting the home key and the user wants to quit, stop now. note: this is necessary bc in PalmOS 2 the appStopEvent was not being quewed correctly
       {
          // was the Home key is pressed?
          if (romVersion >= 0x03000000)
             SysHandleEvent(&event); // cannot be called in PalmOS 2
          callLauncher = 1; // guich@200b4_74
          return -1;
       }
       unlockHeap();
       //{char buf20[20]; xstrprintf(buf20,"%d,%d\n", event.screenX,event.screenY); DbgMessage(buf20); SysTaskDelay(10);}
       #ifdef PALMOS5
       SysTaskDelay(1); // guich@402_23
       #endif
       cont = SysHandleEvent(&event)
           || MenuHandleEvent((void *)0, &event, &err)
           || FrmDispatchEvent(&event);
       lockHeap();
       if (cont)
          return 1;
       if (event.eType == appStopEvent)
          return -1;
    }
    if (!vmStatus.errNum)
       handleMainWinEvent(&event);
    if (vmStatus.errNum > 0) // note: don't put a else here, bc the handleMainWinEvent may throw a fatal error
    {
       // guich@320_9: Is this an AppExitException?
       if(xstrncmp(vmStatus.arg1, "waba/sys/AppExitException", 24) == 0)
       {
          // Exit gracefully.
          vmStatus.errNum = 0;
          return -1;
       }
       // if error occured during processing, display it
       if (!vmGlobals->errorWinDisplayed)
          return -2; //drawErrorWin();
       else // guich@200b4_73: if a penDownEvent was the cause of the error, avoid it
       if (event.eType == penDownEvent)
       {
          callLauncher = 2; // guich@200b4_74 and guich@200b4_108
          return -1;
       }
    }
    return 1;
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
   WObject mainWinObj;
   Err err;
   int runApp;
   int i;
   int doReboot = 0; // guich@401_42
   char *ccmd = (char*)cmdPBP;

   // guich@350_3: added support for custom launchcode handling
   // note that we MUST be the current running app, otherwise, PilotMain
   // will never be called.
   if (nCodes2Launch > 0)
   {
      for (i = 0; i < nCodes2Launch; i++)
         if (codes2Launch[i].launchCode == cmd)
         {
            uint32 ret=0;
            SemaphoreReserve(1);
            SWCallPublic(codes2Launch[i].refNum,PUBLIC_subLaunched,&ret,cmdPBP);
            SemaphoreRelease(1);
            return ret;
         }
      return 0;
   }

   lockHeap();

   // NOTE: set vmStatus to 0 since we may not call VmInit() if there
   // is an device compatibility error or error accessing a class database
   xmemzero((uchar *)&vmStatus, sizeof(vmStatus));

   // Check for a required version of PalmOS - currently PalmOS 2.0
   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

#ifdef PALMOS5
   if (romVersion < 0x04000000) // note: palm os 5 also functions also supports Tungsten W and other 16bpp devices
#else
   if (romVersion < 0x02000000)
#endif
   {
      if ((launchFlags & sysAppLaunchFlagNewGlobals) && (launchFlags & sysAppLaunchFlagUIApp))
      {
         WinDrawChars("Incompatible device!",20,10,10); // guich@340_61
         SysTaskDelay(millisToTicks(2000));
         // PalmOS (all versions) will continuously relaunch
         // unless we switch back to the launcher because this
         // program was launched by another program
         AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch,NULL);
      }
      return sysErrRomIncompatible;
   }

#ifdef USE_STORAGE_HEAP
   createHeapDb();
   if (!heapDmRef)
   {
      //debug("Could not create database heap. Memory full?");
      return 0;
   }
#endif
   initVMGlobals();
   if (!vmGlobals)  // guich@200b4_78: most critical error, fail directly
      return 0;

/*{  // sanity check for xfree,xmalloc and xrealloc - non of these should fail
   int i;
   byte *p;
   uint32 ns = 20;
   byte *b = xmalloc(10);
   for (p=b,i=0; i < 10; i++)
      *p++ = i;
   xfree(b);
   b = xmalloc(10);
   if (!b)
      ErrFatalDisplay("xmalloc failed!");
   for (p=b,i=0; i < 10; i++)
      *p++ = i;
   b = xrealloc(b,&ns);
   if (!b)
      ErrFatalDisplay("xrealloc failed!");
   for (i=10; i < 20; i++)
      *p++ = i;
   for (p=b,i=0; i < 20; i++)
      if (*p++ != i)
         ErrFatalDisplay("xrealloc test failed!");
   xfree(b);
} */

   vmGlobals->romVersion = romVersion;
   vmGlobals->isSony = checkSony();

   // guich@421_50: preparse the command line to see if we want 16bpp screens
   if (ccmd != 0 && ccmd[0] != 0 && ccmd[0] == '/' && ccmd[1] == '1' && ccmd[2] == '6') // guich@421a_64: forgot to test ccmd against null
   {
      vmGlobals->wants16bpp = 1;
      ccmd += 4; // skip this parameter
   }

   // guich@200b4: load the libraries
  	err = SysLibLoad(SWLibTypeID, SWLibCreatorID,  &SWLibRefNum);
   ErrFatalDisplayIf(err,"UNABLE TO LOAD SWNATIVES LIBRARY! Please make sure the file SWNatives.prc is installed. The device will now be resetted...");
   SWLibOpen(SWLibRefNum, vmGlobals);
   vmGlobals->SWLibRefNum = SWLibRefNum;

   // create the 1000 form ...
   vmGlobals->wabaFrmP = FrmInitForm(DefaultFormForm);
   SWCallPublic(SWLibRefNum, PUBLIC_GraphicsInit, 0, 0); // guich@200b3
   SizeForm(vmGlobals->wabaFrmP); // resize the form if needed

#ifndef PALMOS5
   varInCodeSegmentVMGlobals(1,(long)vmGlobals); // guich@200b4_140: store the vmglobals so we can access it further
   vmGlobals->sysTrapWinSetActiveWindowPtr = SysGetTrapAddress(sysTrapWinSetActiveWindow);
   SysSetTrapAddress(sysTrapWinSetActiveWindow, trappedSysTrapWinSetActiveWindowType);

   mustSwap = romVersion < 0x03500000 || vmGlobals->maxColors <= 4;
   if (mustSwap) // guich@300_49
   {
      vmGlobals->evtGetPenTrapPtr = SysGetTrapAddress(sysTrapEvtGetPen);
      SysSetTrapAddress(sysTrapEvtGetPen, trappedEvtPenType);
      vmGlobals->sysTrapFrmCustomAlertPtr = SysGetTrapAddress(sysTrapFrmCustomAlert);
      SysSetTrapAddress(sysTrapFrmCustomAlert, trappedSysTrapFrmCustomAlertType);
   }
#endif

   runApp = 0;
   if (cmd == 40404)
      runApp = 1;
   mainWinObj = startApp(runApp, ccmd);

   if (mainWinObj != 0)
   {
      int32 runThreadsMethodMapNum;
      UtfString methodName, methodDesc;
      vmGlobals->globalMainWin = mainWinObj;
      // cache method map numbers for commonly called methods
      vclass = WOBJ_class(mainWinObj);
      methodName = createUtfString("_postEvent");
      methodDesc = createUtfString("(IIIIII)V");
      vmGlobals->postEventMethodMapNum = getMethodMapNum(vclass,&methodName,&methodDesc, CLASS_SEARCH_ALL,NULL,NULL);
      methodName = createUtfString("_onTimerTick");
      methodDesc = createUtfString("()V");
      vmGlobals->onTimerTickMethodMapNum = getMethodMapNum(vclass,&methodName,&methodDesc, CLASS_SEARCH_ALL,NULL,NULL);
      methodName = createUtfString("_runThreads");
      methodDesc = createUtfString("()V");
      runThreadsMethodMapNum = getMethodMapNum(vclass,&methodName,&methodDesc, CLASS_SEARCH_ALL,NULL,NULL);

      if (vmGlobals->postEventMethodMapNum == -1 || vmGlobals->onTimerTickMethodMapNum == -1 || runThreadsMethodMapNum == -1)
         mainWinObj = 0;

      runThreadsVclass = vclass;
		runThreadsMethod = getMethodByMapNum(runThreadsVclass, &runThreadsVclass, (uint16)runThreadsMethodMapNum);
      runThreadsParams[0].obj = mainWinObj;
		if (runThreadsMethod == NULL)
		   mainWinObj = 0;
   }

   if (mainWinObj == 0)
	{
	   UtfString utf;
	   utf.str = (char *)cmdPBP;
	   utf.len = cmdPBP?xstrlen(cmdPBP):0;
		VmError(ERR_CantInitializeApp, NULL,&utf,NULL); // guich@330_15: add the cmdline to the error message
	}

   // NOTE: we don't get any OS level window repaint events after the program has
   // started because PalmOS uses save-unders when it pops up windows on top of
   // this one (for alerts, launch win, etc.)
   if (vmStatus.errNum == 0)
      drawMainWin(0);

   // NOTE: PalmOS gets unhappy and may crash if you create a socket outside of
   // the "main event loop" so we explicity disallow it
   vmGlobals->globalNetState = NET_READY_FOR_OPEN;

   // activate the 1000 form ...
   FrmSetActiveForm(vmGlobals->wabaFrmP);
   WinSetDrawWindow(&vmGlobals->wabaFrmP->window); // guich@200b3: we must set the draw window again bc it is null until we do it.
   FrmSetEventHandler(vmGlobals->wabaFrmP,formHandler);
   while (1)
   {
      int result = pumpEvents(evtWaitForever);
      if (result == -1)
         break;
      // if error occured during processing, display it
      if (result == -2)
      {
         if (!vmGlobals->errorWinDisplayed)
            drawErrorWin();
         else
         {
            callLauncher = 2;
            break;
         }
      }
   }

   stopApp(mainWinObj);
   FrmDeleteForm(vmGlobals->wabaFrmP);

#ifndef PALMOS5
   SysSetTrapAddress(sysTrapWinSetActiveWindow, vmGlobals->sysTrapWinSetActiveWindowPtr);
   if (mustSwap) // guich@300_49
   {
      SysSetTrapAddress(sysTrapFrmCustomAlert, vmGlobals->sysTrapFrmCustomAlertPtr);
      SysSetTrapAddress(sysTrapEvtGetPen, vmGlobals->evtGetPenTrapPtr);
   }
#endif

   if (vmGlobals->globalNetState == NET_IS_OPEN)
      NetLibClose(vmGlobals->globalSocketLibRefNum, false);

   // guich@200b4: close the libs
   if (vmGlobals->nativeLibs)
      for (i = 0; vmGlobals->nativeLibs[i].ref; i++)
      {
         SWLibClose(vmGlobals->nativeLibs[i].ref);
         SysLibRemove(vmGlobals->nativeLibs[i].ref);
      }
   SWLibClose(SWLibRefNum); // maybe our lib is used by some of the libraries that are being closed?
   SysLibRemove(SWLibRefNum);

   if (vmGlobals->classHooks) xfree(vmGlobals->classHooks);
   if (vmGlobals->nativeLibs) xfree(vmGlobals->nativeLibs);
   destroyHeapDb(); // guich@200b3: bye bye heap. thanks for everything. guich@402_37 - moved from stopApp to here (after the libs are destroyed) - the PDBDriver needs this
   doReboot = vmGlobals->reboot;
   destroyVMGlobals();
   unlockHeap();

   // guich@401_42: reboot the device if requested
   if (doReboot)
      SysReset();

   // guich@200b4_74: in PalmOS 2.0, if we use the SysHandleEvent to call the launcher,
   // the app is switched immediatly, even before the "return 0", but, more important, before it calls
   // SysAppExit (at _startup_ code). This causes the Welcome app to never be called again.
   // To bypass this, we switch to the Preferences app.
   if (callLauncher) // guich@200b4_47 && guich@200b4_74: added callLauncher to explicitly indicate to call the launcher when needed. Removing this will make the vm keep calling the same app if a fatal error is found (or if Vm.exit() is called).
   {
      if (romVersion < 0x03000000 || callLauncher == 2)
      {
          AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch,NULL)
//          AppLaunchWithCommand('SWAB', sysAppLaunchCmdNormalLaunch,NULL) // ben combee problem is here
//          SysAppLauncherDialog();
      }
      else
      {
      	EventType event;
      	event.eType = keyDownEvent;
      	event.data.keyDown.chr = launchChr;
      	event.data.keyDown.modifiers = commandKeyMask;
         SysHandleEvent(&event); // be sure to switch to the launcher. note: in PalmOS2, the app is switched immediatly; the return 0 statement below isnt reached
      }
   }
   return 0;
}
//////////////////////////////////////////////////////////////////////////////
//                          Auxiliary System Functions                      //
//////////////////////////////////////////////////////////////////////////////

static byte isHanderaSilk()
{
#ifdef SUPPORT_HANDERA_HIGHRES
   UInt32 version;
   if (FtrGet(TRGSysFtrID, TRGSilkFtrNum, &version) == 0)
      if (sysGetROMVerMajor(version) >= 1)
      {
         return 1;
      }
#endif
   return 0;
}

void showSip(int16 on, int16 ontop)
{
   if (isHanderaSilk())
   {
      if (!on)
      {
         SilkMinimizeWindow();
         //drawMainWin(1);
      }
      else
      {
/* does not work!
         RectangleType originalAlpha, originalNum;
         SilkRestoreDefaultTemplates();
         if (ontop)
         {
            SilkGetAreas(&originalAlpha, &originalNum);
            originalAlpha.topLeft.y = 0;
            originalNum.topLeft.y = 0;
            SilkSetAreas(&originalAlpha, &originalNum);
         }
         else
         {
            //originalAlpha.topLeft.y = vmGlobals->screenHeight+16 - originalAlpha.extent.y;
            //originalNum.topLeft.y   = vmGlobals->screenHeight+16 - originalAlpha.extent.y; // same extent
         }
         */
         SilkMaximizeWindow();
      }
   }
}

int eventIsAvailable()
{
   int b;
   SemaphoreRelease(1); // guich@450_15: fixed problem of events not being captured
   SysTaskDelay(1);
   b = (int)EvtEventAvail();
   SemaphoreReserve(1);
   return b;
}

static byte checkSony() // returns true if is sony and hr mode is available
{
   SonySysFtrSysInfoP sonySysFtrSysInfoP;
   Err error = 0;

   //Sony ??
   if (!(error = FtrGet(sonySysFtrCreator,sonySysFtrNumSysInfoP, (UInt32*)&sonySysFtrSysInfoP)))
      return sonySysFtrSysInfoP->libr & sonySysFtrSysInfoLibrHR;

   return 0;
}

//static int stateX=170,stateY=170;
void showCommand(int show) // guich@102
{
   if (0 <= vmGlobals->stateX && vmGlobals->stateX <= vmGlobals->screenWidth && 0 <= vmGlobals->stateY && vmGlobals->stateY <= vmGlobals->screenHeight)
   {
   	FontID savedFont;
   	RectangleType rect;
   	// erase the position
   	rect.topLeft.x = vmGlobals->stateX;
   	rect.topLeft.y = vmGlobals->stateY;
   	rect.extent.x = 8;
   	rect.extent.y = 12;
   	WinEraseRectangle(&rect, 0);
   	if (show)
      {
         savedFont = FntSetFont(stdFont);
         WinDrawChars("\26",1,vmGlobals->stateX,vmGlobals->stateY);
      	FntSetFont(savedFont);
      }
      else // if capslock was set before command, show it again
      {
         Boolean capsLockP;
         Boolean numLockP;
         UInt16 tempShiftP;
         Boolean autoShiftedP;
         GrfGetState (&capsLockP,&numLockP,&tempShiftP,&autoShiftedP);
         if (capsLockP)
         {
            GsiSetShiftState (0,0); // erase, since it thinks its being shown
            GsiSetShiftState (glfCapsLock,0);
         }
      }
   }
}

UInt32 getDeviceFreeMem()
{
   UInt32 freeBytesP;
   MemCardInfo (0, 0,0,0,0,0,0, &freeBytesP);
   return freeBytesP;
}

// peterd@200b4_187
static uint32 baseTime;
static uint32 tps;

// this is good as documented
int32 getTimeStamp()
{
    int32 stamp;
    if (tps == 0)
    {
        tps = (uint32)SysTicksPerSecond();
        baseTime = TimGetTicks();
    }
    if ( tps == 100 )
        stamp = (TimGetTicks()-baseTime) * 10;    // common case is faster/better
    else
        stamp = (int32)(((int64)(TimGetTicks()-baseTime) * 1000) / tps);
    // NOTE: Timestamp must be in range from 0..(1 << 30)
    return stamp & 0x3FFFFFFF;    // documented wrap-around
}

uint32 millisToTicks(int32 millis)
{
	if (tps == 0)
		tps = (uint32)SysTicksPerSecond();

	if (millis <= 0)
		return 0;
	millis = millis % (int32)((1L << 30) / 1000L);
	return (millis * tps) / 1000L;
}

// guich@200b2: i putted this outside bc i use it for debug
void tone(int freq, int dur)
{
	SndCommandType sndCmd;
	sndCmd.cmd = sndCmdFreqDurationAmp;
	sndCmd.param1 = freq; // freq
	sndCmd.param2 = dur; // duration
	sndCmd.param3 = sndMaxAmp;
	SemaphoreRelease(1); // guich@421_29: try to make this work in Treo600
	SndDoCmd(NULL, &sndCmd, 0);
	SemaphoreReserve(1);
}
//////////////////////////////////////////////////////////////////////////////
//                             Debug Support                                //
//////////////////////////////////////////////////////////////////////////////
DmOpenRef DmFindAndOpenDB(char *name, uint32 creator, uint32 type, byte readOnly, byte onCard);
DmOpenRef DmFindAndOpenDB(char *name, uint32 creator, uint32 type, byte readOnly, byte onCard) // guich@450_32: now allow to open in read-only mode
{
   static char dbName[32];
	Boolean first;
	UInt cardNo;
	LocalID dbID;
	Err err;
	UInt16 attributes;

	first = true;
	if (type == 'appl' || creator == 'psys' || type == 'rsrc') return NULL; // guich@200b4_116: skip system files.
	while (1)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, type, creator, false, &cardNo, &dbID);
		if (err == dmErrCantFind)
			return NULL;
	 	if (DmDatabaseInfo(cardNo, dbID, dbName, &attributes,	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == 0)
		{
		   if ((attributes & dmHdrAttrResDB) == 0 // guich@200b4_116: skip resource databases
			   && !StrCompare(name, dbName))
				return DmOpenDatabase(cardNo, dbID, readOnly?dmModeReadOnly:dmModeReadWrite);
		}
		first = false;
	}
}
// guich@200 - now sends output to memopad instead of screen.
void systemOutClose();

void systemOutClose()
{
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED || vmGlobals->stateSystemOut == SYSOUT_ERASED || vmGlobals->dmRefSystemOut) // guich@200b4_19
   {
      if (vmGlobals->stateSystemOut == SYSOUT_OPENED)
         debug("======= End =======\n");
      DmCloseDatabase(vmGlobals->dmRefSystemOut);
      vmGlobals->dmRefSystemOut = NULL;
      vmGlobals->stateSystemOut = SYSOUT_CLOSED;
   }
}
static void systemOutOpen()
{
   #define ggetUInt32(b) (uint32)( (uint32)((b)[0])<<24 | (uint32)((b)[1])<<16 | (uint32)((b)[2])<<8 | (uint32)((b)[3]) )
   if (vmGlobals->dmRefSystemOut == 0) // guich@200b4_19
      vmGlobals->dmRefSystemOut = DmFindAndOpenDB("MemoDB",ggetUInt32("memo"),ggetUInt32("DATA"),0,0);
   if (vmGlobals->romVersion >= 0x03000000)
      vmGlobals->maxMemoLen = 32000;
   else
      vmGlobals->maxMemoLen = 4000;
   if (vmGlobals->dmRefSystemOut)
   {
      // finds the position to insert our data
		VoidHand recH;
		VoidPtr recPtr;
      int n = DmNumRecords(vmGlobals->dmRefSystemOut);
      int i;
      vmGlobals->stateSystemOut = SYSOUT_NOTFOUND;
      for (i =0; i < n; i++)
      {
      	recH = DmGetRecord(vmGlobals->dmRefSystemOut, i);
         //DbgMessage((*vmGlobals->int2str)("systemOutOpen handle:",(uint32)recH));
      	if (recH) // not deleted?
      	{
         	recPtr = MemHandleLock(recH);
         	if (!recPtr)
         		DmReleaseRecord(vmGlobals->dmRefSystemOut, i, false);
            else
	     		if (IS_TITLE(recPtr)) // "waba.system.out\n".length
   	  		{
     			   vmGlobals->posSystemOut = i;
		         MemHandleUnlock(recH);
               DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
     			   vmGlobals->stateSystemOut = SYSOUT_OPENED;
     		   	break;
        		}
	     		else
   	  		{
      	      MemHandleUnlock(recH);
         	   DmReleaseRecord(vmGlobals->dmRefSystemOut, i, false);
            }
         }
      }

      if (vmGlobals->stateSystemOut == SYSOUT_NOTFOUND)
      {
         vmGlobals->posSystemOut = n;
         recH = DmNewRecord(vmGlobals->dmRefSystemOut, &vmGlobals->posSystemOut, 1);
         if (!recH)
            vmGlobals->stateSystemOut = SYSOUT_COULDNOTCREATE;
         else
         {
	         recPtr = MemHandleLock(recH);
	         if (!recPtr)
	         {
	            DmDeleteRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
	            vmGlobals->stateSystemOut = SYSOUT_COULDNOTCREATE;
	         }
	         else
	         {
	            // everything is ok. release the record.
		         MemHandleUnlock(recH);
               DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
		         vmGlobals->stateSystemOut = SYSOUT_OPENED;
		         // write the header
		         debug(vmGlobals->titleSystemOut);
		      }
		   }
      }
      if (vmGlobals->stateSystemOut != SYSOUT_OPENED)
         DmCloseDatabase(vmGlobals->dmRefSystemOut); // close the database
   }
}
void debug(char *msg)
{
   static char buf[10];
	char *cr0 = "\n\0";
   int slen = msg?xstrlen(msg):0;

#ifdef ENABLE_TRACE_
{
   static int count=10;
   DbgMessage(msg);
   DbgMessage("\n");
   if (count-- == 0)
   {
      SysTaskDelay(10);
      count = 10;
   }
}
#else
   if (vmGlobals->stateSystemOut == SYSOUT_CLOSED) // start the db
      systemOutOpen();
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED)
   {
		VoidHand recH;
		VoidPtr recPtr;

      if (!xstrncmp(msg,"!erase debug!",13)) // guich@420_18: delete the debug memo?
      {
         DmRemoveRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
         vmGlobals->stateSystemOut = SYSOUT_ERASED;
         systemOutClose();
         return;
      }

     	recH = DmGetRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut);
   	if (recH) // not deleted?
   	{
      	recPtr = MemHandleLock(recH);
      	if (!recPtr)
      		DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
         else
         {
            ULong totalLen = slen+2;  // plus 2 for the cr0
		      ULong lenH = MemHandleSize(recH)-1; // minus 1 for the \0
		      int isTitle = IS_TITLE(msg);
		      totalLen += lenH;
		      if (totalLen > vmGlobals->maxMemoLen) // guich@200b2: dont let the memo grow too much. palm os 2.0 hangs!
		      {
		         lenH = xstrlen(vmGlobals->titleSystemOut)+2; // keep title
		         totalLen = slen+2+lenH;
		      }

		      if (!isTitle) totalLen += 7; // size of timestamp
		      MemHandleUnlock(recH);
		      DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
		     	recH = DmResizeRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, totalLen); // open space to the new string
		   	if (recH)
		   	{
		   	   recH = DmGetRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut); // guich@300_46
		   	   if (recH)
		   	   {
   			   	recPtr = MemHandleLock(recH);
   			   	if (!recPtr)
   				      DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
   				   else
   				   {
   				   	// prepend with timestamp + space
   				   	if (!isTitle)
   				   	{
   					      int32 number = getTimeStamp();
   						   buf[7] = 0;
   						   buf[6] = ' ';
   					  	   buf[5] = (number            % 10) + '0';
   						   buf[4] = (number / 10L      % 10) + '0';
   						   buf[3] = (number / 100L     % 10) + '0';
   						   buf[2] = (number / 1000L    % 10) + '0';
   						   buf[1] = (number / 10000L   % 10) + '0';
   						   buf[0] = (number / 100000L  % 10) + '0';
   						   DmWrite(recPtr, lenH, buf, 7); lenH += 7;
   						}
   				      DmWrite(recPtr, lenH, msg, slen); lenH += slen;
   				      DmWrite(recPtr, lenH, cr0, 2);
   				      MemHandleUnlock(recH);
   				      DmReleaseRecord(vmGlobals->dmRefSystemOut, vmGlobals->posSystemOut, false);
   				   }
   				}
			   }
         }
      }
   }
#endif
}
//////////////////////////////////////////////////////////////////////////////
//                               VM Globals                                 //
//////////////////////////////////////////////////////////////////////////////

static void initVMGlobals()
{
   // guich@200b4: alloc space for the globals
   vmGlobals = xmalloc(sizeof(struct VMGlobals));
   if (!vmGlobals)
   {
      VmQuickError(ERR_CantAllocateMemory, NULL); // guich@200b4_78 - guich@300: removed a call using vmGlobals
      return;
   }
   xmemzero(vmGlobals,sizeof(struct VMGlobals));
   vmGlobals->stateX = 500;
   vmGlobals->stateY = 500;
   vmGlobals->titleSystemOut = "SuperWaba Debug Console";
   vmGlobals->posSystemOut = -1; // index of record being appended
   // initialize the functions table
   // nm<platform>_defines.h
   vmGlobals->xmalloc = xmalloc;
   vmGlobals->xrealloc = xrealloc;
   vmGlobals->xfree_ = xfree_;
   // nm<platform>_bridge.c
   vmGlobals->getTimeStamp = getTimeStamp;
   vmGlobals->millisToTicks = millisToTicks;
   vmGlobals->postStopEvent = postStopEvent;
   vmGlobals->int2str = int2str;
   vmGlobals->debug = debug;
   vmGlobals->showCommand = showCommand;
   vmGlobals->tone = tone;
   vmGlobals->lockWarpRec = lockWarpRec;
   vmGlobals->unlockWarpRec = unlockWarpRec;
   vmGlobals->findAndOpenDB = DmFindAndOpenDB;
   vmGlobals->executePostEvent = executePostEvent;
   vmGlobals->eventIsAvailable = eventIsAvailable;
   vmGlobals->pumpEvents = pumpEvents;
   // xplat_javalang.c
   vmGlobals->toLowerCase = toLowerCase;
   vmGlobals->toUpperCase = toUpperCase;
   vmGlobals->unsigned2hex = unsigned2hex;
   vmGlobals->StringToLowerCase = StringToLowerCase;
   vmGlobals->StringToUpperCase = StringToUpperCase;
   vmGlobals->StringBufferEnsureCapacity = StringBufferEnsureCapacity;
   vmGlobals->StringBufferSetLength = StringBufferSetLength;
   vmGlobals->appendString = appendString;
   vmGlobals->appendCharArray = appendCharArray;
   vmGlobals->appendCharArrayOffLen = appendCharArrayOffLen;
   vmGlobals->appendChar = appendChar;
   vmGlobals->appendInt = appendInt;
   vmGlobals->appendFloat = appendFloat;
   vmGlobals->appendDouble = appendDouble;
   vmGlobals->valueOfDouble = valueOfDouble;
   vmGlobals->StringCopyChars = StringCopyChars;
   vmGlobals->StringEquals = StringEquals;
   vmGlobals->StringCompareTo = StringCompareTo;
   vmGlobals->indexOfChar = indexOfChar;
   vmGlobals->indexOfCharStart = indexOfCharStart;
   vmGlobals->StringIndexOf = StringIndexOf;
   vmGlobals->StringHashCode = StringHashCode;
   vmGlobals->ObjectToStringNative = ObjectToStringNative;
   vmGlobals->ObjectGetClass = ObjectGetClass;
   vmGlobals->ThrowablePrintStackTraceNative = ThrowablePrintStackTraceNative;
   vmGlobals->ClassForName = ClassForName;
   vmGlobals->ClassNewInstance = ClassNewInstance;
   vmGlobals->ClassIsInstance = ClassIsInstance;
   // superwaba.c
   vmGlobals->doubleToStr = doubleToStr;
   vmGlobals->float2str = float2str;
   vmGlobals->int642str = int642str;
   vmGlobals->double2str = double2str;
   vmGlobals->str2int32 = str2int32;
   vmGlobals->str2double = str2double;
   vmGlobals->str2int64 = str2int64;
   vmGlobals->intToString = intToString;
   vmGlobals->stringToUtf16 = stringToUtf16;
   vmGlobals->dumpStackTrace = dumpStackTrace;
   vmGlobals->callMethod = callMethod;
   vmGlobals->createString = createString;
   vmGlobals->stringToUtf = stringToUtf;
   vmGlobals->gc = gc;
   vmGlobals->createArrayObject = createArrayObject;
   vmGlobals->pushObject = pushObject;
   vmGlobals->popObject = popObject;
   vmGlobals->arrayRangeCheck = arrayRangeCheck;
   vmGlobals->VmQuickError = VmQuickError;
   vmGlobals->throwException = throwException;
   vmGlobals->createStringFromCharPtr = createStringFromCharPtr;
   vmGlobals->createStringObject = createStringObject;
   vmGlobals->getClassName = getClassName;
   vmGlobals->getClass = getClass;
   vmGlobals->compatible = compatible;
   vmGlobals->executeMethod = executeMethod;
   vmGlobals->getMethodByMapNum = getMethodByMapNum;
   vmGlobals->getMethodMapNum = getMethodMapNum;
   vmGlobals->printToBuf = printToBuf;
   vmGlobals->nativeLoadClass = nativeLoadClass;
   vmGlobals->loadClassField = loadClassField;
   vmGlobals->constantToVar = constantToVar;
   vmGlobals->loadClassMethod = loadClassMethod;
   vmGlobals->createUtfString = createUtfString;
   vmGlobals->getUtfString = getUtfString;
   vmGlobals->createObject = createObject;
   vmGlobals->arraySize = arraySize;
   vmGlobals->arrayType = arrayType;
   vmGlobals->createMultiArray = createMultiArray;
   vmGlobals->createStringFromUtf = createStringFromUtf;
   vmGlobals->copyArray = copyArray;
   vmGlobals->getField = getField;
   vmGlobals->getClassByIndex = getClassByIndex;
   vmGlobals->getMethod = getMethod;
   vmGlobals->countMethodParams = countMethodParams;
   vmGlobals->compatibleArray = compatibleArray;
   vmGlobals->allocObject = allocObject;
   vmGlobals->expandArray = expandArray;
   vmGlobals->arrayCount = arrayCount;
   vmGlobals->showSip = showSip;
   vmGlobals->vmStatus = &vmStatus;
   vmGlobals->registerLaunchCode = registerLaunchCode;
   vmGlobals->unregisterLaunchCode = unregisterLaunchCode;
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES   
   vmGlobals->BmpGlueSetTransparentValue = BmpGlueSetTransparentValue;
#endif   
}
static void destroyVMGlobals()
{
   if (vmGlobals->className)  xfree(vmGlobals->className);
   xfree(vmGlobals);
}
//////////////////////////////////////////////////////////////////////////////////////
