/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#if defined(PALMOS) && !defined(__arm__)
#define LOCK_MEMORY     { MemSemaphoreReserve(true); }
#define UNLOCK_MEMORY   { MemSemaphoreRelease(true); }
#define PALM_CARD(x)	x,
#endif

#define ALLOW_ACCESS_TO_INTERNALS_OF_CLIPBOARDS
#define ALLOW_ACCESS_TO_INTERNALS_OF_CONTROLS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FIELDS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FINDPARAMS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FORMS
#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_MENUS
#define ALLOW_ACCESS_TO_INTERNALS_OF_PROGRESS
#define ALLOW_ACCESS_TO_INTERNALS_OF_SCROLLBARS
#define ALLOW_ACCESS_TO_INTERNALS_OF_TABLES
#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FONTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_WINDOWS

#include <PalmOS.h>											// Standard Palm stuff
#include <PalmCompatibility.h>								// use new headers
#include <SysEvtMgr.h>
#include <SerialMgr.h>
#include <NetMgr.h>
#include <FloatMgr.h>
#include <DLServer.h>
#include "Screen.h" // Dana support - guich@400_83
#include "TrgChars.h" // handera jogdial support
#include "Silk.h" // handera virtual silk support
#include "vga.h" // handera support
#include "SonyCLIE.h" // clie support
#include "waba_res.h"
#include "PalmChars.h" // FiveWaySDK (see faq)
//#include "PenInputMgr.h"

//#define SUPPORT_HANDERA_HIGHRES // guich note: we cannot support Handera because its virtual keyboard does not support moving to top.

//
// basic types
//

typedef long long int64;
#define uchar unsigned char
#define int32 long
#define uint32 unsigned long
#define float32 float
#define int16 short
#define uint16 unsigned short
#define byte unsigned char

#ifdef PALMOS5
 typedef int32 Pixel;
#else
 typedef byte Pixel;
#endif

typedef union  // added double bits too here
{
   double	 d;	  // for easy assignment of values
   FlpDouble fd;	  // for calling New Floating point manager routines
   uint32	 ul[2]; // for accessing upper and lower longs
   int64 ll;
   long l[2];
} Comp64;

#define _inline_ inline

//
// Register definitions for method parameters
//

#define inA0 :__A0
#define inA1 :__A1
#define inD0 :__D0
#define inD1 :__D1
#define inD2 :__D2

//
// type converters
//

#define getCreator(b) (uint32)( (uint32)((b)[0])<<24 | (uint32)((b)[1])<<16 | (uint32)((b)[2])<<8 | (uint32)((b)[3]) )

static uint32 temp32;
// guich@200b4: we cannot use .w or .l bc the address of *b can be odd and PalmOS can't access odd addresses. :-(
inline asm static uint32 getUInt32(uchar *b: __A0):__D0
{
   lea temp32, a1
   move.b (a0),(a1)		// don't change to (a0)+,(a1)+ ! code smaller but slowest!
   move.b 1(a0),1(a1)
   move.b 2(a0),2(a1)
   move.b 3(a0),3(a1)
   move.l (a1),d0
}
inline asm static int32 getInt32(uchar *b: __A0):__D0
{
   lea temp32, a1
   move.b (a0),(a1)
   move.b 1(a0),1(a1)
   move.b 2(a0),2(a1)
   move.b 3(a0),3(a1)
   move.l (a1),d0
}
inline asm static uint16 getUInt16(uchar *b: __A0):__D0
{
   move.b    (a0)+,d0 // 01
   lsl.w     #8,d0    // 10
   move.b    (a0),d0  // 12
}
inline asm static int16 getInt16(uchar *b: __A0):__D0
{
   move.b    (a0)+,d0 // 01
   lsl.w     #8,d0    // 10
   move.b    (a0),d0  // 12
}

// still havent found a good place to use this code. It makes necessary store a variable in the An register instead of in the Dn register
inline asm static void getu32(uchar *from: __A0, uchar *to: __A1)
{
   move.b (a0),(a1)		// don't change to (a0)+,(a1)+ ! code smaller but slowest!
   move.b 1(a0),1(a1)
   move.b 2(a0),2(a1)
   move.b 3(a0),3(a1)
}
inline asm static void getu16(uchar *from: __A0, uchar *to: __A1)
{
   move.b (a0),(a1)		// don't change to (a0)+,(a1)+ ! code smaller but slowest!
   move.b 1(a0),1(a1)
}

inline static float32 getFloat32(uchar *buf)
{
	uint32 i;
	float32 f;

	// we need to make sure we're aligned before casting
	i = getUInt32(buf);
	f = *((float32 *)&i);
	return f;
}

static double get64bits(uchar *buf)
{
	FlpCompDouble f;
   f.ul[0] = getInt32(buf);
   f.ul[1] = getInt32(buf+4);

   return f.d;
}

inline asm static uint32 ushr(uint32 number:__D0, int32 count:__D1):__D0
{
   lsr.l d1,d0
}

/////// memory locking ///////
#ifndef USE_STORAGE_HEAP
 static void lockHeap() {}
 static void unlockHeap() {}
 #define SemaphoreReserve(t)
 #define SemaphoreRelease(t)
#else
 #define SemaphoreReserve(t) {MemSemaphoreReserve(t);}
 #define SemaphoreRelease(t) {MemSemaphoreRelease(t);}
static int locked=0;
static void lockHeap()
{
   if (locked <= 0)
   {
      SemaphoreReserve(1);
      locked++;
   }
}
static void unlockHeap()
{
//   if (locked)
   {
      SemaphoreRelease(1);
      locked--;
   }
}
#endif

#ifdef PALMOS5
#define GR_ENTER {if (vmGlobals->isHighDensity) WinSetCoordinateSystem(kCoordinatesNative);}
#define GR_EXIT  {if (vmGlobals->isHighDensity) WinSetCoordinateSystem(kCoordinatesStandard);}
#else
#define GR_ENTER
#define GR_EXIT
#endif

//
// x portability functions
//

#define xstrncmp(s1, s2, n) StrNCompare(s1, s2, n)
#define xstrncpy(dst, src, n) StrNCopy(dst, src, (uint32)n)
#define xstrcpy(dst, src) StrCopy(dst, src)
#define xstrlen(s) StrLen(s)
#define xstrcat(dst, src) StrCat(dst, src)
#define xmemmove(dst, src, size) MemMove(dst, src, size)
#define xmemzero(mem, len) MemSet(mem, len, (Byte)0)
#define xmemset(mem, what, len) MemSet(mem, len, (Byte)what)
#define xstrprintf StrPrintF
#define xstrcaselesscmp StrCaselessCompare
#define xstrstr StrStr
#define xmemcmp MemCmp

#ifdef PALMOS5 // In PalmOS 5, those functions are ARM (30x faster)
   #define xfer32(from, to, count) xmemmove(to,from,((count) << 2)) // total of size*4 bytes transfered
   #define xzero32(mem,len) xmemzero(mem,((len) << 2))
#else
   // guich@200b4_86: ultrafast data transfer. note that the total bytes transfered are count*4
   asm static inline void xfer32(byte *from:__A0, byte *to:__A1, uint16 count: __D0)
   {
      subq.l    #1,d0
   loop:
      move.l    (a0)+,(a1)+
      dbf.w     d0,loop
   }
   asm static inline void xzero32(byte *start:__A0, uint16 count: __D0)
   {
      subq.l    #1,d0
   loop:
      clr.l     (a0)+
      dbf.w     d0,loop
   }
#endif

#if defined(__DEFINE_FUNCTIONS) && !defined(__DONTDEFINE_MEMFUNCS)

   extern DmOpenRef heapDmRef;
   extern int GCCreated;

   void *xmalloc(uint32 size);
   void gc_();
   void debug(char *s);
   char *int2str(char *msg, int32 ii);
   #define XMALLOC_MARKSSIZE 6
   #define XMALLOC_EXTRASIZE (sizeof(VoidHand) + 2 + XMALLOC_MARKSSIZE)

   static void *addMemMarks(uint32 size, byte *ptr)
   {
	   // guich@550: detecting begin/end overwrites
      size -= XMALLOC_EXTRASIZE;
	   *((uint16*)ptr)++ = (uint16)size;
	   *ptr++ = 0x01;
	   *ptr++ = 0x01;
	   // p will go here
      *(ptr+size)   = 0x01;
      *(ptr+size+1) = 0x01;
      return ptr;
   }
   static byte* verifyMemMarks(void *p, char*msg)
   {
      if (p)
      {
         byte *ptr = (byte *)p;
         if (*(ptr-1) != 0x01 || *(ptr-2) != 0x01)
         {		              
            char buf[60];
            xstrprintf(buf,"%s - chunk's start was overwritten: %lx",msg, (long)p);
            debug(buf);
            return 0;
         }
         else
         {
            uint16 size = *((uint16*)(ptr-4));
            if (*(ptr+size) != 0x01 || *(ptr+size+1) != 0x01)
            {  
               char buf[70];
               xstrprintf(buf,"%s - chunk's end was overwritten: %lx (%ld)",msg,(long)p,(long)size);
               debug(buf);
               return 0;
            }
         }
         return ptr-4;
      }
      return p;
   }
   

   // resizes the ptr. If newSize is smaller than the original size, always successful.
   // Otherwise, may return null if no more memory. Note that the newSize may not
   // be bigger than 64k-32 (PalmOS limit). Returns a ptr with the new size
   // (maybe with different base address)
   // note that this routine does not change where the pointer is allocated.
   // If it was in storage heap, it remains in storage heap; if it was in
   // dynamic heap, it remains in there.
   // if growing, the new area allocated is zeroed out (keeping the old data intact).
   // if p is null, memory is allocated using xmalloc.
   void *xrealloc(void *p, uint32 *newSize);
   void *xrealloc(void *p, uint32 *newSize)
   {
	   VoidHand memH;
	   uchar *ptr=NULL;
	   int firstTry = 1;
	   uint16 pos;
	   uint32 oldSize=0;
	   if (*newSize < 1024) *newSize = 1024; // when on Palm, if using storage heap, allocing something under 1024 would be stored in the main memory, but a new reallocation to above 1024 would make it fail bc it would try to alloc on the storage heap
      if (p != NULL)
      {
         ptr = verifyMemMarks(p, "xrealloc");
   	   ptr -= 2;
   	   pos = *((uint16 *)ptr);
   	   ptr = (byte *)ptr - sizeof(VoidHand);
   	   memH = *((VoidHand *)ptr);
   	   oldSize = MemHandleSize(memH); // oldsize already includes extra space
   	   *newSize += XMALLOC_EXTRASIZE; // new size here does not contain the extra space; add it now
   	   if (oldSize == *newSize)
   	   {
   	      *newSize -= XMALLOC_EXTRASIZE;
   	      return p; // nothing to do
   	   }
   	   if (*newSize > 65510) // too big
   	   {
   	      *newSize -= XMALLOC_EXTRASIZE;
   	      return NULL;
   	   }
   	   MemHandleUnlock(memH);
tryAgain:
   	   #ifdef USE_STORAGE_HEAP
   	   if (pos != (uint16)-1) // storage heap?
   	   {
   	      DmReleaseRecord(heapDmRef, pos, 0);
   	      memH = DmResizeRecord(heapDmRef, pos, (uint32)*newSize); // resize the record to have only 8 bytes
            memH = DmGetRecord(heapDmRef, pos); // andy@340_4: Sets the busy bit for this record.
         }
   	   else
   	   #endif
	 	   if (memH && MemHandleResize(memH, *newSize) != 0) // guich@550_39: added memH first check
    	      memH = 0;
   	   if (!memH)
   	   {
   	      if (firstTry && GCCreated) // guich@400_14
   	      {
   	         gc_(); // gc may free some memory blocks
   	         firstTry--;
   	         goto tryAgain;
   	      }
   	      *newSize -= XMALLOC_EXTRASIZE;
   		   return NULL;
   		}
  	 	   // locks memory and store again the pointers
         ptr = MemHandleLock(memH);
         *((VoidHand *)ptr) = memH;
         ptr += sizeof(VoidHand);
         *((uint16 *)ptr) = pos;
         ptr += 2;
         if (*newSize > oldSize)
            xmemzero(ptr+oldSize-(sizeof(VoidHand)+2),*newSize-oldSize);
         ptr = addMemMarks(*newSize, ptr);
         *newSize -= XMALLOC_EXTRASIZE;
      }
      else
      {
         ptr = xmalloc(*newSize);
         xmemzero(ptr,*newSize);
      }
      return ptr;
   }

   static uint16 getNextHeapPos() // guich@421a_66
   {
      int32 i;
      MemHandle h;
      for (i=DmNumRecords(heapDmRef)-1; i >= 0; i--)
         if ((h=DmQueryRecord(heapDmRef,i)) && MemHandleSize(h) == 8)
            DmRemoveRecord(heapDmRef,i);
         else
            break;
      return (uint16)(i+1);
   }

   void *xmalloc(uint32 size)
   {
	   VoidHand memH;
	   uchar *ptr;
	   int firstTry = 1;
	   uint16 pos = (uint16)-1;

	   // we stick the handle in the first bytes and then return a pointer
	   // inside the memory block. Then when we free it, we can get the
	   // memory handle back to unlock and free it without having to track it.
	   // guich@200b3: we stick also the position of the chunk
	   size += XMALLOC_EXTRASIZE;
tryAgain:
#ifdef USE_STORAGE_HEAP
	   if (size >= 1024) // guich@200b3: if size is greater than 1k, alloc space at the storage heap
	   {
	      pos = getNextHeapPos(); // guich@421a_66: reuse last added records. DmNumRecords(heapDmRef); // add always at the end
		   memH = DmNewRecord(heapDmRef, &pos, size);
      }
      else
#endif
         memH = MemHandleNew(size);
	   if (!memH)
	   {
	      if (firstTry && GCCreated) // guich@400_14
	      {
	         gc_(); // gc may free some memory blocks
	         firstTry--;
	         goto tryAgain;
	      }
		   return NULL;
		}
	   ptr = MemHandleLock(memH);

	   // find memory leaks
      // to find a memory leak, you must:
      // 1. get the number that Pose warns
      // 2. put a breakpoint in nmpalm_b.c at SysAppExit
      // 3. when breakpoint is hit, do a "hd 0" in the Palm Debug Console (CodeWarrior)
      // 4. find the handle associated with the number given in 1.
      // 5. Use this handle in the comparision here and set the breakpoint in the "pos" dummy line.
   //	if ((int16)memH == 0x1AF8)
   //	   pos = pos * 2 / 2;

	   *((VoidHand *)ptr) = memH;
	   ptr += sizeof(VoidHand);
	   *((uint16 *)ptr) = pos;
	   ptr += 2;
	   
	   return addMemMarks(size, ptr);
   }
   void xfree_(void *p);
   void xfree_(void *p)
   {
	   VoidHand memH;
	   uchar *ptr;
	   uint16 pos;
	   p = verifyMemMarks(p,"xfree");
      if (p)
      {
   	   ptr = (byte*)p - 2;
   	   pos = *((uint16 *)ptr);
   	   ptr = (byte *)ptr - sizeof(VoidHand);
   	   memH = *((VoidHand *)ptr);
   	   
   	   MemHandleUnlock(memH);
   	   #ifdef USE_STORAGE_HEAP
   	   if (pos != (uint16)-1) // storage heap?
   	   {
   	      DmReleaseRecord(heapDmRef, pos, 0);

   	      if (pos == (DmNumRecords(heapDmRef)-1)) // guich@421a_66
   	         DmRemoveRecord(heapDmRef,pos);
   	      else
   	         DmResizeRecord(heapDmRef, pos, (uint32)8); // resize the record to have only 8 bytes (we can't resize to 0)
   	   }
   	   else
   	   #endif
   	      MemHandleFree(memH);
      }
   }
#endif // __DEFINE_FUNCTIONS