/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

// guich@200: I really wont like to see this code running in another JAVA virtual machine for the PalmOS. if you want do so, please ask me before.

static ULong oldBPP;
static int usePalmOSFuncs=0; // use internal routines?
static UInt16 sonyRefNum;
static uint32 *lookupY;

// If they can dificult, why make easier? Sony created HR functions in libraries
// and we need to open the library, use the function, and close it. Why didnt they made
// like HandEra? Just detect the highres mode and use the appropriate func. To bypass having
// to change all methods, i created this.

/* Important note: Palm os added support for color and grayscale starting in
 * version 3.5. Before that, there was no o.s. functions to handle it. so
 * i made the Palm Gray Primitives to handle it in palm os prior to 3.5, which
 * is detected at runtime.
 */

////////////////////////////////////////////////////////////////////////////
//                        PALM COLOR PRIMITIVES                           //
////////////////////////////////////////////////////////////////////////////
// note: for optimization, colors in the primitive routines are always treated by their indexes, not by rgb values.

#define COLOR_HASHSIZE 255
struct htElemColor
{
    uint32 key;
    int32 value;
    struct htElemColor *next;
};

static struct htElemColor **colorHashTable; // guich@400_33        /* The storage array for the methods */
#define htHash_Color(key) (unsigned)((key) & COLOR_HASHSIZE)

/* return the htElemColor found, else NULL if key doesn't exist */
static struct htElemColor *htGetColor(uint32 key)
{
   struct htElemColor *np;
   for (np = colorHashTable[htHash_Color(key)]; np; np = np->next)
      if (np->key == key)
         return np;
   return NULL;
}
/* return struct containing value or NULL if unable to store */
static struct htElemColor *htPutColor(uint32 key, int32 value)
{
   struct htElemColor *np;
   unsigned hashval;
   if ((np = htGetColor(key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElemColor *) (*vmGlobals->xmalloc) (sizeof(struct htElemColor));
      if (np != NULL)
      {
          hashval = htHash_Color(key);
          /* point to last w/ this hash */
          np->next = colorHashTable[hashval];
          np->key = key;
          np->value = value;
          /* attach to hashtab array */
          colorHashTable[hashval] = np;
      }
   }
   return np;
}
static void disposeColorHashtable(byte freeTable)
{
   // dispose the hashtable
   int i;
   for (i = 0; i <= COLOR_HASHSIZE; i++)
   {
      struct htElemColor *np;
      np = colorHashTable[i];
      while (np)
      {
         struct htElemColor *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      colorHashTable[i] = NULL;
   }
   if (freeTable) xfree(colorHashTable);
}

static int32 *lastPal;
static int lastPalStart=-2,lastPalEnd=-1; // guich@230_18
static void setPalette(int32 *pal, int start, int end)
{
   int len = end-start+1;
   if (isColor)
   {
      if (start != lastPalStart || end != lastPalEnd || MemCmp(pal+start,lastPal+start,len*4) != 0) // guich@230_18
      {
         WinHandle old = WinSetDrawWindow(GRHANDLE_WinHandle(mainGrHandle)); // guich@200b4_22
   	   WinPalette(winPaletteSet,start,len,(RGBColorType *)pal);
   	   WinSetDrawWindow(old);
         if (len > 16 || start < 200) disposeColorHashtable(0); // cached colors are now invalid.
         lastPal = pal;
         lastPalStart = start;
         lastPalEnd = end;
      }
	}
   xmemmove(palette+start,pal,4*len); // copy this palette to the global palette
}

int mustSwap = 0;
static IndexedColorType rgb2index(int r, int g, int b)
{
	RGBColorType rgb;
	WinHandle old;
	IndexedColorType idx;
	rgb.r = r;
	rgb.g = g;
	rgb.b = b;
   if (mustSwap) old = WinSetDrawWindow(GRHANDLE_WinHandle(mainGrHandle)); // guich@200b4_160
	idx = WinRGBToIndex(&rgb);
   if (mustSwap) WinSetDrawWindow(old);
   return idx;
}
// this method returns the index in the current palette, not the color
static inline int32 getColor_Color(int red, int green, int blue)
{
   if (isColor)
   {
      IndexedColorType idx;
  	   uint32 key = ((int32)red << 16) | ((int32)green << 8) | blue;
   	struct htElemColor *elem = htGetColor(key);
      if (elem != NULL)
         return elem->value;
      idx = rgb2index(red,green,blue);
      htPutColor(key,idx);
      return idx;
   }
   else
   {
      int lum = (red+green+blue)/3;//(30 * red + 59 * green + 11 * blue) / 100; // converts to luminance
      return 15-lum/16;
   }
}
////////////////////////////////////////////////////////////////////////////
static void getBitmapRow_Color(GrHandle h, int x, int y, int w, byte *row)
{
   if (GRHANDLE_IsBitmap(h))
      xmemmove(row,GRHANDLE_MemAddr(h) + GRHANDLE_WidthInBPP(h)*(y) + x, w);
   else
   {
	   WinHandle oldH = WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
	   if (sonyRefNum)
   	   while (w--) *row++ = HRWinGetPixel(sonyRefNum,x++,y);
   	else
   	   while (w--) *row++ = WinGetPixel(x++,y);
	   WinSetDrawWindow(oldH);
	}
}
////////////////////////////////////////////////////////////////////////////
static void drawScanLine(int dstX, int dstY, int width, Pixel *pixels)
{
   byte lastColor = pixels[0];
   int lastI=0, i=1;
   for (; i < width; i++)
      if (pixels[i] != lastColor)
      {
         (*drawHLineNOCT)(dstX+lastI,dstY,i-lastI,(int)lastColor);
         lastI = i;
         lastColor = pixels[i];
      }
   (*drawHLineNOCT)(dstX+lastI,dstY,i-lastI,(int)lastColor);
}
////////////////////////////////////////////////////////////////////////////
// cuts the area x,y,width,height of the bitmap bmp with dimensions bmpW,bmpH to the (current active) screen location dstX,dstY
static void drawBitmap_Color(GrHandle h, int x, int y, int width, int height, int dstX, int dstY, int drawOp, int32 backColor, int32 foreColor, int doClip)
{
   int bmpW = GRHANDLE_Width(h);
   int bmpH = GRHANDLE_Height(h);
   int originalWidth = width;
   int originalHeight = height;
   int i;
   RectangleType r1;
   WinDrawOperation op;

   dstX += TRANSX;
   dstY += TRANSY;
   if (!doClip)
   {
      /*
      | Although doClip is not set, we make sure that the area of the bitmap
      | that we want to copy IS inside its area
      | If doClip is set, then the clipping computation below ensures
      | the sanity of the operation (and is quite as fast).
      | DoClip is important for game sprites where programmers ensure that the
      | sprite will never pass screen boundaries.
      */
      if ((x <= -width) || (x >= bmpW) || (y <= -height) || (y >= bmpH))
         return;
   }
   else
   {
      /* clip the source rectangle to the source surface */
      if (x < 0)
      {
         width += x;
         dstX -= x;
         x = 0;
      }
      i = bmpW - x;
      if (width > i)
         width = i;
      if (y < 0)
      {
         height += y;
         dstY -= y;
         y = 0;
      }
      i = bmpH - y;
      if (height > i)
         height = i;

      /* clip the destination rectangle against the clip rectangle */
      if (dstX < CLIPX1)
      {
         i = CLIPX1 - dstX;
         dstX = CLIPX1;
         x += i;
         width -= i;
      }
      if ((dstX + width) > CLIPX2)
         width = CLIPX2 - dstX;
      if (dstY < CLIPY1)
      {
         i = CLIPY1 - dstY;
         dstY = CLIPY1;
         y += i;
         height -= i;
      }
      if ((dstY + height) > CLIPY2)
         height = CLIPY2 - dstY;

      /* check the validity */
      if ((width <= 0) || (height <= 0))
         return;
   }

   makeRect(x,y,width,height,r1);
   //{char buf[80]; xstrprintf(buf, "%d,%d,%d,%d", (int)x, (int)y,(int)width,(int)height); (*vmGlobals->debug)(buf);}

   if (usePalmOSFuncs && GRHANDLE_Palette(h)) // guich@200b3: the palette is not set back; otherwise, bye bye bitmap colors
      setPalette(GRHANDLE_Palette(h),0,GRHANDLE_PaletteSize(h)-1);

   switch (drawOp)
   {
      case WIN_ERASE:        op = winErase; break;
      case WIN_MASK:         op = winMask; break;
      case WIN_INVERT:       op = winInvert; break;
      case WIN_OVERLAY:      op = winOverlay; break;
      case WIN_PAINT_INVERSE:op = winPaintInverse; break;
      case WIN_REPLACE_COLOR: // guich@200b4: since palm doenst have this kind of operation, i have to do it by myself
         {
            int i,j;
            int *bmpPt;
            int *bmpRow = (*vmGlobals->xmalloc)(2*width);
            if (!bmpRow) return;
            for (j = 0; j < height; j++)
            {
               // get the entire row
               WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
               bmpPt = bmpRow;
               if (sonyRefNum)
                  for (i = 0; i < width; i++) *bmpPt++ = HRWinGetPixel(sonyRefNum,x+i,y+j);
               else
                  for (i = 0; i < width; i++) *bmpPt++ = WinGetPixel(x+i,y+j);
               bmpPt = bmpRow;
               WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(activeGrHandle));
               // change the pixels for that row
               for (i = 0; i < width; i++)
                  if (*bmpPt++ != backColor)
                     (*setPixelNOCT)(dstX+i,dstY+j,foreColor);
            }
            xfree(bmpRow);
         }
         return;
      case WIN_SWAP_COLORS:
         WinSetBackColor(backColor);
         WinSetForeColor(foreColor);
         op = winPaint; // first we have to paint, and after that swap colors
         break;
      case WIN_SPRITE:
         //if (bmpPt != backColor) screenPt = bmpPt;
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
         (*vmGlobals->BmpGlueSetTransparentValue)((BitmapType *)GRHANDLE_BmpHandle(h),backColor); // guich@340_2: use now the Palm OS Glue
#endif         
         // no break here!
      case WIN_PAINT:
      default:
         op = winPaint;
         break;
   }

   if (sonyRefNum)
      HRWinCopyRectangle(sonyRefNum,(WinHandle)GRHANDLE_WinHandle(h),(WinHandle)GRHANDLE_WinHandle(activeGrHandle), &r1,dstX,dstY, op);
   else
      WinCopyRectangle((WinHandle)GRHANDLE_WinHandle(h),(WinHandle)GRHANDLE_WinHandle(activeGrHandle), &r1,dstX,dstY, op);
   if (drawOp == WIN_SWAP_COLORS) // now we swap the colors
   {
      makeRect(dstX,dstY,width,height,r1);
      WinSetDrawMode(winSwap);
      if (sonyRefNum)
         HRWinPaintRectangle(sonyRefNum,&r1,0);
      else
         WinPaintRectangle(&r1,0);
      WinSetDrawMode(winPaint);
   } else
	if (drawOp == WIN_SPRITE)
      ((BitmapType *)GRHANDLE_BmpHandle(h))->flags.hasTransparency = 0; // guich@340_2: updated the flag name
}
////////////////////////////////////////////////////////////////////////////
//                        COLOR (8BPP) PRIMITIVES                         //
////////////////////////////////////////////////////////////////////////////
// gets the bit value of the pixel, using the current translation. returns -1 if error (out of clip bounds)
asm static int32 getPixel_8bpp (int x, int y)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     error        // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bgt.s     error        // if x > x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     error        // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bgt.s     error        // if y > y2 end

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0        // + x
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel
    moveq     #0,d0        // clear d0, if not, it may return invalid bits
    move.b    (a0),d0      // get the pixel contents
    bra.s     end
error:
    moveq.w   #-1,d0
end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invPixelNOCT_8bpp (int x, int y) // ps: this is a palette based invert, not a color based
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ~ *scr;
    not.b     (a0)
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invVLineNOCT_8bpp (int x, int y, int h) // ps: this is a palette based invert, not a color based
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0

    lea       widthInBPP,a1// get width in bpp - note that here we save in a1 bc we're going to reuse it
    sub.l     d0,d0        // store in a1
    move.w    (a1),d0
    move.l    d0,a1

    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    move.w    8(a7),d0     // d0 = h
    beq.s     end          // end if h = 0
    subq.w    #1,d0        // h--
loop:
    // *scr = ~ *scr;
    not.b     (a0)
    add.l     a1,a0      // scr += widthInBPP
    dbf.w     d0,loop      // while (--h);

end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// erase one color to the other, but does not translates or clip rect
asm static void eraseVLineNOCT_8bpp (int x, int y, int h, int32 cfrom, int32 cto)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    lea       widthInBPP,a1// get width in bpp - note that here we save in a1 bc we're going to reuse it
    sub.l     d0,d0        // store in a1
    move.w    (a1),d0
    move.l    d0,a1

    // *scr = color
    move.w    12(a7),d2    // d2 = cfrom
    move.w    16(a7),d1    // d1 = cto
    move.w    8(a7),d0     // d0 = h
    beq.s     end          // end if h = 0
    subq.w    #1,d0        // h--
loop:
    cmp.b     (a0),d2      // are the bits the same?
    bne.s     cont         // no, the current color dont match the old one
    move.b    d1,(a0)
cont:
    add.l     a1,a0      // scr += widthInBPP
    dbf.w     d0,loop      // while (--h);

end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value, but does not translates or clip rect
asm static void setPixelNOCT_8bpp (int x, int y, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0  // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = color
    move.w    10(a7),d2     // d2 = color
    move.b    d2,(a0)
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the upper clip bounds
asm static void setPixelXM_8bpp (int x, int y, int32 index)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // verify clip x upper bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = color
    move.w    10(a7),d2     // d2 = color
    move.b    d2,(a0)
end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the clip bounds, using the current translation
asm static void setPixel_8bpp (int x, int y, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     end          // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge.s     end          // if y >= y2 end

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = color
    move.w    10(a7),d2     // d2 = color
    move.b    d2,(a0)
end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1, without translate or clip rect
asm static void drawHLineNOCT_8bpp (int x, int y, int w, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    move.w    8(a7),d1     // get w
    tst.w     d1           // == 0 ?
    beq       end          // if yes, bye bye

    subq.w    #1,d1        // d1 -= 1 due to dbf
    move.w    12(a7),d0    // d0 = color
    // do *scr++ = color;  while (--howManyFullPixels)

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// fills a rect from x,y to x+w-1,y+h-1, without translate or clip rect
asm static void fillRectNOCT_8bpp (int x, int y, int w, int h, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // save the variables
    lea       widthInBPP,a1  // get width in bpp
    sub.l     d2,d2          // store the widthInBpp in d2
    move.w    (a1),d2
    move.l    a0,a1        // save a0 in a1

    move.w    8(a7),d1     // get w

extLoop:
    subq.w    #1,d1        // d1 -= 1 due to dbf
    move.w    14(a7),d0    // d0 = color
    // do *scr++ = color;  while (--howManyFullPixels)

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

end:
    subq.w    #1, 10(a7)       // dec height
    tst.w     10(a7)       // height == 0?
    beq.s     realEnd      // if yes, end
    add.l     d2, a1
    move.l    a1,a0
    move.w    8(a7), d1
	 bra.s     extLoop      // and draw again

realEnd:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1
asm static void drawHLine_8bpp (int x, int y, int w, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip y bounds - in the h line, if outside y bounds, end here
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt       end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge       end          // if y >= y2 end

    // verify clip x bounds - sets x = max(x,clipX1), w = min(w,clipX2-x)
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0        // d0-d2 >= 0 -> d0 >= d2
    bge.s     xIsGreaterX1 // if x >= x1, continue
    // since x < x1 (lower clip) test if x+w < x1. if true, exit: the line is outside the clip
    sub.w     d0,d2        // d2 = d2 - d0 (x1 = x1 - x)
    cmp.w     8(a7),d2     // d2-w >= 0
    bge       end          // if (x1-x >= w) then the line is smaller than clipX1. end
    sub.w     d2,8(a7)     // subtract from w what we are going to add to x
    move.l    (a0),d0      // else x = x1
xIsGreaterX1:
    swap      d2           // get x2
    sub.w     d0,d2        // clipW = x2 - x
    ble.s     end          // end if clipW < 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipW > w, continue
    move.w    d2,8(a7)     // else w = clipW

goAhead:
    // move the computed values back to the stack
  //  move.w    d0,4(a7)     // x
    andi.l    #0x0000FFFF,d0 // clear d0 high bits

    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // if w <= 8 use the normal setPixel. this will speedup the process
    move.w    8(a7),d1     // get w
    tst.w     d1           // == 0 ?
    beq       end          // if yes, bye bye

    subq.w    #1,d1        // d1 -= 1 due to dbf
    move.w    12(a7),d0    // d0 = color
    // do *scr++ = color;  while (--howManyFullPixels)

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

end:
    rts                    // and return
}
////////////////////////////////////////////////////////////////////////////
// clears the screen with the specified 8-bit pattern. clears the hole screen, ignoring translate and clip
asm static void clearScreen_8bpp (int32 color) // color must be 8 bits
{
    sub.l     d1,d1        // clear d1
    lea       maxY,a0      // get maxY
    move.w    (a0),d1      // store in d1

    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d0   // get y * widthInBPP

    lsr.l     #2,d0        // divide by 4
    subi.l    #1,d0        // minus 1

    move.l    4(a7),d2     // gets the pattern to fill
    move.b    d2,d1        // d1=000000XX
    lsl.w     #8,d1        // d1=0000XX00
    move.b    d2,d1        // d1=0000XXXX
    move.w    d1,d2        // saves d1 in d2
    swap      d1           // d1=XXXX0000
    move.w    d2,d1        // d1=XXXXXXXX
    lea       activeScr,a0
    movea.l   (a0),a0
loop:
    move.l    d1,(a0)+
    dbf.w     d0,loop

    rts
}
////////////////////////////////////////////////////////////////////////////
// draws a line from y to y+h-1
asm static void drawVLine_8bpp (int x, int y, int h, int32 color)
{
    sub.l     d1,d1        // clear d1
    sub.l     d0,d0
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds - in the v line, if outside x bounds, end here
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt       end          // if x < clipx1 end
    swap      d2           // get clipx2
    cmp.w     d2,d0
    bgt       end          // if x > clipx2 end

    // verify clip y bounds - sets y = max(y,clipY1), h = min(h,clipY2-y)
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    bge.s     yIsGreaterY1 // if y >= clipy1, continue
    // guich@200b4_121: was not testing if  y+h < clipy1
    add.w     8(a7),d1     // y += h
    cmp.w     d2,d1        // compare y+h with clipy1
    ble.s     end          // if <=, nothing to draw; end
    move.w    d1,8(a7)		// guich@400_75 - h will hold the difference of the clipping. 1) h = y+h
    neg.w     d2				// 2) clipy1 = -clipy1
    add.w     d2,8(a7)     // 3) h = y+h + -clipy1
    neg.w     d2           // 4) --clipy1 = clipy1
    move.w    d2,d1        // else y = clipy1
yIsGreaterY1:
    swap      d2           // get clipy2
    sub.w     d1,d2        // clipH = clipy2 - y
    ble.s     end          // end if clipH <= 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipH > h, continue
    move.w    d2,8(a7)     // else h = clipH

goAhead:
    // uchar *scr = &activeScr[widthInBPP*y + x];
    lea       lookupY,a0    // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1         // index = y * 4
    move.l    (a0,d1),d1	 // get y * widthInBPP
    add.l     d1,d0
    lea       activeScr,a0  // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0         // increment a0 to point to the pixel

    // *scr = color
    move.w    12(a7),d2     // d2 = color
    move.w    8(a7),d0      // d0 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d0         // h--
    lea       widthInBPP,a1 // store widthInBPP in d1
    sub.l     d1,d1
    move.w    (a1),d1
loop:
    move.b    d2,(a0)
    add.l     d1,a0         // scr += widthInBPP
    dbf.w     d0,loop       // while (--h);
end:
    rts
}
////////////////////////////////////////////////////////////////////////////
//                        GrayScale (4BPP) PRIMITIVES                     // guich@200b4_20
////////////////////////////////////////////////////////////////////////////
// gets the bit value of the pixel, using the current translation. returns -1 if error (out of clip bounds)
asm static int32 getPixel_4bpp (int x, int y)
{
    sub.l     d0,d0
    sub.l     d1,d1
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     error        // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bgt.s     error        // if x > x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     error        // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bgt.s     error        // if y > y2 end

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1

    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP
    add.l     d1,d0

    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // pixel = (pixel >> data[x&1]) & F
    moveq     #0,d0        // clear d0, if not, it may return invalid bits
    move.b    (a0),d0      // d0 = pixel
    andi.w    #0x1,d2      // d2 = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    move.b    (a1),d1      // d1 = value to right shift

    lsr.b     d1,d0        // *scr >> 2
    andi.b    #0xF,d0      // erase the rest of the bits
    bra.s     end
error:
    moveq.w   #-1,d0
end:
    rts                    // and return

data:
    dc.b   4,0
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invVLineNOCT_4bpp (int x, int y, int h) // ps: this is a palette based invert, not a color based
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a1   // get the lookup y table
    movea.l   (a1),a1
    lsl.w     #2,d1        // index = y * 4
    move.l    (a1,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    // prepare the eor operator
    move.w    #0xF,d2      // d2 = color mask
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]

    // prepare the loop parameters
    move.w    8(a7),d1      // d1 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d1         // h--
    lea       widthInBPP,a1  // get width in bpp

loop:
    eor.b     d2,(a0)      // add the and with the shift left
    add.w     (a1),a0       // scr += 80
    dbf.w     d1,loop       // while (--h);

end:
    rts                    // and return

data:
    dc.b   4,0
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invPixelNOCT_4bpp (int x, int y) // ps: this is a palette based invert, not a color based
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    move.w    #0xF,d2      // d2 = color mask

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    eor.b     d2,(a0)      // add the and with the shift left

    rts                    // and return

data:
    dc.b   4,0
}
////////////////////////////////////////////////////////////////////////////
// erase one color to the other, but does not translates or clip rect
asm static void eraseVLineNOCT_4bpp (int x, int y, int h, int32 cfrom, int32 cto)
{
    movem.l   d3-d4/a2,-(a7) // save 2 registers
    sub.l     d0,d0
    sub.l     d1,d1

    // prepare the loop parameters
    move.w    20(a7),d3      // d3 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d3         // h--

    // gets x and y
    move.w    16(a7),d0     // x
    move.w    18(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a2   // get the lookup y table
    movea.l   (a2),a2
    lsl.w     #2,d1        // index = y * 4
    move.l    (a2,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]

    lea       widthInBPP,a2// get width in bpp
    move.b    (a1),d1      // d1 = value to left shift
    move.w    24(a7),d2    // d2 = old color
    move.w    28(a7),d4    // d4 = new color
    lsl.b     d1,d2        // old color << data[x&1] - put the color bits in same position
    lsl.b     d1,d4        // new color << data[x&1]

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
loop:
    move.b    (a0),d0      // d0 = current color
    and.b     4(a1),d0     // erase the other bits
    cmp.b     d2,d0        // are the bits the same?
    bne.s     cont         // no, the current color dont match the old one

    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // erase those bits; *scr & data[x&1+2]
    or.b      d4,(a0)      // add the and with the shift left (set the new color)
cont:
    add.w     (a2),a0       // scr += 80
    dbf.w     d3,loop       // while (--h);

end:
    movem.l   (a7)+,d3-d4/a2 // restore saved registers
    rts                    // and return

data:
    dc.b   4,0
    dc.b   0x0F,0xF0
    dc.b   0xF0,0x0F
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value, but does not translates or clip rect
asm static void setPixelNOCT_4bpp (int x, int y, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    move.w    10(a7),d2     // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+2]
    or.b      d2,(a0)      // add the and with the shift left

end:
    rts                    // and return

data:
    dc.b   4,0
    dc.b   0x0F,0xF0
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the upper clip bounds
asm static void setPixelXM_4bpp (int x, int y, int32 index)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    move.w    10(a7),d2     // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+2]
    or.b      d2,(a0)      // add the and with the shift left

end:
    rts                    // and return

data:
    dc.b   4,0
    dc.b   0x0F,0xF0
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the clip bounds, using the current translation
asm static void setPixel_4bpp (int x, int y, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     end          // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge.s     end          // if y >= y2 end

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // x = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]
    move.w    10(a7),d2     // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+2]
    or.b      d2,(a0)      // add the and with the shift left

end:
    rts                    // and return

data:
    dc.b   4,0
    dc.b   0x0F,0xF0
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1, without translate or clip rect
asm static void drawHLineNOCT_4bpp (int x, int y, int w, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // saves x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // if w <= 4 use the normal setPixel. this will speedup the process
    move.w    8(a7),d0     // get w
    tst.w     d0           // == 0 ?
    beq       end          // if yes, bye bye
    cmpi.w    #4,d0        // compares to 4
    bgt.s     fits         // continue with normal process if > 4
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #1,d2        // x = x&1  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&1]
    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+2]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #1,d2        // is modulus 2
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    rts  // guich@200b4: ! changed 2 rts to only one!

data3:
    dc.b   4,0
    dc.b   0x0F,0xF0

    // draws the pixels if they arent modulus 2
fits:
    andi.w    #1,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 2

    lea       data2(d2),a1 // a1 = data[w&1]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    2(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    4(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 1;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #1,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    12(a7),d0
    muls.w    #0x11,d0     // color = 0x11*color
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 2
    // *scr = (data[w&1+2]*color) | (data[w&1] & *scr)
    andi.w    #1,d2        // w = w&1  (d2 already has w)
    beq.s     end          // no more pixels remain
    // only 1 pixel remain
    and.b     #0x0F,(a0)
    move.w    12(a7),d0
    lsl.b     #4,d0
    or.b      d0,(a0)

end:
    rts                    // and return

data2:
    dc.b   0, 0xF0     // and
    dc.b   0, 1        // or
    dc.b   0, 1        // sub
}
////////////////////////////////////////////////////////////////////////////
// fills a rect from x,y to x+w-1,y+h-1, without translate or clip rect
asm static void fillRectNOCT_4bpp (int x, int y, int w, int h, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // saves x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // save the variables
    lea       widthInBPP,a1  // get width in bpp
    move.w    (a1),d0        //
    lea       dataA0,a1
    move.w    d0,10(a1)
    move.l    a0,(a1)      // save a0
    move.w    d2,4(a1)     // save x
    move.w    8(a7),d0     // get w
    move.w    d0,6(a1)     // save w

extLoop:
    // if w <= 4 use the normal setPixel. this will speedup the process
    cmpi.w    #4,d0        // compares to 4
    bgt.s     fits         // continue with normal process if > 4
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #1,d2        // x = x&1  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&1]
    move.w    14(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+2]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #1,d2        // is modulus 2
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    bra.s       end  // guich@200b4: ! changed 2 rts to only one!

data3:
    dc.b   4,0
    dc.b   0x0F,0xF0
data2:
    dc.b   0, 0xF0     // and
    dc.b   0, 1        // or
    dc.b   0, 1        // sub

    // draws the pixels if they arent modulus 2
fits:
    andi.w    #1,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 2

    lea       data2(d2),a1 // a1 = data[w&1]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    2(a1),d1     // d1 = value to mul
    muls.w    14(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    4(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 1;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #1,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    14(a7),d0
    muls.w    #0x11,d0     // color = 0x11*color
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 2
    // *scr = (data[w&1+2]*color) | (data[w&1] & *scr)
    andi.w    #1,d2        // w = w&1  (d2 already has w)
    beq.s     end          // no more pixels remain
    // only 1 pixel remain
    and.b     #0x0F,(a0)
    move.w    14(a7),d0
    lsl.b     #4,d0
    or.b      d0,(a0)

end:
    subq.w    #1, 10(a7)       // dec height
    tst.w     10(a7)       // height == 0?
    beq.s     realEnd      // if yes, end
    lea       dataA0,a1
    move.w    6(a1),d0      // restore w
    move.w    4(a1),d2      // restore x
    move.l    (a1),a0       // restore a0
    add.l     8(a1),a0      // go to next line
    move.l    a0,(a1)      // save a0 again
    move.w    d2,4(a7)
    move.w    d0,8(a7)
	 bra       extLoop      // and draw again

realEnd:
    rts                    // and return

dataA0:
    dc.l   0 // a0
    dc.w   0 // x
    dc.w   0 // w
    dc.l   0 // widthInBPP
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1
asm static void drawHLine_4bpp (int x, int y, int w, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip y bounds - in the h line, if outside y bounds, end here
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt       end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge       end          // if y >= y2 end

    // verify clip x bounds - sets x = max(x,clipX1), w = min(w,clipX2-x)
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0        // d0-d2 >= 0 -> d0 >= d2
    bge.s     xIsGreaterX1 // if x >= x1, continue
    // since x < x1 (lower clip) test if x+w < x1. if true, exit: the line is outside the clip
    sub.w     d0,d2        // d2 = d2 - d0 (x1 = x1 - x)
    cmp.w     8(a7),d2     // d2-w >= 0
    bge       end          // if (x1-x >= w) then the line is smaller than clipX1. end
    sub.w     d2,8(a7)     // subtract from w what we are going to add to x
    move.l    (a0),d0      // else x = x1
xIsGreaterX1:
    swap      d2           // get x2
    sub.w     d0,d2        // clipW = x2 - x
    ble.s     end2          // end if clipW < 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipW > w, continue
    move.w    d2,8(a7)     // else w = clipW

goAhead:
    // move the computed values back to the stack
    move.w    d0,4(a7)     // x
    andi.l    #0x0000FFFF,d0 // clear d0 high bits

    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // saves x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // if w <= 8 use the normal setPixel. this will speedup the process
    move.w    8(a7),d0     // get w
    tst.w     d0           // == 0 ?
    beq       end          // if yes, bye bye
    cmpi.w    #4,d0        // compares to 8
    bgt.s     fits         // continue with normal process if > 8
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #1,d2        // x = x&1  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&1]
    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&1]
    move.b    2(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&1+4]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #1,d2        // is modulus 2
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    rts

data3:
    dc.b   4,0
    dc.b   0x0F,0xF0

    // draws the pixels if they arent modulus 2
fits:
    andi.w    #1,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 2

    lea       data2(d2),a1 // a1 = data[w&1]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&1] & *scr
    move.b    2(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&1+4]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    4(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 1;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #1,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    12(a7),d0
    muls.w    #0x11,d0     // color = 0x11*color (propagate color to all pixels)
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 2
    // *scr = (data[w&1+2]*color) | (data[w&1] & *scr)
    andi.w    #1,d2        // w = w&1  (d2 already has w)
    beq.s     end          // no more pixels remain
    // only 1 pixel remain
    and.b     #0x0F,(a0)
    move.w    12(a7),d0
    lsl.b     #4,d0
    or.b      d0,(a0)

end:
    rts                    // and return

data2:
    dc.b   0, 0xF0     // and
    dc.b   0, 1        // or
    dc.b   0, 1        // sub
}
////////////////////////////////////////////////////////////////////////////
// clears the screen with the specified 8-bit pattern. clears the whole screen, ignoring translate and clip
asm static void clearScreen_4bpp (int32 color) // color must be 8 bits
{
    sub.l     d1,d1        // clear d1
    lea       maxY,a0      // get maxY
    move.w    (a0),d1      // store in d1

    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d0   // get y * widthInBPP


    lsr.l     #2,d0        // divide by 8 (each 32 bits has 8 colors)
    subi.l    #1,d0        // minus 1

    move.l    4(a7),d2     // gets the pattern to fill
    move.b    d2,d1        // d1=0000000X
    lsl.b     #4,d1        // d1=000000X0		 guich@200b4_197: fixed color being wrongly computed
    or.w      d2,d1        // d1=000000XX
    move.w    d1,d2        // save again in d1
    lsl.w     #8,d1        // d1=0000XX00
    move.b    d2,d1        // d1=0000XXXX
    move.w    d1,d2        // saves d1 in d2
    swap      d1           // d1=XXXX0000
    move.w    d2,d1        // d1=XXXXXXXX
    lea       activeScr,a0
    movea.l   (a0),a0
loop:
    move.l    d1,(a0)+
    dbf.w     d0,loop

    rts
}
////////////////////////////////////////////////////////////////////////////
// draws a line from y to y+h-1
asm static void drawVLine_4bpp (int x, int y, int h, int32 color)
{
    sub.l     d0,d0
    sub.l     d1,d1
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds - in the v line, if outside x bounds, end here
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt       end          // if x < clipx1 end
    swap      d2           // get clipx2
    cmp.w     d2,d0
    bgt       end          // if x > clipx2 end

    // verify clip y bounds - sets y = max(y,clipY1), h = min(h,clipY2-y)
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    bge.s     yIsGreaterY1 // if y >= clipy1, continue
    // guich@200b4_121: was not testing if  y+h < clipy1
    add.w     8(a7),d1     // y += h
    cmp.w     d2,d1        // compare y+h with clipy1
    ble.s     end          // if <=, nothing to draw; end
    move.w    d1,8(a7)		// guich@400_75 - h will hold the difference of the clipping. 1) h = y+h
    neg.w     d2				// 2) clipy1 = -clipy1
    add.w     d2,8(a7)     // 3) h = y+h + -clipy1
    neg.w     d2           // 4) --clipy1 = clipy1
    move.w    d2,d1        // else y = clipy1
yIsGreaterY1:
    swap      d2           // get clipy2
    sub.w     d1,d2        // clipH = clipy2 - y
    ble.s     end          // end if clipH <= 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipH > h, continue
    move.w    d2,8(a7)     // else h = clipH

goAhead:
    // uchar *scr = &activeScr[80*y + (x>>1)];
    move.w    d0,d2        // save x in d2
    lsr.w     #1,d0        // x>>1
    lea       lookupY,a0   // get the lookup y table
    movea.l   (a0),a0
    lsl.w     #2,d1        // index = y * 4
    move.l    (a0,d1),d1   // get y * widthInBPP

    add.l     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.l    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    andi.w    #0x1,d2      // d2 = x&1  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&1]

    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // saves the lsl param in d1
    lsl.b     d1,d2        // color = (color<<data[x&1]);

    move.w    8(a7),d0      // d0 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d0         // h--

    move.b    2(a1),d1      // saves the and param in d1

    // *scr = ((*scr & data[x&1+2]) + color << data[x&1]);
    lea       widthInBPP,a1
loop:
    and.b     d1,(a0)       // *scr & 0x0F
    or.b      d2,(a0)
    add.w     (a1),a0       // scr += 80
    dbf.w     d0,loop       // while (--h);
end:
    rts
data:
    dc.b   4,0
    dc.b   0x0F,0xF0
}
////////////////////////////////////////////////////////////////////////////
//                   PALM GRAYSCALE (2BPP) PRIMITIVES                     //
////////////////////////////////////////////////////////////////////////////
static int32 getColor_Gray(int red, int green, int blue)
{
   int lum = (((red << 5) + (green << 6) + (blue << 2)) / 100) >> 6;
   //int lum = ((red + 59 * green + 11 * blue) / 100; // converts to luminance
   return 3 - lum;
}
////////////////////////////////////////////////////////////////////////////
// gets the bit value of the pixel, using the current translation. returns -1 if error (out of clip bounds)
asm static int32 getPixel_2bpp (int x, int y)
{
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     error        // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bgt.s     error        // if x > x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     error        // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bgt.s     error        // if y > y2 end

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0// get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // pixel = (pixel >> data[x&3]) & 3
    moveq     #0,d0        // clear d0, if not, it may return invalid bits
    move.b    (a0),d0      // d0 = pixel
    andi.w    #0x3,d2      // d2 = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]
    move.b    (a1),d1      // d1 = value to right shift

    lsr.b     d1,d0        // *scr >> 2
    andi.b    #0x3,d0      // erase the rest of the bits
    bra.s     end
error:
    moveq.w   #-1,d0
end:
    rts                    // and return

data:
    dc.b   6,4,2,0
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invVLineNOCT_2bpp (int x, int y, int h)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a1  // get width in bpp
    mulu.w    (a1),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // x = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]
    // prepare the eor operator
    move.w    #3,d2         // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]

    // prepare the loop parameters
    move.w    8(a7),d1      // d1 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d1         // h--
    lea       widthInBPP,a1  // get width in bpp

loop:
    eor.b     d2,(a0)      // add the and with the shift left
    add.w     (a1),a0       // scr += 40
    dbf.w     d1,loop       // while (--h);

end:
    rts                    // and return

data:
    dc.b   6,4,2,0
}
////////////////////////////////////////////////////////////////////////////
// inverts the pixel, but does not translates or clip rect
asm static void invPixelNOCT_2bpp (int x, int y)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // x = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]
    move.w    #3,d2         // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    eor.b     d2,(a0)      // add the and with the shift left

    rts                    // and return

data:
    dc.b   6,4,2,0
}
////////////////////////////////////////////////////////////////////////////
// erase one color to the other, but does not translates or clip rect
asm static void eraseVLineNOCT_2bpp (int x, int y, int h, int32 cfrom, int32 cto)
{
    movem.l   d3-d4/a2,-(a7) // save 2 registers

    // prepare the loop parameters
    move.w    20(a7),d3      // d3 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d3         // h--

    // gets x and y
    move.w    16(a7),d0     // x
    move.w    18(a7),d1     // y

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a2// get width in bpp
    mulu.w    (a2),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0 // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // x = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]

    move.b    (a1),d1      // d1 = value to left shift
    move.w    24(a7),d2    // d2 = old color
    move.w    28(a7),d4    // d4 = new color
    lsl.b     d1,d2        // old color << data[x&3] - put the color bits in same position
    lsl.b     d1,d4        // new color << data[x&3]

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
loop:
    move.b    (a0),d0      // d0 = current color
    and.b     8(a1),d0     // erase the other bits
    cmp.b     d2,d0        // are the bits the same?
    bne.s     cont         // no, the current color dont match the old one

    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // erase those bits; *scr & data[x&3+4]
    or.b      d4,(a0)      // add the and with the shift left (set the new color)
cont:
    add.w     (a2),a0       // scr += 40
    dbf.w     d3,loop       // while (--h);

end:
    movem.l   (a7)+,d3-d4/a2 // restore saved registers
    rts                    // and return

data:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC
    dc.b   0xC0,0x30,0xC,0x3
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the upper clip bounds
asm static void setPixelXM_2bpp (int x, int y, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // x = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]
    move.w    10(a7),d2     // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&3+4]
    or.b      d2,(a0)      // add the and with the shift left

end:
    rts                    // and return

data:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the clip bounds, using the current translation
asm static void setPixel_2bpp (int x, int y, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt.s     end          // if x < x1 end
    swap      d2           // get x2
    cmp.w     d2,d0
    bge.s     end          // if x >= x2 end

    // verify clip y bounds
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt.s     end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge.s     end          // if y >= y2 end

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // x = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]
    move.w    10(a7),d2     // d2 = color

    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&3+4]
    or.b      d2,(a0)      // add the and with the shift left

end:
    rts                    // and return

data:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1, without translate or clip rect
asm static void drawHLineNOCT_2bpp (int x, int y, int w, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // saves x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // if w <= 8 use the normal setPixel. this will speedup the process
    move.w    8(a7),d0     // get w
    tst.w     d0           // == 0 ?
    beq       end          // if yes, bye bye
    cmpi.w    #8,d0        // compares to 8
    bgt.s     fits         // continue with normal process if > 8
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #3,d2        // x = x&3  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&3]
    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&3+4]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #3,d2        // is modulus 4
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    rts

data3:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC

    // draws the pixels if they arent modulus 4
fits:
    andi.w    #3,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 4

    lea       data2(d2),a1 // a1 = data[w&3]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    8(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 2;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #2,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    12(a7),d0
    muls.w    #0x55,d0     // color = 0x55*color
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 4
    // *scr = (data[w&3+3]*color) | (data[w&3] & *scr)
    andi.w    #3,d2        // w = w&3  (d2 already has w)
    beq.s     end          // no more pixels remain
    lea       data(d2),a1  // a1 = data[w&3]

    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)      // *scr = or results
end:
    rts                    // and return

data:
    dc.b   0,0x3F,0xF,0x3 // and
    dc.b   0,64,80,84     // or
data2:
    dc.b   0, 0xC0,0xF0,0xFC // and
    dc.b   0, 21,5, 1        // or
    dc.b   0, 3, 2, 1        // sub
}
////////////////////////////////////////////////////////////////////////////
// fills a rect from x,y to x+w-1,y+h-1, without translate or clip rect
asm static void fillRectNOCT_2bpp (int x, int y, int w, int h, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    lea       dataA0,a1
    // uchar *scr = &activeScr[40*y + (x>>2)];
    lea       widthInBPP,a0  // get width in bpp
    move.w    (a0), d2
    move.w    d2,10(a1)     // store the widthInBPP

    mulu.w    d2,d1      // 40*y
    move.w    d0,d2        // saves x in d2
    lsr.w     #2,d0        // x>>2
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    move.l    a0,(a1)      // save a0
    move.w    d2,4(a1)     // save x
    move.w    8(a7),d0     // get w
    move.w    d0,6(a1)     // save w

extLoop:

    // if w <= 8 use the normal setPixel. this will speedup the process
    cmpi.w    #8,d0        // compares to 8
    bgt.s     fits         // continue with normal process if > 8
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #3,d2        // x = x&3  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&3]
    move.w    14(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&3+4]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #3,d2        // is modulus 4
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    bra.s     end

data3:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC
data2:
    dc.b   0, 0xC0,0xF0,0xFC // and
    dc.b   0, 21,5, 1        // or
    dc.b   0, 3, 2, 1        // sub
data:
    dc.b   0,0x3F,0xF,0x3 // and
    dc.b   0,64,80,84     // or

    // draws the pixels if they arent modulus 4
fits:
    andi.w    #3,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 4

    lea       data2(d2),a1 // a1 = data[w&3]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    14(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    8(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 2;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #2,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    14(a7),d0
    muls.w    #0x55,d0     // color = 0x55*color
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 4
    // *scr = (data[w&3+3]*color) | (data[w&3] & *scr)
    andi.w    #3,d2        // w = w&3  (d2 already has w)
    beq.s     end          // no more pixels remain
    lea       data(d2),a1  // a1 = data[w&3]

    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    14(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)      // *scr = or results

end:
    subq.w    #1, 10(a7)       // dec height
    tst.w     10(a7)       // height == 0?
    beq.s     realEnd      // if yes, end
    lea       dataA0,a1
    move.w    6(a1),d0      // restore w
    move.w    4(a1),d2      // restore x
    move.l    (a1),a0       // restore a0
    add.l     8(a1),a0      // go to next line
    move.l    a0,(a1)      // save a0 again
    move.w    d2,4(a7)
    move.w    d0,8(a7)
	 bra       extLoop      // and draw again

realEnd:
    rts                    // and return

dataA0:
    dc.l   0
    dc.w   0
    dc.w   0
    dc.l   0
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1
asm static void drawHLine_2bpp (int x, int y, int w, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip y bounds - in the h line, if outside y bounds, end here
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    blt       end          // if y < y1 end
    swap      d2           // get y2
    cmp.w     d2,d1
    bge       end          // if y >= y2 end

    // verify clip x bounds - sets x = max(x,clipX1), w = min(w,clipX2-x)
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0        // d0-d2 >= 0 -> d0 >= d2
    bge.s     xIsGreaterX1 // if x >= x1, continue
    // since x < x1 (lower clip) test if x+w < x1. if true, exit: the line is outside the clip
    sub.w     d0,d2        // d2 = d2 - d0 (x1 = x1 - x)
    cmp.w     8(a7),d2     // d2-w >= 0
    bge       end          // if (x1-x >= w) then the line is smaller than clipX1. end
    sub.w     d2,8(a7)     // subtract from w what we are going to add to x
    move.l    (a0),d0      // else x = x1
xIsGreaterX1:
    swap      d2           // get x2
    sub.w     d0,d2        // clipW = x2 - x
    ble.s     end2          // end if clipW < 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipW > w, continue
    move.w    d2,8(a7)     // else w = clipW

goAhead:
    // move the computed values back to the stack
    move.w    d0,4(a7)     // x

    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // saves x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // if w <= 8 use the normal setPixel. this will speedup the process
    move.w    8(a7),d0     // get w
    tst.w     d0           // == 0 ?
    beq       end          // if yes, bye bye
    cmpi.w    #8,d0        // compares to 8
    bgt.s     fits         // continue with normal process if > 8
    subq.w    #1,d0        // d0 -= 1 due to dbf
    move.w    4(a7),d2     // x
    andi.w    #3,d2        // x = x&3  (d2 already has x)
loop2:
    lea       data3(d2),a1 // a1 = data[x&3]
    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // d1 = value to left shift
    lsl.b     d1,d2        // color << data[x&3]
    move.b    4(a1),d1     // d1 = value to and
    and.b     d1,(a0)      // *scr & data[x&3+4]
    or.b      d2,(a0)      // add the and with the shift left

    addq.w    #1,4(a7)     // inc x
    move.w    4(a7),d2     // move x to d1
    andi.w    #3,d2        // is modulus 4
    bne.s     dontInc
    addq.l    #1,a0
dontInc:
    dbf.w     d0,loop2

end2:
    rts

data3:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC

    // draws the pixels if they arent modulus 4
fits:
    andi.w    #3,d2        // d2 already has x
    beq.s     cont         // continue: starting pixel is % 4

    lea       data2(d2),a1 // a1 = data[w&3]
    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)+     // *scr = or results
    // w -= drawed pixels
    move.b    8(a1),d2
    sub.w     d2,8(a7)
    ble.s     end          // end if w <= 0

cont:
    // int howManyFullPixels = w >> 2;
    move.w    8(a7),d1     // w
    beq.s     end          // end if w = o
    move.w    d1,d2        // save w in d2
    lsr.w     #2,d1        // d1 = howManyFullPixels
    beq.s     remain       // if (howManyFullPixels == 0) goto remain

    move.w    12(a7),d0
    muls.w    #0x55,d0     // color = 0x55*color
    // do *scr++ = color;  while (--howManyFullPixels)
    subq.w    #1,d1

loop:
    move.b    d0,(a0)+
    dbf.w     d1,loop

remain:
    // howManyPixelsRemain = w % 4
    // *scr = (data[w&3+3]*color) | (data[w&3] & *scr)
    andi.w    #3,d2        // w = w&3  (d2 already has w)
    beq.s     end          // no more pixels remain
    lea       data(d2),a1  // a1 = data[w&3]

    move.b    (a1),d1      // d1 = value to and
    and.b     d1,(a0)      // *scr = data[v&3] & *scr
    move.b    4(a1),d1     // d1 = value to mul
    muls.w    12(a7),d1    // d1 = data[w&3+3]*color
    or.b      d1,(a0)      // *scr = or results
end:
    rts                    // and return

data:
    dc.b   0,0x3F,0xF,0x3 // and
    dc.b   0,64,80,84     // or
data2:
    dc.b   0, 0xC0,0xF0,0xFC // and
    dc.b   0, 21,5, 1        // or
    dc.b   0, 3, 2, 1        // sub
}
////////////////////////////////////////////////////////////////////////////
// clears the screen with the specified 8-bit pattern. clears the hole screen, ignoring translate and clip
asm static void clearScreen_2bpp (int32 color) // color must be 8 bits
{
//    move.w    #1599,d0 // 160*10-1 full bytes
    lea       maxX,a0      // get maxX
    move.w    (a0),d0      // value
    lea       maxY,a0      // get maxY
    mulu.w    (a0),d0      // multiply maxX/4*maxY
    lsr.w     #4,d0        // divide by 16
    subi.w    #1,d0        // minus 1

    move.l    4(a7),d2     // gets the pattern to fill
    move.b    d2,d1        // d1=000000XX
    lsl.w     #8,d1        // d1=0000XX00
    move.b    d2,d1        // d1=0000XXXX
    move.w    d1,d2        // saves d1 in d2
    swap      d1           // d1=XXXX0000
    move.w    d2,d1        // d1=XXXXXXXX
    lea       activeScr,a0
    movea.l   (a0),a0
loop:
    move.l    d1,(a0)+
    dbf.w     d0,loop

    rts
}
////////////////////////////////////////////////////////////////////////////
// draws a line from y to y+h-1
asm static void drawVLine_2bpp (int x, int y, int h, int32 color)
{
    // gets x and y
    move.w    4(a7),d0     // x
    move.w    6(a7),d1     // y

    // translate
    lea       transXY,a0   // get transXY
    move.l    (a0),d2      // move to d0
    add.w     d2,d0        // x += transX
    swap      d2           // get the y part
    add.w     d2,d1        // y += transY

    // verify clip x bounds - in the v line, if outside x bounds, end here
    lea       clipX,a0     // get clipX. lower 16 bits has the lower bound (x1), upper 16 bits has the upper bound (x2)
    move.l    (a0),d2
    cmp.w     d2,d0
    blt       end          // if x < clipx1 end
    swap      d2           // get clipx2
    cmp.w     d2,d0
    bgt       end          // if x > clipx2 end

    // verify clip y bounds - sets y = max(y,clipY1), h = min(h,clipY2-y)
    lea       clipY,a0     // get clipY. lower 16 bits has the lower bound (y1), upper 16 bits has the upper bound (y2)
    move.l    (a0),d2
    cmp.w     d2,d1
    bge.s     yIsGreaterY1 // if y >= clipy1, continue
    // guich@200b4_121: was not testing if  y+h < clipy1
    add.w     8(a7),d1     // y += h
    cmp.w     d2,d1        // compare y+h with clipy1
    ble.s     end          // if <=, nothing to draw; end
    move.w    d1,8(a7)		// guich@400_75 - h will hold the difference of the clipping. 1) h = y+h
    neg.w     d2				// 2) clipy1 = -clipy1
    add.w     d2,8(a7)     // 3) h = y+h + -clipy1
    neg.w     d2           // 4) --clipy1 = clipy1
    move.w    d2,d1        // else y = clipy1
yIsGreaterY1:
    swap      d2           // get clipy2
    sub.w     d1,d2        // clipH = clipy2 - y
    ble.s     end          // end if clipH <= 0
    cmp.w     8(a7),d2
    bgt.s     goAhead      // if clipH > h, continue
    move.w    d2,8(a7)     // else h = clipH

goAhead:
    // uchar *scr = &activeScr[40*y + (x>>2)];
    move.w    d0,d2        // save x in d2
    lsr.w     #2,d0        // x>>2
    lea       widthInBPP,a0  // get width in bpp
    mulu.w    (a0),d1      // 40*y
    add.w     d1,d0
    lea       activeScr,a0   // get screen address
    movea.l   (a0),a0
    adda.w    d0,a0        // increment a0 to point to the pixel

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    andi.w    #0x3,d2      // d2 = x&3  (d2 already has x)
    lea       data(d2),a1  // a1 = data[x&3]

    move.w    12(a7),d2    // d2 = color
    move.b    (a1),d1      // saves the lsl param in d1
    lsl.b     d1,d2        // color = (color<<data[x&3]);

    move.w    8(a7),d0      // d0 = h
    beq.s     end           // end if h = 0
    subq.w    #1,d0         // h--

    move.b    4(a1),d1      // saves the and param in d1

    // *scr = ((*scr & data[x&3+4]) + color << data[x&3]);
    lea       widthInBPP,a1
loop:
    and.b     d1,(a0)       // *scr & 0x3F
    or.b      d2,(a0)
    add.w     (a1),a0       // scr += 40
    dbf.w     d0,loop       // while (--h);
end:
    rts
data:
    dc.b   6,4,2,0
    dc.b   0x3F,0xCF,0xF3,0xFC
}
////////////////////////////////////////////////////////////////////////////
// rol byte <b> <count> bits
inline static asm byte rol(byte b: __D0, byte count: __D1): __D0
{
   rol.b    d1,d0
}
inline static asm byte ror(byte b: __D0, byte count: __D1): __D0
{
   ror.b    d1,d0
}
////////////////////////////////////////////////////////////////////////////
static void getBitmapRow_2bpp(GrHandle h, int x, int y, int w, byte *row)
{
   byte *pt = GRHANDLE_MemAddr(h)+GRHANDLE_WidthInBPP(h)*y;
   while (w--)
   {
      *row++ = rol(*(pt + (x >> 2)),((x&3)+1)*2) & 3;
      x++;
   }
}
////////////////////////////////////////////////////////////////////////////
// guich@200b4: speedup drawBitmap in gray scale.
static asm void bitblt_Gray(byte *pScr, byte *pBmp, int dstX, int x, int width, int height, int32 stepScr, int32 stepBmp, int drawOp, int backColor, int foreColor)
{
   // sintax: _ = parameter   __ = local var
   #define _pScr        40(a7)
   #define _pBmp        44(a7)
   #define _dstX        48(a7)
   #define _x           50(a7)
   #define _width       52(a7)
   #define _height      54(a7)
   #define _stepScr     56(a7)
   #define _stepBmp     60(a7)
   #define _drawOp      64(a7)
   #define _backColor   66(a7)
   #define _foreColor   68(a7)
   #define __temp       d0
   #define __backColor  d1
   #define __screenPt   d2
   #define __curBitScr  d3
   #define __curBitBmp  d4
   #define __bScr       d5
   #define __bBmp       d6
   #define __width      d7 // copy of _width that will be decremented
   #define __pScr       a2
   #define __pBmp       a3
   #define __drawOpAddr a4
   #define __foreColor  a6
   MACHINE 68000

   movem.l   d3-d7/a2-a4/a6,-(a7) // save all registers

/*   move.l _pScr,d0
   move.l _pBmp,d0
   move.w _dstX,d0
   move.w _x,d0
   move.w _width,d0
   move.w _height,d0
   move.l _stepScr,d0
   move.l _stepBmp, d0
   move.w _drawOp, d0
   move.w _backColor,d0
   move.w _foreColor,d0*/


   // compute curBitScr = dstX = ~dstX & 3
   move.w    _dstX,__curBitScr
   not.w     __curBitScr
   andi.w    #3,__curBitScr
   move.w    __curBitScr,_dstX // store back the computed value so we can use it again without needing to recalculate it.
   // compute curBitBmp = x = ~x & 3
   move.w    _x,__curBitBmp
   not.w     __curBitBmp
   andi.w    #3,__curBitBmp
   move.w    __curBitBmp,_x // store back the computed value so we can use it again without needing to recalculate it.

   // store some parameters in their registers
   move.w    _backColor,__backColor
   move.w    _foreColor,__foreColor
   movea.l   _pScr,__pScr
   movea.l   _pBmp,__pBmp

   // compute the address of the drawOp case
   moveq.l   #0,__temp
   move.w    _drawOp,__temp
   add.w     __temp,__temp    // * 2
   lea       caseJumps,__drawOpAddr
   add.l     __temp,__drawOpAddr
   move.w    (__drawOpAddr),__drawOpAddr

   // while (height--)
   // {
   bra.w     heightLoopEnd
caseJumps:
   dc.w 0x0054, 0x0000, 0x0008, 0x0010, 0x0016, 0x001C, 0x0022, 0x0030, 0x003E

heightLoop:

   move.w    _dstX,__curBitScr
   move.w    _x,__curBitBmp
   move.w    _width,__width

   // bScr = rol(*pScr,2*(4-curBitScr)); // align the 1st pixel to the least significative bits
   move.b    __curBitScr,__temp
   subq.b    #4,__temp // curBit - 4
   neg.b     __temp    // - (curBit - 4)
   add.b     __temp,__temp // * 2
   move.b    (__pScr),__bScr // *pScr -> bScr
   rol.b     __temp,__bScr  // rol(bScr,temp1)

   // bBmp = rol(*pBmp,2*(4-curBitBmp));
   move.b    __curBitBmp,__temp
   subq.b    #4,__temp // curBit - 4
   neg.b     __temp    // - (curBit - 4)
   add.b     __temp,__temp // * 2
   move.b    (__pBmp),__bBmp // *pBmp -> bBmp
   rol.b     __temp,__bBmp  // rol(bBmp,temp1)

   // while (width--)
   // {
   bra       widthLoopEnd

widthLoop:

   // switch (drawOp)
   // {

   jmp       (pc,__drawOpAddr)

   // case WIN_ERASE:         screenPt = bScr & bBmp;
   move.b    __bBmp,__screenPt
   and.b     __bScr,__screenPt
   bra.s     caseEnd   // break;

   // case WIN_MASK:          screenPt = bScr & ~bBmp;
   move.b    __bBmp,__screenPt
   not.b     __screenPt
   and.b     __bScr,__screenPt
   bra.s     caseEnd   // break;

   // case WIN_INVERT:        screenPt = bScr ^ bBmp;
   move.b    __bBmp,__screenPt
   eor.b     __bScr,__screenPt
   bra.s     caseEnd   // break;

   // case WIN_OVERLAY:       screenPt = bScr | bBmp;
   move.b    __bBmp,__screenPt
   or.b     __bScr,__screenPt
   bra.s     caseEnd   // break;

   // case WIN_PAINT_INVERSE: screenPt = ~bBmp;
   move.b    __bBmp,__screenPt
   not.b     __screenPt
   bra.s     caseEnd   // break;

   // case WIN_SPRITE_COLOR:  if ((bBmp & 3) != backColor)
   move.b    __bBmp,__temp
   andi.b    #0x3,__temp
   cmp.b     __temp,__backColor // else goto cont; break;
   beq.s     dontPutPixel
   // screenPt = bBmp;
   move.b    __bBmp,__screenPt
   bra.s     caseEnd

   // case WIN_REPLACE_COLOR:   if ((bBmp & 3) != backColor)
   move.b    __bBmp,__temp
   andi.b    #0x3,__temp
   cmp.b     __temp,__backColor // else goto cont; break;
   beq.s     dontPutPixel
   // screenPt = foreColor;
   move.w    __foreColor,__screenPt
   bra.s     caseEnd   // break;

   // case WIN_SWAP_COLORS:
   move.b    __bBmp,__temp
   andi.w    #0x3,__temp
   cmp.b     __temp,__backColor // if ((bBmp & 3) == backColor)
   bne.s     testwithfore
   move.w    __foreColor,__screenPt // screenPt = foreColor
   bra.s     caseEnd   // break;
testwithfore:
   cmp.w     __temp,__foreColor  // else if ((bBmp & 3) == foreColor)
   bne.s     dontPutPixel
   move.w    __backColor,__screenPt // screenPt = backColor
   bra.s     caseEnd   // break;

   // case WIN_PAINT:         screenPt = bBmp;
   move.b    __bBmp,__screenPt
   // }

caseEnd:

   // bScr = (bScr & b11111100) | (screenPt & 3); // set the new pixel
   andi.b     #3,__screenPt
   andi.b    #0xFC,__bScr // 11111100b
   or.b      __screenPt,__bScr

dontPutPixel:
   // if (!curBitScr--)  // advance to next byte?
   // {
   move.b    __curBitScr,__temp
   subq.b    #1,__curBitScr
   tst.b     __temp
   bne.s     bitScrDidntOverflowed
   //    curBitScr = 3;
   moveq.w   #3,__curBitScr
   //    *pScr = bScr;
   move.b    __bScr,(__pScr)
   //    bScr = *++pScr;
   addq.l    #1,__pScr
   move.b    (__pScr),__bScr
   // }

bitScrDidntOverflowed:

   // bScr = rol(bScr,2);
   rol.b     #2,__bScr

   // if (!curBitBmp--)  // advance to next byte?
   // {
   move.b    __curBitBmp,__temp
   subq.b    #1,__curBitBmp
   tst.b     __temp
   bne.s     bitBmpDidntOverflowed
   // curBitBmp = 3;
   moveq.w   #3,__curBitBmp
   // bBmp = *++pBmp;
   addq.l    #1,__pBmp
   move.b    (__pBmp),__bBmp
   // }

bitBmpDidntOverflowed:

   // bBmp = rol(bBmp,2);
   rol.b     #2,__bBmp
   // }
   // reset the values

widthLoopEnd:

   move.w    __width,__temp
   subq.w    #1,__width
   tst.w     __temp
   bne.w     widthLoop

   // if (curBitScr >= 0)
   tst.w     __curBitScr
   blt.s     lastByteAlreadyMoved

   // *pScr = rol(bScr,curBitScr*2); // move back the last byte
   move.b    __curBitScr,__temp
   add.b     __temp,__temp
   rol.b     __temp,__bScr // bScr wont be used again
   move.b    __bScr,(__pScr)

lastByteAlreadyMoved:

   // advance to next row
   movea.l   _pScr,__pScr         // get the original address
   adda.l    _stepScr,__pScr      // pScr += stepScr;
   move.l    __pScr,_pScr         // save back

   movea.l   _pBmp,__pBmp         // get the original address
   adda.l    _stepBmp,__pBmp      // pBmp += stepBmp;
   move.l    __pBmp,_pBmp         // save back

   // }

heightLoopEnd:

   move.w    _height,__temp
   subq.w    #1,_height
   tst.w     __temp
   bne.w     heightLoop

   movem.l   (a7)+,d3-d7/a2-a4/a6
   rts
}

// guich@200b4_86: ultrafast data transfer. note that the total bytes transfered are count*4
asm static inline void xfer32(byte *from:__A0, byte *to:__A1, uint16 count: __D0)
{
   subq.l    #1,d0
loop:
   move.l    (a0)+,(a1)+
   dbf.w     d0,loop
}
// cuts the area x,y,width,height of the bitmap bmp with dimensions bmpW,bmpH to the (current active) screen location dstX,dstY
static void drawBitmap_Gray(GrHandle h, int x, int y, int width, int height, int dstX, int dstY, int drawOp, int32 backColor, int32 foreColor, int doClip)
{
   int bmpW = GRHANDLE_Width(h);
   int bmpH = GRHANDLE_Height(h);
   int originalWidth = width;
   int originalHeight = height;
   RectangleType r1, r2;
   //int32 t = -(*vmGlobals->getTimeStamp)();
   byte *pScr,*pBmp;

   if (doClip) // guich@340_14
   {
      int k;
      byte panX,panY;

      // guich@200b4_65: remade this drawBitmap routine
      // 1. make sure that the area of the bitmap that we want to copy IS inside its area
      makeRect(x,y,width,height,r1);
      makeRect(0,0,bmpW,bmpH,r2);
      if (!intersects(&r1,&r2)) return;

      // 2 test if we're panning a bitmap - guich@240_1
      panX = bmpW > CLIPX2 && dstX <= CLIPX1;
      panY = bmpH > CLIPY2 && dstY <= CLIPY1;
      if (panX || panY)
      {
         if (panX)
         {
            x -= dstX;
            dstX = 0;
            width = min(CLIPX2-CLIPX1,bmpW-x);
         }
         if (panY)
         {
            y -= dstY;
            dstY = 0;
            height = min(CLIPY2-CLIPY1,bmpH-y);
         }
      }
      // 3. translate the dstXY
      dstX += TRANSX;
      dstY += TRANSY;

      // 4. if not panning, intersect the source bitmap with the screen (Window saveBehind)
      if ((!panX && !panY) || x < 0 || y < 0 // guich@300_15: if moving a popup window to the left of the screen, it gets here still with x < 0 (or y < 0).
          || (x+width) > MaxScreenX || (y+height) > MaxScreenY) // guich@341_8: fixed problem when a MessageBox has width > than screenWidth, and it was dragged to the right and the saveBehind method would make this fail bc it was trying to move more pixels at the right than possible
      {
         if (dstX == 0 && dstY == 0) // guich@200b4_72: without this test, if you create a bitmap greater than the screen widths, it will not be displayed correctly.
         {
            makeRect(0,0,MaxScreenX,MaxScreenY,r2); // r1 already set
            intersectWith(&r1,&r2);

            k = r1.extent.x - width;
            if (k < 0 && r1.topLeft.x == r2.topLeft.x)
               dstX -= k;
            k = r1.extent.y - height;
            if (k < 0 && r1.topLeft.y == r2.topLeft.y)
               dstY -= k;

            x = r1.topLeft.x;
            y = r1.topLeft.y;
            width = r1.extent.x;
            height = r1.extent.y;
         }
      }
      // 5. intersect the destination bitmap with the clip area (Commander/Sprite)
      makeRect(dstX,dstY,width,height,r1);
      makeRect(CLIPX1,CLIPY1,CLIPX2-CLIPX1,CLIPY2-CLIPY1,r2);
      intersectWith(&r1,&r2);

      if (r1.extent.x <= 0 || r1.extent.y <= 0) // nothing to draw?
         return;

      k = r1.extent.x - width;
      if (!panX && k < 0 && r1.topLeft.x == r2.topLeft.x)
         x -= k;
      k = r1.extent.y - height;
      if (!panY && k < 0 && r1.topLeft.y == r2.topLeft.y)
         y -= k;

      dstX   = r1.topLeft.x;
      dstY   = r1.topLeft.y;
      width  = r1.extent.x;
      height = r1.extent.y;
      if (originalWidth < width) // guich@341_15: fix a problem when drawing sub-images of an animation image
         width = originalWidth;
      if (originalHeight < height)
         height = originalHeight;
   }
   else
   {
      dstX += TRANSX;
      dstY += TRANSY;
   }

   pScr = activeScr+widthInBPP*dstY + (dstX >> 2);
   pBmp = GRHANDLE_MemAddr(h)+GRHANDLE_WidthInBPP(h)*y + (x >> 2);

   // 5. can we speedup the process?
   if (TRANSX == 0 && CLIPX1 == 0 && CLIPY1 == 0 && CLIPX2 == MaxScreenX && CLIPY2 == MaxScreenY && x == 0 && /*y == 0 && */dstX == 0 && /*dstY == 0 && */width == bmpW && height <= bmpH && bmpW == MaxScreenX && drawOp == WIN_PAINT)
      xfer32(pBmp,pScr,(int32)width * (int32)height / 16); // 4 pixels per byte, 4 bytes being transferred
   else
      bitblt_Gray(pScr,pBmp, dstX, x, width, height, widthInBPP, GRHANDLE_WidthInBPP(h), drawOp, (int)backColor, (int)foreColor);
   //t += (*vmGlobals->getTimeStamp)();
   //(*vmGlobals->debug)((*vmGlobals->int2str)("drawBitmap",t));
}
// copies the screen width
static void drawScreen_(GrHandle hSrc, GrHandle hDest, int ySrc, int yDest, int height)
{
   int32 width;
   width = GRHANDLE_WidthInBPP(hSrc);
   if (width == GRHANDLE_WidthInBPP(hDest))  // do a basic check
   {
      byte *pSrc,*pDest;
      pSrc  = GRHANDLE_MemAddr(hSrc)  + width * ySrc;
      pDest = GRHANDLE_MemAddr(hDest) + width * yDest;
      xfer32(pSrc,pDest,(width * (int32)height) >> 2); // 4 bytes being transfered
   }
}

////////////////////////////////////////////////////////////////////////////
//        Routines that can use the underlying OS to draw faster          //
////////////////////////////////////////////////////////////////////////////
static void eraseRect(int x, int y, int w, int h, int cfrom, int cto)
{
   if (translateAndClip(&x,&y,&w,&h))
      while (w--)
         (*eraseVLineNOCT)(x++,y,h,cfrom,cto);
}
////////////////////////////////////////////////////////////////////////////
static void drawCursor(int x, int y, int w, int h)
{
   if (translateAndClip(&x,&y,&w,&h))
      while (w--)
         (*invVLineNOCT)(x++,y,h);
}
////////////////////////////////////////////////////////////////////////////
static void drawCursorOutline(int x, int y, int w, int h)
{
   if (translateAndClip(&x,&y,&w,&h))
   {
      int y2 = y + h -1;
      (*invVLineNOCT)(x,y,h);
      (*invVLineNOCT)(x+w-1,y,h);
      while (w--)
      {
         (*invPixelNOCT)(x,y);
         (*invPixelNOCT)(x,y2);
         x++;
      }
   }
}
////////////////////////////////////////////////////////////////////////////
static void drawDottedCursor(int x, int y, int w, int h)
{
   int x2,y2,xx,yy;
   x += TRANSX;
   y += TRANSY;
   x2 = x + w -1;
   y2 = y + h -1;
   xx = x;
   yy = y;
   // TODO a better clipping algotithm
   // vertical lines
   for (; h > 0; h-=2,yy+=2)
      if (CLIPY1 <= yy && yy < CLIPY2)
      {         
         if (CLIPX1 <= xx && xx < CLIPX2)
            invPixelNOCT(x,yy);
         if (CLIPX1 <= x2 && x2 < CLIPX2)
            invPixelNOCT(x2,yy);
      }
   // horizontal lines         
   for (; w > 0; w-=2,xx+=2)
      if (CLIPX1 <= xx && xx < CLIPX2) 
      {
         if (CLIPY1 <= y && y < CLIPY2)
            invPixelNOCT(xx,y);
         if (CLIPY1 <= y2 && y2 < CLIPY2)
            invPixelNOCT(xx,y2);
      }
}
////////////////////////////////////////////////////////////////////////////
void drawRect(int x, int y, int w, int h, int32 c);
void drawRect(int x, int y, int w, int h, int32 c)
{
   (*drawHLine)(x,y,w,c);
   (*drawHLine)(x,y+h-1,w,c);
   (*drawVLine)(x,y,h,c);
   (*drawVLine)(x+w-1,y,h,c);
}
////////////////////////////////////////////////////////////////////////////
void fillRect(int x, int y, int w, int h, int32 c);
void fillRect(int x, int y, int w, int h, int32 c)
{
   x += TRANSX;
   y += TRANSY;

  // if (CLIPY1 <= (y+h) && y < CLIPY2 && CLIPX1 <= (x+w) && x < CLIPX2) // line must lie inside y clip bounds, must not end before clip x1 and must not start after clip x2
   {
      if (x < CLIPX1) {w -= CLIPX1-x; x = CLIPX1;} // line start before clip x1
      if ((x+w) > CLIPX2) w = CLIPX2-x; // line stops after clip x2
      if (y < CLIPY1) {h -= CLIPY1-y; y = CLIPY1;} // line start before clip y1
      if ((y+h) > CLIPY2) h = CLIPY2-y; // line stops after clip y2

      if (h > 0 && w > 0)
         (*fillRectNOCT)(x,y,w,h,c); // guich@300_51
   }
}
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
// sets the gray mode and the window bitmap
// use as:
// 1. oldSSA = setgrey(1, activeScr, MaxScreenX/8);  // sets the gray mode
// 2. setgrey(0, oldSSA, 10);

static const void *oldSSA=0;
char startedInGrayMode = -1;
static const void *setGrayMode(int grey, const void *newSSA, byte vpw)
{
    static byte pixcSaved=0;
    const void *oldSSA;
    const void * volatile *SSA = (const void**)0xfffffa00;
    volatile byte *CKCON = (byte *)0xfffffa27;
    volatile byte *VPW  = (byte *)0xfffffa05;
    volatile byte *PICF = (byte *)0xfffffa20;
    volatile byte *LBAR = (byte *)0xfffffa29;
    volatile byte *FRCM = (byte *)0xfffffa31;
    volatile byte *POSR = (byte *)0xfffffa2d;
    volatile Word *GPMR = (Word *)0xfffffa32;
    volatile byte *PIXC = (byte *)0xfffffa25; // lcd pixel clock
    // guich: these are the possible palettes.
    Word gray0 = 0, gray1 = 2, gray2 = 4, gray3 = 6; // linear - dark (flickers in some IIIe devices
//    Word gray0 = 0, gray1 = 1, gray2 = 3, gray3 = 6; // logarithm - bright
//    Word gray0 = 0, gray1 = 1, gray2 = 5, gray3 = 6; // darker - its displayed as 0556

    Word grayPal = gray2 + (gray3<<4) + (gray0<<8) + (gray1<<12);

    oldSSA = *SSA;
    *CKCON &= 0x7f;    // clocking control
    *SSA = newSSA;     // screen starting address
    if (grey || startedInGrayMode == 1) // guich@200b3: if we already was in gray mode, let it stay there when exiting.
    {
       if (startedInGrayMode == -1)
          startedInGrayMode = *LBAR == 22;
       *VPW = vpw;     // virtual page width
       *PICF |= 0x01;  // panel interface configuration
       *LBAR = 22;     // last buffer address register
       *FRCM = 0xb9;   // frame-rate modulation control
       *GPMR = grayPal; // gray palette mapping register
       *POSR &= 0xf0;  // panning offset register
    }
    else
    {
       *VPW = vpw;
       *PICF &= 0xfe;
       *LBAR = 10;
       *POSR &= 0xf0;
    }
    SysTaskDelay(2);
    *CKCON |= 0x80;    // clocking control

    if (grey) // guich: fixes gray flicker!
    {
      pixcSaved = *PIXC;
		*PIXC = 0;
	 }
 	 else
		*PIXC = pixcSaved;

    return oldSSA;
}
////////////////////////////////////////////////////////////////////////////
//                             COMMON CODE                                //
// The routines presented here are not called intensively, so they may    //
// simply check the right code to call with an 'if'                       //
////////////////////////////////////////////////////////////////////////////
int scaleImage(GrHandle srcG, uint16 width, uint16 height, GrHandle dstG, uint16 newWidth, uint16 newHeight) // guich@420_14
{
   GrHandle last = lastGrHandle;

   int32 h = newHeight << 1;
   int32 hi = height << 1;
   int32 hf = height / h;
   int32 wf = 0;
   int32 w = newWidth << 1;
   int32 wi = width << 1;
   int32 x,y,oldIY=-1;
   Pixel *p;
   byte lowMemoryMode = 0;

   Pixel *buf = (Pixel *)(*vmGlobals->xmalloc)(sizeof(Pixel)*(int32)width*(int32)height);
   if (buf == NULL) // guich@401_30: now if there's no memory to allocate the full buffer, use a slower routine
   {
      lowMemoryMode = 1;
      buf = (Pixel *)(*vmGlobals->xmalloc)(sizeof(Pixel)*(int32)width);
      if (buf == NULL)
         return 0;
   }
   GR_ENTER // guich@400_59
   if (lowMemoryMode)
   {
      for (y=0; y < newHeight; y++, hf += hi)
      {
         int32 iy = hf/h;
         if (iy != oldIY) // guich@402_44: only scan if needed
         {
            oldIY = iy;
            SetDrawWindow(srcG);
            p = buf;
            for (x = 0; x < width; x++)
               *p++ = (*getPixel)(x,iy); // guich@402_43: replaced y by iy
            // now set all pixels
            SetDrawWindow(dstG);
         }
         wf = width/w;
         for (x=0; x < newWidth; x++,wf+=wi)
            (*setPixel)(x,y,buf[wf/w]);
      }
   }
   else
   {
      // first get all pixels
      SetDrawWindow(srcG);
      p = buf;
      for (y = 0; y < height; y++)
         for (x = 0; x < width; x++)
            *p++ = (*getPixel)(x,y);
      // now set all pixels
      SetDrawWindow(dstG);
      for (y=0; y < newHeight; y++, hf += hi)
      {
         int32 iy = hf/h;
         wf = width/w;
         p = &buf[iy*width]; // guich@401_30: changed to int32
         for (x=0; x < newWidth; x++,wf+=wi)
            (*setPixel)(x,y,p[wf/w]);
      }
   }
   GR_EXIT
   xfree(buf);
   SetDrawWindow(last?last:mainGrHandle); // guich@421_5: check if last was not null
   return 1;
}

extern uint32 *colorTable;
int read8bppImage(int32 width, int32 height, byte *in);

int read8bppImage(int32 width, int32 height, byte *in) // guich@420_15
{
   BitmapType *bmp;
   Err err;
   byte *data;
   int32 i,scanline,scanline2;

   if (!usePalmOSFuncs) return 0; // only works in Palm OS >= 3.5

   // first, we create a 8bpp bitmap
   bmp = BmpCreate(width,height,8,NULL,&err);
   if (err != errNone)
      return 0;

  	data = BmpGetBits((BitmapPtr)bmp);
   scanline = ((width + 3)>>2)<<2;
   scanline2 = ((width + 1)>>1)<<1;
   // now we fastly move the original bmp to the bitmap
   for (i=0; i < height; i++) // note: we must always allocate a buffer bc the bmp is stored upsidedown (who was the "genius" who invented this???)
      xmemmove(data + (height-i-1)*scanline2, in + i * scanline, width);

   // and draw the bmp into the target
   WinDrawBitmap(bmp,0,0);
   // free resources and return
   BmpDelete(bmp);

   return 1;
}

static void fillInRow(GrHandle h, byte *buf, int y)
{
   int w = GRHANDLE_Width(h);
   byte *bmp = GRHANDLE_MemAddr(h)+GRHANDLE_WidthInBPP(h)*y;
   int x;

   switch (displayBPP)
   {
      case 2:
   	   for (x =0; x < w; x++)
   			*buf++ = (rol(*(bmp + (x >> 2)),((x&3)+1)*2) & 3);
   	   break;
   	case 4:
   	   for (x =0; x < w; x++)
	   		*buf++ = (rol(*(bmp + (x >> 1)),((x&1)+1)*4) & 0xF);
   	   break;
   	case 8:
         xmemmove(buf,bmp,w);
         break;
   }
}

static void clearScreen(int32 color)
{
   if (!usePalmOSFuncs)
   {
		int c = (int)color;
	   c *= 0x55;
	   if (((maxX*maxY) & 15) == 0) // must be at least 16 bytes long bc clearscreen fills 16 colors (a 32bit) at once
	   	clearScreen_2bpp(c);
	   else
	   {
	      int size = widthInBPP * maxY;
	      byte *b = activeScr;
	      while (size--)
	         *b++ = c;
	   }
	}
	else
	{
      if (displayBPP == 4)
         clearScreen_4bpp(color);
      else
         clearScreen_8bpp(color);
   }
}
static void imageChangeColors(GrHandle h, int width, int height, int32 colorFrom, int32 colorTo)
{
   if (!usePalmOSFuncs)
   {
   	byte *bptr,*by;
   	int y,x,xx;
      static byte data[8] = {6,4,2,0, 0x3F,0xCF,0xF3,0xFC};
      for (y = 0; y < height; y++)
      {
         by = GRHANDLE_MemAddr(h)+GRHANDLE_WidthInBPP(h)*y;
      	for (x =0; x < width; x++)
   	   {
   	      bptr = by + (x >> 2);
   		   xx = x & 3;
   		   if ((rol(*bptr,(xx+1)*2) & 3) == colorFrom)
   		      *bptr = (byte)((colorTo << data[xx]) | (*bptr & data[xx+4]));
   	   }
   	}
   }
   else
   {
      WinHandle oldH = WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
      int x,y;
      int colorFromIndex = colorFrom;
      WinSetForeColor(colorTo);
      if (sonyRefNum)
         for (y = 0; y < height; y++)
            for (x = 0; x < width; x++)
               if ((HRWinGetPixel(sonyRefNum,x,y)) == colorFromIndex)
                  HRWinDrawPixel(sonyRefNum,x,y);
      else
         for (y = 0; y < height; y++)
            for (x = 0; x < width; x++)
               if ((WinGetPixel(x,y)) == colorFromIndex)
                  WinDrawPixel(x,y);
      WinSetDrawWindow(oldH);
   }
}
static void eraseWindow(WinHandle wh, int width, int height)
{
   int y=0;
   WinHandle oldwh = &vmGlobals->wabaFrmP->window;
   WinSetDrawWindow(wh);
   WinSetForeColor(rgb2index(0xFF,0xFF,0xFF));
   if (sonyRefNum)
      while (height--)
      {
         HRWinDrawLine(sonyRefNum,0,y,width-1,y);
         y++;
      }
   else
      while (height--)
      {
         WinDrawLine(0,y,width-1,y);
         y++;
      }
   WinSetDrawWindow(oldwh);
}
int graphicsActive = 0;
static byte isHandEra()
{
   UInt32 version;
   if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0)
      return sysGetROMVerMajor(version) >= 1;
   return 0;
}
static byte isSony() // returns true if is sony and hr mode is available
{
   SonySysFtrSysInfoP sonySysFtrSysInfoP;
   Err error = 0;

   //Sony ??
   if (!(error = FtrGet(sonySysFtrCreator,sonySysFtrNumSysInfoP, (UInt32*)&sonySysFtrSysInfoP)))
      return sonySysFtrSysInfoP->libr & sonySysFtrSysInfoLibrHR;
   return 0;
}
static byte isDana() // guich@402_39
{
   uint32 version=0;
   int is = _ScreenFeaturePresent(&version);
/*   if (is) : 0x1002001 = 1.0?
      vmGlobals->debug(vmGlobals->int2str("is dana",version));
   else
      vmGlobals->debug("isnt dana");*/
   return is;
}

static uint32 lastLYSize = 0;
inline static void initLookupY()
{
	uint32 *p;
	int32 count = maxY+1; // guich@310_10: changed from MaxScreenY to maxY (bug occurs when loading an image greater than the screen resolution
	uint32 sum=0;
   uint32 size = count << 2;
   // create the lookup table for y coord
   if (size > lastLYSize)
   {
      lastLYSize = size;
      if (lookupY) xfree(lookupY);
      lookupY = (uint32 *)(*vmGlobals->xmalloc)(size);
   }
   p = lookupY;
   while (count--)
   {
      *p++ = sum;
      sum += widthInBPP;
   }
}
void graphicsStart()
{
	if (!graphicsActive && !vmGlobals->errorWinDisplayed)
	{
   	Word error;
      if (usePalmOSFuncs)
      {
         if (isSony())
         {
            UInt32 width,height;
            UInt32 depth;
            Boolean enableColorP;
            /* HR available */
            if ((error = SysLibFind(sonySysLibNameHR, &sonyRefNum)))
            {
               if (error == sysErrLibNotFound)
               {
                  /* couldn't find lib */
                  error = SysLibLoad( 'libr', sonySysFileCHRLib, &sonyRefNum );
               }
            }
            if (!error )
            {
               error = HROpen(sonyRefNum);
               width = hrWidth; height = hrHeight;
               depth = displayBPP;//8; // (in color mode of 256 colors)
               enableColorP = isColor;//true;
               error = HRWinScreenMode ( sonyRefNum, winScreenModeSet,&width, &height, &depth, &enableColorP );
               // high-resolution mode
            }
            else sonyRefNum = 0;
         }
         else
         {
            WinScreenMode(winScreenModeSet, 0, 0, &displayBPP, 0);
            if (isHandEra()) VgaSetScreenMode(screenMode1To1, rotateModeNone);
         }
      }
      mainGrHandle = CreateOffscreenWindow(0, MaxScreenX,MaxScreenY,usePalmOSFuncs?0:1,&error);
      if (error != 0)
      {
         (*vmGlobals->debug)("could not start graphics mode");
         return;
      }
      SetDrawWindow(mainGrHandle);
      if (!usePalmOSFuncs)
         oldSSA = setGrayMode(1, GRHANDLE_MemAddr(mainGrHandle), GRHANDLE_Width(mainGrHandle)/8);
      else
         eraseWindow(GRHANDLE_WinHandle(mainGrHandle), GRHANDLE_Width(mainGrHandle),GRHANDLE_Height(mainGrHandle));
/*{
ScrnSystemStateType p;
ScrnGetSystemState(&p);
vmGlobals->debug("init");
vmGlobals->debug(vmGlobals->int2str("screenMode",p.screenMode));
vmGlobals->debug(vmGlobals->int2str("curAppScreenMode",p.curAppScreenMode));
vmGlobals->debug(vmGlobals->int2str("offset_X",p.offset_X));
vmGlobals->debug(vmGlobals->int2str("offset_Y",p.offset_Y));
vmGlobals->debug(vmGlobals->int2str("draw_override",p.draw_override));
} */
   }
   graphicsActive = 1;
}
void graphicsSwap(int on)
{
   if (on && !graphicsActive && !vmGlobals->errorWinDisplayed)
   {
      if (!usePalmOSFuncs)
         oldSSA = setGrayMode(1, GRHANDLE_MemAddr(mainGrHandle), GRHANDLE_Width(mainGrHandle)/8);
/*{ used in Dana tests
RectangleType rect;
ScrnSystemStateType p;
ScrnGetSystemState(&p);
vmGlobals->debug(vmGlobals->int2str("screenMode",p.screenMode));
vmGlobals->debug(vmGlobals->int2str("curAppScreenMode",p.curAppScreenMode));
vmGlobals->debug(vmGlobals->int2str("offset_X",p.offset_X));
vmGlobals->debug(vmGlobals->int2str("offset_Y",p.offset_Y));
vmGlobals->debug(vmGlobals->int2str("draw_override",p.draw_override));

   rect.topLeft.x = 0;
   rect.topLeft.y = 0;

   WinGetDisplayExtent(&rect.extent.x, &rect.extent.y);
   WinSetWindowBounds(FrmGetWindowHandle(vmGlobals->wabaFrmP), &rect);
} */
      graphicsActive = 1;
   }
   else
   if (!on && graphicsActive)
   {
      if (!usePalmOSFuncs)
         setGrayMode(0, oldSSA, MaxScreenX>>4);
      graphicsActive = 0;
   }
}
void graphicsReset() // guich@300_60: called when another form is set active: the screen address may had changed - this is only called on PalmOS >= 3.5
{
   WinHandle wh = (WinHandle)GRHANDLE_WinHandle(mainGrHandle);
   byte *scr = usePalmOSFuncs?(byte *)BmpGetBits(WinGetBitmap(wh)):GRHANDLE_MemAddr(mainGrHandle); // guich#320_12
//   byte *mainScr = GRHANDLE_MemAddr(mainGrHandle);
//   if (scr != mainScr) // address changed? guich@402_45: if PrintBoy dialog "out of range" appears, this isnt called, resulting in a reset
   {
      GRHANDLE_MemAddr(mainGrHandle) = scr;
      if (wh == GRHANDLE_WinHandle(activeGrHandle))
         activeScr = GRHANDLE_MemAddr(mainGrHandle);
      SemaphoreReserve(1);
      (*vmGlobals->executePostEvent)(-6000,0,0,0,0,0);
      SemaphoreRelease(1);
   }
}

// the following variables must be set in this method: MaxScreenX, MaxScreenY, displayBPP, isColor, maxColors
void graphicsSetup()
{
   Coord cx,cy;

   WinGetDisplayExtent(&cx,&cy);

   displayBPP = 2;      // bits per pixel for the display
   maxColors = 4;
   MaxScreenX = cx;
   MaxScreenY = cy;
   isColor = 0;

   usePalmOSFuncs = (vmGlobals->romVersion >= 0x03500000)?1:0;
   palette = (int32*)(*vmGlobals->xmalloc)(1024);
   if (!palette)
      return;
   xmemmove(palette,systemPalette,4*256); // copy this palette to the global palette
   colorHashTable = (struct htElemColor **)((*vmGlobals->xmalloc)(sizeof(struct htElemColor)*(COLOR_HASHSIZE+1))); // guich@400_33
   if (!colorHashTable)
      return;
   xmemzero(colorHashTable,sizeof(struct htElemColor)*(COLOR_HASHSIZE+1));
   if (usePalmOSFuncs)
   {
      UInt32 depth,pow2,width,height;
      int bpp;
      Boolean enableColor;

      // get the screen sizes and deph
	   WinScreenMode(winScreenModeGet, &width, &height, &depth, 0);
	   if (isHandEra())
	   {
#ifdef SUPPORT_HANDERA_HIGHRES
	      MaxScreenX = width;
	      MaxScreenY = height-16; // exclude minimized silk area
#else
         MaxScreenX = MaxScreenY = width;	// exclude grafitti area
#endif
	   }
	   else
      if (isSony())
      {
         MaxScreenX = hrWidth;
         MaxScreenY = hrHeight;
      }
      else
	   {
   	   MaxScreenX = width;
   	   MaxScreenY = height;
   	}
	   oldBPP = depth;

	   // get if color is supported or only grayscale
	   WinScreenMode(winScreenModeGetSupportsColor, 0, 0, &depth, &enableColor);
	   isColor = (int)enableColor;
      // get the supported depths
      if (isDana()) // guich@402_39
      {
         displayBPP = 4;
         maxColors = 16;
      }
      else
      {
         WinScreenMode(winScreenModeGetSupportedDepths, 0, 0, &depth, 0);

         if (width == 160 && height == 160 && depth == 3) // guich@200b4_101: mono devices must use internal routines.
            usePalmOSFuncs = 0;
         else
         {
      	   // an array of bits is returned. scans for the maximum depth.
      	   pow2 = (UInt32)1 << (UInt32)31;
      	   bpp = 32;
      	   while (pow2)
      	      if (pow2 & depth)
      	      {
      	         displayBPP = bpp;
      	         maxColors = pow2*2;
      	         break;
      	      }
      	      else
      	      {
      	         pow2 /= 2;
      	         bpp -= 1; // available modes expected: 32,24,16,8,4,2
      	      }
      	}
      }
	}
	vmGlobals->screenBPP = displayBPP; // set to real displayBPP before changing it
   // guich@200b3: for some reason, the palette is not being correctly applied in 16bpp. so, we narrow down the bpp.
   if (displayBPP > 8)
   {
      displayBPP = 8;
      maxColors = 256;
   }
	vmGlobals->maxColors = maxColors;		// used in settings
	vmGlobals->screenWidth = MaxScreenX;
	vmGlobals->screenHeight = MaxScreenY;
}
////////////////////////////////////////////////////////////////////////////
typedef void (*drawTextCNOTProcType)(uint16 *text, uint16 len, int16 x, int16 beforeX, int16 clipX2, int16 height, byte *scr0, byte *bitmapTable, int16 *bitIndexTable, int16 offsetY, int32 rowWidthInBytes, int16 firstChar, int16 lastChar, byte color0);
drawTextCNOTProcType drawTextCNOT;
static struct CharBits cbits;

// clip, but not translate
static void drawTextCNOT_2bpp(uint16 *text, uint16 len, int16 x,
   int16 beforeX, // how many pixels to skip, because they are before the clip region
   int16 clipX2,  // since we do not know the string width, here we just need to check if we're after the screen bounds
   int16 height,  // how many char rows to write
   byte *scr0,
   byte *bitmapTable,
   int16 *bitIndexTable,
   int16 offsetY,
   int32 rowWidthInBytes, // from here to below: let the system find the position for us
   int16 firstChar,
   int16 lastChar,
   byte color0)
{
   byte charAnds[8] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
   byte scrAnds[4] = {0x3F,0xCF,0xF3,0xFC};

   int16 offset;
   int32 index;
   int16 width,h;
   byte *charStart,*scrStart;
   byte charStartBit,scrStartBit;
   uint32 scrWidthInBytes = widthInBPP;
   byte scrAnd,charAnd;
   byte color;
   uint32 scrRollup = height*scrWidthInBytes;
   uint32 charRollup = height*rowWidthInBytes;
   struct UserFont *uf = currentUserFont;
   struct CharBits *bits = &cbits;
   byte isUff = IS_UNICODE_FONT(uf);
   color0 = ror(color0,2);

   scrStart     = scr0 + (x>>2);
   scrStartBit  = x&3;
   color        = ror(color0, scrStartBit<<1);

   if (!isUff)
   {
      charRollup = height*rowWidthInBytes;
      bitmapTable += offsetY * rowWidthInBytes;
   }

   while (len--)
   {
      if (*text == 13)  // guich@341_9: fixed <enter> problems with MultiEdit
         goto next;
      if (isUff)
      {
         uffSetCharBits(*text, uf, bits);  // pgr@402_50
         width = bits->width;
         if (width == 0)
            goto skip;
         bitmapTable = bits->bitmapTable + (int32)offsetY * bits->rowWIB;
         rowWidthInBytes = bits->rowWIB;
         charRollup = height * rowWidthInBytes;
         offset = bits->offset;
      }
		else
		{
         index = (byte)(*text) - firstChar;
         ////////////// Char validation
         if (index < 0 || index > lastChar)
         {
            // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
            goto skip;
         }
         else
         {
            // Get the source x coordinate and width of the character
            offset    = bitIndexTable[index];
            width     = bitIndexTable[index+1] - offset;
            if (width == 0) // guich@340_46: show the MCS char if no width (glyph not present)
            {
skip:
               x += UF_MCS_WIDTH; // guich@401_12: just skip 4 pixels
               beforeX -= UF_MCS_WIDTH;
               scrStart     = scr0 + (x>>2);
               scrStartBit  = x&3;
               color        = ror(color0, scrStartBit<<1);
               goto next;
            }
         }
      }
      // draw the char in vertical stripes
      if (beforeX > 0)
      {
         while (beforeX > 0 && width > 0)
         {
            x++;
            beforeX--;
            width--;
            offset++;
         }
         scrStart     = scr0 + (x>>2);
         scrStartBit  = x&3;
         color        = ror(color0, scrStartBit<<1);
      }
      if (x >= clipX2) // beyond clipping region?
         return;

      charStart    = bitmapTable + (offset >> 3); // char: 8 pixels per byte
      charStartBit = offset & 7;

      while (width-- > 0) // something of the char remain to be drawn?
      {
         scrAnd  = scrAnds[scrStartBit];
         charAnd = charAnds[charStartBit];

         for (h = height; h-- > 0;) // draw each row
         {
            if (*charStart & charAnd)
               *scrStart = (*scrStart & scrAnd) | color;
            charStart += rowWidthInBytes;
            scrStart  += scrWidthInBytes;
         }
         // verify bit bounds
         if (++charStartBit == 8) // finished this byte?
         {
            charStartBit = 0; // reset counter
            charStart++;      // inc current byte
         }
         if (++scrStartBit == 4)
         {
            scrStartBit = 0;
            scrStart++;
         }
         // increment x-dependant values
         x++;
         if (x >= clipX2) // beyond clipping region?
            return;
         color = ror(color,2);
         scrStart -= scrRollup;
         charStart -= charRollup;
      }
next:
      text++;
   }
}

static void drawTextCNOT_4bpp(uint16 *text, uint16 len, int16 x,
   int16 beforeX, // how many pixels to skip, because they are before the clip region
   int16 clipX2,  // since we do not know the string width, here we just need to check if we're after the screen bounds
   int16 height,  // how many char rows to write
   byte *scr0,
   byte *bitmapTable,
   int16 *bitIndexTable,
   int16 offsetY,
   int32 rowWidthInBytes, // from here to below: let the system find the position for us
   int16 firstChar,
   int16 lastChar,
   byte color0)
{
   byte charAnds[8] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
   byte scrAnds[4] = {0x0F,0xF0};

   int16 offset;
   int32 index;
   int16 width,h;
   byte *charStart,*scrStart;
   byte charStartBit,scrStartBit;
   uint32 scrWidthInBytes = widthInBPP;
   byte scrAnd,charAnd;
   byte color;
   uint32 scrRollup = height*scrWidthInBytes;
   uint32 charRollup;
   struct UserFont *uf = currentUserFont;
   struct CharBits *bits = &cbits;
   byte isUff = IS_UNICODE_FONT(uf);
   color0 = ror(color0,4);

   scrStart     = scr0 + (x>>1);
   scrStartBit  = x&1;
   color        = ror(color0, scrStartBit<<2);

   if (!isUff)
   {
      charRollup = height*rowWidthInBytes;
      bitmapTable += offsetY * rowWidthInBytes;
   }

   while (len--)
   {
      if (*text == 13)  // guich@341_9: fixed <enter> problems with MultiEdit
         goto next;
      if (isUff)
      {
         uffSetCharBits(*text, uf, bits);  // pgr@402_50
         width = bits->width;
         if (width == 0)
            goto skip;
         bitmapTable = bits->bitmapTable + (int32)offsetY * bits->rowWIB;
         rowWidthInBytes = bits->rowWIB;
         charRollup = height * rowWidthInBytes;
         offset = bits->offset;
      }
      else
      {
         index = (byte)(*text) - firstChar;
         ////////////// Char validation
         if (index < 0 || index > lastChar)
         {
            // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
            goto skip;
         }
         else
         {
            // Get the source x coordinate and width of the character
            offset    = bitIndexTable[index];
            width     = bitIndexTable[index+1] - offset;
            if (width == 0) // guich@340_46: show the MCS char if no width (glyph not present)
            {
skip:
               x += UF_MCS_WIDTH; // guich@401_12: just skip 4 pixels
               beforeX -= UF_MCS_WIDTH;
               scrStart     = scr0 + (x>>1);
               scrStartBit  = x&1;
               color        = ror(color0, scrStartBit<<2);
               goto next;
            }
         }
      }
      // draw the char in vertical stripes
      if (beforeX > 0)
      {
         while (beforeX > 0 && width > 0)
         {
            x++;
            beforeX--;
            width--;
            offset++;
         }
         scrStart     = scr0 + (x>>1);
         scrStartBit  = x&1;
         color        = ror(color0, scrStartBit<<2);
      }
      if (x >= clipX2) // beyond clipping region?
         return;

      charStart    = bitmapTable + (offset >> 3); // char: 8 pixels per byte
      charStartBit = offset & 7;

      while (width-- > 0) // something of the char remain to be drawn?
      {
         scrAnd  = scrAnds[scrStartBit];
         charAnd = charAnds[charStartBit];

         for (h = height; h-- > 0;) // draw each row
         {
            if (*charStart & charAnd)
               *scrStart = (*scrStart & scrAnd) | color;
            charStart += rowWidthInBytes;
            scrStart  += scrWidthInBytes;
         }
         // verify bit bounds
         if (++charStartBit == 8) // finished this byte?
         {
            charStartBit = 0; // reset counter
            charStart++;      // inc current byte
         }
         if (++scrStartBit == 2)
         {
            scrStartBit = 0;
            scrStart++;
         }
         // increment x-dependant values
         x++;
         if (x >= clipX2) // beyond clipping region?
            return;
         color = ror(color,4);
         scrStart -= scrRollup;
         charStart -= charRollup;
      }
next:
      text++;
   }
}

static void drawTextCNOT_8bpp(uint16 *text, uint16 len, int16 x,
   int16 beforeX, // how many pixels to skip, because they are before the clip region
   int16 clipX2,  // since we do not know the string width, here we just need to check if we're after the screen bounds
   int16 height,  // how many char rows to write
   byte *scr0,
   byte *bitmapTable,
   int16 *bitIndexTable,
   int16 offsetY,
   int32 rowWidthInBytes, // from here to below: let the system find the position for us
   int16 firstChar,
   int16 lastChar,
   byte color0)
{
   byte charAnds[8] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

   int16 offset;
   int32 index;
   int16 width,h;
   byte *charStart,*scrStart;
   byte charStartBit;
   uint32 scrWidthInBytes = widthInBPP;
   byte charAnd;
   byte color=color0;
   uint32 scrRollup = height*scrWidthInBytes;
   uint32 charRollup = height*rowWidthInBytes;
   struct UserFont *uf = currentUserFont;
   struct CharBits *bits = &cbits;
   byte isUff = IS_UNICODE_FONT(uf);

   scrStart = scr0 + x;

   if (!isUff)
   {
      charRollup = height*rowWidthInBytes;
      bitmapTable += offsetY * rowWidthInBytes;
   }

   while (len--)
   {
      if (*text == 13)  // guich@341_9: fixed <enter> problems with MultiEdit
         goto next;
      if (isUff)
      {
         uffSetCharBits(*text, uf, bits);  // pgr@402_50
         width = bits->width;
         if (width == 0)
            goto skip;
         bitmapTable = bits->bitmapTable + (int32)offsetY * bits->rowWIB;
         rowWidthInBytes = bits->rowWIB;
         charRollup = height * rowWidthInBytes;
         offset = bits->offset;
      }
		else
		{
         index = (byte)(*text) - firstChar;
         ////////////// Char validation
         if (index < 0 || index > lastChar)
         {
            // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
            goto skip;
         }
         else
         {
            // Get the source x coordinate and width of the character
            offset    = bitIndexTable[index];
            width     = bitIndexTable[index+1] - offset;
            if (width == 0) // guich@340_46: show the MCS char if no width (glyph not present)
            {
skip:
               x += UF_MCS_WIDTH; // guich@401_12: just skip 4 pixels
               beforeX -= UF_MCS_WIDTH;
               scrStart     = scr0 + x;
               goto next;
            }
         }
      }
      // draw the char in vertical stripes
      if (beforeX > 0)
      {
         while (beforeX > 0 && width > 0)
         {
            x++;
            beforeX--;
            width--;
            offset++;
         }
         scrStart     = scr0 + x;
      }
      if (x >= clipX2) // beyond clipping region?
         return;

      charStart    = bitmapTable + (offset >> 3); // char: 8 pixels per byte
      charStartBit = offset & 7;

      while (width-- > 0) // something of the char remain to be drawn?
      {
         charAnd = charAnds[charStartBit];

         for (h = height; h-- > 0;) // draw each row
         {
            if (*charStart & charAnd)
               *scrStart = color;
            charStart += rowWidthInBytes;
            scrStart  += scrWidthInBytes;
         }
         // verify bit bounds
         if (++charStartBit == 8) // finished this byte?
         {
            charStartBit = 0; // reset counter
            charStart++;      // inc current byte
         }
         scrStart++;
         // increment x-dependant values
         x++;
         if (x >= clipX2) // beyond clipping region?
            return;
         scrStart -= scrRollup;
         charStart -= charRollup;
      }
next:
      text++;
   }
}

static void drawText(uint16 *text, uint32 len, int x, int y, int32 color)
{
   struct UserFont *font = currentUserFont;
   FontPtr fontP = font->fontP;
   int16 height = (byte)fontP->fRectHeight,y2;
   int16 beforeX=0;
   int32 offsetY=0;
   int16 cy1 = CLIPY1;
   int16 cy2 = CLIPY2;
   // Y validation
   y += TRANSY;
   y2 = y + height;
   // do we cross the bounds?
   if (y < cy1 || y2 >= cy2)
   {
      // first, verify if we are fully before or completely after the clip Y
      if (y2 < cy1 || y >= cy2)
         return;
      // ok, we're inside the clipping region. now verify if we cross the boundaries
      if (y < cy1) // text starts before clip: set offsetY, and decrease height
      {
         offsetY = cy1 - y;
         height -= offsetY;
         y += offsetY;
         y2 = y + height;
      } // don't put an else here!
      if (y2 >= cy2) // text goes beyond clip: decrease height
         height -= y2 - cy2;
   }
   // X validation
   x += TRANSX; // we do not have the text width
   // now we verify if we're before the clip X
   beforeX = CLIPX1 - x; // stores how many pixels we'll skip
   if (beforeX < 0)
      beforeX = 0; // no pixels to skip

   (*drawTextCNOT)(text, (uint16)len, (int16)x,
      beforeX, CLIPX2, height,
      activeScr + lookupY[y],
      font->bitmapTable,
      font->bitIndexTable, offsetY, font->rowWidthInBytes,
      fontP->firstChar, fontP->lastChar - fontP->firstChar, (byte)color);
}

// called by the system to let the port register the primitives
static void registerPrimitives(GrHandle h)
{
   if (usePalmOSFuncs)
   {
      if (displayBPP == 4) // guich@200b4: corrects a bug in the PalmOS
      {
         if (drawVLine != &drawVLine_4bpp) // guich@200b4: avoid setting again with the same things
         {
   	      drawVLine      = &drawVLine_4bpp;
   	      drawHLine      = &drawHLine_4bpp;
   	      setPixel       = &setPixel_4bpp;
   	      getPixel       = &getPixel_4bpp;
   	      setPixelXM     = &setPixelXM_4bpp;
   	      setPixelNOCT   = &setPixelNOCT_4bpp;
   	      invVLineNOCT   = &invVLineNOCT_4bpp;
   	      invPixelNOCT   = &invPixelNOCT_4bpp;
   	      eraseVLineNOCT = &eraseVLineNOCT_4bpp;
   	      drawHLineNOCT  = &drawHLineNOCT_4bpp;
   	      fillRectNOCT   = &fillRectNOCT_4bpp;
            drawTextCNOT   = &drawTextCNOT_4bpp;
   	   }
	   }
      else
      {
         if (drawVLine != &drawVLine_8bpp)
         {
   	      drawVLine      = &drawVLine_8bpp;
   	      drawHLine      = &drawHLine_8bpp;
   	      setPixel       = &setPixel_8bpp;
   	      getPixel       = &getPixel_8bpp;
   	      setPixelXM     = &setPixelXM_8bpp;
   	      setPixelNOCT   = &setPixelNOCT_8bpp;
   	      invVLineNOCT   = &invVLineNOCT_8bpp;
   	      invPixelNOCT   = &invPixelNOCT_8bpp;
   	      eraseVLineNOCT = &eraseVLineNOCT_8bpp;
   	      drawHLineNOCT  = &drawHLineNOCT_8bpp;
   	      fillRectNOCT   = &fillRectNOCT_8bpp;
            drawTextCNOT   = &drawTextCNOT_8bpp;
   	   }
	   }
      getBitmapRow      = &getBitmapRow_Color;
      getColor          = &getColor_Color;
      drawBitmap        = &drawBitmap_Color;
      drawScreen        = &drawScreen_;
   }
   else
   if (getBitmapRow != &getBitmapRow_2bpp)
   {
      getBitmapRow      = &getBitmapRow_2bpp;
      drawVLine         = &drawVLine_2bpp;
      drawHLine         = &drawHLine_2bpp;
      setPixel          = &setPixel_2bpp;
      getPixel          = &getPixel_2bpp;
      setPixelXM        = &setPixelXM_2bpp;
      setPixelNOCT      = NULL; // not used in 2bpp
      invPixelNOCT      = &invPixelNOCT_2bpp;
      invVLineNOCT      = &invVLineNOCT_2bpp;
      eraseVLineNOCT    = &eraseVLineNOCT_2bpp;
      drawHLineNOCT     = &drawHLineNOCT_2bpp;
      fillRectNOCT      = &fillRectNOCT_2bpp;
      getColor          = &getColor_Gray;
      drawBitmap        = &drawBitmap_Gray;
      drawScreen        = &drawScreen_;
      drawTextCNOT      = &drawTextCNOT_2bpp;
   }
}
static int getWidthInBPP(int width)
{
   if (!usePalmOSFuncs)
      return (width+3)/4; // in bytes
   else
   if (displayBPP == 4)
      return ((int)((width+3)/4)) * 2;
   else
      return ((int)(width+1)/2) * 2; // always 8 bpp , odd width
}
#define BLOCK_SIZE     (128L)
#define BOUNDARY_MASK   (long)(~(long)(BLOCK_SIZE * 1024 - 1))
#define BLOCK_OVERHEAD (20L)
static byte *allocAvoiding128kbBoundary(int32 size)
{
   byte *failed[10];
   int count=0,i;
   byte *ptr;
   //MemHeapCompact(0);
   while (1)
   {
      long start, end;
      ptr = (byte*)(*vmGlobals->xmalloc)(size);
      start = (((long)ptr) - BLOCK_OVERHEAD)   & BOUNDARY_MASK;
      end   = (((long)ptr) + size) & BOUNDARY_MASK;

      if (ptr == NULL)
         break;
      if (start != end   ||         // the block crosses the boundary, or the end touches the boundary
         start == (((long)ptr)-BLOCK_OVERHEAD)) // the start of the block touches the boundary
      {
         // the allocated block crosses the 128kb limit. Add it to the list of allocated blocks and try again
         if (count == 10)
         {
            for (i=0; i < count; i++)
              xfree(failed[i]);
             count = 0;
         }
         failed[count++] = ptr;
         // try again
      }
      else break;
    }
    if (count > 0) // free the failed blocks
       for (i=0; i < count; i++)
          xfree(failed[i]);
    return ptr;
}
static byte *allocWinMemory(int32 width, int32 height, VoidPtr *ptr1, VoidPtr *ptr2)
{
   byte *b = NULL;
   if (usePalmOSFuncs)
   {
      Err error;
      int bpp = displayBPP;
      if (sonyRefNum)
   		*ptr1 = HRBmpCreate(sonyRefNum,width,height,bpp,NULL,&error);
      else
   		*ptr1 = BmpCreate(width,height,bpp,NULL,&error);
		if (error) //sysErrNoFreeResource
	      return NULL;
		if (sonyRefNum)
   		*ptr2 = HRWinCreateBitmapWindow(sonyRefNum,(BitmapType *)*ptr1, &error);
		else
   		*ptr2 = WinCreateBitmapWindow((BitmapType *)*ptr1, &error);
		if (error)
		{
		   BmpDelete(*ptr1); // guich@421_47
		   *ptr1 = NULL;
	      return NULL;
	   }
		b = BmpGetBits((BitmapType *)*ptr1);
   }
   else
   {
	   int32 size = (int32)getWidthInBPP(width)*height+2; // guich@200b4_54: alloc 1 more byte for the image to avoid the buggy bitblt_Gray that may write the last byte inside the memory structures.
		b = !mainGrHandle?allocAvoiding128kbBoundary(size):(*vmGlobals->xmalloc)(size); // guich@240_22: the video memory has to avoid the 128kb limit bc of the problem of corrupted screens.
		if (b) xmemzero(b,size);
   }
   return b;
}
static void FreeGrHandle(GrHandle h);
static void DestroyOffscreenWindow(GrHandle h)
{
   if (h && GRHANDLE_MemAddr(h))
   {
	   if (GRHANDLE_Palette(h))
	      xfree(GRHANDLE_Palette(h));
	   if (GRHANDLE_WinHandle(h))
	      WinDeleteWindow((WinHandle)GRHANDLE_WinHandle(h),false);
	   if (GRHANDLE_BmpHandle(h))
	      BmpDelete((BitmapType *)GRHANDLE_BmpHandle(h));
      if (!GRHANDLE_IsBitmap(h) && (!mainGrHandle || (GRHANDLE_MemAddr(h) != GRHANDLE_MemAddr(mainGrHandle))))
         xfree(GRHANDLE_MemAddr(h));
      FreeGrHandle(h);
   }
}
// if allocWinMem is true, a new "screen" will be allocated. if is false, same "screen" of MainWin is used, but with its own clip and translate
GrHandle CreateOffscreenWindow(WObject obj, int width, int height, int allocWinMem, Word *error)
{
   GrHandle h;
   VoidPtr ptr1=NULL,ptr2=NULL;
   byte *bmp=NULL;
   *error = 1;
   if (obj && (*vmGlobals->pushObject)(obj) == -1) return NULL; // GC can be called on createOffscreenWindow
   if (allocWinMem)
      bmp = allocWinMemory(width,height,&ptr1,&ptr2); // even if the bitmaps are not handled directly, something different of null must be returned here
   *error = 1;
   if (!bmp && allocWinMem)
   {
      (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "Not enough memory for bit map"); // guich@200b4_78
      goto freeAndExit;
   }
   h = (GrHandle)(*vmGlobals->xmalloc)(sizeof(GraphicsContext));
   if (!h)
   {
      (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "Not enough memory for GraphicsContext"); // guich@200b4_78
      goto freeAndExit;
   }
   GRHANDLE_WidthInBPP(h)   = getWidthInBPP(width); // saves the size in bytes
   GRHANDLE_BmpHandle(h)    = ptr1;
   GRHANDLE_WinHandle(h)    = allocWinMem?ptr2:&vmGlobals->wabaFrmP->window;
   GRHANDLE_MemAddr(h)      = allocWinMem?bmp:(usePalmOSFuncs?(byte *)BmpGetBits(WinGetBitmap(GRHANDLE_WinHandle(h))):GRHANDLE_MemAddr(mainGrHandle));
   GRHANDLE_TransX(h)       = 0;
   GRHANDLE_TransY(h)       = 0;
   GRHANDLE_Width(h)        = width;
   GRHANDLE_Height(h)   	 = height;
   GRHANDLE_ClipX1(h)       = 0;
   GRHANDLE_ClipY1(h)       = 0;
   GRHANDLE_ClipX2(h)       = width;
   GRHANDLE_ClipY2(h)       = height;
   GRHANDLE_PaletteSize(h)  = 0;
   GRHANDLE_Palette(h)      = NULL;
   *error = 0;
   if (obj) (*vmGlobals->popObject)(obj);
   return h;
freeAndExit:
   if (obj) (*vmGlobals->popObject)(obj);
   if (bmp) xfree(bmp);
   return NULL;
}
int lastWidthInBPP = -1;
int lastMaxY = -1;
void SetDrawWindow(GrHandle h)
{
   WinHandle wh;
   if (!activeGrHandle)
      currentUserFont = vmGlobals->allFonts+0;
   wh = (WinHandle)GRHANDLE_WinHandle(h);
   if (wh)
      WinSetDrawWindow(wh);
   activeGrHandle = h;
   registerPrimitives(h);
   transXY = MAKE32(GRHANDLE_TransX(h),GRHANDLE_TransY(h));
   clipX=MAKE32(GRHANDLE_ClipX1(h),GRHANDLE_ClipX2(h));
   clipY=MAKE32(GRHANDLE_ClipY1(h),GRHANDLE_ClipY2(h));

   // guich@300_47: after a new form was called, the main address may change, so we make sure we always use it
   activeScr = usePalmOSFuncs?(byte *)BmpGetBits(WinGetBitmap(GRHANDLE_WinHandle(h))):GRHANDLE_MemAddr(h); // guich@400_50: now we always get the winhandle, bc it may change in different objects

   maxX = GRHANDLE_Width(h);
   maxY = GRHANDLE_Height(h);
   widthInBPP = GRHANDLE_WidthInBPP(h);
   mustSwap = displayBPP == 8 && !BmpColortableSize(WinGetBitmap(wh)); // guich@200b4_160
   if (widthInBPP != lastWidthInBPP || maxY > lastMaxY) // guich@310_10: test also if maxY was changed  - guich@340_12: only if maxY is greater.
   {
      initLookupY();
      lastWidthInBPP = widthInBPP;
      lastMaxY = maxY;
   }
}
////////////////////////////////////////////////////////////////////////////
void graphicsStop()
{
   if (graphicsActive)
   {
      if (!usePalmOSFuncs)
         setGrayMode(0, oldSSA, MaxScreenX>>4);
      else
        	WinScreenMode(winScreenModeSet, 0, 0, &oldBPP, 0);
    	xfree(mainGrHandle); // guich@200b4_61: free this mainGrHandle too.
      xfree(palette);
   	disposeColorHashtable(1);
    	xfree(lookupY);
    	if (sonyRefNum)
         HRClose(sonyRefNum);

      graphicsActive = 0;
   }
   xfree(systemPalette); // guich@421_20
   releaseAllFonts();
}
 