/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

///////////////////////////////////////////////////////////
//          New Unicode Support

// Load and cache used ranges. pending: unload unused ranges
// Return index in range arrays, or -1 if MCS
static int32 uffEnsureRange(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // check if the range is already in the cache - if so, it is sure it is already loaded.
   if (ch <= uff->recentLastChar && ch >= uff->recentFirstChar)
      return uff->recentRangeIndex;
   else
   {
      int32 rindex = -1;

      // Find range. Perform a binary search here. This whould help if a lot of ranges.
      int32 low = 0;
      int32 hi = uff->nParts - 1;
      int32 mid = 0;

      while (low <= hi)
      {
         mid = (low + hi) >> 1;
         if (ch >= uff->uffFirstChar[mid] && ch <= uff->uffLastChar[mid])
         {
            rindex = mid;
            break;
         }
         else
         if (ch < uff->uffFirstChar[mid])
            hi = mid - 1;
         else
            low = mid + 1;
      }

      if (rindex != -1)
      {
         if (uff->ranges[rindex].rowWIB == 0) // not yet loaded
         {
            int32 size;
            struct UffUserFont *uff = uf->uff;
            struct UserFontRange *range = &uff->ranges[rindex];
            byte *recPtr;
            int32 *bit;
            range->recH = DmQueryRecord(uf->dmRef, uff->recNo[rindex]); // load record
            recPtr = (byte*)MemHandleLock(range->recH);
            range->rowWIB = getUInt32(recPtr); recPtr += 4;

            size = (int32)range->rowWIB * (int32)uf->fontP->fRectHeight;
            range->bitmapTable = recPtr; recPtr += size;

            // unfortunately, we can't point to the bitIndexTable due to memory alinment
            size = uff->uffLastChar[rindex] - uff->uffFirstChar[rindex] + 1 + 1;
            range->bitIndexTable = bit = (int32*)vmGlobals->xmalloc(size<<2);
            if (bit == NULL)
            {
               MemHandleUnlock(range->recH);
               return -1;
            }
            for (; size--; recPtr+=4)
               *bit++ = getInt32(recPtr);
         }
         uff->recentFirstChar  = uff->uffFirstChar[rindex];
         uff->recentLastChar   = uff->uffLastChar[rindex];
         uff->recentRangeIndex = rindex;
      }
      return rindex;
   }
}

// Get the width of a character
static int uffCharWidth(char16 ch, struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   // ensure that range is loaded
   int32 rindex = uffEnsureRange(ch,uf);
   if (rindex == -1)
      return UF_MCS_WIDTH;
   else
   {
      struct UserFontRange *range = &uff->ranges[rindex];
      int32 index = (int32)ch - uff->recentFirstChar;
      return range->bitIndexTable[index + 1] - range->bitIndexTable[index];
   }
}

// Update a CharBits object to reflect the specificities of a given character.
void uffSetCharBits(char16 ch, struct UserFont *uf, struct CharBits *bits)
{
   struct UffUserFont *uff = uf->uff;
   int32 rindex = uffEnsureRange(ch,uf);
	if (rindex == -1 && ch != 0xFFFD) // eisvogel@421_44: The character could not be loaded by the ensureRange function. In Unicode there are several replacement characters. The interesting one is U+FFFD. If a font can't display a requested character it should display U+FFFD.
	{
		ch = 0xFFFd; // unicode replacement character
		rindex = uffEnsureRange(ch,uf);
	}
	if (rindex == -1 && ch != 0x0) // eisvogel@421_44: Neither the requested character nor the replacement character was found in the font. Usuall TTF and BDF fonts define a so called ".nodef" character at the codepoint 0x0. This is the very last fallback for characters not supported by the font.
   {
		ch=0x0; // ".notdef" character in ttf and bdf fonts
		rindex = uffEnsureRange(ch,uf);
	}
   if (rindex == -1)
      bits->width = 0;
   else
   {
      int32 cindex = (int32) ch - uff->recentFirstChar;
      struct UserFontRange *range = &uff->ranges[rindex];
      bits->offset = range->bitIndexTable[cindex];
      bits->width = range->bitIndexTable[cindex + 1] - bits->offset;
      if (bits->width != 0)
      {
         bits->rowWIB = range->rowWIB;
         bits->bitmapTable = range->bitmapTable;
      }
   }
}
static void destroyUff(struct UserFont *uf)
{
   struct UffUserFont *uff = uf->uff;
   if (uff->ranges) // guich@552_14: check against out of memory on this member
   {
      int16 nParts = uff->nParts;
      while (--nParts >= 0)
      {
         xfree(uff->ranges[nParts].bitIndexTable);
         if (uff->ranges[nParts].recH)
            MemHandleUnlock(uff->ranges[nParts].recH);
      }
   }
   xfree(uf->fontP);
   xfree(uff->recNo);
   xfree(uff->uffFirstChar);
   xfree(uff->uffLastChar);
   xfree(uff->ranges);
   xfree(uf->uff);
   //can't release uf->recH!
   if (uf->dmRef)
      DmCloseDatabase(uf->dmRef);
}

static int loadUff(DmOpenRef dmRef, struct UserFont*uf)
{
   VoidHand recH;
   uint16 nParts;
   byte *recPtr;
   uint16 *recNo,*fchar,*lchar;
   struct UffUserFont *uff;
   FontPtr f;
   uf->dmRef = dmRef;
   // load header record -  the name's already stored
  	recH = DmQueryRecord(dmRef, 0);
  	recPtr = (byte*)MemHandleLock(recH);
   recPtr += 2; // magic is 12003
   recPtr += 2+getUInt16(recPtr); // fcName is the short name, e.g. "gnuuni", "myarial"
   recPtr += 2; // fcSize is the point size, e.g. 10, 12
   recPtr += 2; // fcSlant is 0 for normal font; 1, 2, 3 for bold, italic, bold-italic
   recPtr += 2+getUInt16(recPtr); // comment
   uf->fontP = f = (FontPtr)vmGlobals->xmalloc(sizeof(FontType));
   uf->uff = NULL;
   xmemzero(f,sizeof(FontType));
   if (f != NULL)
   {
      f->fontType     = getUInt16(recPtr);          recPtr += 2;
      f->maxWidth     = getUInt16(recPtr);          recPtr += 2;
      f->kernMax      = getUInt16(recPtr);          recPtr += 2;
      f->nDescent     = getUInt16(recPtr);          recPtr += 2;
      f->fRectWidth   = getUInt16(recPtr);          recPtr += 2;
      f->fRectHeight  = getUInt16(recPtr);          recPtr += 2;
      f->ascent       = getUInt16(recPtr);          recPtr += 2;
      f->descent      = getUInt16(recPtr);          recPtr += 2;
      f->leading      = getUInt16(recPtr);          recPtr += 2;
      // versus UserFonts, we miss owTLoc and rowWords
      uf->uff           = uff = (struct UffUserFont*)vmGlobals->xmalloc(sizeof(struct UffUserFont));
   }
   if (uf->uff == NULL)
   {
      MemHandleUnlock(recH);
      DmCloseDatabase(dmRef);
      return 0;
   }
   xmemzero(uf->uff,sizeof(struct UffUserFont));
   uff->nParts       = nParts = getUInt16(recPtr); recPtr += 2;

   uff->recNo        = recNo = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffFirstChar = fchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->uffLastChar  = lchar = (uint16 *)vmGlobals->xmalloc(nParts<<1);
   uff->ranges       = (struct UserFontRange *)vmGlobals->xmalloc((int32)nParts*(int32)sizeof(struct UserFontRange));
   if (recNo == NULL || fchar == NULL || lchar == NULL || uff->ranges == NULL)
   {
      MemHandleUnlock(recH); // guich@552_14: also unlock the font
      destroyUff(uf);
      vmGlobals->debug(vmGlobals->int2str("The UFF font has too many range parts and cannot be loaded. The max allowed is 4093 but it has",nParts)); // guich@552_14
      return 0;
   }
   xmemzero(uff->ranges,nParts*sizeof(struct UserFontRange));

   while (nParts-- > 0)
   {
      *recNo++ = getUInt16(recPtr);  recPtr += 2;
      *fchar++ = (uint16)getUInt32(recPtr);  recPtr += 4; // not sure why the original version is not 16 bits
      *lchar++ = (uint16)getUInt32(recPtr);  recPtr += 4;
   }

   uf->recH = (VoidHand)1; // can't be 0, otherwise the font won't be displayed - guich@552_14: moved to here
   MemHandleUnlock(recH);
   return 1;
}
//
///////////////////////////////////////////////////////////
int nrFonts;
void releaseAllFonts()
{
   int i;
   struct UserFont *uf;
   for (i=nrFonts-1; i >= 0; i--)
      if ((uf=&vmGlobals->allFonts[i])->recH)
      {
         if (IS_UNICODE_FONT(uf))
            destroyUff(uf);
         else
         {
            MemHandleUnlock(uf->recH);
            //DmReleaseRecord(uf->dmRef, uf->recIndex, false); // guich@340_47 - guich@400_26: now we do a queryrecord; no need to release the record
            if (uf->recIndex == 0) DmCloseDatabase(uf->dmRef);
         }
      }
   if (vmGlobals->allFonts) xfree(vmGlobals->allFonts);
}

#define FM_STRINGWIDTH 1
#define FM_CHARARRAYWIDTH 2
#define FM_CHARWIDTH 3
static int charWidth(int16 ch, struct UserFont *uf)
{
   int16 index,width;
   if (IS_UNICODE_FONT(uf)) // guich@402_50: ufolib?
      return uffCharWidth(ch, uf);

   if (ch > 0xFF) ch = 0; // eisvogel@421_44: ignore unicode chars.
   index = ch - uf->fontP->firstChar;
   width;
   if (ch == 13) // guich@341_9
      return 0;
   if (index < 0 || ch > uf->fontP->lastChar)
   {
      // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
      return UF_MCS_WIDTH; //guich@401_12: directly return the value.  index = uf->numberOfChars;//guich@340_46 - (int)('?' - uf->fontP->firstChar);
   }
#ifdef PALMOS5
   if (!uf->recH) // internal font? use Palm OS routines
      width = FntWCharWidth(ch);
   else
#endif
   width = uf->bitIndexTable[index+1] - uf->bitIndexTable[index];
   if (width == 0) // guich@340_46
      width = UF_MCS_WIDTH;
   return width;
}

static void FontMetricsGetWidth(int type, Var stack[])
{
	WObject font, fontMetrics; // guich@450_36: surface is ignored from now on
	int32 width=0;
	struct UserFont *uf;

	fontMetrics = stack[0].obj;
	font = WOBJ_FontMetricsFont(fontMetrics);
	if (font == 0)
	{
		RETURN32(0)
	}
	// surface is unused and user fonts are supported!
   uf = WOBJ_FontUserFont(font);
   GR_ENTER
#ifdef PALMOS5
   if (FntGetFont() != uf->fid)
   	FntSetFont(uf->fid);
#endif

	switch (type)
	{
		case FM_CHARWIDTH:
		{
			width = (int32)charWidth((int16)stack[1].intValue,uf);
			break;
		}
		case FM_STRINGWIDTH:
		case FM_CHARARRAYWIDTH:
		{
			WObject string, charArray;
 			int32 start, count;
			uint16 *chars;

			width = 0;
			if (type == FM_STRINGWIDTH)
			{
				string = stack[1].obj;
				if (string == 0)
					break;
				charArray = WOBJ_StringCharArrayObj(string);
				if (charArray == 0)
					break;
				start = 0;
				count = WOBJ_arrayLen(charArray);
			}
			else // FM_CHARARRAYWIDTH
			{
				charArray = stack[1].obj;
				start = stack[2].intValue;
				count = stack[3].intValue;
				if ((*vmGlobals->arrayRangeCheck)(charArray, start, count) == 0)
					break; // array null or range invalid
			}
			chars = (uint16 *)WOBJ_arrayStart(charArray);
			chars = &chars[start];
			while (count-- > 0)
				width += (int32)charWidth(*chars++, uf); // guich@200b2: must handle unicode chars in the future
			break;
		}
	}
	GR_EXIT
	RETURN32(width)
}

static int getUserFontIndex(char *name, char *sufix)
{
   int i,len;
   char buf[41];
   xmemzero(buf,sizeof(buf));
   len = xstrlen(name);
   xstrncpy(buf,name,len);
   if (sufix) xstrncpy(buf+len,sufix,xstrlen(sufix));
   for (i =0; i < nrFonts; i++)
      if (vmGlobals->allFonts[i].name && !StrCaselessCompare(buf,vmGlobals->allFonts[i].name))
         return i;
   return -1;
}
static FontPtr getFontPtr(FontID fontID)
{
	FontID prevFontID;
	FontPtr fontPtr;
   // sets the main font
   prevFontID = FntSetFont(fontID);
   // gets the current font ptr
   fontPtr = FntGetFontPtr();
   // since we dont use the standard routines, restore the original font
   FntSetFont(prevFontID);
   return fontPtr;
}
static void storeUserFont(struct UserFont *uf, FontPtr fontP, char *name, VoidHand recH, DmOpenRef dmRef, uint16 recIndex, uint16 idx)
{
   int32 bitmapTableSize;
	xstrncpy(uf->name,name,min(xstrlen(name),40));
	uf->recH = recH;
	uf->dmRef = dmRef;
	uf->fontP = fontP;
	uf->recIndex = recIndex;
   uf->rowWidthInBytes = fontP->rowWords * 2;
   uf->numberOfChars   = fontP->lastChar - fontP->firstChar + 1;
   uf->bitmapTable     = (byte*)fontP + (int32)sizeof(FontType);
   bitmapTableSize     = (int32)uf->rowWidthInBytes * (int32)fontP->fRectHeight;
   uf->bitIndexTable   = (int16 *)(uf->bitmapTable + bitmapTableSize);
#ifdef PALMOS5
   if (idx < 4) // internal fonts?
   {
      FontID ids[4] = {stdFont,boldFont,largeFont,largeBoldFont};
      uf->fid = ids[idx];
   }
   else
   {
      Err err;
      uf->fid = fntAppFontCustomBase+idx;
      if ((err=FntDefineFont(uf->fid,fontP)) != 0)
      {
         (*vmGlobals->debug)("Error: cannot define font");
         (*vmGlobals->debug)(name);
      }
   }
#endif
}

byte loadAllFonts()
{
	DmSearchStateType state;
	DmOpenRef dmRef;
	Boolean first;
	UInt cardNo;
	LocalID dbID;
	char dbName[41]; // guich@340_47
	Err err;
	ULong crid = getCreator("Font");
	ULong cridu = 'UFFo';
	ULong typeu = 'DATA';
	int installed=0;
	int unicodeInstalled = 0;
	int i=0;
	int internal=4;
   int index = 0,specialIndex; // guich@400_41: fixed problem with storing internal fonts

   GR_ENTER
   // first find how many installed fonts
	first = true;
	while (1)
	{
	   uint32 numRecords;
		err = DmGetNextDatabaseByTypeCreator(first, &state, crid, crid, false, &cardNo, &dbID);
		if (err == dmErrCantFind)
			break;
	 	if (/*!vmGlobals->isHighDensity && */DmDatabaseInfo(cardNo, dbID, dbName, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == 0 // guich@401_23: always delete this file from now on
	 	  && dbName[0] == '5' && xstrncmp(dbName,"5SW",3) == 0) // guich@400_24
	      DmDeleteDatabase(cardNo,dbID);
	   else
	   if (DmDatabaseSize(cardNo, dbID, &numRecords, NULL, NULL) == errNone) // guich@340_47
	      installed += numRecords;
	   else
  	      installed++;
		first = false;
	}
	// guich@402_50: now finds how many unicode fonts are installed
	first = true;
	while (1)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, typeu, cridu, false, &cardNo, &dbID);
		if (err == dmErrCantFind)
			break;
      unicodeInstalled++; // unicode fonts has many records inside it, but count as only one font
		first = false;
	}
	// alloc mem
	nrFonts = unicodeInstalled+installed+internal;
	vmGlobals->allFonts = (struct UserFont *) (*vmGlobals->xmalloc)(sizeof(struct UserFont)*nrFonts); // important! allFonts can't be changed at runtime, bc fonts are directly mapped to it.
	xmemzero(vmGlobals->allFonts,sizeof(struct UserFont)*nrFonts);

	// get the internal fonts
	storeUserFont(vmGlobals->allFonts + SWSTANDARD_FONT, getFontPtr(vmGlobals->isSony?hrStdFont:stdFont),"SWSmall", 0,0,0,0);
	storeUserFont(vmGlobals->allFonts + SWSTANDARD_BOLD_FONT, getFontPtr(vmGlobals->isSony?hrBoldFont:boldFont),"SWSmallBold", 0,0,0,1);
   if (vmGlobals->romVersion >= 0x03000000) // has large fonts?
   {
		storeUserFont(vmGlobals->allFonts + SWLARGE_FONT, getFontPtr(vmGlobals->isSony?hrLargeFont:largeFont),"SWLarge", 0,0,0,2);
		storeUserFont(vmGlobals->allFonts + SWLARGE_BOLD_FONT, getFontPtr(vmGlobals->isSony?hrLargeBoldFont:largeBoldFont),"SWLargeBold", 0,0,0,3);
   }
	// now get the installed fonts
	first = true;
	for (i = 0; i < installed;)
	{
	   uint16 attr;
	   int idx;
		err = DmGetNextDatabaseByTypeCreator(first, &state, crid, crid, false, &cardNo, &dbID);
	 	if (DmDatabaseInfo(cardNo, dbID, dbName, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == 0)
		{
  		   VoidHand recH;
		   CharPtr recPtr;
		   char *dbAlias = dbName;
		   uint32 recCount; // guich@340_47
		   uint32 j;
			dmRef = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
			recCount = DmNumRecords(dmRef);
         for (j = 0; j < recCount; j++) // guich@340_47
         {
         	recH = DmQueryRecord(dmRef, j); // guich@400_26
           	recPtr = MemHandleLock(recH);
           	if (recCount > 1) // packaged font file? get the name
           	{
           	   int len = xstrlen((char *)recPtr)+1;
           	   if (len < 4) // bad font file?
           	   {
                  MemHandleUnlock(recH);
                  //DmReleaseRecord(dmRef, j, false); guich@402_51
                  DmCloseDatabase(dmRef);
                  break;
           	   }
               len = (int)((len+3) / 4)*4; // 4 byte align
           	   dbAlias = (char *)recPtr;
           	   recPtr += len; // point to the font structure
           	}
        	   if (*(dbAlias+1) == 'S' && *(dbAlias+2) == 'W')
        	      dbAlias++;
            specialIndex = -1;
           	if (!StrCompare("SWSmall", dbAlias))
           	   specialIndex = SWSTANDARD_FONT; else
           	if (!StrCompare("SWSmallBold", dbAlias))
           	   specialIndex = SWSTANDARD_BOLD_FONT;
           	if (!StrCompare("SWLarge", dbAlias))
           	   specialIndex = SWLARGE_FONT; else
           	if (!StrCompare("SWLargeBold", dbAlias))
           	   specialIndex = SWLARGE_BOLD_FONT;
				idx = (specialIndex!=-1) ? specialIndex : (internal + index++);
      	   storeUserFont(&vmGlobals->allFonts[idx], (FontPtr)recPtr, dbAlias, recH, dmRef, j, idx);
      	}
      	i += recCount;
		}
		first = false;
	}
	// get the unicode ones
	first = true;
	index += internal;
	while (1)
	{
		err = DmGetNextDatabaseByTypeCreator(first, &state, typeu, cridu, false, &cardNo, &dbID);
		if (err == dmErrCantFind)
			break;
	   DmDatabaseInfo(cardNo, dbID, (char*)&vmGlobals->allFonts[index].name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		dmRef = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
		loadUff(dmRef,vmGlobals->allFonts+index++);
		first = false;
	}
	GR_EXIT
   return 1;
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Font static void fontCreate(Font obj)
void FontCreate(Var stack[]) // (Lwaba/fx/Font;)V
{
	WObject obj = stack[0].obj; // class object
   int bold = WOBJ_FontStyle(obj) == Font_BOLD;
   int large = WOBJ_FontSize(obj) >= 14;
	UtfString fontName;
	int index;
   char *c;

   //static char *sufixes[4] = {"Small","SmallBold","Large","LargeBold"};
   char *sufix = 0;
   GR_ENTER
   fontName = (*vmGlobals->stringToUtf)(WOBJ_FontName(obj), STU_USE_STATIC | STU_NULL_TERMINATE);
   c = xstrstr(fontName.str,"Small");
   if (c != NULL && c != fontName.str)
      *c = 0;
   c = xstrstr(fontName.str,"Large");
   if (c != NULL && c != fontName.str)
      *c = 0;

   if (fontName.len > 3 && (fontName.str[1] == fontName.str[2]) && (fontName.str[0] == 'U' || fontName.str[1] == 'u') && (fontName.str[1] == 'F' || fontName.str[1] == 'f')) // is it uff?
      sufix = (MaxScreenX==160)?"_L":"_H";
   else
      switch (large*2+bold)
      {
         case 0: sufix = "Small";     break;
         case 1: sufix = "SmallBold"; break;
         case 2: sufix = "Large";     break;
         case 3: sufix = "LargeBold"; break;
      }
   // first, try to load the font itself using the sw font pattern
   index = getUserFontIndex(fontName.str,sufix);
   if (index == -1)
   {
      // ok, the pattern didnt work, use the font name given
      if (!(xstrlen(fontName.str) > 2 && fontName.str[0] == 'S' && fontName.str[1] == 'W')) // is it the internal ones? dont try the name itself
         index = getUserFontIndex(fontName.str,NULL);
      if (index == -1)
      {
         // ok, the font is not found, use the internal ones
         index = getUserFontIndex(/*(MaxScreenX>=320)?(vmGlobals->isSony?"HSW":"5SW"):(MaxScreenX>=240)?"MSW":"LSW"*/"SW",sufix);
         if (index == -1)
            index = bold?1:0; // internal ones not found (maybe large in palmOS 2.0), use the default small font.
         // change the font name
         WOBJ_FontName(obj) = (*vmGlobals->createString)(vmGlobals->allFonts[index].name);
      }
   }
   WOBJ_FontUserFont(obj) = vmGlobals->allFonts+index;
   GR_EXIT
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static void fontMetricsCreate(FontMetrics obj)
void FontMetricsCreate(Var stack[]) // (Lwaba/fx/FontMetrics;)V
{
	WObject font, fontMetrics; // guich@450_36: surface is ignored from now on
	struct UserFont *uf;

	fontMetrics = stack[0].obj;
	font = WOBJ_FontMetricsFont(fontMetrics);
	if (font == 0)
	{
		WOBJ_FontMetricsAscent(fontMetrics) = 0;
		WOBJ_FontMetricsDescent(fontMetrics) = 0;
		WOBJ_FontMetricsLeading(fontMetrics) = 0;
		return;
	}
	// surface is unused and user defined fonts are supported!
	uf = WOBJ_FontUserFont(font);
#ifdef PALMOS5 // guich@401_23: we need to double values in P5
   if (!uf->recH && vmGlobals->isHighDensity && MaxScreenX != 160) // guich@402_1: fixed problem for lowres devices
   {
   	WOBJ_FontMetricsAscent(fontMetrics) = uf->fontP->ascent<<1;
   	WOBJ_FontMetricsDescent(fontMetrics) = uf->fontP->descent<<1;
   	WOBJ_FontMetricsLeading(fontMetrics) = uf->fontP->leading<<1;
   }
   else
#endif
   {
   	WOBJ_FontMetricsAscent(fontMetrics) = uf->fontP->ascent;
   	WOBJ_FontMetricsDescent(fontMetrics) = uf->fontP->descent;
   	WOBJ_FontMetricsLeading(fontMetrics) = uf->fontP->leading;
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetCharWidth(FontMetrics obj, char c)
void FontMetricsGetCharWidth(Var stack[]) // (Lwaba/fx/FontMetrics;C)I
{
	FontMetricsGetWidth(FM_CHARWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, String s)
void FontMetricsGetTextWidth_String(Var stack[]) // (Lwaba/fx/FontMetrics;Ljava/lang/String;)I
{
	FontMetricsGetWidth(FM_STRINGWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/FontMetrics static int fontMetricsGetTextWidth(FontMetrics obj, char chars[], int start, int count)
void FontMetricsGetTextWidth(Var stack[]) // (Lwaba/fx/FontMetrics;[CII)I
{
	FontMetricsGetWidth(FM_CHARARRAYWIDTH, stack);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Sound static void soundBeep()
void SoundBeep(Var stack[]) // ()V
{
	SemaphoreRelease(1);
	SndPlaySystemSound(sndInfo);
	SemaphoreReserve(1);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Sound static void soundTone(int freq, int duration)
void SoundTone(Var stack[]) // (II)V
{
   (*vmGlobals->tone)(stack[0].intValue,stack[1].intValue);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Sound static void soundSetEnabled(boolean on)
void SoundSetEnabled(Var stack[]) // (Z)V
{
	UInt16 amp = stack[0].intValue ? vmGlobals->defSoundAmp : 0; // if turn on, set it to the default user selection
   SndSetDefaultVolume (NULL, &amp, NULL);
}
//////////////////////////////////////////////////////////////////////////
#define WOBJ_SoundClipPath(o) (o)[1].obj
#define WOBJ_SoundClipLoaded(o) (o)[2].intValue
// waba/fx/SoundClip static boolean soundClipPlay(SoundClip obj)
void SoundClipPlay(Var stack[]) // (Lwaba/fx/SoundClip;)Z
{
#ifdef sysTrapSndPlayResource	// compiling in cw8 gives error
	WObject sound, pathString;
	UtfString path;
	byte *wav;
	byte *rec;
	int32 v=0;
	// Until the sampled sound feature exists, check the trap address instead.
	if (SysGetTrapAddress(sysTrapSndPlayResource) == SysGetTrapAddress(sysTrapSysUnimplemented))
		RETURN32(0)
	sound = stack[0].obj;
	pathString = WOBJ_SoundClipPath(sound);
	path = (*vmGlobals->stringToUtf)(pathString, STU_NULL_TERMINATE | STU_USE_STATIC);
	rec = (*vmGlobals->lockWarpRec)(path.str, path.len, NULL);
	if (rec != NULL)
	{
	   wav = rec;
	   if ((int32)wav & 1) wav++; // guich@421_76: warp now aligns it for us.
	   SemaphoreRelease(1); // guich@400_11
	   v = SndPlayResource(wav, sndGameVolume, sndFlagSync);  // must use sync because we're going to free the buffer right away
	   SemaphoreReserve(1);
	   (*vmGlobals->unlockWarpRec)(rec, path.len);
	}
	RETURN32(v)
#endif
}
void SoundClipFree(Var stack[]) // (Lwaba/fx/SoundClip;)V
{
}

