/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

 // defines port numbers
#define SerialPort_DEFAULT 0x0000
#define SerialPort_IRCOMM  0x1000
#define SerialPort_SIR     0x1001
#define SerialPort_USB     0x1002
#define SerialPort_BLUETOOTH 0x1003

// Defines used to Enable Hardware IrDA on old palm devices
#define M328BASE        0xFFFFF000
#define M328_UARTMISC   (M328BASE+0x908)
#define M328_UARTSTCNT  (M328BASE+0x900)
#define UART_IRDA_ENABLE   0x0020
#define UART_IRDA_DISABLE  0xFFDF
#define UART_POLARITY_INVERTED  0x000C
#define UART_POLARITY_NORMAL    0xFFF3
#define UART_RX_ENABLED       0x4000
#define UART_RX_DISABLED      0xBFFF
#define UART_TX_ENABLED       0x2000
#define UART_TX_DISABLED      0xDFFF

#define NET_NOT_READY_FOR_OPEN 0
#define NET_READY_FOR_OPEN 1
#define NET_IS_OPEN 2
#define NET_OPEN_FAILED 3

static byte hsKeybEnabled=1; // guich@320_23: avoid calling it if already open/closed

static int32 _SerialPortClose(WObject port)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   UInt refNum;
   // Hardware Constants
   UInt16 *UART_MISC_REG = (UInt16 *) M328_UARTMISC;
   UInt32 ftrHSVersion;
   int32 portNumber = WOBJ_SerialPortNumber(port);
   int directHardware = WOBJ_SerialPortDirectHardware(port);
   int ret;

   // dodah12@200b4: If on Handspring device, enable the keyboard
   // thread before closing the serial library.
   if (!FtrGet ('hsEx', 0, &ftrHSVersion) && !hsKeybEnabled)
   {
      HsExtKeyboardEnable (true);
      hsKeybEnabled = 1;
   }

   if (WOBJ_SerialPortRefNum(port) == -1)
      return 0;

   refNum = (UInt)WOBJ_SerialPortRefNum(port);

   WOBJ_SerialPortRefNum(port) = -1;
   // check for presence of new serial mgr
   if (WOBJ_SerialPortNewSerialMgr(port))
   {
      UNLOCK_MEMORY; // guich@401_39
      // new serial mgr
      SrmSendWait(refNum);
      // SIR Port
      if (portNumber == SerialPort_SIR )
         SrmControl(refNum, srmCtlIrDADisable, 0, 0);
      SrmSetReceiveBuffer(refNum, NULL, 0); // guich@200b4_184
      ret = SrmClose(refNum) == 0;
      LOCK_MEMORY;
      return ret;
   }
   else
   {
      UNLOCK_MEMORY;
      // old serial mgr
      SerSendWait(refNum, -1); // flush buffer

      // SIR Port
      if (portNumber == SerialPort_SIR )
      {
         if (directHardware)
         {
            // Direct HARDWARE Access
            // try to disable irda support via hardware
            *UART_MISC_REG &= UART_IRDA_DISABLE & UART_POLARITY_NORMAL;
         }
         else
         {
            // control call
            SerControl( refNum, serCtlIrDADisable, 0, 0);
         }
      }
      SerSetReceiveBuffer(refNum, NULL, 0); // guich@200b4_184: try to fix close problems in serial
      ret = SerClose(refNum) == 0;
      LOCK_MEMORY;
      return ret;
   }
#endif
}


// guich@120
static Var SerialPortCreateIt(WObject port, int32 number, int32 baudRate, int32 bits, int32 parity, int32 stopBits, int32 timeOut)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   SerSettingsType settings;
   UInt refNum;
   Var v;
#ifndef PALMOS5
   UInt32 ftrHSVersion;
#endif

   // NewSerialMgr present
   UInt32 value;
   // Rom Version
   UInt32 romVersion;
   // Hardware Constants
   UInt16 *UART_MISC_REG = (UInt16 *) M328_UARTMISC;
   // err
   Err   err;

   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
   v.obj = 0;
   WOBJ_SerialPortLastErr(port) = 0;

   // dodah12@200b4: If on Handspring device, disable the keyboard
   // thread before opening the serial library.
#ifndef PALMOS5
   if (!FtrGet ('hsEx', 0, &ftrHSVersion) && hsKeybEnabled)
   {
      HsExtKeyboardEnable (false);
      hsKeybEnabled = 0;
   }
#endif

   WOBJ_SerialPortRefNum(port) = -1;
   WOBJ_SerialPortNumber(port) = number;

   // default baud rate
   if (baudRate == 0)
      baudRate = 9600;

   // check for presence of new serial mgr
   if ( FtrGet(sysFileCSerialMgr, sysFtrNewSerialPresent, &value) != NULL
     || (romVersion < 0x04100000 && number == SerialPort_SIR)) // guich@241_19: since of a bug in PalmOS < 4.1 new Serial Manager for SIR, we avoid using it (change requested by Matthias Ringwald and Teddy Roerby)
   {
      // not present
      WOBJ_SerialPortNewSerialMgr(port) = false;
      switch (number)
      {
         case SerialPort_IRCOMM:
            if (SysLibFind("SerIrCOMM Lib", &refNum) != 0)
               if (SysLibLoad('libr', 'ircm', &refNum) != 0)
               {
                  WOBJ_SerialPortLastErr(port) = -11;
                  return v;
               }

            if ((err=SerOpen(refNum, 0, baudRate)) != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         case SerialPort_SIR:
            if (SysLibFind("Serial Library", &refNum) != 0)
            {
               WOBJ_SerialPortLastErr(port) = -11;
               return v;
            }
            if ((err=SerOpen(refNum, 0, baudRate)) != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }

            // check for rom version >= OS 3.0
            if (romVersion >= 0x03000000)
            {
               // Control Call
               WOBJ_SerialPortDirectHardware(port) = false;
               if ((err=SerControl( refNum, serCtlIrDAEnable, 0, 0)) != 0)
               {
                  WOBJ_SerialPortLastErr(port) = err;
                  return v;
               }
            }
            else
            {
               // Direct HARDWARE Access
               // try to enable irda support via hardware
               WOBJ_SerialPortDirectHardware(port) = true;
               *UART_MISC_REG = *UART_MISC_REG | UART_IRDA_ENABLE | UART_POLARITY_INVERTED;
            }
            break;

         case SerialPort_USB:
            if (SysLibFind("USB Library", &refNum) != 0)
            {
               WOBJ_SerialPortLastErr(port) = -11;
               return v;
            }
            if ((err=SerOpen(refNum, usbLibPortHotsync, baudRate)) != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         default:
            if (SysLibFind("Serial Library", &refNum) != 0)
            {
               WOBJ_SerialPortLastErr(port) = -11;
               return v;
            }
            if ((err=SerOpen(refNum, 0, baudRate)) != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;
      }

   }
   else
   {
      // yes - use new serial mgr routines
      // no
      WOBJ_SerialPortNewSerialMgr(port) = true;

      switch (number)
      {

         // SIR
         case SerialPort_SIR:
            // use port serPortIrPort and call IrDA Enable
            UNLOCK_MEMORY; // guich@400_27: released semaphore in all ports to avoid reset in Palm OS 5.
            err=SrmOpen( serPortIrPort, baudRate, &refNum);
            LOCK_MEMORY;
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            UNLOCK_MEMORY;
            err=SrmControl(refNum, srmCtlIrDAEnable, 0, 0 );
            LOCK_MEMORY;
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         // IrCOMM
         case SerialPort_IRCOMM:
            // use port 'ircm'
            UNLOCK_MEMORY;
				err=SrmOpen( 'ircm', baudRate, &refNum);
            LOCK_MEMORY;
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         // USB
         case SerialPort_USB:
            // use port usbLibSrmDriverIDDesktop
            UNLOCK_MEMORY;
            err=SrmOpen( usbLibSrmDriverIDDesktop, baudRate, &refNum);
            LOCK_MEMORY;
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;
         // serial port
         case SerialPort_DEFAULT:
            // use port serPortCradlePort
            UNLOCK_MEMORY;
            err=SrmOpen( serPortCradlePort, baudRate, &refNum);
            LOCK_MEMORY;
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         // guich@330_34: bluetooth
         case SerialPort_BLUETOOTH:
            UNLOCK_MEMORY;
            err=SrmOpen( 'rfcm', baudRate, &refNum); // a dialog appears here
            LOCK_MEMORY;
            // use port specified by user
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;

         // andrzejjan@330_34: Separated the SerialPort_DEFAULT and default cases to support Bluetooth and other ports
         default:
            UNLOCK_MEMORY;
            err=SrmOpen( number, baudRate, &refNum);
            LOCK_MEMORY;
            // use port specified by user
            if (err != 0)
            {
               WOBJ_SerialPortLastErr(port) = err;
               return v;
            }
            break;
      }
   }
   settings.baudRate = baudRate;
   settings.flags = 0;

   // no handshake via SIR
   if (number != SerialPort_SIR)
      settings.flags |= serSettingsFlagRTSAutoM | serSettingsFlagCTSAutoM;

   if (bits == 8)
      settings.flags |= serSettingsFlagBitsPerChar8;
   else if (bits == 7)
      settings.flags |= serSettingsFlagBitsPerChar7;
   else if (bits == 6)
      settings.flags |= serSettingsFlagBitsPerChar6;
   else if (bits == 5)
      settings.flags |= serSettingsFlagBitsPerChar5;
// if (parity != 0)
//    settings.flags |= serSettingsFlagParityEvenM;

   if (parity != 0)
   {
      settings.flags |= serSettingsFlagParityOnM;
      settings.flags |= (parity==1)?serSettingsFlagParityEvenM:serSettingsFlagXonXoffM;
   }

   if (stopBits == 1)
      settings.flags |= serSettingsFlagStopBits1;
   else if (stopBits == 2)
      settings.flags |= serSettingsFlagStopBits2;
      settings.ctsTimeout = timeOut;

   if (WOBJ_SerialPortNewSerialMgr(port))
   {
      // do settings
      UInt16   sizePtr = sizeof(Int32);
      UNLOCK_MEMORY; // guich@401_39
      err = SrmControl(refNum, srmCtlSetBaudRate,  &settings.baudRate, &sizePtr);
      if (err == 0)
         err = SrmControl(refNum, srmCtlSetFlags,  &settings.flags,  &sizePtr);
      if (err == 0)
         err = SrmControl(refNum, srmCtlSetCtsTimeout,&settings.ctsTimeout,   &sizePtr);
      LOCK_MEMORY;
      if (err)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SrmClose(refNum);
         return v;
      }
   }
   else
   {
      if ((err=SerSetSettings(refNum, &settings)) != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SerClose(refNum);
         return v;
      }
   }

   WOBJ_SerialPortRefNum(port) = (int32)refNum;
   WOBJ_SerialPortTimeout(port) = (*vmGlobals->millisToTicks)(100);
   return v;
#endif
}

void SerialPortDestroy(WObject port)
{
   _SerialPortClose(port);
}

static void SerialPortReadWriteBytes(Var stack[], int isRead)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
#ifdef PALMOS5 // guich@401_4: this code fixes problems in P5
   WObject port, byteArray;
   int32 start, count;
   byte *bytes,*bytes0,*temp;
   ULong numRW;
   UInt refNum;
   Err err=0;
   VoidHand tempH;
   // Hardware Constants
   UInt16 *UART_STCNT_REG = (UInt16 *) M328_UARTSTCNT;
   int32 timeout,portNumber,directHardware,newSerialMgr;

   port = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = err;
      RETURN32(-1) // port not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SerialPortLastErr(port) = -14;
      RETURN32(-1) // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   bytes = &bytes[start];
   refNum = (UInt)WOBJ_SerialPortRefNum(port);
   newSerialMgr = WOBJ_SerialPortNewSerialMgr(port);
   timeout = WOBJ_SerialPortTimeout(port);
   portNumber = WOBJ_SerialPortNumber(port);
   directHardware = WOBJ_SerialPortDirectHardware(port);

   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_SerialPortLastErr(port) = -99;
      RETURN32(-1) // no memory to allocate temp buffer
   }
continueReading:
   temp = bytes;
	bytes = MemHandleLock(tempH);
	bytes0 = bytes; // bytes will be incremented later
	if (!isRead)
   	xmemmove(bytes,temp,count); // copy our sending string to a internal buffer, because we're going to release the semaphore, and so, we can't access the user's buffer

   UNLOCK_MEMORY; // release it now, bc PalmOS 5 doesnt like the release/reserve/release/reserve/... approach

   if (newSerialMgr)
   {
      if (isRead)
         numRW = SrmReceive(refNum, (VoidPtr)bytes, (ULong)count, timeout, &err);
      else
      {
         if (portNumber == SerialPort_SIR)
         {
            // disable rx via srmcontrol
            SrmControl( refNum, srmCtlRxDisable, 0, 0);
         }
         numRW = SrmSend(refNum, (VoidPtr)bytes, (ULong)count, &err);
         if (portNumber == SerialPort_SIR)
         {
            // wait till tx finished
            SrmSendWait( refNum);
            // re-enable IRDA
            SrmControl(refNum, srmCtlIrDAEnable, 0, 0); // egr@330_54: Need this for using SIR (new Ser.Man.)
            // disable rx via srmcontrol
            SrmControl( refNum, srmCtlRxEnable, 0, 0);
         }
      }
      LOCK_MEMORY;
      if (err != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SrmClearErr(refNum);
      }
   }
   else
   {
      if (isRead)
         numRW = SerReceive(refNum, (VoidPtr)bytes, (ULong)count, timeout, &err);
      else
      {
         if (portNumber == SerialPort_SIR)
         {
            if ( directHardware )
            {
               *UART_STCNT_REG &= UART_RX_DISABLED;
            }
            else
            {
               SerControl(refNum, serCtlIrDAEnable, 0, 0);   // egr@330_54: Need this for using SIR (old Ser.Man.)
               SerControl(refNum, serCtlRxDisable, 0, 0);   // guich@330_26: changed serCtlIrDADisable to serCtlRxDisable
            }
         }
         numRW = SerSend(refNum, (VoidPtr)bytes, (ULong)count, &err);
         if (portNumber == SerialPort_SIR)
         {
            // wait till tx finished
            SerSendWait(refNum, -1);
            if ( directHardware )
            {
               *UART_STCNT_REG |= UART_RX_ENABLED;
            }
            else
            {
               SerControl(refNum, serCtlRxEnable, 0, 0); // guich@330_26: changed serCtlIrDAEnable to serCtlRxEnable
            }
         }
      }
      LOCK_MEMORY;
      if (err != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SerClearErr(refNum);
      }
   }
   xmemmove(temp,bytes0,numRW);
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   RETURN32(numRW)

#else
   WObject port, byteArray;
   int32 start, count;
   uchar *bytes;
   ULong numRW;
   UInt refNum;
   Err err=0;
   // Hardware Constants
   UInt16 *UART_STCNT_REG = (UInt16 *) M328_UARTSTCNT;

   port = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = err;
      RETURN32(-1) // port not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SerialPortLastErr(port) = -14;
      RETURN32(-1) // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   refNum = (UInt)WOBJ_SerialPortRefNum(port);

   if (WOBJ_SerialPortNewSerialMgr(port))
   {
      if (isRead)
      {
         numRW = SrmReceive(refNum, (VoidPtr)&bytes[start], (ULong)count,
         WOBJ_SerialPortTimeout(port), &err);
      }
      else
      {
         if (WOBJ_SerialPortNumber(port) == SerialPort_SIR)
         {
            // disable rx via srmcontrol
            SrmControl( refNum, srmCtlRxDisable, 0, 0);
         }
         numRW = SrmSend(refNum, (VoidPtr)&bytes[start], (ULong)count, &err);
         if (WOBJ_SerialPortNumber(port) == SerialPort_SIR)
         {
            // wait till tx finished
            SrmSendWait( refNum);
            // re-enable IRDA
            SrmControl(refNum, srmCtlIrDAEnable, 0, 0); // egr@330_54: Need this for using SIR (new Ser.Man.)
            // disable rx via srmcontrol
            SrmControl( refNum, srmCtlRxEnable, 0, 0);
         }
      }
      if (err != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SrmClearErr(refNum);
      }
   }
   else
   {
      if (isRead)
      {
         numRW = SerReceive(refNum, (VoidPtr)&bytes[start], (ULong)count,
         WOBJ_SerialPortTimeout(port), &err);
      }
      else
      {
         if (WOBJ_SerialPortNumber(port) == SerialPort_SIR)
         {
            if ( WOBJ_SerialPortDirectHardware(port) )
            {
               *UART_STCNT_REG &= UART_RX_DISABLED;
            }
            else
            {
               SerControl(refNum, serCtlIrDAEnable, 0, 0);   // egr@330_54: Need this for using SIR (old Ser.Man.)
               SerControl(refNum, serCtlRxDisable, 0, 0);   // guich@330_26: changed serCtlIrDADisable to serCtlRxDisable
            }
         }
         numRW = SerSend(refNum, (VoidPtr)&bytes[start], (ULong)count, &err);
         if (WOBJ_SerialPortNumber(port) == SerialPort_SIR)
         {
            // wait till tx finished
            SerSendWait(refNum, -1);
            if ( WOBJ_SerialPortDirectHardware(port) )
            {
               *UART_STCNT_REG |= UART_RX_ENABLED;
            }
            else
            {
               SerControl(refNum, serCtlRxEnable, 0, 0); // guich@330_26: changed serCtlIrDAEnable to serCtlRxEnable
            }
         }
      }
      if (err != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         SerClearErr(refNum);
      }
   }
   RETURN32(numRW)
#endif // PALMOS5
#endif
}
static int32 _SocketClose(WObject socket)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   NetSocketRef socketRef;
   SDWord timeout;
   Err err;
   uint16 refNum = vmGlobals->globalSocketLibRefNum;
   int32 ok;

   socketRef = (NetSocketRef)WOBJ_SocketRef(socket);
   if (socketRef <= 0) // guich@330_39: socketref can be 0 when using SerialSocket
      return 0;
   WOBJ_SocketRef(socket) = -1;
   timeout = (*vmGlobals->millisToTicks)(1500);

   UNLOCK_MEMORY; // guich@401_10
   ok = NetLibSocketClose(refNum, socketRef, timeout, &err) == 0;
   LOCK_MEMORY;
   return ok;
#endif
}
void SocketDestroy(WObject socket)
{
   _SocketClose(socket);
}

static int32 getLocalIP()
{
   Word index=0, instance, len=sizeof(DWord);
   DWord creator, ip;
   while (NetLibIFGet(vmGlobals->globalSocketLibRefNum, index++, &creator, &instance) != netErrInvalidInterface)
   {
      if (creator==netIFCreatorPPP)
      {
         if (NetLibIFSettingGet(vmGlobals->globalSocketLibRefNum, creator, instance, netIFSettingActualIPAddr, &ip, &len) == 0)
            return ip;
      }
   }
   return 0;
}

static void CatalogReadWriteBytes(Var stack[], int isRead)
{
   WObject cat, byteArray;
   int32 start, count, pos, recOffset;
   uchar *bytes, *recBytes;

   cat = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      RETURN32(-1) // no current record
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1) // array null or range invalid
   recBytes = WOBJ_CatalogCurRecPtr(cat);
   recOffset = WOBJ_CatalogCurRecOffset(cat);
   if (recOffset + count > WOBJ_CatalogCurRecLen(cat))
      RETURN32(-1) // past end of record
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
      xmemmove(&bytes[start], &recBytes[recOffset], count);
   else
   {
      if (DmWrite(recBytes, recOffset, &bytes[start], count) != 0)
         RETURN32(-1)
      WOBJ_CatalogCurRecModified(cat) = 1;
   }
   WOBJ_CatalogCurRecOffset(cat) = recOffset + count;
   RETURN32(count)
}
static void _RecClose(WObject cat)
{
   DmOpenRef dmRef;
   VoidHand recH;
   int32 pos;
   Boolean dirty;

   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      return; // no current record
   recH = WOBJ_CatalogCurRecHandle(cat);
   MemHandleUnlock(recH);
   dmRef = WOBJ_CatalogDmRef(cat);
   if (WOBJ_CatalogCurRecModified(cat) != 0)
   {
      dirty = true;
      WOBJ_CatalogSetBackup(cat) = true; // guich@200
   }
   else
      dirty = false;
   DmReleaseRecord(dmRef, pos, dirty);
   WOBJ_CatalogCurRecPos(cat) = -1;
}

static int32 _CatalogClose(WObject cat)
{
   DmOpenRef dmRef;

   _RecClose(cat); // release any open records
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      return 0;
   // guich@200 - sets the backup bit if the file was modified
   if (WOBJ_CatalogSetBackup(cat))
   {
      LocalID dbID;
      UInt cardNo;
      Word attributes;

      DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, &cardNo, NULL);
      DmDatabaseInfo(cardNo, dbID, NULL, &attributes,NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      attributes |= dmHdrAttrBackup;
      DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes,NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   }
   if (dmRef != vmGlobals->dmRefSystemOut) // guich@200b4_19: only close if is not the memo db
      DmCloseDatabase(dmRef); // guich@200b2: removed from before the last block to correct the "address error" from pose
   WOBJ_CatalogDmRef(cat) = 0;
   return 1;
}

void CatalogDestroy(WObject cat)
{
   _CatalogClose(cat);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static void serialPortCreate(SerialPort obj, int number, int baudRate, int bits, int parity, int stopBits)
void SerialPortCreate(Var stack[]) // (Lwaba/io/SerialPort;IIIZI)V
{
   SerialPortCreateIt(stack[0].obj, stack[1].intValue,  stack[2].intValue,  stack[3].intValue,  stack[4].intValue,  stack[5].intValue, 2 * sysTicksPerSecond);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortClose(SerialPort obj)
void SerialPortClose(Var stack[]) // (Lwaba/io/SerialPort;)Z
{
   WObject port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;
   RETURN32(_SerialPortClose(port))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortIsOpen(SerialPort obj)
void SerialPortIsOpen(Var stack[]) // (Lwaba/io/SerialPort;)Z
{
   WObject port = stack[0].obj;
   RETURN32((WOBJ_SerialPortRefNum(port) == -1)?0:1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortSetFlowControl(SerialPort obj, boolean on)
void SerialPortSetFlowControl(Var stack[]) // (Lwaba/io/SerialPort;Z)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   WObject port;
   int32 flowOn;
   UInt refNum;
   SerSettingsType settings;
   UInt16 sizePtr;
   int32 v=0;

   port = stack[0].obj;
   flowOn = stack[1].intValue;

   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = -10;
      RETURN32(0)
   }
   refNum = (UInt)WOBJ_SerialPortRefNum(port);

   if (WOBJ_SerialPortNumber(port) != SerialPort_SIR)
   {
      if (WOBJ_SerialPortNewSerialMgr(port))
      {
         UNLOCK_MEMORY; // guich@401_39
         sizePtr = sizeof(Int32);
         if (SrmControl(refNum, srmCtlGetFlags, &settings.flags, &sizePtr) != 0)
            goto err;

         if (flowOn == 1)
            settings.flags |= srmSettingsFlagRTSAutoM | srmSettingsFlagCTSAutoM;
         else
            settings.flags &= 0xffff - (srmSettingsFlagRTSAutoM | srmSettingsFlagCTSAutoM);

         if ( SrmControl(refNum, srmCtlSetFlags, &settings.flags, &sizePtr) != 0)
            goto err;
      }
      else
      {
         UNLOCK_MEMORY; // guich@401_39
         if (SerGetSettings(refNum, &settings) != 0)
            goto err;

         if (flowOn == 1)
            settings.flags |= serSettingsFlagRTSAutoM | serSettingsFlagCTSAutoM;
         else
            settings.flags &= 0xffff - (serSettingsFlagRTSAutoM |
         serSettingsFlagCTSAutoM);

         if (SerSetSettings(refNum, &settings) != 0)
            goto err;
      }
   }
   v = 1;
err:
   LOCK_MEMORY;
   RETURN32(v)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortSetReadTimeout(SerialPort obj, int millis)
void SerialPortSetReadTimeout(Var stack[]) // (Lwaba/io/SerialPort;I)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   WObject port;
   int32 millis;

   port = stack[0].obj;
   millis = stack[1].intValue;
   WOBJ_SerialPortLastErr(port) = 0;

   if (millis < 0 || WOBJ_SerialPortRefNum(port) == -1)
      RETURN32(0)
   WOBJ_SerialPortTimeout(port) = (*vmGlobals->millisToTicks)(millis);
   RETURN32(1)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortReadBytes(SerialPort obj, byte buf[], int start, int count)
void SerialPortReadBytes(Var stack[]) // (Lwaba/io/SerialPort;[BII)I
{
   SerialPortReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortReadCheck(SerialPort obj)
void SerialPortReadCheck(Var stack[]) // (Lwaba/io/SerialPort;)I
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   WObject port;
   UInt refNum;
   ULong numBytes;
   Err err;

   port = stack[0].obj;
   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = -10;
      RETURN32(-1) // port not open
   }
   refNum = (UInt)WOBJ_SerialPortRefNum(port);
   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortNewSerialMgr(port))
   {
      UNLOCK_MEMORY; // guich@401_39
      err=SrmReceiveCheck(refNum, &numBytes);
      LOCK_MEMORY;
      if (err != 0)  // terryf@200b4_179: changed SrmSendCheck to SrmReceiveCheck
      {
         WOBJ_SerialPortLastErr(port) = err;
         RETURN32(-1)
      }
   }
   else
   {
      UNLOCK_MEMORY; // guich@401_39
      SerClearErr(refNum); // guich@200b4_175
      err = SerReceiveCheck(refNum, &numBytes);
      LOCK_MEMORY;
      if (err != 0)
      {
         WOBJ_SerialPortLastErr(port) = err;
         RETURN32(-1)
      }
   }
   RETURN32(numBytes)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortWriteBytes(SerialPort obj, byte buf[], int start, int count)
void SerialPortWriteBytes(Var stack[]) // (Lwaba/io/SerialPort;[BII)I
{
   SerialPortReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
static void SocketReadWriteBytes(Var stack[], int isRead)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
#ifdef PALMOS5 // guich@400_49: in PalmOS 5, we need to keep the semaphore released all the time. For this, we need to create a buffer that will receive the bytes, and then copy this to the output.
   WObject socket, byteArray;
   NetSocketRef socketRef;
   int32 millis, start, count, countSoFar, n;
   SDWord timeout;
   byte *bytes,*bytes0,*temp;
   Err err;
   uint16 refNum = vmGlobals->globalSocketLibRefNum;
   VoidHand tempH;
   byte refreshBeforeEachRead;
   int32 v = -1;

   socket = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   socketRef = (NetSocketRef)WOBJ_SocketRef(socket);
   WOBJ_SocketLastErr(socket) = 0;
   if (socketRef == -1)
   {
      WOBJ_SocketLastErr(socket) = -13;
      RETURN32(-1) // socket not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SocketLastErr(socket) = -14;
      RETURN32(-1) // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   bytes = &bytes[start];
   countSoFar = 0;
   millis = WOBJ_SocketTimeout(socket);
   timeout = (*vmGlobals->millisToTicks)(millis);
   refreshBeforeEachRead = WOBJ_SocketRefreshBeforeEachRead(socket);

   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_SocketLastErr(socket) = -99;
      RETURN32(-1) // no memory to allocate temp buffer
   }
continueReading:

   temp = bytes;
	bytes = MemHandleLock(tempH);
	bytes0 = bytes; // bytes will be incremented later
	if (!isRead)
   	xmemmove(bytes,temp,count); // copy our sending string to a internal buffer, because we're going to release the semaphore, and so, we can't access the user's buffer

   UNLOCK_MEMORY; // release it now, bc PalmOS 5 doesnt like the release/reserve/release/reserve/... approach
   // guich@330_14
   if (refreshBeforeEachRead)
   {
      Word temp=0;
      byte allUp;
      err=NetLibConnectionRefresh(refNum, true, &allUp, &temp);
   }

   while (countSoFar < count)
   {
      int chunkSize;
      chunkSize = count - countSoFar;
      if (chunkSize > 0x7000)
         chunkSize = 0x7000;
      if (isRead)
      {
         n = NetLibReceive(refNum, socketRef, bytes, chunkSize, 0, 0, 0, timeout, &err);
         if (chunkSize > 0 && (n == 0 || err == netErrSocketNotOpen))
         {
            LOCK_MEMORY;
            //(*vmGlobals->debug)("error, closing socket");
            _SocketClose(socket); // server closed connection
            UNLOCK_MEMORY;
         }
      }
      else
         n = NetLibSend(refNum, socketRef, bytes,  chunkSize, 0, NULL, 0, timeout, &err);

      if (n > 0)
      {
         countSoFar += n;
         bytes += n;
      }
      else if (n <= 0)
      {
         LOCK_MEMORY;
         WOBJ_SocketLastErr(socket) = err;
         // NOTE: n == 0 under PalmOS when server closes connection
         if (countSoFar != 0)
            v = countSoFar;
         else if (n < 0 && (err == netErrTimeout || err == netErrWouldBlock)) // pierre@401_2
            v = 0;
         else
            v = -1;
         goto end;
      }
   }
   v = count;
   LOCK_MEMORY;
end:
   xmemmove(temp,bytes0,countSoFar);
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   RETURN32(v)


#else


   WObject socket, byteArray;
   NetSocketRef socketRef;
   int32 millis, start, count, countSoFar, n;
   SDWord timeout;
   uchar *bytes;
   Err err;
   int32 v;
   uint16 refNum = vmGlobals->globalSocketLibRefNum;

   v = -1;
   socket = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   socketRef = (NetSocketRef)WOBJ_SocketRef(socket);
   WOBJ_SocketLastErr(socket) = 0;
   if (socketRef == -1)
   {
      WOBJ_SocketLastErr(socket) = -13;
      RETURN32(-1) // socket not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SocketLastErr(socket) = -14;
      RETURN32(-1) // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   bytes = &bytes[start];
   countSoFar = 0;
   millis = WOBJ_SocketTimeout(socket);
   timeout = (*vmGlobals->millisToTicks)(millis);
   // guich@330_14
   if (WOBJ_SocketRefreshBeforeEachRead(socket))
   {
      Word temp=0;
      byte allUp;
      UNLOCK_MEMORY;
      err=NetLibConnectionRefresh(refNum, true, &allUp, &temp);
      LOCK_MEMORY;
   }

   while (countSoFar < count)
   {
      int chunkSize;
      chunkSize = count - countSoFar;
      if (chunkSize > 0x7000)
         chunkSize = 0x7000;
      if (isRead)
      {
         n = NetLibReceive(refNum, socketRef, bytes, chunkSize, 0, 0, 0, timeout, &err);
         if (chunkSize > 0 && (n == 0 || err == netErrSocketNotOpen))
            _SocketClose(socket); // server closed connection
      }
      else
         n = NetLibSend(refNum, socketRef, bytes,  chunkSize, 0, NULL, 0, timeout, &err);

      if (n > 0)
      {
         countSoFar += n;
         bytes += n;
      }
      else if (n <= 0)
      {
         WOBJ_SocketLastErr(socket) = err;
         // NOTE: n == 0 under PalmOS when server closes connection
         if (countSoFar != 0)
            v = countSoFar;
         else if (n < 0 && (err == netErrTimeout || err == netErrWouldBlock)) // pierre@401_2
            v = 0;
         else
            v = -1;

         RETURN32(v)
      }
   }
   RETURN32(count)
#endif // PalmOS5
#endif // debug
}

/* This is a working version of sockets. It runs in TT and in Kyocera.
   If you hear a beep, the send/receive worked; otherwise, if silence, it failed.

static void testSock()
{
   Err myErr, myIFErr;
   int sockfd;
   char recdata[256];
   UInt netrefnum;
   Word AppNetRefnum;
   NetHostInfoBufType hostInfo;
   NetHostInfoPtr hostPtr;
   Err errno;
   NetSocketAddrINType sockAddr;
   uint16 timeout = sysTicksPerSecond * 100;

   UNLOCK_MEMORY;
   myErr = SysLibFind("Net.lib", &netrefnum);
   if (myErr)
     return;
   AppNetRefnum = (Word)netrefnum;

   myErr = NetLibOpen(AppNetRefnum, &myIFErr);
   if ((myErr && myErr != netErrAlreadyOpen) || myIFErr)
   {
     if (!myErr) NetLibClose(AppNetRefnum, 0);
     AppNetRefnum = 0;
     return;
   }

   hostPtr = NetLibGetHostByName(AppNetRefnum,"www.ig.com.br",&hostInfo, timeout, &errno);
   sockAddr.family = netSocketAddrINET;
   sockAddr.port = NetHToNS(80);
   sockAddr.addr = *((NetIPAddr *)hostInfo.hostInfo.addrListP[0]);
   sockfd = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, timeout, &errno);
   myErr = NetLibSocketConnect(AppNetRefnum, sockfd, (void *)&sockAddr, sizeof(sockAddr), timeout, &errno);
   myErr = NetLibSend(AppNetRefnum, sockfd, "GET / HTTP/1.0\n\n", 16, 0,0,0, timeout, &errno);
   if (myErr == 16 && NetLibReceive(AppNetRefnum, sockfd, recdata, 32, 0, 0, 0, timeout, &errno) == 32 && errno == 0)
   {
   	SndCommandType sndCmd;
   	sndCmd.cmd = sndCmdFreqDurationAmp;
   	sndCmd.param1 = 1500; // freq
   	sndCmd.param2 = 500; // duration
   	sndCmd.param3 = sndMaxAmp;
   	SndDoCmd(NULL, &sndCmd, 0);
   }
   myErr = NetLibSocketClose(AppNetRefnum, sockfd, timeout,&errno);
   myErr = NetLibClose(AppNetRefnum, 1);
   LOCK_MEMORY;
}*/

static Err openNetLib()
{
#if 1 //@t ndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   DWord version=0;
   Word refNum=0, ifErr=0;

   Err err = FtrGet(netFtrCreator, netFtrNumVersion, &version);
   // Find the net library
   if (err == 0 )
      err = SysLibFind("Net.lib", &refNum);

   // This section rewritten by vik@asi.org to suddenly close NetLib
   // in the event of a failure, and to use the system NetLibOpenCount
   // to determine if the Net library is already opened. This is done
   // because the old method used its own state variable, which was not
   // cleared when the PDA powered down and the Net library closed
   // automagically or the session closed itself by timeout
   // Note: I don't think anyone actually closes it...

   if (!err)
   {
      // guich@300_48: local variables were not being initialized correctly, leading to various errors
      Boolean allUp=0;      // vik@200b4r8_168 Moved allUp to widen its scope.
      // If the library isn't open, open it.
      Word OpenCount=0;
      NetLibOpenCount(refNum,&OpenCount);
      UNLOCK_MEMORY; // guich@300_22
      if (OpenCount==0)
      {
         err = NetLibOpen(refNum, &ifErr);
         allUp = ifErr == 0; // guich@300_48: fix allUp when openCount = 0
      }
      else
      {
         // The net library is already open. Refresh any down connections.
         // vik@200b4r8_168 Moved allUp from here to widen its scope.
         Word dummy=0;
         err=NetLibConnectionRefresh(refNum, true, &allUp, &dummy);
      }

      // If we managed this, flag it as open.
      // vik@200b4r8_168 Fail the open if all interfaces don't come up.
      //  This forces a close of the netlib later.
      if ((err == 0 || err == netErrAlreadyOpen) && allUp) // guich@400_25: if netErrAlreadyOpen, we can continue
      {
         vmGlobals->globalSocketLibRefNum = refNum;
         vmGlobals->globalNetState = NET_IS_OPEN;
         err = 0;
      }
      else vmGlobals->globalNetState = NET_OPEN_FAILED;

      // If we didn't open it properly, abruptly terminate the netlib.
      if (vmGlobals->globalNetState != NET_IS_OPEN)
      {
         NetLibClose(refNum,true);
         err = ifErr?ifErr:err; // guich@400_25: instead of err was -13
      }
      LOCK_MEMORY; // guich@300_22
   }
   else
   {
      // We couldn't find NetLib and don't even have a refNum.
      // Return with error.
      err = -11;
   }
   return err;
#endif
}
// waba/io/NativeMethods void socketCreate(Socket obj, String host, int port, int timeout, boolean noLinger)
void SocketCreate(Var stack[]) // (Lwaba/io/Socket;Ljava/lang/String;IIZ)V
{
#if 1 // @t ndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   WObject sock, host;
   int32 port;
   UtfString s;
   NetSocketRef socketRef;
   NetSocketAddrINType sockAddr;
   SDWord timeout;
   Err err;
   int status;
   int noLinger; // guich@330_18
   uint16 refnum;

   sock = stack[0].obj;
   host = stack[1].obj;
   port = stack[2].intValue;
   timeout = (*vmGlobals->millisToTicks)(stack[3].intValue); // guich@200b4: now the open timeout is passed.
   noLinger = (int)stack[4].intValue;

   WOBJ_SocketLastErr(sock) = 0;
   WOBJ_SocketRef(sock) = -1;
   WOBJ_SocketTimeout(sock) = 5000; // guich@200b3: read timeout mathias ringwald asked to change this

   if (vmGlobals->globalNetState == NET_NOT_READY_FOR_OPEN)
   {
      WOBJ_SocketLastErr(sock) = -12;
      return; // not ready
   }
   else
   if ((err = openNetLib()) != 0)
   {
      WOBJ_SocketLastErr(sock) = err;
      return;
   }
   refnum = vmGlobals->globalSocketLibRefNum; // guich@330_50: store in a temp variable to avoid accessing vmGlobals during SemaphoreReserve
   s = (*vmGlobals->stringToUtf)(host, STU_NULL_TERMINATE | STU_USE_STATIC);
   if (s.len == 0)
   {
      WOBJ_SocketLastErr(sock) = -10;
      return;
   }
   // set up sockAddr structure
   UNLOCK_MEMORY; // guich@300_22
   xmemzero((char *)&sockAddr, sizeof(sockAddr));
   sockAddr.family = netSocketAddrINET;
   sockAddr.port = NetHToNS((Word)port);
   // is the given str an ip address?
   sockAddr.addr = NetLibAddrAToIN(refnum, s.str);
   if (sockAddr.addr == -1)
   {
      // no, not an ip, but a real address
      NetHostInfoBufType hostInfo;
      NetHostInfoPtr hostPtr;

      hostPtr = NetLibGetHostByName(refnum, s.str, &hostInfo, timeout, &err);
      if (hostPtr == NULL)
      {
         LOCK_MEMORY; // guich@300_22
         WOBJ_SocketLastErr(sock) = err;
         return;
      }
      sockAddr.addr = *((NetIPAddr *)hostInfo.hostInfo.addrListP[0]);
   }
   socketRef = NetLibSocketOpen(refnum, netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, timeout, &err);
   if (socketRef == -1)
   {
      LOCK_MEMORY; // guich@310_16
      WOBJ_SocketLastErr(sock) = err;
      return;
   }
   if (noLinger) // guich@330_18
   {
      NetSocketLingerType linger;
      linger.onOff = true;
      linger.time  = 0;
      NetLibSocketOptionSet(refnum, socketRef, netSocketOptLevelSocket, netSocketOptSockLinger, &linger, sizeof(linger),-1,&err);
   }
   //LOCK_MEMORY;
   status = NetLibSocketConnect(refnum, socketRef, (NetSocketAddrType *)&sockAddr, sizeof(sockAddr), timeout, &err);
   //UNLOCK_MEMORY;
   if (status == -1)
   {
      Err tempErr = err;
      NetLibSocketClose(refnum, socketRef, timeout, &err);
      LOCK_MEMORY; // guich@300_22
      WOBJ_SocketLastErr(sock) = tempErr; // guich@310_16
      return;
   }

   LOCK_MEMORY; // guich@300_22
   WOBJ_SocketRef(sock) = socketRef;
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Socket static boolean socketClose(Socket obj)
void SocketClose(Var stack[]) // (Lwaba/io/Socket;)Z
{
   WObject socket;

   socket = stack[0].obj;
   WOBJ_SocketLastErr(socket) = 0;
   RETURN32(_SocketClose(socket))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Socket static boolean socketIsOpen(Socket obj)
void SocketIsOpen(Var stack[]) // (Lwaba/io/Socket;)Z
{
   WObject socket;

   socket = stack[0].obj;
   RETURN32((WOBJ_SocketRef(socket) == -1)?0:1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Socket static boolean socketSetReadTimeout(Socket obj, int millis)
void SocketSetReadTimeout(Var stack[]) // (Lwaba/io/Socket;I)Z
{
   WObject socket;
   int32 millis;

   socket = stack[0].obj;
   millis = stack[1].intValue;
   WOBJ_SocketLastErr(socket) = 0;
   if (millis < 0)
   {
      RETURN32(0)
   }
   WOBJ_SocketTimeout(socket) = millis;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Socket static int socketReadBytes(Socket obj, byte buf[], int start, int count)
void SocketReadBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   SocketReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Socket static int socketWriteBytes(Socket obj, byte buf[], int start, int count)
void SocketWriteBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   SocketReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean socketDisconnect()
void SocketDisconnect(Var stack[]) // ()Z
{
   Err err;
   int32 v = 0;

   if (vmGlobals->globalNetState == NET_IS_OPEN)
   {
      uint16 refnum = vmGlobals->globalSocketLibRefNum;
      UNLOCK_MEMORY; // guich@400_49
      do
      {
         err = NetLibClose(refnum,true);
      } while (err == netErrStillOpen); // guich@512_4: keep calling until there's surely no more instances
      NetLibFinishCloseWait(refnum); // guich@512_4: mortal bullet
      LOCK_MEMORY;
      if (err == netErrNotOpen) err = 0;
      v = err == 0;
      if (err) (*vmGlobals->debug)((*vmGlobals->int2str)("err on SocketDisconnect",err));
      vmGlobals->globalNetState = NET_READY_FOR_OPEN;
      EvtEnqueueKey(0x030D,0x030D, commandKeyMask); // guich@300_38: add this to cut the line in the kyocera phone
   }
   RETURN32(v)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj, String name, int mode)
void CatalogCreate(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;I)V
{
   #define ggetUInt32(b) (uint32)( (uint32)((b)[0])<<24 | (uint32)((b)[1])<<16 | (uint32)((b)[2])<<8 | (uint32)((b)[3]) )
   UtfString s;
   WObject cat, name;
   int32 mode;
   DmOpenRef dmRef;
   ULong dbType, dbCreator;
   byte isMemo=0;

   cat = stack[0].obj;
   name = stack[1].obj;
   mode = stack[2].intValue;
   WOBJ_CatalogDmRef(cat) = 0;
   WOBJ_CatalogCurRecPos(cat) = -1;
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      return;
   dbCreator = vmGlobals->appCreatorId;
   dbType = 'DATA';
   if (s.len > 10 && s.str[s.len - 10] == '.' && s.str[s.len - 5] == '.')
   {
      isMemo = !xstrncmp(s.str,"MemoDB.memo.DATA",s.len); // are we trying to open the memo db?
      s.str[s.len - 10] = 0;
      dbCreator = ggetUInt32((uchar *)&s.str[s.len - 9]);
      dbType = ggetUInt32((uchar *)&s.str[s.len - 4]);
   }
   // guich@200b4_19: fixed the memo and debug concurrent use
   if (isMemo && vmGlobals->dmRefSystemOut) // memo already opened?
      dmRef = vmGlobals->dmRefSystemOut;
   else
   {
      dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType, mode == Catalog_READ_ONLY, 0);
      if (isMemo && dmRef)
         vmGlobals->dmRefSystemOut = dmRef; // assign to the global pos
   }

   if (mode == Catalog_CREATE && !dmRef)
   {
      if (DmCreateDatabase(0, s.str, dbCreator, dbType, false) != 0)
         return;
      dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType,0,0);
      // set the backup bit on the database
      if (dmRef != 0)
      {
         LocalID dbID;
         UInt cardNo;
         Word attributes;

         DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, &cardNo, NULL);
         DmDatabaseInfo(cardNo, dbID, NULL, &attributes,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
         attributes |= dmHdrAttrBackup;
         DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      }
   }
   if (dmRef != 0)
      WOBJ_CatalogDmRef(cat) = dmRef;
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean catalogRename(Catalog obj, String newName)
void CatalogRename(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;)Z
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID dbID;
   UInt cardNo;
   UtfString s;
   WObject name;
   ULong dbType, dbCreator;
   Err err;
   byte isMemo=0;

   cat = stack[0].obj;
   name = stack[1].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      RETURN32(0)
   if (s.len < 10 || s.str[s.len - 10] != '.' || s.str[s.len - 5] != '.')
      RETURN32(0)
   isMemo = !xstrncmp(s.str,"MemoDB.memo.DATA",s.len); // are we trying to open the memo db?
   if (isMemo) // can't rename to the memo database!
      RETURN32(0)

   s.str[s.len - 10] = 0;
   dbCreator = ggetUInt32((uchar *)&s.str[s.len - 9]);
   dbType = ggetUInt32((uchar *)&s.str[s.len - 4]);
   DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, &cardNo, NULL);
   err = DmSetDatabaseInfo(cardNo, dbID, s.str, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &dbType, &dbCreator);
   RETURN32(err == errNone)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size)
void CatalogAddRecord(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   UInt pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmRemoveRecord(dmRef, pos); // guich@400_16: changed from DmDeleteRecord
      RETURN32(-1)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32(pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size, int pos)
void CatalogAddRecord2(Var stack[]) // (Lwaba/io/Catalog;II)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   UInt pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = stack[2].intValue;//DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmRemoveRecord(dmRef, pos); // guich@400_16: changed from DmDeleteRecord
      RETURN32(-1)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32(pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogResizeRecord(Catalog obj, int size)
void CatalogResizeRecord(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   int32 pos, size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   if (size <= 0)
      RETURN32(0)
   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      RETURN32(0)
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   recH = DmResizeRecord(dmRef, pos, (ULong)size);
   if (recH == NULL)
      RETURN32(0)
   recH = DmGetRecord(dmRef, pos); // guich@310_12: fixed Record Not Locked bug
   if (recH == NULL)
      RETURN32(0)
   recPtr = MemHandleLock(recH);
   if (recPtr == NULL)
      RETURN32(0)
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogClose(Catalog obj)
void CatalogClose(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;

   cat = stack[0].obj;
   RETURN32(_CatalogClose(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDelete(Catalog obj)
void CatalogDelete(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID localID;
   UInt cardNo;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   DmOpenDatabaseInfo(dmRef, &localID, NULL, NULL, &cardNo, NULL);
   _CatalogClose(cat);
   if (DmDeleteDatabase(cardNo, localID) != 0)
      RETURN32(0)
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
static void _ULongToChars(ULong l, char *buf)
{
   buf[0] = l >> 24;
   buf[1] = LOWER8(l >> 16);
   buf[2] = LOWER8(l >> 8);
   buf[3] = LOWER8(l);
}
// waba/io/NativeMethods String []catalogListCatalogs(int creatorId, int type)
void CatalogListCatalogs(Var stack[]) // (II)[Ljava/lang/String;
{
   WObject stringArray, *strings;
   DmSearchStateType state;
   Boolean first;
   UInt cardNo;
   LocalID dbID;
   ULong dbType, dbCreator;
   int n, len;
   char name[32 + 10];
   int32 type,crtr;

   crtr = stack[0].intValue;
   type = stack[1].intValue;
   n = 0;
   first = true;
   while (DmGetNextDatabaseByTypeCreator(first,
      &state, type, crtr, false, &cardNo, &dbID) == 0)
   {
      first = false;
      n++;
   }
   if (n == 0)
      RETURN32(0)
   stringArray = (*vmGlobals->createArrayObject)(1, n);
   if (stringArray == NULL || (*vmGlobals->pushObject)(stringArray) == -1)
      RETURN32(0)
   strings = (WObject *)WOBJ_arrayStart(stringArray);
   n = 0;
   first = true;
   while (DmGetNextDatabaseByTypeCreator(first, &state,
         type, crtr, false, &cardNo, &dbID) == 0)
   {
      first = false;
      name[0] = 0;
      DmDatabaseInfo(cardNo, dbID, name, NULL, NULL, NULL, NULL,
         NULL, NULL, NULL, NULL, &dbType, &dbCreator);
      len = xstrlen(name);
      name[len++] = '.';
      _ULongToChars(dbCreator, &name[len]);
      len += 4;
      name[len++] = '.';
      _ULongToChars(dbType, &name[len]);
      len += 4;
      name[len++] = 0;
      strings[n++] = (*vmGlobals->createString)(name);
   }
   (*vmGlobals->popObject)(); // stringArray
   RETURN32(stringArray)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDeleteRecord(Catalog obj)
void CatalogDeleteRecord(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (pos == -1)
      RETURN32(0)
   _RecClose(cat);
   // NOTE: Was DmDeleteRecord() but this didn't match with CE
   // and was too difficult to use in actual programs since it
   // didn't delete the record immediately
   if (DmRemoveRecord(dmRef, pos) != 0)
      RETURN32(0)
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordCount(Catalog obj)
void CatalogGetRecordCount(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0)?-1:DmNumRecords(dmRef))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordSize(Catalog obj)
void CatalogGetRecordSize(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0 || pos == -1)?-1:WOBJ_CatalogCurRecLen(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogIsOpen(Catalog obj)
void CatalogIsOpen(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0)?0:1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogSetRecordPos(Catalog obj, int pos)
void CatalogSetRecordPos(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   int32 pos, count;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;

   cat = stack[0].obj;
   pos = stack[1].intValue;
   _RecClose(cat);
   if (pos < 0)
      RETURN32(0)
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED && dmRef == vmGlobals->dmRefSystemOut && pos == vmGlobals->posSystemOut) // guich@200b4_19: don't let him use the debug record - guich@200final_8: forgot to test the dmRef
      RETURN32(0)
   count = (int32)DmNumRecords(dmRef);
   if (pos >= count)
      RETURN32(0)
   recH = DmGetRecord(dmRef, pos);
   if (!recH)
      RETURN32(0)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmReleaseRecord(dmRef, pos, false);
      RETURN32(0)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = MemHandleSize(recH);
   WOBJ_CatalogCurRecModified(cat) = 0;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogReadBytes(Catalog obj, byte buf[], int start, int count)
void CatalogReadBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   CatalogReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogSkipBytes(Catalog obj, int count)
void CatalogSkipBytes(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   WObject cat;
   int32 count, pos, offset;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   count = stack[1].intValue;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   if (pos == -1)
      RETURN32(-1)
   offset = WOBJ_CatalogCurRecOffset(cat)+count;
   if (offset < 0 || offset > WOBJ_CatalogCurRecLen(cat)) // guich@210_11
      RETURN32(-1)
   WOBJ_CatalogCurRecOffset(cat) = offset;    // guich@230_23: changed count to offset
   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogWriteBytes(Catalog obj, byte buf[], int start, int count)
void CatalogWriteBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   CatalogReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogInspectRecord(Catalog obj, byte buf[], int recPosition)
void CatalogInspectRecord(Var stack[]) // (Lwaba/io/Catalog;[BI)I
{
   WObject cat, byteArray;
   int32 count, pos, size;
   uchar *bytes;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;

   cat = stack[0].obj;
   byteArray = stack[1].obj;
   pos = stack[2].intValue;
   count = WOBJ_arrayLen(byteArray);
   bytes = (uchar *)WOBJ_arrayStart(byteArray);

   // gets the record
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   recH = DmQueryRecord(dmRef, pos);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   size = MemHandleSize(recH);
   if (count > size) count = size;

   // transfers the data. ps: i did not made a direct reference to the record because i dont have where to store the memhandle params.
   xmemmove(bytes, recPtr, count);
   MemHandleUnlock(recH);

   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static byte catalogGetRecordAttributes(Catalog obj, int recordPos)
void CatalogGetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;I)B
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;
   UInt16 attr;
   Err err;

   cat = stack[0].obj;
   pos = stack[1].intValue; // guich@400_42: now the user send the record index
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (pos < 0)
      RETURN32(0)
   err = DmRecordInfo (dmRef, pos, &attr, NULL, NULL);
   RETURN32(err?-1:LOWER8(attr)) // returns byte
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetRecordAttributes(Catalog obj, int recordPos, byte attr)
void CatalogSetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;IB)V
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;
   UInt16 attr;
   Err err;

   cat = stack[0].obj;
   pos = stack[1].intValue; // guich@400_42
   attr = (UInt16)stack[2].intValue;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      return;
   if (pos < 0)
      return;
   err = DmSetRecordInfo (dmRef, pos, &attr, NULL);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetAttributes(Catalog obj)
void CatalogGetAttributes(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID dbID;
   UInt cardNo;
   Word attributes;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)

   DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, &cardNo, NULL);
   DmDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   RETURN32(attributes)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetAttributes(Catalog obj, int i)
void CatalogSetAttributes(Var stack[]) // (Lwaba/io/Catalog;I)V
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID dbID;
   UInt cardNo;
   Word attributes;

   cat = stack[0].obj;
   attributes = LOWER16(stack[1].intValue);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef != 0)
   {
      DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, &cardNo, NULL);
      DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordPos(Catalog obj)
void CatalogGetRecordPos(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat = stack[0].obj; // class object
   RETURN32(WOBJ_CatalogCurRecPos(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj)
void CatalogCreate_NoFunc(Var stack[]) // (Lwaba/io/Catalog;)V
{
   (*vmGlobals->debug)("This Catalog constructor has no effect!");
}
//////////////////////////////////////////////////////////////////////////
static uint16 getVN()
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   uint32 it = vfsIteratorStart;
   uint16 vn,lastValid=0;
   Err err;
   UNLOCK_MEMORY;
//   VFSVolumeEnumerate(&vn, &it);
   while (it != vfsIteratorStop)
   {
      err = VFSVolumeEnumerate(&vn, &it);
      if (err == errNone)
         lastValid = vn;
      else
         break;
   }
   LOCK_MEMORY;
   return lastValid;
#endif
}
void FileDestroy(WObject file)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WOBJ_FileLastError(file) = 0;
	if (WOBJ_FileHandle(file))
	{
	   Err err;
	   int32 handle = WOBJ_FileHandle(file);
	   UNLOCK_MEMORY;
	   err = VFSFileClose(handle);
	   LOCK_MEMORY;
	   WOBJ_FileLastError(file) = err;
	   WOBJ_FileHandle(file) = NULL;
	}
#endif
}
static void FileReadWriteBytes(Var stack[], int isRead)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   int32 v = -1;
   WObject file, byteArray;
   int32 start, count;
   uint32 offset, size, n=0;
   byte *bytes,*temp;
   FileRef fref;
   Err err;
   VoidHand tempH;

   file = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   fref = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   UNLOCK_MEMORY;
   err = VFSFileTell(fref,&offset);
   LOCK_MEMORY;
   if (err != errNone)
   {
      WOBJ_FileLastError(file) = err;
      RETURN32(-1)
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1) // array null or range invalid

   // create a temp buffer
   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_FileLastError(file) = -99;
      RETURN32(-1) // no memory to allocate temp buffer
   }
continueReading:
	temp = MemHandleLock(tempH);
   bytes = (byte *)WOBJ_arrayStart(byteArray);

   if (isRead)
   {
      UNLOCK_MEMORY;
      err = VFSFileRead(fref, count, temp, &n);
      LOCK_MEMORY;
      if (err == 10759/*FileEOF*/) // guich@512_11: if err is fileeof, there's no error!
         err = errNone;
      if (err != errNone)
         WOBJ_FileLastError(file) = err;
      else
         xmemmove(&bytes[start],temp,n);
   }
   else
   {
      xmemmove(temp,&bytes[start],count);
      UNLOCK_MEMORY;
      // The write operation is more complex. First, we simply try to write
      err = VFSFileWrite(fref, count, temp, &n);
      LOCK_MEMORY;
      if (err != errNone) // error? now we do something more complex: expand the file and write to it
      {
         // failed...
         WOBJ_FileLastError(file) = err; // store the first error
         UNLOCK_MEMORY;
         err = VFSFileSize(fref,&size);
         LOCK_MEMORY;
         if (err != errNone) // cannot get file size? return the previous error, this is probably a consequence of it
            goto releaseAndReturn;
         if (offset + count <= size) // the file does not need to be expanded
            goto releaseAndReturn; // a real error happened
         // otherwise, the file need to be explicitly expanded
         UNLOCK_MEMORY;
         if ((err = VFSFileResize(fref, offset+count)) == errNone)
            err = VFSFileWrite(fref, count, temp, &n);
         LOCK_MEMORY;
         WOBJ_FileLastError(file) = err;
      }
   }
   if (err == errNone) // guich@512_10: changed != to ==
      v = n;

releaseAndReturn:
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   RETURN32(v)
#endif
}
#define FILE_CREATE_DIR 1
#define FILE_IS_DIR     2
#define FILE_DELETE     3
#define FILE_RENAME     4
#define FILE_EXISTS     5
#define FILE_GET_ATTR   6
#define FILE_SET_ATTR   7

#define ATTR_ARCHIVE   1
#define ATTR_HIDDEN    2
#define ATTR_READ_ONLY 4
#define ATTR_SYSTEM    8

static void FileOp(Var stack[], int op)
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	int32 v=0;
	WObject file = stack[0].obj; // class object
	Err err=0;
	uint16 vn = WOBJ_FileSlot(file);
	UtfString fileName = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
	WOBJ_FileLastError(file) = 0;

	switch (op)
	{
	   case FILE_CREATE_DIR:
         UNLOCK_MEMORY;
      	err = VFSDirCreate(vn, fileName.str);
	      break;
	   case FILE_DELETE:
      	FileDestroy(file); // close it first - also reserves the semaphore
         UNLOCK_MEMORY;
      	err = VFSFileDelete(vn, fileName.str);
			break;
		case FILE_EXISTS:
		{
      	FileRef fref;
         UNLOCK_MEMORY;
			err = VFSFileOpen(vn, fileName.str, vfsModeRead, &fref);
         v = err == errNone || err == vfsErrFileStillOpen || err == vfsErrFilePermissionDenied;
		   VFSFileClose(fref);
         LOCK_MEMORY;
         RETURN32(v) // don't remove this from here; different value is returned!
	   }
	   case FILE_RENAME:
	   {
      	UtfString fileNameTo;
      	char *to,*s;
      	fileNameTo = (*vmGlobals->stringToUtf)(stack[1].obj, STU_CREATE | STU_USE_STATIC | STU_NULL_TERMINATE);
      	// strip the path from the file
      	s = &fileNameTo.str[fileNameTo.len-1];
      	while (s >= fileNameTo.str && *s != '/') s--;
      	if (s >= fileNameTo.str) // found a / ?
      	   s++;
      	else
      	   s = fileNameTo.str; // else use the whole string

      	FileDestroy(file); // close it first
    	   // allocate a temp buffer bc we need to release the semaphore, and the fileNameTo variable is allocated in the global memory
    	   fileNameTo.len = xstrlen(s);
     	   to = (char*)vmGlobals->xmalloc(fileNameTo.len+1);
     	   if (to == NULL)
     	   {
     	      err = -10;
     	      break;
     	   }
     	   xmemmove(to,s,fileNameTo.len+1);
         UNLOCK_MEMORY;
      	err = VFSFileRename(vn, fileName.str, to);
      	xfree(to);
      	break;
	   }
	   case FILE_IS_DIR:
	   {
      	FileRef fref;
      	uint32 attr;
         UNLOCK_MEMORY;
			err = VFSFileOpen(vn, fileName.str, vfsModeRead, &fref);
      	if (err == errNone && (err=VFSFileGetAttributes(fref, &attr)) == errNone)
      	{
         	v = (attr & vfsFileAttrDirectory) != 0;
            VFSFileClose(fref);
         }
         else goto error;
         LOCK_MEMORY;
         RETURN32(v)  // don't remove this from here; different value is returned!
	   }
	}

	if (err == errNone)
	   v = 1;
error:
   LOCK_MEMORY;
  	WOBJ_FileLastError(file) = err;
   RETURN32(v)
#endif
}

// waba/io/NativeMethods boolean fileIsAvailable()
void FileIsAvailable(Var stack[]) // ()Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	UInt32 vfsMgrVersion;
	uint32 ret=0;
	UNLOCK_MEMORY;
	ret = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion) == errNone;
	LOCK_MEMORY;
	RETURN32(ret)
#else
   RETURN32(0)
#endif
}
// waba/io/NativeMethods boolean fileIsInserted(int slot)
void FileIsInserted(Var stack[]) // (I)Z
{
   int32 slot = stack[0].intValue;
   uint16 vn = slot==-1 ? getVN() : (uint16)slot;
   uint32 used,total,ret;
   ret = VFSVolumeSize(vn, &used, &total) == errNone;
   RETURN32(ret)
}
//////////////////////////////////////////////////////////////////////////
#define DONT_OPEN 0
#define READ_ONLY 1
#define WRITE_ONLY 2
#define READ_WRITE 3
#define CREATE 4
// waba/io/NativeMethods void fileCreate(File file, String path, int mode, int slot)
void FileCreate(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;II)V
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj;
	int32   mode = stack[2].intValue;
	int32   slot = stack[3].intValue;
	uint16 vn = slot==-1 ? getVN() : (uint16)slot;
	WOBJ_FileSlot(file) = vn;

	if (mode != DONT_OPEN)
	{
		FileRef fref=0;
		uint16 imode=0;
		uint32 attr;
		Err err=0;
		byte isDir;
		UtfString path;
		path = (*vmGlobals->stringToUtf)(stack[1].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
		UNLOCK_MEMORY;
		switch (mode)
		{
		   case WRITE_ONLY:
		      imode = vfsModeWrite;
		      break;
		   case CREATE:
	   	   err = VFSFileCreate(vn, path.str); // no break!
	   	   if (err != errNone && err != vfsErrFileAlreadyExists)
	   	      goto error;
		   case READ_WRITE:
		      imode = vfsModeReadWrite;
		      break;
		   case READ_ONLY:
		   default:
	   	   imode = vfsModeRead;
	   	   break;
	   }
	   err = VFSFileOpen(vn, path.str, imode, &fref);
	   if (err != errNone)
	      goto error;
	
	   VFSFileGetAttributes(fref, &attr);
	   isDir = attr & vfsFileAttrDirectory;
	   if (isDir) // if is a directory, close it immediately
	      VFSFileClose(fref);
	   LOCK_MEMORY;
	   if (!isDir)
	      WOBJ_FileHandle(file) = fref;
	error:
	   if (!fref) LOCK_MEMORY;
	   WOBJ_FileLastError(file) = err;
	}
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileClose(File obj)
void FileClose_(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	FileDestroy(stack[0].obj);
	RETURN32(1)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileCreateDir(File obj)
void FileCreateDir(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileOp(stack, FILE_CREATE_DIR);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileDelete(File obj)
void FileDelete_(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileOp(stack, FILE_DELETE);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileExists(File obj)
void FileExists(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileOp(stack, FILE_EXISTS);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetSize(File obj)
void FileGetSize(Var stack[]) // (Lwaba/io/File;)I
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj; // class object
	uint32 size;
	int32 handle = WOBJ_FileHandle(file);
	Err err;
   int32 ret;

	UNLOCK_MEMORY;
	err = VFSFileSize(handle, &size);
	LOCK_MEMORY;
	if (err != errNone) // guich@421_38: if its the root dir, return the free disk space
	{
	   // check if its the root dir
   	UtfString fileName = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
	   if (fileName.len == 1 && fileName.str[0] == '/') // get free space
	   {
	      uint32 total;
      	UNLOCK_MEMORY;
	      err = VFSVolumeSize(WOBJ_FileSlot(file),&size,&total);
      	LOCK_MEMORY;
	      size = total - size;
	   }
	}
	WOBJ_FileLastError(file) = err;
	ret = err == errNone ? (int32)size : -1;
	RETURN32(ret)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsDir(File obj)
void FileIsDir(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileOp(stack, FILE_IS_DIR);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsOpen(File obj)
void FileIsOpen(Var stack[]) // (Lwaba/io/File;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj; // class object
	WOBJ_FileLastError(file) = 0;
	RETURN32(WOBJ_FileHandle(file) != NULL)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods String []fileListFiles(File obj)
void FileListFiles(Var stack[]) // (Lwaba/io/File;)[Ljava/lang/String;
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file, stringArray, *strings;
	UtfString path;
   uint32 it;
   uint16 vn;
   FileInfoType fileinf;
   FileRef dir;
   char fileName[128];
	int count=0,len;
	WObject v=0;

   file = stack[0].obj;
   // get the first volume
   vn = WOBJ_FileSlot(file);
   // get the path
   path = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
	WOBJ_FileLastError(file) = 0;

   xmemzero(&fileinf,sizeof(fileinf)); // in the first try we don't need to store the names
   UNLOCK_MEMORY;
   VFSFileOpen (vn, path.str, vfsModeRead, &dir);
   // first, count how many entries
   it = vfsIteratorStart;
   while (true)
   {
      if (VFSDirEntryEnumerate (dir, &it, &fileinf))
         break;
      count++;
   }
   LOCK_MEMORY;
   if (count == 0) // no items
      goto error;

   // now, store the entries
   fileinf.nameBufLen = sizeof(fileName);
   fileinf.nameP = fileName;
	stringArray = (*vmGlobals->createArrayObject)(1, count);
	if (!stringArray || (*vmGlobals->pushObject)(stringArray) == -1)
	   goto error;
   strings = (WObject *)WOBJ_arrayStart(stringArray);
   it = vfsIteratorStart;
   while (true)
   {
      UNLOCK_MEMORY;
      if (VFSDirEntryEnumerate(dir, &it, &fileinf))
         break;

      if (fileinf.attributes & vfsFileAttrDirectory)
      {
         // if file is a directory, append a / to it
         len = xstrlen(fileName);
         fileName[len++] = '/';
         fileName[len] = 0;
      }
      else
      if (fileinf.attributes == 6 && fileName[0] == 'V' && xstrncmp(&fileName[1],"OLUME.NAM",9) == NULL)// fileinf.attributes & vfsFileAttrVolumeLabel) // is this the volume label? replace it by the real volume label
      {
         VFSVolumeGetLabel(vn, &fileName[1], sizeof(fileName)-3); // guich@552_9: changed -2 to -3
  			fileName[0] = '[';
  			len = xstrlen(fileName);
  			fileName[len++] = ']';
  			fileName[len] = 0;
      }
      LOCK_MEMORY;
		*strings++ = (*vmGlobals->createString)(fileName);
   }
   LOCK_MEMORY;
	(*vmGlobals->popObject)(); // stringArray

	v = stringArray;
error:
   UNLOCK_MEMORY;
   VFSFileClose(dir);
   LOCK_MEMORY;
	RETURN32(v)

#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileReadBytes(File obj, byte []b, int off, int len)
void FileReadBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileReadWriteBytes(stack, 1);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileRename(File obj, String path)
void FileRename(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileOp(stack, FILE_RENAME);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileSetPos(File obj, int pos)
void FileSetPos(Var stack[]) // (Lwaba/io/File;I)Z
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj; // class object
	Err err;
	int32 handle = WOBJ_FileHandle(file);
	UNLOCK_MEMORY;
	err = VFSFileSeek(handle, vfsOriginBeginning, stack[1].intValue);
   if (err == 10759/*FileEOF*/) // guich@520_1: if err is fileeof, there's no error!
      err = errNone;
	LOCK_MEMORY;
	WOBJ_FileLastError(file) = err;
	RETURN32(err == errNone)
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileWriteBytes(File obj, byte []b, int off, int len)
void FileWriteBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   FileReadWriteBytes(stack, 0);
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods void fileSetAttributes(File obj, int attr)
void FileSetAttributes(Var stack[]) // (Lwaba/io/File;I)V
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj; // class object
	uint32 attr,userAttr;
	Err err;
	int32 handle = WOBJ_FileHandle(file);
	userAttr = stack[1].intValue;
   UNLOCK_MEMORY;
	if ((err=VFSFileGetAttributes(handle, &attr)) == errNone)
	{
	   // change only attributes that the user can change
      if (userAttr & ATTR_ARCHIVE)
         attr |= vfsFileAttrArchive;
      else
         attr &= ~vfsFileAttrArchive;

      if (userAttr & ATTR_HIDDEN)
         attr |= vfsFileAttrHidden;
      else
         attr &= ~vfsFileAttrHidden;

      if (userAttr & ATTR_READ_ONLY)
         attr |= vfsFileAttrReadOnly;
      else
         attr &= ~vfsFileAttrReadOnly;

      if (userAttr & ATTR_SYSTEM)
         attr |= vfsFileAttrSystem;
      else
         attr &= ~vfsFileAttrSystem;

	   err = VFSFileSetAttributes(handle, attr);
   }
   LOCK_MEMORY;
  	WOBJ_FileLastError(file) = err;
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetAttributes(File obj)
void FileGetAttributes(Var stack[]) // (Lwaba/io/File;)I
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
	WObject file = stack[0].obj; // class object
	uint32 attr;
	int32 v = -1;
	int32 handle = WOBJ_FileHandle(file);
	Err err;
	UNLOCK_MEMORY;
	err = VFSFileGetAttributes(handle, &attr);
	LOCK_MEMORY;
	if (err == errNone)
	   v = ((attr & vfsFileAttrArchive ) ? ATTR_ARCHIVE   : 0) |
          ((attr & vfsFileAttrHidden  ) ? ATTR_HIDDEN    : 0) |
          ((attr & vfsFileAttrReadOnly) ? ATTR_READ_ONLY : 0) |
          ((attr & vfsFileAttrSystem  ) ? ATTR_SYSTEM    : 0);
	WOBJ_FileLastError(file) = err;
	RETURN32(v)
#endif
}
//////////////////////////////////////////////////////////////////////////
#define TIME_CREATED  1
#define TIME_MODIFIED 2
#define TIME_ACCESSED 4
// waba/io/NativeMethods void fileSetTime(File obj, byte whichTime, waba.sys.Time time)
void FileSetTime(Var stack[]) // (Lwaba/io/File;BLwaba/sys/Time;)V
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   WObject file = stack[0].obj;
   FileRef fref = WOBJ_FileHandle(file);
   Err err;
   WOBJ_FileLastError(file) = 0;
   if (fref != NULL)
   {
      DateTimeType dt;
      uint32 date;
      int which = stack[1].intValue;
      WObject time = stack[2].obj;
      dt.year   = WOBJ_TimeYear(time);
      dt.month  = WOBJ_TimeMonth(time);
      dt.day    = WOBJ_TimeDay(time);
      dt.hour   = WOBJ_TimeHour(time);
      dt.minute = WOBJ_TimeMinute(time);
      dt.second = WOBJ_TimeSecond(time);
      date = TimDateTimeToSeconds(&dt);
      UNLOCK_MEMORY;
      if (which & TIME_CREATED)  err = VFSFileSetDate(fref, vfsFileDateCreated, date);
      if (which & TIME_ACCESSED) err = VFSFileSetDate(fref, vfsFileDateAccessed, date);
      if (which & TIME_MODIFIED) err = VFSFileSetDate(fref, vfsFileDateModified, date);
      LOCK_MEMORY;
      WOBJ_FileLastError(file) = err;
   }
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods waba.sys.Time fileGetTime(File obj, byte whichTime)
void FileGetTime(Var stack[]) // (Lwaba/io/File;B)Lwaba/sys/Time;
{
#ifndef ENABLE_DEBUG_BY_COMENTING_IO_ROUTINES
   static WClass *timeClass;
	WObject file = stack[0].obj; // class object
	uint32 date;
	int which = stack[1].intValue;
   WObject time=0;
   Err err;
   UNLOCK_MEMORY;
	err = VFSFileGetDate(WOBJ_FileHandle(file), (which == TIME_CREATED)  ? vfsFileDateCreated  :
	                                                (which == TIME_MODIFIED) ? vfsFileDateModified :
	                                                (which == TIME_ACCESSED) ? vfsFileDateAccessed :
	                                                -1, &date);
	LOCK_MEMORY;
	WOBJ_FileLastError(file) = err;
	if (err == errNone)
	{
      DateTimeType dt;
      // create the waba.sys.Time object
      if (timeClass == NULL)
         timeClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/sys/Time"),1);
      time = (*vmGlobals->createObject)(timeClass);
      if (time)
      {
         TimSecondsToDateTime(date, &dt);
         WOBJ_TimeYear(time)   = dt.year;
         WOBJ_TimeMonth(time)  = dt.month;
         WOBJ_TimeDay(time)    = dt.day;
         WOBJ_TimeHour(time)   = dt.hour;
         WOBJ_TimeMinute(time) = dt.minute;
         WOBJ_TimeSecond(time) = dt.second;
      }
	}
	RETURN32(time)
#endif
}
