/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#include "superwaba.h"

void doubleToStr(char dest[], double val, int totalLength, int decimal, int options);
WObject double2str(double d, int places);
int32 str2int32(char *str);
double str2double(char *str);
char *intToString(int32 i);
WObject int642str(int64 i);
WObject float2str(float f);
static double log(double x);
static double exp(double x);
static double pow(double x, double y);
int64 str2int64(char *str);

//////////////////////////////// Math ////////////////////////////////////
// guich@330_28: moved from nmpalm_wabasys.c to shrink SWNatives.prc size

int64 str2int64(char *str)
{
   int64 r = 0;
   int64 m = 1;
   char *c = str;
   while (*c) // test and go to end of string
   {
      if (*c != '+' && *c != '-' && (*c < '0' || *c > '9')) // guich@300_37: fixed int2str and long2str when an invalid char appeared.
         return r;
      c++; // go to end of string
   }
   c--;
   while (c >= str)
   {
      if (*c == '+') break;
      if (*c == '-') {r *= -1; break;}
      r += m * (*c-'0');
      c--;
      if (m == 1)
         m = 10;
      else
         m *= 10;
   }
   return r;
}

WObject float2str(float f)
{
	FlpCompDouble cFloat;
	char buf[20];
	int len;

	cFloat.d = (double)f;
	FlpFToA(cFloat.fd, buf);
	len = xstrlen(buf);
	if (len > 3)
	{
	   char *c = &buf[len - 3];
	   if (*c == 'e' && *(c+1) == '0' && *(c+2) == '0')
   		buf[len - 3] = 0;
   }
	return (*vmGlobals->createString)(buf);
}

WObject double2str(double d, int places)
{
	char buf[40];

	doubleToStr(buf,d,0,places,0);
	return (*vmGlobals->createString)((char *)buf);
}

char *intToString(int32 i)
{
	static char buf[20];
   char *c = buf+sizeof(buf)-2;
	xzero32((byte *)buf, 20>>2);
   if (i == 0)
      *c = '0';
   else
   {
      byte negative = (i < 0);
      if (!negative)
         i = -i;
      while (i <= -10)
      {
         *c-- = -(i % 10) + '0';
         i /= 10;
      }
      *c-- = -i + '0';
      if (negative)
         *c-- = '-';
      c++;
   }
   return c;
}

WObject int642str(int64 i)
{
	static char buf[60];
   char *c = buf+sizeof(buf)-2;
	xzero32((byte *)buf, sizeof(buf)>>2);
   if (i == 0)
      *c = '0';
   else
   {
      byte negative = (i < 0);
      if (!negative)
         i = -i;
      while (i <= -10)
      {
         *c-- = -(i % 10) + '0';
         i /= 10;
      }
      *c-- = -i + '0';
      if (negative)
         *c-- = '-';
      c++;
   }

   return (*vmGlobals->createString)(c); // guich@421_73: changed buf to c
}

static int _isnan(double value)
{
   Comp64 v;
   v.d = value;
   return (v.ul[1] == 0) && (v.ul[0] == 0x7ff80000L);
}

int32 str2int32(char *str)
{
   int32 r = 0;
   int32 m = 1;
   char *c = str;
   while (*c) // test and go to end of string
   {
      if (*c != '+' && *c != '-' && (*c < '0' || *c > '9'))
         return r;
      c++; // go to end of string
   }
   c--;
   while (c >= str)
   {
      if (*c == '+') break;
      if (*c == '-') {r *= -1; break;}
      r += m * (*c-'0');
      c--;
      if (m == 1)
         m = 10;
      else
         m *= 10;
   }
   return r;
}

/** recognizes exponencial term (e) */
double str2double(char *str)
{
   char *e;
   if (*str == ' ' || *str == 0) return 0; // empty string or invalid char?
   if (*str == '+') str++; // guich@350_14: fixes problem when converting numbers that starts with +

   e = StrStr(str,"E"); // guich@350_13: handle exponent in uppercase 'E'
   if (e != NULL)
      *e = 'e';

   if (StrStr(str, "e")) // scientific notation?
   {
		int  i, start, length, punctPos, exponent, ePos;
		double  result, sign, fractPart;
		char *ePtr;

		result = fractPart = 0.0;
		length = punctPos = xstrlen(str);
		ePtr = StrStr(str, "e");
		ePos = ePtr - str;

		start = 0;
		sign = 1.0;
		if (str[0] == '-')
		{
			sign = -1.0;
			start = 1;
		}

		for (i=start; i < length; i++)				// parse the string from left to right converting the integer part
		{
			if (str[i] != ',' && str[i] != '.' && str[i] != 'e')
			{
				if ('0' <= str[i] && str[i] <= '9')
					result = result * 10.0 + (str[i] - '0');
				else
				   return 0.0;
			}
			else
			{
				punctPos = i;
				break;
			}
		}

		if (str[punctPos] == '.' || str[punctPos] == ',')				// parse the string from the end to the '.' converting the fractional part
		{
			if (ePtr)
				start = ePos;
			else
				start = length;
			for (i=start-1; i>punctPos; i--)
			{
				if ('0' <= str[i] && str[i] <= '9')
					fractPart = fractPart / 10.0 + (str[i] - '0');
				else
					return 0.0;
			}
			result += fractPart / 10.0;
		}

		exponent = 0;
		if (ePtr) exponent = str2int32(ePtr+1);			// adjust the result for the exponent
		return result * pow(10,exponent) * sign;
   }
   else
   {
	   double i=0,d=0;
	   byte m=1;
	   int64 d2 =0;
	   int32 f = 1; // guich@330_11: changed from int16 to int32 to fix a problem when converting 2.000001
	   char *c = str;
	   if (*c == '-') // guich@240_25
	   {
	      c++;
	      m = -1;
	   }
	   // finds the decimal point
	   while (*c != 0 && *c != '.') c++;
	   if (*c == 0) // dont have decimal point - only integer
	      return (double)str2int64(str);
	   // convert the integer part
	   *c++ = 0; // truncate the string
	   i = str2int64(str);
	   // converts the decimal part
	   while (*c == '0') {f = (f==1)?10:(f*10); c++;} // if have zeros before the first significant number, count it
	   d = str2int64(c);
	   d2 = (int64)d;
	   while (d2 > 0) {d2 /= 10; d /= 10;} // converts to under 0
	   if (f > 1) d /= f; // add less significand zeroes
	   return m==1?(i+d):(i-d);
	}
}

#ifdef PALMOS
#define POSITIVE_INFINITY_VALUE 0x7ff0000000000000L
#define NEGATIVE_INFINITY_VALUE 0xfff0000000000000L
#define NAN_VALUE 0x7ff8000000000000L

#define LN2   0.693147180559945309417
#define INV_LN2   1.44269504088896338700e+00
#define Lg1  6.666666666666735130e-01
#define Lg2  3.999999999940941908e-01
#define Lg3  2.857142874366239149e-01
#define Lg4  2.222219843214978396e-01
#define Lg5  1.818357216161805012e-01
#define Lg6  1.531383769920937332e-01
#define Lg7  1.479819860511658591e-01
#define P1    1.66666666666666019037e-01
#define P2   -2.77777777770155933842e-03
#define P3    6.61375632143793436117e-05
#define P4   -1.65339022054652515390e-06
#define P5    4.13813679705723846039e-08

static double log(double x)
{
   Comp64 c64;
   int32 hx;
   double f,s,z,w,r;
   int32 k;

   c64.d = x;
   hx = c64.ul[0];

   if (hx < 0x00100000)
   {
      if (hx < 0)
         return NAN_VALUE;
      else
         return NEGATIVE_INFINITY_VALUE;
   }
   else if (hx >= 0x7ff00000)
   {
      if (hx == 0x7ff00000)
         return POSITIVE_INFINITY_VALUE;
      else
         return NAN_VALUE;
   }

   if ((hx & 0x000fffff) <= 0x6a09c)
      k = (hx>>20)-1023;
   else
      k = (hx>>20)-1022;

   c64.ll = (c64.ll - (((long long) k) << 52));
   f = c64.d - 1.0;
   s = f / (2.0 + f);
   z = s*s;
   w = z*z;
   r = w*(Lg2 + w*(Lg4 + w*Lg6)) + z*(Lg1 + w*(Lg3 + w*(Lg5 + w*Lg7)));
   return k * LN2 + f - s*(f - r);
}
static double exp(double x)
{
   Comp64 c64;
   int32 hx;
   int32 ix;
   int32 k;
   double rx,t,c,y;
   long long yl;

   c64.d = x;
   hx = c64.ul[0];
   ix = hx & 0x7fffffff;

   // |x| >= 709.78, infinite, NaN: non-finite result
   if (ix >= 0x40862e42)
   {
      if (hx == 0x7ff80000)
         return NAN_VALUE;
      else if (hx < 0)
         return 0.0;
      else
         return POSITIVE_INFINITY_VALUE;
   }

   k = (int32)(((int32) ((x * INV_LN2 * 10)+5)) / 10); // round
   rx = x - k * LN2;
   t = rx * rx;
   c = rx - t*(P1 + t*(P2 + t*(P3 + t*(P4 + t*P5))));
   y = 1.0 - ((rx*c) / (c - 2.0) - rx);
   if (k==0)
      return y;
   else
   {
      c64.d = y;
      yl = c64.ll + (((long long) k) << 52);
      // overflow/underflow result in yl < 0,
      // when all exponent bits = 0: "denormalized numbers"
      if (yl < 0x0010000000000000L)
         return (k > 0 ? POSITIVE_INFINITY_VALUE : 0.0);
      else
      {
         c64.ll = yl;
         return c64.d;
      }
   }
}
static double pow(double x, double y)
{
    if (_isnan(x) || _isnan(y))
      return NAN_VALUE;

    if (x < 0.0)
    {
      Comp64 c64;
      long long lly;
      int yIntType;
      int32 expY;

      c64.d = y;
      lly = c64.ll;
      expY = ((int32) ((lly & 0x7ff0000000000000L) >> 52)) - 0x3ff;

      // 0=non-int, 1=odd, 2=even
      if (y == 0.0)
         yIntType = 2;
      else if (expY < 0)
         yIntType = 0;
      else if (expY <= 54)
      {
         long long temp = ((((lly & 0x000fffffffffffffL) + 0x0010000000000000L) << expY) & 0x001fffffffffffffL);
         if (temp == 0L)
            yIntType = 2;
         else if (temp == 0x0010000000000000L)
           yIntType = 1;
         else
           yIntType = 0;
      }
      else
         yIntType = 2;

      if (yIntType == 0)
         return NAN_VALUE;
      else if (yIntType == 1)
         return -exp (y * log (-x));
      else return exp (y * log (-x));
    }
    else
      return exp (y * log (x));
}
#endif
//////////// from here to below code taken from Michael Brereton's EWE //////////////
static char decimalPoint = '.', plusSign = '+', negSign = '-';

#define EXP_NOTATION 0x10
#define NO_EXP_NOTATION 0x20

#define LN10 2.3025850929940455
#define MAX_FLOAT 1.15292150460684698E18 // pow(2,60);
#define MIN_FLOAT 8.881784197001252E-16  // pow(2,-50);

#define bits(VAL) (*(int64 *)(&VAL))
static char Infinity [] = "+Inf";
static char NaN [] = "NaN";
#define MAX_DECIMALS 14 // DONT CHANGE THIS!

void doubleToStr(char dest[],double val,int totalLength,int decimal, int options)
{
	int neg = val<0;
	if (neg) val *= -1;

	if (_isnan(val))
	{
		xstrncpy(dest,NaN,3);
		return;
	} else
	if (bits(val) == POSITIVE_INFINITY_VALUE || bits(val) == NEGATIVE_INFINITY_VALUE)
	{
		xstrncpy(dest,Infinity,4);
		dest[0] = neg ? negSign : plusSign;
		return;
	}

	if (val != 0)
   	// Should we do Sci. Notation?
   	if (((val >= MAX_FLOAT  || val <= MIN_FLOAT || ((options & EXP_NOTATION) != 0)) && ((options & NO_EXP_NOTATION) == 0)))
   	{
   	   // Have to do e version.
   		double power = log(val)/LN10;
   		double man;
   		int negpow = power < 0;
   		if (negpow) power *= -1;
   		if (negpow)
   		{
   			man = 1-(power-(int64)power);
   			power = (double)(1+(int64)power);
   			if (man >= 1)
   			{
   				man -= 1;
   				power -= 1;
   			}
   		}
   		else
   		{
   			man = power-(int64)power;
   			power = (double)(int64)power;
   		}
   		man = pow(10,man);
   		if (neg) man *= -1;
   		if (totalLength == 0)
   			doubleToStr(dest,man,0,decimal,options|NO_EXP_NOTATION);
   		else
   			doubleToStr(dest,man,1,0,options|NO_EXP_NOTATION);
   		while(*dest != 0) dest++;
   		*dest++ = 'e';
   		*dest++ = negpow ? negSign : plusSign;
   		doubleToStr(dest,power,0,0,NO_EXP_NOTATION);
   		return;
   	}
   ///// end of EWE stuff ////
   {double d;
   int64 i;
   char *c = dest;
	static char buf2[40];
   char *cc = buf2;
   int n;
   static double rounds[] = {5e-1,5e-2,5e-3,5e-4,5e-5,5e-6,5e-7,5e-8,5e-9,5e-10,5e-11,5e-12,5e-13,5e-14,5e-15,5e-16,5e-17,5e-18};
   byte floatingPoint = decimal < 0;
   int intcount = 0; // guich@240_7

   if (neg) // guich@200b3: corrected negative value convertion.
      *c++ = negSign; // value already negated
   i = (int64)val;
   d = (val - (double)i);

   // guich@240_7: count the number of digits before the .
   while (i > 0)
   {
      intcount++;
      i /= 10;
   }

   i = (int64)val;
   if (d != 0)
   {
      if (floatingPoint) // free flow
      {
         double temp = d;
   		for (n = 1; n <= MAX_DECIMALS; n++)
   		{
   			temp *= 10.0;
   			if (temp >= 1.0) decimal = n;
   			temp -= (int64)temp;
   		}
   		if (decimal == MAX_DECIMALS) // 12.04 = 12 + 0.03999999.... in this case, we just round it and try again
   		{
            d += rounds[MAX_DECIMALS-intcount];
            temp = d;
      		for (n = 1; n <= MAX_DECIMALS; n++)
      		{
      			temp *= 10.0;
      			if (temp >= 1.0) decimal = n;
      			temp -= (int64)temp;
      		}
         }
      }

      if (decimal >= MAX_DECIMALS) decimal = MAX_DECIMALS-intcount; else
      if (decimal < 0) decimal = 0; // guich@512_3
      d += rounds[decimal];
      if (d >= 1.0d)
      {
         i += (int64)d; // guich@240_5
         d -= 1.0d;
      }
   }
   // convert integer part
   do
   {
      n = i % 10;
      *cc++ = '0'+n;
      i /= 10;
   } while (i > 0);
   // reverse the string
   n = cc - buf2;
   cc--;
   while (n--)
      *c++ = *cc--;
   // convert the decimal part
   if (decimal > 0)
   {
      *c++ = decimalPoint;
      while (decimal--)
      {
         d *= 10;
         n = (int)d%10;
         *c++ = '0'+n;
         d -= n; // dont let the number grow above 10
      }
      if (floatingPoint) // if floating point, strip out the leading zeros
      {
         while (*(c-1) == '0')
            c--;
         if (*(c-1) == decimalPoint) // guich@240_7
            c--;
      }
   }

   *c = 0;
   }
}
