/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Convert.c,v 1.9 2005/07/06 01:37:46 fdie Exp $ */

#include "superwaba.h"
#include "SW_natives.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static String convertUnsigned2hex(int b, int places)
void ConvertUnsigned2hex(Var stack[]) // (II)Ljava/lang/String;
{
   char buf[9];
   uint32 b = stack[0].intValue;
   int places = stack[1].intValue;
   vmGlobals->unsigned2hex(b,places,buf);
   RETURN32(vmGlobals->createStringFromCharPtr(buf,places))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static char convertToLowerCase(char c)
void ConvertToLowerCase(Var stack[]) // (C)C
{
   RETURN32(vmGlobals->toLowerCase(stack[0].intValue))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static char convertToUpperCase(char c)
void ConvertToUpperCase(Var stack[]) // (C)C
{
   RETURN32(vmGlobals->toUpperCase(stack[0].intValue))
}

//////////////////////////////////////////////////////////////////////////
static double stack2double(Var stack) // internal use only
{
   UtfString s;
   double d;

   d = 0;
   s = (*vmGlobals->stringToUtf)(stack.obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len > 0) // ATT: FlpAToF has a flaw! if u convert 54321.9876543 using it returns invalid result!
      d = (*vmGlobals->str2double)(s.str);
   return d;
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static int convertToInt(String s)
void ConvertToInt(Var stack[]) // (Ljava/lang/String;)I
{
   WObject string = stack[0].obj;
   UtfString s = (*vmGlobals->stringToUtf)(string, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      RETURN32(0)
   RETURN32((*vmGlobals->str2int32)(s.str))
}
//////////////////////////////////////////////////////////////////////////
// guich@200: corrected to correctly handle international chars
// waba/sys/Convert static String convertToString(char c)
void ConvertToString_Char(Var stack[]) // (C)Ljava/lang/String;
{
   WObject obj, charArrayObj;
   uint16 *charStart;
   int len = 1;

   // create and fill char array
   charArrayObj = (*vmGlobals->createArrayObject)(5, len);
   if (!charArrayObj)
      RETURN32(0)
   if ((*vmGlobals->pushObject)(charArrayObj) == -1)
      RETURN32(0)
   charStart = (uint16 *)WOBJ_arrayStart(charArrayObj);
   charStart[0] = (uint16)stack[0].intValue;
   // create String object and set char array
   obj = (*vmGlobals->createStringObject)();
   if (obj != 0)
      WOBJ_StringCharArrayObj(obj) = charArrayObj;
   (*vmGlobals->popObject)(); // charArrayObj
   RETURN32(obj)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static int convertToIntBitwise(float f)
void ConvertToIntBitwise(Var stack[]) // (F)I
{
   RETURN32(stack[0].intValue)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static float convertToFloatBitwise(int i)
void ConvertToFloatBitwise(Var stack[]) // (I)F
{
   RETURN32(stack[0].intValue)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static String convertToString(float f)
void ConvertToString_Float(Var stack[]) // (F)Ljava/lang/String;
{
   RETURN32((*vmGlobals->float2str)(stack[0].floatValue))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static String convertToString(int i)
void ConvertToString_Int(Var stack[]) // (I)Ljava/lang/String;
{
   RETURN32((*vmGlobals->createString)((*vmGlobals->intToString)(stack[0].intValue)))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static String convertToString(double d, int precision)
void ConvertToString_Double(Var stack[]) // (DI)Ljava/lang/String;
{
   double d;
   copy64(stack,(Var*)&d);
   RETURN32((*vmGlobals->double2str)(d,(int)stack[2].intValue))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static double convertToDouble(String s)
void ConvertToDouble(Var stack[]) // (Ljava/lang/String;)D
{
   UtfString s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   // ATT: FlpAToF has a flaw! if u convert 54321.9876543 using it returns invalid result!
   double d = (s.len > 0)? (*vmGlobals->str2double)(s.str) : 0;
   RETURN64_FROM_VALUE(d)
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static String convertToString(float f, int d)
void ConvertToString_Float_Int(Var stack[]) // (FI)Ljava/lang/String;
{
   RETURN32((*vmGlobals->double2str)((double)stack[0].floatValue,(int)stack[1].intValue))
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static float convertToFloat(String s)
void ConvertToFloat(Var stack[]) // (Ljava/lang/String;)F
{
   double d = stack2double(stack[0]);
   float32 f = (float32)d;
   RETURN32_FLOAT(f)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static String convertToString(String doubleValue, int n)
void ConvertToString_String_Int(Var stack[]) // (Ljava/lang/String;I)Ljava/lang/String;
{
   RETURN32((*vmGlobals->double2str)(stack2double(stack[0]),(int)stack[1].intValue))
}

// fdie@512_13: Little Endian mixed mode targets encode doubles (8 bytes) as B4B5B6B7B0B1B2B3
// Full little endian is B0B1B2B3B4B5B6B7
// On mixed little endian targets both dwords have to be swapped.

static __inline void copy64_(Var *src, Var *dest)
{
#if SW_DOUBLES_MIXED  // Mixed!
   *dest = *(src+1);
   *(dest+1) = *src;
#else   
   *dest = *src;
   *(dest+1) = *(src+1);
#endif   
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static long convertDoubleToLongBits(double value)
void ConvertDoubleToLongBits(Var stack[]) // (D)J
{
   double d;
   int64 l;
   copy64_(stack, (Var*)&d); //fdie@512_13
   l = *((int64*)&d);
   _LOG_DEBUG((LOG_CAT_MATH, "ConvertDoubleToLongBits(%lf)=0x%08lx%08lx", d, (long)(l>>32), (long)(l & 0xFFFFFFFFL)));
   RETURN64_FROM_VALUE(l)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Convert static double convertLongBitsToDouble(long bits)
void ConvertDoubleFromLongBits(Var stack[]) // (J)D
{
   double d;
   int64 l;
   copy64_(stack, (Var*)&l); //fdie@512_13
   d = *((double*)&l);
   _LOG_DEBUG((LOG_CAT_MATH, "ConvertDoubleFromLongBits(0x%08lx%08lx)=%lf", (long)(l>>32), (long)(l & 0xFFFFFFFFL), d));
   RETURN64_FROM_VALUE(d)
}

/*===========================================================================*/
