/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_File.c,v 1.26 2005/11/10 01:44:11 guich Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"
#include "SW_natives.h"
#include <stdio.h>

#define INVALID_FILEPTR_VALUE (void *)0

#define FILE_CREATE_DIR 1
#define FILE_IS_DIR     2
#define FILE_DELETE     3
#define FILE_RENAME     4
#define FILE_EXISTS     5
#define FILE_GET_ATTR   6
#define FILE_SET_ATTR   7

#define ATTR_ARCHIVE   1
#define ATTR_HIDDEN    2
#define ATTR_READ_ONLY 4
#define ATTR_SYSTEM    8

//////////////////////////////////////////////////////////////////////////
// guich@350_19: added File support (grabbed code from SuperWaba 1.21)
#define File_DONT_OPEN 0
#define File_READ_ONLY 1
#define File_WRITE_ONLY 2
#define File_READ_WRITE 3
#define File_CREATE 4

#define DONT_OPEN 0
#define READ_ONLY 1
#define WRITE_ONLY 2
#define READ_WRITE 3
#define CREATE 4

#if defined PALMOS

static uint16 getVN()
{
   uint32 it = vfsIteratorStart/* | vfsIncludePrivateVolumes*/;
   uint16 vn,lastValid=0;
   Err err;
   UNLOCK_MEMORY;
//   VFSVolumeEnumerate(&vn, &it);
   while (it != vfsIteratorStop)
   {
      err = VFSVolumeEnumerate(&vn, &it);
      if (err == errNone)
         lastValid = vn;
      else
         break;
   }
   LOCK_MEMORY;
   return lastValid;
}

void fileDestroy(WObject file)
{
   WOBJ_FileLastError(file) = 0;
   if (WOBJ_FileHandle(file))
   {
      Err err;
      int32 handle = WOBJ_FileHandle(file);
      UNLOCK_MEMORY;
      err = VFSFileClose(handle);
      LOCK_MEMORY;
      WOBJ_FileLastError(file) = err;
      WOBJ_FileHandle(file) = 0;
   }
}

void FileDestroy(WObject file)
{
   fileDestroy(file);
}

static void FileReadWriteBytes(Var stack[], int isRead)
{
   int32 v = -1;
   WObject file, byteArray;
   int32 start, count;
   uint32 offset, size, n=0;
   byte *bytes,*temp;
   FileRef fref;
   Err err;
   VoidHand tempH;

   file = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   fref = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   UNLOCK_MEMORY;
   err = VFSFileTell(fref,&offset);
   LOCK_MEMORY;
   if (err != errNone)
   {
      WOBJ_FileLastError(file) = err;
      RETURN32(-1);
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1); // array null or range invalid

   // create a temp buffer
   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_FileLastError(file) = -99;
      RETURN32(-1) // no memory to allocate temp buffer
   }
continueReading:
   temp = MemHandleLock(tempH);
   bytes = (byte *)WOBJ_arrayStart(byteArray);

   if (isRead)
   {
      UNLOCK_MEMORY;
      err = VFSFileRead(fref, count, temp, &n);
      LOCK_MEMORY;
      if (err == 10759/*FileEOF*/) // guich@512_11: if err is fileeof, there's no error!
         err = errNone;
      if (err != errNone)
         WOBJ_FileLastError(file) = err;
      else
         xmemmove(&bytes[start],temp,n);
   }
   else
   {
      xmemmove(temp,&bytes[start],count);
      UNLOCK_MEMORY;
      // The write operation is more complex. First, we simply try to write
      err = VFSFileWrite(fref, count, temp, &n);
      LOCK_MEMORY;
      if (err != errNone) // error? now we do something more complex: expand the file and write to it
      {
         // failed...
         WOBJ_FileLastError(file) = err; // store the first error
         UNLOCK_MEMORY;
         err = VFSFileSize(fref,&size);
         LOCK_MEMORY;
         if (err != errNone) // cannot get file size? return the previous error, this is probably a consequence of it
            goto releaseAndReturn;
         if (offset + count <= size) // the file does not need to be expanded
            goto releaseAndReturn; // a real error happened
         // otherwise, the file need to be explicitly expanded
         UNLOCK_MEMORY;
         if ((err = VFSFileResize(fref, offset+count)) == errNone)
            err = VFSFileWrite(fref, count, temp, &n);
         LOCK_MEMORY;
         WOBJ_FileLastError(file) = err;
      }
   }
   if (err == errNone) // guich@512_10: changed != to ==
      v = n;

releaseAndReturn:
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   RETURN32(n);
}

static void FileOp(Var stack[], int op)
{
   int32 v = 0;
   WObject file = stack[0].obj; // class object
   Err err=0;
   uint16 vn = WOBJ_FileSlot(file);
   UtfString fileName = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
   WOBJ_FileLastError(file) = 0;

   switch (op)
   {
      case FILE_CREATE_DIR:
         UNLOCK_MEMORY;
         err = VFSDirCreate(vn, fileName.str);
         break;
      case FILE_DELETE:
         FileDestroy(file); // close it first - also reserves the semaphore
         UNLOCK_MEMORY;
         err = VFSFileDelete(vn, fileName.str);
         break;
      case FILE_EXISTS:
      {
         FileRef fref;
         UNLOCK_MEMORY;
         err = VFSFileOpen(vn, fileName.str, vfsModeRead, &fref);
         v = err == errNone || err == vfsErrFileStillOpen || err == vfsErrFilePermissionDenied;
         VFSFileClose(fref);
         LOCK_MEMORY;
         RETURN32(v) // don't remove this from here; different value is returned!
      }
      case FILE_RENAME:
      {
         UtfString fileNameTo;
         char *to,*s;
         fileNameTo = (*vmGlobals->stringToUtf)(stack[1].obj, STU_CREATE | STU_USE_STATIC | STU_NULL_TERMINATE);
         // strip the path from the file
         s = &fileNameTo.str[fileNameTo.len-1];
         while (s >= fileNameTo.str && *s != '/') s--;
         if (s >= fileNameTo.str) // found a / ?
            s++;
         else
            s = fileNameTo.str; // else use the whole string

         FileDestroy(file); // close it first
         // allocate a temp buffer bc we need to release the semaphore, and the fileNameTo variable is allocated in the global memory
         fileNameTo.len = xstrlen(s);
         to = (char*)vmGlobals->xmalloc(fileNameTo.len+1);
         if (to == NULL)
         {
            err = -10;
            break;
         }
         xmemmove(to,s,fileNameTo.len+1);
         UNLOCK_MEMORY;
         err = VFSFileRename(vn, fileName.str, to);
         xfree(to);
         break;
      }
      case FILE_IS_DIR:
      {
         FileRef fref;
         uint32 attr;
         UNLOCK_MEMORY;
         err = VFSFileOpen(vn, fileName.str, vfsModeRead, &fref);
         if (err == errNone && (err=VFSFileGetAttributes(fref, &attr)) == errNone)
         {
            v = (attr & vfsFileAttrDirectory) != 0;
            VFSFileClose(fref);
         }
         else goto error;
         LOCK_MEMORY;
         RETURN32(v);  // don't remove this from here; different value is returned!
      }
   }

   if (err == errNone)
      v = 1;
error:
   LOCK_MEMORY;
   WOBJ_FileLastError(file) = err;
   RETURN32(v);
}


// waba/io/NativeMethods boolean fileIsAvailable()
void FileIsAvailable(Var stack[]) // ()Z
{
	UInt32 vfsMgrVersion;
	uint32 ret;
	UNLOCK_MEMORY;
	ret = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion) == errNone;
	LOCK_MEMORY;
	RETURN32(ret)
}
// waba/io/NativeMethods boolean fileIsInserted(int slot)
void FileIsInserted(Var stack[]) // (I)Z
{
   int32 slot = stack[0].intValue;
   uint16 vn = slot==-1 ? getVN() : (uint16)slot;
   uint32 used,total,ret;
   ret = VFSVolumeSize(vn, &used, &total) == errNone;
   RETURN32(ret)
}

// waba/io/NativeMethods void fileCreate(File file, String path, int mode, int slot)
void FileCreate(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;II)V
{
   WObject file = stack[0].obj;
   int32   mode = stack[2].intValue;
	int32   slot = stack[3].intValue;
	uint16 vn = slot==-1 ? getVN() : (uint16)slot; // guich@552_15
	WOBJ_FileSlot(file) = vn;

   if (mode != DONT_OPEN) // guich@552_15: this must always be called bc the getVN must be computed
   {
	   FileRef fref=0;
	   uint16 imode=0;
	   uint32 attr;
	   Err err=0;
	   byte isDir;
	   UtfString path;
	   path = (*vmGlobals->stringToUtf)(stack[1].obj, STU_USE_STATIC | STU_NULL_TERMINATE);   
	   UNLOCK_MEMORY;
	   switch (mode)
	   {
	      case WRITE_ONLY:
	         imode = vfsModeWrite;
	         break;
	      case CREATE:
	         err = VFSFileCreate(vn, path.str); // no break!
	         if (err != errNone && err != vfsErrFileAlreadyExists)
	            goto error;
	      case READ_WRITE:
	         imode = vfsModeReadWrite;
	         break;
	     case READ_ONLY:
	     default:
	         imode = vfsModeRead;
	         break;
	   }
	   err = VFSFileOpen(vn, path.str, imode, &fref);
	   if (err != errNone)
	      goto error;
	
	   VFSFileGetAttributes(fref, &attr);
	   isDir = attr & vfsFileAttrDirectory;
	   if (isDir) // if is a directory, close it immediately
	      VFSFileClose(fref);
	   LOCK_MEMORY;
	   if (!isDir)
	      WOBJ_FileHandle(file) = fref;
	error:
	   if (!fref) LOCK_MEMORY;
	   WOBJ_FileLastError(file) = err;
	}
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileClose(File obj)
void FileClose_(Var stack[]) // (Lwaba/io/File;)Z
{
  FileDestroy(stack[0].obj);
  RETURN32(true);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileCreateDir(File obj)
void FileCreateDir(Var stack[]) // (Lwaba/io/File;)Z
{
   FileOp(stack, FILE_CREATE_DIR);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileDelete(File obj)
void FileDelete_(Var stack[]) // (Lwaba/io/File;)Z
{
   FileOp(stack, FILE_DELETE);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileExists(File obj)
void FileExists(Var stack[]) // (Lwaba/io/File;)Z
{
   FileOp(stack, FILE_EXISTS);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetSize(File obj)
void FileGetSize(Var stack[]) // (Lwaba/io/File;)I
{
   WObject file = stack[0].obj; // class object
   uint32 size;
   int32 handle = WOBJ_FileHandle(file);
   Err err;
   int32 ret;

   UNLOCK_MEMORY;
   err = VFSFileSize(handle, &size);
   LOCK_MEMORY;
   if (err != errNone) // guich@421_38: if its the root dir, return the free disk space
   {
      // check if its the root dir
      UtfString fileName = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
      if (fileName.len == 1 && fileName.str[0] == '/') // get free space
      {
         uint32 total;
         UNLOCK_MEMORY;
         err = VFSVolumeSize(WOBJ_FileSlot(file),&size,&total);
         LOCK_MEMORY;
         size = total - size;
      }
   }
   WOBJ_FileLastError(file) = err;
   ret = err == errNone ? (int32)size : -1;
   RETURN32(ret);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsDir(File obj)
void FileIsDir(Var stack[]) // (Lwaba/io/File;)Z
{
   FileOp(stack, FILE_IS_DIR);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsOpen(File obj)
void FileIsOpen(Var stack[]) // (Lwaba/io/File;)Z
{
   WObject file = stack[0].obj; // class object
   WOBJ_FileLastError(file) = 0;
   RETURN32(WOBJ_FileHandle(file) != 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods String []fileListFiles(File obj)
void FileListFiles(Var stack[]) // (Lwaba/io/File;)[Ljava/lang/String;
{
   WObject file, stringArray, *strings;
   UtfString path;
   uint32 it;
   uint16 vn;
   FileInfoType fileinf;
   FileRef dir;
   char fileName[128];
   int count=0,len;
   WObject v=0;

   file = stack[0].obj;
   // get the first volume
   vn = WOBJ_FileSlot(file);
   // get the path
   path = (*vmGlobals->stringToUtf)(WOBJ_FileName(file), STU_USE_STATIC | STU_NULL_TERMINATE);
   WOBJ_FileLastError(file) = 0;

   xmemzero(&fileinf,sizeof(fileinf)); // in the first try we don't need to store the names
   UNLOCK_MEMORY;
   VFSFileOpen (vn, path.str, vfsModeRead, &dir);
   // first, count how many entries
   it = vfsIteratorStart;
   while (true)
   {
      if (VFSDirEntryEnumerate (dir, &it, &fileinf))
         break;
      count++;
   }
   LOCK_MEMORY;
   if (count == 0) // no items
      goto error;

   // now, store the entries
   fileinf.nameBufLen = sizeof(fileName);
   fileinf.nameP = fileName;
   stringArray = (*vmGlobals->createArrayObject)(1, count);
   if (!stringArray || (*vmGlobals->pushObject)(stringArray) == -1)
      goto error;
   strings = (WObject *)WOBJ_arrayStart(stringArray);
   it = vfsIteratorStart;
   while (true)
   {
      UNLOCK_MEMORY;
      if (VFSDirEntryEnumerate(dir, &it, &fileinf))
         break;

      if (fileinf.attributes & vfsFileAttrDirectory)
      {
         // if file is a directory, append a / to it
         len = xstrlen(fileName);
         fileName[len++] = '/';
         fileName[len] = 0;
      }
      else
      if (fileinf.attributes == 6 && fileName[0] == 'V' && xstrncmp(&fileName[1],"OLUME.NAM",9) == 0)// fileinf.attributes & vfsFileAttrVolumeLabel) // is this the volume label? replace it by the real volume label
      {
         VFSVolumeGetLabel(vn, &fileName[1], sizeof(fileName)-3); // guich@552_9: changed -2 to -3
         fileName[0] = '[';
         len = xstrlen(fileName);
         fileName[len++] = ']';
         fileName[len] = 0;
      }
      LOCK_MEMORY;
      *strings++ = (*vmGlobals->createString)(fileName);
   }
   LOCK_MEMORY;
   (*vmGlobals->popObject)(); // stringArray

   v = stringArray;
error:
   UNLOCK_MEMORY;
   VFSFileClose(dir);
   LOCK_MEMORY;
   RETURN32(v);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileReadBytes(File obj, byte []b, int off, int len)
void FileReadBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
   FileReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileRename(File obj, String path)
void FileRename(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;)Z
{
   FileOp(stack, FILE_RENAME);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileSetPos(File obj, int pos)
void FileSetPos(Var stack[]) // (Lwaba/io/File;I)Z
{
   WObject file = stack[0].obj; // class object
   Err err;
   int32 handle = WOBJ_FileHandle(file);
   UNLOCK_MEMORY;
   err = VFSFileSeek(handle, vfsOriginBeginning, stack[1].intValue);
   if (err == 10759/*FileEOF*/) // guich@520_1: if err is fileeof, there's no error!
      err = errNone;
   LOCK_MEMORY;
   WOBJ_FileLastError(file) = err;
   RETURN32(err == errNone);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileWriteBytes(File obj, byte []b, int off, int len)
void FileWriteBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
   FileReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods void fileSetAttributes(File obj, int attr)
void FileSetAttributes(Var stack[]) // (Lwaba/io/File;I)V
{
   WObject file = stack[0].obj; // class object
   uint32 attr,userAttr;
   Err err;
   int32 handle = WOBJ_FileHandle(file);
   userAttr = stack[1].intValue;
   UNLOCK_MEMORY;
   if ((err=VFSFileGetAttributes(handle, &attr)) == errNone)
   {
      // change only attributes that the user can change
      if (userAttr & ATTR_ARCHIVE)
         attr |= vfsFileAttrArchive;
      else
         attr &= ~vfsFileAttrArchive;

      if (userAttr & ATTR_HIDDEN)
         attr |= vfsFileAttrHidden;
      else
         attr &= ~vfsFileAttrHidden;

      if (userAttr & ATTR_READ_ONLY)
         attr |= vfsFileAttrReadOnly;
      else
         attr &= ~vfsFileAttrReadOnly;

      if (userAttr & ATTR_SYSTEM)
         attr |= vfsFileAttrSystem;
      else
         attr &= ~vfsFileAttrSystem;

      err = VFSFileSetAttributes(handle, attr);
   }
   LOCK_MEMORY;
   WOBJ_FileLastError(file) = err;
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetAttributes(File obj)
void FileGetAttributes(Var stack[]) // (Lwaba/io/File;)I
{
   WObject file = stack[0].obj; // class object
   uint32 attr;
	int32 v = -1;
   int32 handle = WOBJ_FileHandle(file);
   Err err;
   UNLOCK_MEMORY;
   err = VFSFileGetAttributes(handle, &attr);
   LOCK_MEMORY;
	if (err == errNone)
	   v = ((attr & vfsFileAttrArchive ) ? ATTR_ARCHIVE   : 0) |
          ((attr & vfsFileAttrHidden  ) ? ATTR_HIDDEN    : 0) |
          ((attr & vfsFileAttrReadOnly) ? ATTR_READ_ONLY : 0) |
          ((attr & vfsFileAttrSystem  ) ? ATTR_SYSTEM    : 0);
	WOBJ_FileLastError(file) = err;
	RETURN32(v)
}
//////////////////////////////////////////////////////////////////////////
#define TIME_CREATED  1
#define TIME_MODIFIED 2
#define TIME_ACCESSED 4
// waba/io/NativeMethods void fileSetTime(File obj, byte whichTime, waba.sys.Time time)
void FileSetTime(Var stack[]) // (Lwaba/io/File;BLwaba/sys/Time;)V
{
   WObject file = stack[0].obj;
   FileRef fref = WOBJ_FileHandle(file);
   Err err = 0;
   WOBJ_FileLastError(file) = 0;
   if (fref != 0)
   {
      DateTimeType dt;
      uint32 date;
      int which = stack[1].intValue;
      WObject time = stack[2].obj;
      dt.year   = WOBJ_TimeYear(time);
      dt.month  = WOBJ_TimeMonth(time);
      dt.day    = WOBJ_TimeDay(time);
      dt.hour   = WOBJ_TimeHour(time);
      dt.minute = WOBJ_TimeMinute(time);
      dt.second = WOBJ_TimeSecond(time);
      date = TimDateTimeToSeconds(&dt);
      UNLOCK_MEMORY;
      if (which & TIME_CREATED)  err = VFSFileSetDate(fref, vfsFileDateCreated, date);
      if (which & TIME_ACCESSED) err = VFSFileSetDate(fref, vfsFileDateAccessed, date);
      if (which & TIME_MODIFIED) err = VFSFileSetDate(fref, vfsFileDateModified, date);
      LOCK_MEMORY;
      WOBJ_FileLastError(file) = err;
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods waba.sys.Time fileGetTime(File obj, byte whichTime)
void FileGetTime(Var stack[]) // (Lwaba/io/File;B)Lwaba/sys/Time;
{
   static WClass *timeClass;
   WObject file = stack[0].obj; // class object
   uint32 date;
   int which = stack[1].intValue;
   WObject time=0;
   Err err;
   UNLOCK_MEMORY;
   err = VFSFileGetDate(WOBJ_FileHandle(file), (which == TIME_CREATED)  ? vfsFileDateCreated  :
                                                  (which == TIME_MODIFIED) ? vfsFileDateModified :
                                                  (which == TIME_ACCESSED) ? vfsFileDateAccessed :
                                                  -1, &date);
   LOCK_MEMORY;
   WOBJ_FileLastError(file) = err;
   if (err == errNone)
   {
      DateTimeType dt;
      // create the waba.sys.Time object
      if (timeClass == NULL)
         timeClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/sys/Time"),1);
      time = (*vmGlobals->createObject)(timeClass);
      if (time)
      {
         TimSecondsToDateTime(date, &dt);
         WOBJ_TimeYear(time)   = dt.year;
         WOBJ_TimeMonth(time)  = dt.month;
         WOBJ_TimeDay(time)    = dt.day;
         WOBJ_TimeHour(time)   = dt.hour;
         WOBJ_TimeMinute(time) = dt.minute;
         WOBJ_TimeSecond(time) = dt.second;
      }
   }
   RETURN32(time);
}

#else // defined PALMOS


#ifdef WINCE
  #define errno GetLastError()
  #define _tmkdir(path) CreateDirectory(path, 0)
#elif PALMOS
  #include <errno.h>
#else
  #include <errno.h>
  #include <time.h>
  #include <sys/types.h>
  #include <sys/stat.h>
#endif

#ifdef WIN32
  #include "win32/dirent.h"
  #ifdef WINCE
    #define _trename MoveFile
  #else
    #include <direct.h>
    #include <sys/timeb.h>
    #include <sys/utime.h>
    #define stat _stat
  #endif
#elif PALMOS

  #include <stdio.h>

  #define _tmkdir(path) mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO)
  #define _tstat stat
  #define _tunlink unlink
  #define _trename rename
  #define _tchmod chmod
  #define _tfopen fopen
  #define _trmdir rmdir

#else
  // POSIX
  #include <sys/time.h>
  #include <unistd.h>
  #include <dirent.h>
  #if !defined __SYMBIAN32__    // utime is unknown there.  investigating...
     #include <utime.h>
  #endif
  #ifdef UNICODE
  #define _tmkdir >>>PGR: FIXME
  #define _tstat  >>>PGR: FIXME
  #define _tunlink >>>PGR: FIXME
  #define _trename >>>PGR: FIXME
  #define _tchmod >>>PGR: FIXME
  #define _tfopen >>>PGR: FIXME
  #define _trmdir >>>PGR: FIXME
  #else
  #define _tmkdir(path) mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO)
  #define _tstat stat
  #define _tunlink unlink
  #define _trename rename
  #define _tchmod chmod
  #define _tfopen fopen
  #define _trmdir rmdir
  #endif
#endif

void fileDestroy(WObject file)
{
   FILE * fileH = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   if (fileH != INVALID_FILEPTR_VALUE)
   {
      fclose(fileH);
      WOBJ_FileHandle(file) = INVALID_FILEPTR_VALUE;
   }
}

static TCHAR * fileAllocPathName(WObject string)
{
   WObject charArray;
   TCHAR *s;
   int32 len, i;
   int16 *chars;

   if (string == 0)
      return NULL;
   charArray = WOBJ_StringCharArrayObj(string);
   if (charArray == 0)
      return NULL;
   len = WOBJ_arrayLen(charArray);
   chars = (int16 *)WOBJ_arrayStart(charArray);
   s = (TCHAR *)xmalloc((len + 1) * sizeof(TCHAR));
   if (s == NULL)
      return NULL;
   for (i = 0; i < len; i++)
   {
      s[i] = (TCHAR)chars[i];
      #if PATH_SEPARATOR == '/'
         if (s[i] == '\\') s[i] = PATH_SEPARATOR;
      #else
         if (s[i] == '/') s[i] = PATH_SEPARATOR;
      #endif
   }
   s[i] = 0;
   //printf ("fileAllocPathName => %s\n", s);
   return s;
}

static void fileFreePathName(TCHAR const * path)
{
   if (path != NULL)
      xfree_((void *)path);
}

#ifdef __ECOS
#define DIR_FLAG __stat_mode_DIR
#else
#define DIR_FLAG S_IFDIR
#endif

static void fileOp(Var stack[], int op)
{
   WObject file = stack[0].obj;
   TCHAR * path;
   int32 ret=0;
   byte ok = 1;

   WOBJ_FileLastError(file) = 0;
   path = fileAllocPathName(WOBJ_FileName(file));
   if (path == NULL)
      RETURN32(0)

   switch (op)
   {
      case FILE_CREATE_DIR:
         if (_tmkdir(path))
            ret = 1;
         else
            WOBJ_FileLastError(file) = errno;
         break;
      case FILE_RENAME:
      {
         TCHAR const * dstPath = fileAllocPathName(stack[1].obj);
         if (dstPath != NULL)
         {
            fileDestroy(file);
            if (_trename(path, dstPath) != 0)
               WOBJ_FileLastError(file) = errno;
            else
               ret = 1;
            fileFreePathName(dstPath);
         }
         break;
      }
#ifdef WINCE // not common functions between wince and win32
      case FILE_IS_DIR:
      {
         DWORD attr = GetFileAttributes(path);
         if (attr == 0xFFFFFFFF)
            WOBJ_FileLastError(file) = GetLastError();
         else
            ret = (attr & FILE_ATTRIBUTE_DIRECTORY) != 0;
         break;
      }
      case FILE_DELETE:
      {
         DWORD attr;
         fileDestroy(file);
         attr = GetFileAttributes(path);
         if (attr == 0xFFFFFFFF)
            ok = 0;
         else
         if (attr & FILE_ATTRIBUTE_DIRECTORY)
            ok = (byte)RemoveDirectory(path);
         else
            ok = (byte)DeleteFile(path);
         if (!ok)
            WOBJ_FileLastError(file) = GetLastError();
         else
            ret = 1;
         break;
      }
      case FILE_EXISTS:
      {
         DWORD attr = GetFileAttributes(path);
         if (attr == 0xFFFFFFFF)
            WOBJ_FileLastError(file) = GetLastError();
         else
            ret = 1;
         break;
      }
      case FILE_GET_ATTR:
      {
         DWORD attr = GetFileAttributes(path);
         if (attr == 0xFFFFFFFF)
         {
            ret = -1;
            WOBJ_FileLastError(file) = GetLastError();
         }
         else
         {
            ret = ((attr & FILE_ATTRIBUTE_ARCHIVE ) ? ATTR_ARCHIVE   : 0) |
                  ((attr & FILE_ATTRIBUTE_HIDDEN  ) ? ATTR_HIDDEN    : 0) |
                  ((attr & FILE_ATTRIBUTE_READONLY) ? ATTR_READ_ONLY : 0) |
                  ((attr & FILE_ATTRIBUTE_SYSTEM  ) ? ATTR_SYSTEM    : 0);
         }
         break;
      }
      case FILE_SET_ATTR:
      {
         DWORD userAttr = stack[1].intValue;
         DWORD attr = GetFileAttributes(path);

         if (attr == 0xFFFFFFFF)
            WOBJ_FileLastError(file) = GetLastError();
         else
         {
            // change only attributes that the user can change
            if (userAttr & ATTR_ARCHIVE)
               attr |= FILE_ATTRIBUTE_ARCHIVE;
            else
               attr &= ~FILE_ATTRIBUTE_ARCHIVE;

            if (userAttr & ATTR_HIDDEN)
               attr |= FILE_ATTRIBUTE_HIDDEN;
            else
               attr &= ~FILE_ATTRIBUTE_HIDDEN;

            if (userAttr & ATTR_READ_ONLY)
               attr |= FILE_ATTRIBUTE_READONLY;
            else
               attr &= ~ FILE_ATTRIBUTE_READONLY;

            if (userAttr & ATTR_SYSTEM)
               attr |= FILE_ATTRIBUTE_SYSTEM;
            else
               attr &= ~FILE_ATTRIBUTE_SYSTEM;

            if (!SetFileAttributes(path, attr))
               WOBJ_FileLastError(file) = GetLastError();
         }
         break;
      }
#else
      case FILE_IS_DIR:
      {
         struct stat statData;
         if (_tstat(path, &statData) != 0)
            WOBJ_FileLastError(file) = errno;
         else
            ret = ((statData.st_mode & DIR_FLAG) != 0);
         break;
      }
      case FILE_DELETE:
      {
         struct stat statData;

         fileDestroy(file);
         if (_tstat(path, &statData) != 0)
            ok = 0;
         else
         {
            if (statData.st_mode & DIR_FLAG)
               ok = (_trmdir(path)  == 0);
            else
               ok = (_tunlink(path) == 0);
         }
         if (!ok)
            WOBJ_FileLastError(file) = errno;
         else
            ret = 1;
         break;
      }
      case FILE_EXISTS:
      {
         struct stat statData;
         int l = xstrlen(path);
         if (l > 3 && path[l-1] == PATH_SEPARATOR) path[l-1] = 0; // guich@510_3 - guich@551_1: changed > 1 to > 3, because c:\ requires the slash!
         if (_tstat(path, &statData) != 0)
            WOBJ_FileLastError(file) = errno;
         else
            ret = 1;
         break;
      }
      case FILE_GET_ATTR:
      {
         struct stat statData;
         if (_tstat(path, &statData) != 0)
         {
            ret = -1;
            WOBJ_FileLastError(file) = errno;
         }
         else
         {
            // ATTR_HIDDEN and ATTR_ARCHIVE are not POSIX
            #if defined __ECOS
            ret = ((statData.st_mode & S_IWUSR)         ? 0 : ATTR_READ_ONLY) |
                  ((statData.st_mode & __stat_mode_REG) ? 0 : ATTR_SYSTEM);
            #elif defined WIN32
            ret = ((statData.st_mode & S_IWRITE) ? 0 : ATTR_READ_ONLY) |
                  ((statData.st_mode & S_IFREG)  ? 0 : ATTR_SYSTEM);
            #else
            ret = ((statData.st_mode & S_IWUSR)  ? 0 : ATTR_READ_ONLY) |
                  ((statData.st_mode & S_IFREG)  ? 0 : ATTR_SYSTEM);
            #endif
         }
         break;
      }
      case FILE_SET_ATTR:
      {
         int userAttr = stack[1].intValue;
         struct stat statData;
         if (_tstat(path, &statData) != 0)
            WOBJ_FileLastError(file) = errno;
         else
         {
            // ATTR_HIDDEN and ATTR_ARCHIVE are not POSIX
            if (userAttr & ATTR_READ_ONLY)
            {
               #if defined WIN32
               statData.st_mode &= ~S_IWRITE;
               #else
               statData.st_mode &= ~S_IWUSR;
               #endif
            }
            else
            {
               #if defined WIN32
               statData.st_mode |= S_IWRITE;
               #else
               statData.st_mode |= S_IWUSR;
               #endif
            }
            if (_tchmod(path, statData.st_mode))
               ret = 1;
            else
               WOBJ_FileLastError(file) = errno;
         }
         break;
      }
#endif
   }
   fileFreePathName(path);
   RETURN32(ret)
}

static void fileReadWriteBytes(Var stack[], int isRead)
{
   WObject file, byteArray;
   int32 start, count;
   uchar *bytes;
   FILE * fileH;
   int numRW;
   int ok;

   file = stack[0].obj;
   fileH = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   if (fileH == INVALID_FILEPTR_VALUE)
      RETURN32(-1)
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1) // array null or range invalid
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
   {
      numRW = fread(bytes+start, 1, count, fileH);
      ok = ((numRW == count) || (ferror(fileH) == 0));
   }
   else
   {
      numRW = fwrite(bytes+start, 1, count, fileH);
      ok = (numRW == count);
   }
   if (!ok)
   {
      WOBJ_FileLastError(file) = errno;
      numRW = -1; // guich@402_49: now returns -1.
   }
   RETURN32(numRW)
}

static int isDirectoryEntry(struct dirent * entry, TCHAR const * path)
{
   #if defined WIN32 // and BSD as well - LINUX (I think)
      return (entry->d_type & DT_DIR);
   #else  // EPOC and SYMBIAN do not have the d_type field.
   {
      TCHAR fullPath[MAX_PATH];
      struct stat statData;
      lstrcpy(fullPath, path);
      if (path[lstrlen(path)-1] != (TCHAR)'/')
         lstrcat(fullPath, TEXT("/"));
      lstrcat(fullPath, entry->d_name);
      if (stat(fullPath, &statData) != 0) return -1;  /* error: skip this */
      #ifdef __ECOS
      return (statData.st_mode & __stat_mode_DIR);
      #else
      return (statData.st_mode & S_IFDIR);
      #endif
   }
   #endif
}

// waba/io/NativeMethods boolean fileIsAvailable()
void FileIsAvailable(Var stack[]) // ()Z
{
   RETURN32(1); // always available!
}
// waba/io/NativeMethods boolean fileIsInserted(int slot)
void FileIsInserted(Var stack[]) // (I)Z
{
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods void fileCreate(File file, String path, int mode, int slot)
void FileCreate(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;II)V
{
   WObject file = stack[0].obj;
   int32 fileMode = stack[2].intValue;
//   int32 slot = stack[3].intValue;  - currently ignored
   TCHAR * rwMode;
   TCHAR const * path;
   FILE * fileH;

   WOBJ_FileLastError(file) = 0;
   WOBJ_FileHandle(file) = INVALID_FILEPTR_VALUE;
   switch (fileMode)
   {
      case File_READ_ONLY : rwMode = TEXT("rb");   break; // guich@501_1: changed all flags
      case File_WRITE_ONLY: rwMode = TEXT("wb");   break; // TODO: write_only should be non-destructive
      case File_READ_WRITE: rwMode = TEXT("rb+");  break;
      case File_CREATE    : rwMode = TEXT("rb+");  break;
      default:
         return;
   }

   path = fileAllocPathName(stack[1].obj);
   if (path == NULL)
      return;
   fileH = _tfopen(path, rwMode);
   if (!fileH && fileMode == File_CREATE) // guich@501_1: force file creation
      fileH = _tfopen(path, TEXT("wb+"));
   if (fileH)
      WOBJ_FileHandle(file) = (void *)fileH;
   else
   {
      fileH = INVALID_FILEPTR_VALUE;
      WOBJ_FileLastError(file) = errno;
   }
   fileFreePathName(path);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileClose(File obj)
void FileClose_(Var stack[]) // (Lwaba/io/File;)Z
{
   fileDestroy(stack[0].obj);
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileDelete(File obj)
void FileDelete_(Var stack[]) // (Lwaba/io/File;)Z
{
   fileOp(stack, FILE_DELETE);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileExists(File obj)
void FileExists(Var stack[]) // (Lwaba/io/File;)Z
{
   fileOp(stack, FILE_EXISTS);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetSize(File obj)
void FileGetSize(Var stack[]) // (Lwaba/io/File;)I
{
   WObject file = stack[0].obj;
   TCHAR * path = fileAllocPathName(WOBJ_FileName(file));
   int32 ret=0;

   #ifdef WINCE
   WIN32_FIND_DATA FindFileData;
   HANDLE hFind = FindFirstFile(path, &FindFileData);
   if (hFind == INVALID_HANDLE_VALUE)
   {
      ret = 0;
      WOBJ_FileLastError(file) = GetLastError();
   }
   else
   {
      ret = FindFileData.nFileSizeLow;
      FindClose(hFind);
   }

   #else
   struct stat statData;
   WOBJ_FileLastError(file) = 0;
   if (_tstat(path, &statData) == 0)
      ret = statData.st_size;
   else
   {
      ret = 0;
      WOBJ_FileLastError(file) = errno;
   }
   #endif

   fileFreePathName(path);
   RETURN32(ret)
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileCreateDir(File obj)
void FileCreateDir(Var stack[]) // (Lwaba/io/File;)Z
{
   fileOp(stack, FILE_CREATE_DIR);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsDir(File obj)
void FileIsDir(Var stack[]) // (Lwaba/io/File;)Z
{
   fileOp(stack, FILE_IS_DIR);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileIsOpen(File obj)
void FileIsOpen(Var stack[]) // (Lwaba/io/File;)Z
{
   WObject file = stack[0].obj;
   FILE * fileH = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   RETURN32((fileH != INVALID_FILEPTR_VALUE))
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods String []fileListFiles(File obj)
void FileListFiles(Var stack[]) // (Lwaba/io/File;)[Ljava/lang/String;
{
   WObject file, stringArray;
   WObject * strings;
   TCHAR * path;
   int count = 0;
   TCHAR const * fileName = 0;
   struct dirent * entry;
   DIR * dir;

   file = stack[0].obj;
   WOBJ_FileLastError(file) = 0;

   // first, count how many entries
   path = fileAllocPathName(WOBJ_FileName(file));
   dir = opendir(path);
   if (!dir)
      WOBJ_FileLastError(file) = errno;
   else
   {
      while ((entry = readdir(dir)))
      {
         fileName = entry->d_name;
         if (!(fileName[0] == '.' && (fileName[1] == 0 || (fileName[1] == '.' && fileName[2] == 0))))  // skip . and ..
            ++count;
      }
      closedir(dir);
   }
   if (!count)  // no items?
   {
      fileFreePathName(path);
      RETURN32(0)
   }

   // now, store the entries
   stringArray = (*vmGlobals->createArrayObject)(1, count);
   if ((*vmGlobals->pushObject)(stringArray) == -1)
   {
      fileFreePathName(path);
      RETURN32(0)
   }

   strings = (WObject *)WOBJ_arrayStart(stringArray);
   dir = opendir(path);

   if (!dir)
      WOBJ_FileLastError(file) = errno;
   else
   {
      while ((entry = readdir(dir)))
      {
         fileName = entry->d_name;
         if (!(fileName[0] == '.' && (fileName[1] == 0 || (fileName[1] == '.' && fileName[2] == 0))))  // skip . and ..
         {
            int isDir = isDirectoryEntry(entry, path);
            int len;
            WObject string;

            if (isDir == -1) continue;  // probably bad name - skip this
            len = lstrlen(fileName);
            if (isDir)
            {
               ++len;
            }
            string = (*vmGlobals->createStringFromUnicode)(fileName, len);
            if (isDir)
            {
               uint16 * array = (uint16 *)WOBJ_arrayStart(WOBJ_StringCharArrayObj(string));
               array[len-1] = (TCHAR)'/';
            }
            *strings++ = string;
         }
      }
      closedir(dir);
   }
   fileFreePathName(path);

   (*vmGlobals->popObject)(); // stringArray
   RETURN32(stringArray)
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileReadBytes(File obj, byte []b, int off, int len)
void FileReadBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
   fileReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileRename(File obj, String path)
void FileRename(Var stack[]) // (Lwaba/io/File;Ljava/lang/String;)Z
{
   fileOp(stack, FILE_RENAME);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean fileSetPos(File obj, int pos)
void FileSetPos(Var stack[]) // (Lwaba/io/File;I)Z
{
   WObject file;
   int32 pos;
   FILE * fileH;

   file = stack[0].obj;
   pos = stack[1].intValue;
   fileH = WOBJ_FileHandle(file);
   WOBJ_FileLastError(file) = 0;
   if (fileH != INVALID_FILEPTR_VALUE)
   {
      if ((fseek(fileH, pos, SEEK_SET) != 0))
         WOBJ_FileLastError(file) = errno;
      else
         RETURN32(1)
   }
   RETURN32(0)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileWriteBytes(File obj, byte []b, int off, int len)
void FileWriteBytes(Var stack[]) // (Lwaba/io/File;[BII)I
{
   fileReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods void fileSetAttributes(File obj, int attr)
void FileSetAttributes(Var stack[]) // (Lwaba/io/File;I)V
{
   fileOp(stack, FILE_SET_ATTR);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods int fileGetAttributes(File obj)
void FileGetAttributes(Var stack[]) // (Lwaba/io/File;)I
{
   fileOp(stack, FILE_GET_ATTR);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods void fileSetTime(File obj, byte whichTime, waba.sys.Time time)

enum {
   TIME_CREATED   = 1,
   TIME_MODIFIED  = 2,
   TIME_ACCESSED  = 4
};

void FileSetTime(Var stack[]) // (Lwaba/io/File;BLwaba/sys/Time;)V
{
#if !defined __SYMBIAN32__    // utime is unknown there.  investigating...
   WObject file = stack[0].obj;
   FILE * fileH = WOBJ_FileHandle(file);
   int which = stack[1].intValue;

   WOBJ_FileLastError(file) = 0;
   if ((fileH != INVALID_FILEPTR_VALUE) && which)
   {
      WObject time = stack[2].obj;

#if defined WINCE
         SYSTEMTIME tm;
         FILETIME fileTime;
         tm.wYear = (WORD)WOBJ_TimeYear(time);
         tm.wMonth = (WORD)WOBJ_TimeMonth(time);
         tm.wDay = (WORD)WOBJ_TimeDay(time);
         tm.wHour = (WORD)WOBJ_TimeHour(time);
         tm.wMinute = (WORD)WOBJ_TimeMinute(time);
         tm.wSecond = (WORD)WOBJ_TimeSecond(time);
         tm.wMilliseconds = (WORD)WOBJ_TimeMillis(time);
         SystemTimeToFileTime(&tm, &fileTime);
         if (!SetFileTime(fileH, (which & TIME_CREATED)  ? &fileTime : NULL,
                                 (which & TIME_ACCESSED) ? &fileTime : NULL,
                                 (which & TIME_MODIFIED) ? &fileTime : NULL))
         {
            WOBJ_FileLastError(file) = GetLastError();
         }

#elif !defined PALMOS  // WIN32 and POSIX  (*not* PALMOS)
         if (which & TIME_CREATED)
         {
            WOBJ_FileLastError(file) = EACCES;
            // not allowed to modify the creation date.
         }
         else
         {
            TCHAR * path = fileAllocPathName(WOBJ_FileName(file));
            if (!path)
               WOBJ_FileLastError(file) = ENOMEM;
            else
            {
               struct tm tm;
               struct __utimbuf timbuf;
               time_t t;

               tm.tm_year = WOBJ_TimeYear(time) - 1900;
               tm.tm_mon = WOBJ_TimeMonth(time) - 1;
               tm.tm_mday = WOBJ_TimeDay(time);
               tm.tm_hour = WOBJ_TimeHour(time);
               tm.tm_min = WOBJ_TimeMinute(time);
               tm.tm_sec = WOBJ_TimeSecond(time);
               tm.tm_isdst = -1;
               t = mktime(&tm);
               if (which & TIME_ACCESSED)
                  timbuf.actime = t;
               if (which & TIME_MODIFIED)
                  timbuf.modtime = t;
               if (__utime(path, &timbuf) != 0)
                  WOBJ_FileLastError(file) = errno;
               else
                  WOBJ_FileLastError(file) = 0;
               fileFreePathName(path);
            }
         }
#endif
   }
#endif // SYMBIAN32
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods waba.sys.Time fileGetTime(File obj, byte whichTime)
void FileGetTime(Var stack[]) // (Lwaba/io/File;B)Lwaba/sys/Time;
{
   WObject file = stack[0].obj;
   TCHAR * path = fileAllocPathName(WOBJ_FileName(file));
   WObject ret=0;

   #ifdef WINCE
   WIN32_FIND_DATA FindFileData;
   HANDLE hFind = FindFirstFile(path, &FindFileData);

   WOBJ_FileLastError(file) = 0;
   if (hFind == INVALID_HANDLE_VALUE)
      WOBJ_FileLastError(file) = GetLastError();
   else
   {
      FILETIME * whichTime;
      switch (stack[1].intValue)
      {
         case TIME_CREATED:
            whichTime = &FindFileData.ftCreationTime;
            break;
         case TIME_MODIFIED:
            whichTime = &FindFileData.ftLastWriteTime;
            break;
         case TIME_ACCESSED:
            whichTime = &FindFileData.ftLastAccessTime;
            break;
         default:
            whichTime = 0;
            break;
      }
      // create the waba.sys.Time object
      if (whichTime)
      {
         static WClass * timeClass = 0;
         WObject time;
         if (!timeClass)
            timeClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/sys/Time"),1);
         // create the waba.sys.Time object
         time = (*vmGlobals->createObject)(timeClass);
         if (time)
         {
            SYSTEMTIME sysTime;
            FileTimeToSystemTime(whichTime, &sysTime);
            WOBJ_TimeYear(time)   = sysTime.wYear;
            WOBJ_TimeMonth(time)  = sysTime.wMonth;
            WOBJ_TimeDay(time)    = sysTime.wDay;
            WOBJ_TimeHour(time)   = sysTime.wHour;
            WOBJ_TimeMinute(time) = sysTime.wMinute;
            WOBJ_TimeSecond(time) = sysTime.wSecond;
            WOBJ_TimeMillis(time) = sysTime.wMilliseconds;
            ret = time;
         }
      }
      FindClose(hFind);
   }

   #elif PALMOS

   #else
   struct stat statData;

   WOBJ_FileLastError(file) = 0;
   if (_tstat(path, &statData) != 0)
      WOBJ_FileLastError(file) = errno;
   else
   {
      time_t *  whichTime;
      switch (stack[1].intValue)
      {
         case TIME_CREATED:
            whichTime = &statData.st_ctime;
            break;
         case TIME_MODIFIED:
            whichTime = &statData.st_mtime;
            break;
         case TIME_ACCESSED:
            whichTime = &statData.st_atime;
            break;
         default:
            whichTime = 0;
            break;
      }
      if (whichTime)
      {
         static WClass * timeClass = 0;
         WObject time;
         if (!timeClass)
            timeClass = (*vmGlobals->getClass)((*vmGlobals->createUtfString)("waba/sys/Time"),1);
         // create the waba.sys.Time object
         time = (*vmGlobals->createObject)(timeClass);
         if (time)
         {
            struct tm * tm = localtime(whichTime);
            WOBJ_TimeYear(time)   = tm->tm_year + 1900;
            WOBJ_TimeMonth(time)  = tm->tm_mon + 1;
            WOBJ_TimeDay(time)    = tm->tm_mday;
            WOBJ_TimeHour(time)   = tm->tm_hour;
            WOBJ_TimeMinute(time) = tm->tm_min;
            WOBJ_TimeSecond(time) = tm->tm_sec;
            WOBJ_TimeMillis(time) = 0;
            ret = time;
         }
      }
   }
   #endif

   fileFreePathName(path);
   RETURN32(ret)
}

#endif // defined PALMOS
