/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Socket.c,v 1.16 2005/11/14 18:53:31 guich Exp $ */

#if defined(PALMOS) && defined(__arm__)

#define NetLibOpen              NetLibOpenm68k
#define NetLibClose             NetLibClosem68k
#define NetLibOpenConfig        NetLibOpenConfigm68k
#define NetLibOpenCount         NetLibOpenCountm68k
#define NetLibFinishCloseWait   NetLibFinishCloseWaitm68k
#define NetLibConnectionRefresh NetLibConnectionRefreshm68k
#define NetLibSocketConnect     NetLibSocketConnectm68k
#define NetLibSocketClose       NetLibSocketClosem68k
#define NetLibGetHostByName     NetLibGetHostByNamem68k
#define NetLibSocketOpen        NetLibSocketOpenm68k
#define NetLibSocketClose       NetLibSocketClosem68k
#define NetLibSocketOptionSet   NetLibSocketOptionSetm68k
#define NetLibSocketOptionGet   NetLibSocketOptionGetm68k
#define NetLibSend              NetLibSendm68k
#define NetLibReceive           NetLibReceivem68k
#define NetLibAddrINToA         NetLibAddrINToAm68k
#define NetLibAddrAToIN         NetLibAddrAToINm68k

#include <PalmOSARM.h>
#include <PalmCompatibility.h>
#undef NetLibOpen
#undef NetLibClose
#undef NetLibOpenConfig
#undef NetLibOpenCount
#undef NetLibFinishCloseWait
#undef NetLibConnectionRefresh
#undef NetLibSocketConnect
#undef NetLibSocketClose
#undef NetLibGetHostByName
#undef NetLibSocketOpen
#undef NetLibSocketClose
#undef NetLibSocketOptionSet
#undef NetLibSocketOptionGet
#undef NetLibSend
#undef NetLibReceive
#undef NetLibAddrINToA
#undef NetLibAddrAToIN

#define PALM_REFNUM(x)

//=====================================================
// Macros that convert native integers to and from
//  big-endian (network) order which is the order used to store
//  variable length integers by the BitMove utilities.
//====================================================

//#if CPU_TYPE == CPU_x86
#if defined(i386) || defined(__arm__)

#define _NetSwap16(x) \
    ((((x) >> 8) & 0xFF) | \
     (((x) & 0xFF) << 8))

#define _NetSwap32(x) \
    ((((x) >> 24) & 0x00FF) | \
     (((x) >>  8) & 0xFF00) | \
     (((x) & 0xFF00) <<  8) | \
     (((x) & 0x00FF) << 24))

#define NetHToNS(x) _NetSwap16(x)
#define NetHToNL(x) _NetSwap32(x)
#define NetNToHS(x) _NetSwap16(x)
#define NetNToHL(x) _NetSwap32(x)

//#elif CPU_TYPE == CPU_68K
#elif defined(m68k)

#define NetHToNS(x) (x)
#define NetHToNL(x) (x)
#define NetNToHS(x) (x)
#define NetNToHL(x) (x)

#else
// We'll define these macros for any other architectures needed as we come
// across them.
#endif

Err NetLibOpen                  (UInt16 *netIFErrsP);
Err NetLibClose                 (UInt16 immediate);
Err NetLibOpenCount             (UInt16 *countP);
Err NetLibOpenConfig            (UInt16 configIndex, UInt32 openFlags, UInt16 *netIFErrP);
Err NetLibFinishCloseWait       ();
Err NetLibConnectionRefresh     (Boolean refresh, UInt8 *allInterfacesUpP, UInt16 *netIFErrP);
Int16 NetLibSocketConnect       (NetSocketRef socket, NetSocketAddrType *sockAddrP, Int16 addrLen,
                                    Int32 timeout, Err *errP);
Int16 NetLibSocketClose         (NetSocketRef socket, Int32 timeout, Err *errP);
NetHostInfoPtr NetLibGetHostByName
                                (const Char *nameP, NetHostInfoBufPtr bufP, Int32 timeout, Err *errP);
NetSocketRef NetLibSocketOpen   (NetSocketAddrEnum domain, NetSocketTypeEnum type,
                                    Int16 protocol, Int32 timeout, Err *errP);
Int16 NetLibSocketClose         (NetSocketRef socket, Int32 timeout, Err *errP);
Int16 NetLibSocketOptionSet     (NetSocketRef socket, UInt16 /*NetSocketOptLevelEnum*/ level,
                                    UInt16 /*NetSocketOptEnum*/ option, void *optValueP, UInt16 optValueLen,
                                    Int32 timeout, Err *errP);
Int16 NetLibSocketOptionGet     (NetSocketRef socket, UInt16 /*NetSocketOptLevelEnum*/ level,
                                UInt16 /*NetSocketOptEnum*/ option, void *optValueP, UInt16 *optValueLenP,
                                Int32 timeout, Err *errP);

Int16 NetLibSend                (NetSocketRef socket, void *bufP, UInt16 bufLen, UInt16 flags,
                                    void *toAddrP, UInt16 toLen, Int32 timeout, Err *errP);
Int16 NetLibReceive             (NetSocketRef socket, void *bufP, UInt16 bufLen, UInt16 flags,
                                    void *fromAddrP, UInt16 *fromLenP, Int32 timeout, Err *errP);

Char *NetLibAddrINToA           (NetIPAddr inet, Char *spaceP);
NetIPAddr NetLibAddrAToIN       (const Char *a);

typedef struct
{
   UInt32  pilotMainBranch;
   UInt32  dispatchTableOffset;
   UInt32  shlIndex;
   UInt32  version;
   UInt32  entries;
   UInt32  more[4];
}
ModuleHeaderType;

typedef void *NetLibAPIType;
NetLibAPIType gNETLink = NULL;

#include "superwaba.h"

#define netLibCreator       'netl'
#define netLibType          'libr'
#define REVISION            0x01
#define ENTRIES             0x44
#define sysResTModuleCode   'amdc'

Err LinkModule(UInt32 type, UInt32 creator, NetLibAPIType *dispatchTableP, UInt32 *numEntriesP)
{
   UInt32 refNum;

   Err err = SysFindModule(type, creator, 0, 0, &refNum);
   if (err) err = SysLoadModule(type, creator, 0, 0, &refNum);
   if (err) return err;

   _LOG_INFO((LOG_CAT_SOCK, "DmOpenDatabaseByTypeCreator"));
   DmOpenRef dbP = DmOpenDatabaseByTypeCreator(type, creator, dmModeReadOnly);
   if (dbP == NULL)
   {
      err = DmGetLastErr();
      _LOG_ERROR((LOG_CAT_SOCK, "error=0x%x", err));
      return err;
   }

   _LOG_DEBUG((LOG_CAT_SOCK, "found db! open code resource"));
   MemHandle resH = DmGet1Resource(sysResTModuleCode, 0);
   if (resH != NULL)
   {
      _LOG_DEBUG((LOG_CAT_SOCK, "found resource!"));
      void *resP = MemHandleLock(resH);
      if (resP)
      {
         ModuleHeaderType *header = (ModuleHeaderType *)resP;
         _LOG_DEBUG((LOG_CAT_SOCK, "header=%08lx", header));
         *dispatchTableP = (UInt8 *)resP + header->dispatchTableOffset;
         _LOG_INFO((LOG_CAT_SOCK, "dispatchTableP=%08lx, %d entries", *dispatchTableP, header->entries));
         if (numEntriesP) *numEntriesP = header->entries;
         MemHandleUnlock(resH);
      }
      DmReleaseResource(resH);
   }
   return err;
}

#else

#include "superwaba.h"

#endif //defined(PALMOS) && defined(__arm__)

#include "SW_natives.h"

#if defined(WINCE)
#include <ras.h>
#endif

#if defined(PALMOS)

static Err openNetLib()
{
   DWord version=0;
   Word refNum=0, ifErr=0;

   _LOG_INFO((LOG_CAT_SOCK, "openNetLib"));
   Err err = FtrGet(netFtrCreator, netFtrNumVersion, &version);

   // Find the net library
#if defined(__arm__) // fdie@ on PalmOS/ARM we use the PalmOS5RE stub library linked statically
   if (err == 0 )
   {
      err = LinkModule(netLibType, netLibCreator, &gNETLink, NULL);
   }
#else
   if (err == 0 )
      err = SysLibFind("Net.lib", &refNum);
#endif

   // This section rewritten by vik@asi.org to suddenly close NetLib
   // in the event of a failure, and to use the system NetLibOpenCount
   // to determine if the Net library is already opened. This is done
   // because the old method used its own state variable, which was not
   // cleared when the PDA powered down and the Net library closed
   // automagically or the session closed itself by timeout
   // Note: I don't think anyone actually closes it...

   if (!err)
   {
      // guich@300_48: local variables were not being initialized correctly, leading to various errors
      Boolean allUp=0;      // vik@200b4r8_168 Moved allUp to widen its scope.
      // If the library isn't open, open it.
      Word OpenCount = 0;
      _LOG_DEBUG((LOG_CAT_SOCK, "call NetLibOpenCount"));
      NetLibOpenCount(PALM_REFNUM(refNum) &OpenCount);
      _LOG_DEBUG((LOG_CAT_SOCK, "NetLibOpenCount=%d", OpenCount));
      UNLOCK_MEMORY; // guich@300_22
      if (OpenCount == 0)
      {
         err = NetLibOpen(PALM_REFNUM(refNum) &ifErr);
         allUp = ifErr == 0; // guich@300_48: fix allUp when openCount = 0
         _LOG_INFO((LOG_CAT_SOCK, "NetLibOpen=0x%x", err));
      }
      else
      {
         // The net library is already open. Refresh any down connections.
         // vik@200b4r8_168 Moved allUp from here to widen its scope.
         Word dummy=0;
         err=NetLibConnectionRefresh(PALM_REFNUM(refNum) true, &allUp, &dummy);
      }

      // If we managed this, flag it as open.
      // vik@200b4r8_168 Fail the open if all interfaces don't come up.
      //  This forces a close of the netlib later.
      if ((err == 0 || err == netErrAlreadyOpen) && allUp) // guich@400_25: if netErrAlreadyOpen, we can continue
      {
         vmGlobals->globalSocketLibRefNum = refNum;
         vmGlobals->globalNetState = NET_IS_OPEN;
         err = 0;
      }
      else vmGlobals->globalNetState = NET_OPEN_FAILED;

      // If we didn't open it properly, abruptly terminate the netlib.
      if (vmGlobals->globalNetState != NET_IS_OPEN)
      {
         NetLibClose(PALM_REFNUM(refNum) true);
         err = ifErr?ifErr:err; // guich@400_25: instead of err was -13
      }
      LOCK_MEMORY; // guich@300_22
   }
   else
   {
      // We couldn't find NetLib and don't even have a refNum.
      // Return with error.
      err = -11;
   }
   _LOG_INFO((LOG_CAT_SOCK, "openNetLib=0x%x", err));
   return err;
}

static int32 _SocketClose(WObject socket)
{
   NetSocketRef socketRef;
   SDWord timeout;
   Err err;
   _unused_ uint16 refNum = vmGlobals->globalSocketLibRefNum;
   int32 ok;

   _LOG_INFO((LOG_CAT_SOCK, "_SocketClose"));
   socketRef = (NetSocketRef)WOBJ_SocketRef(socket);
   if (socketRef <= 0) // guich@330_39: socketref can be 0 when using SerialSocket
      return 0;
   WOBJ_SocketRef(socket) = -1;
   timeout = (*vmGlobals->millisToTicks)(1500);

   UNLOCK_MEMORY
   ok = NetLibSocketClose(PALM_REFNUM(refNum) socketRef, timeout, &err) == 0;
   LOCK_MEMORY

   return ok;
}

static void SocketReadWriteBytes(Var stack[], int isRead)
{
   // guich@400_49: in PalmOS 5, we need to keep the semaphore released all the time. For this, we need to create a buffer that will receive the bytes, and then copy this to the output.
   WObject socket, byteArray;
   NetSocketRef socketRef;
   int32 millis, start, count, countSoFar, n;
   SDWord timeout;
   byte *bytes,*bytes0,*temp;
   Err err;
   _unused_ uint16 refNum = vmGlobals->globalSocketLibRefNum;
   VoidHand tempH;
   byte refreshBeforeEachRead;
   int32 v = -1;

   socket = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   socketRef = (NetSocketRef)WOBJ_SocketRef(socket);
   WOBJ_SocketLastErr(socket) = 0;

   _LOG_INFO((LOG_CAT_SOCK, "Socket%sBytes(%ld,%ld)", isRead ? "Read":"Write", start, count));
   if (socketRef == -1)
   {
      WOBJ_SocketLastErr(socket) = -13;
      RETURN32(-1) // socket not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SocketLastErr(socket) = -14;
      RETURN32(-1) // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   bytes = &bytes[start];
   countSoFar = 0;
   millis = WOBJ_SocketTimeout(socket);
   timeout = (*vmGlobals->millisToTicks)(millis);
   refreshBeforeEachRead = WOBJ_SocketRefreshBeforeEachRead(socket);

   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_SocketLastErr(socket) = -99;
      RETURN32(-1) // no memory to allocate temp buffer
   }
continueReading:

   temp = bytes;
   bytes = MemHandleLock(tempH);
   bytes0 = bytes; // bytes will be incremented later
   if (!isRead)
   xmemmove(bytes,temp,count); // copy our sending string to a internal buffer, because we're going to release the semaphore, and so, we can't access the user's buffer

   UNLOCK_MEMORY; // release it now, bc PalmOS 5 doesnt like the release/reserve/release/reserve/... approach
   // guich@330_14
   if (refreshBeforeEachRead)
   {
      Word temp=0;
      byte allUp;
      err=NetLibConnectionRefresh(PALM_REFNUM(refNum) true, &allUp, &temp);
   }

   while (countSoFar < count)
   {
      int32 chunkSize;
      chunkSize = count - countSoFar;
      if (chunkSize > 0x7000)
         chunkSize = 0x7000;
      if (isRead)
      {
         n = NetLibReceive(PALM_REFNUM(refNum) socketRef, bytes, chunkSize, 0, 0, 0, timeout, &err);
         if (chunkSize > 0 && (n == 0 || err == netErrSocketNotOpen))
         {
            LOCK_MEMORY;
            //(*vmGlobals->debug)("error, closing socket");
            _SocketClose(socket); // server closed connection
            UNLOCK_MEMORY;
         }
      }
      else
         n = NetLibSend(PALM_REFNUM(refNum) socketRef, bytes,  chunkSize, 0, NULL, 0, timeout, &err);

      if (n > 0)
      {
         countSoFar += n;
         bytes += n;
      }
      else if (n <= 0)
      {
         LOCK_MEMORY;
         WOBJ_SocketLastErr(socket) = err;
         // NOTE: n == 0 under PalmOS when server closes connection
         if (countSoFar != 0)
            v = countSoFar;
         else if (n < 0 && (err == netErrTimeout || err == netErrWouldBlock)) // pierre@401_2
            v = 0;
         else
            v = -1;
         goto end;
      }
   }
   v = count;
   LOCK_MEMORY;
end:
   xmemmove(temp,bytes0,countSoFar);
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   RETURN32(v)
}

/*--------------------------------------------------------------socketDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void socketDestroy(WObject socketObj)
{
   _SocketClose(socketObj);
}

/*---------------------------------------------------------------SocketCreate-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/NativeMethods void socketCreate(Socket obj, String host, int port, int timeout, boolean noLinger)
void SocketCreate(Var stack[]) // (Lwaba/io/Socket;Ljava/lang/String;IIZ)V
{
   WObject sock, host;
   int32 port;
   UtfString s;
   NetSocketRef socketRef;
   NetSocketAddrINType sockAddr;
   int32 timeout;
   Err err;
   int status;
   int noLinger; // guich@330_18
   uint16 refnum;

   _LOG_INFO((LOG_CAT_SOCK, "SocketCreate"));
   sock = stack[0].obj;
   host = stack[1].obj;
   port = stack[2].intValue;
   timeout = (*vmGlobals->millisToTicks)(stack[3].intValue); // guich@200b4: now the open timeout is passed.
   noLinger = (int)stack[4].intValue;

   WOBJ_SocketLastErr(sock) = 0;
   WOBJ_SocketRef(sock) = -1;
   WOBJ_SocketTimeout(sock) = 5000; // guich@200b3: read timeout mathias ringwald asked to change this

   if (vmGlobals->globalNetState == NET_NOT_READY_FOR_OPEN)
   {
      WOBJ_SocketLastErr(sock) = -12;
      return; // not ready
   }
   else if ((err = openNetLib()) != 0)
   {
      WOBJ_SocketLastErr(sock) = err;
      return;
   }
   refnum = vmGlobals->globalSocketLibRefNum; // guich@330_50: store in a temp variable to avoid accessing vmGlobals during SemaphoreReserve
   s = (*vmGlobals->stringToUtf)(host, STU_NULL_TERMINATE | STU_USE_STATIC);
   if (s.len == 0)
   {
      WOBJ_SocketLastErr(sock) = -10;
      return;
   }
   _LOG_INFO((LOG_CAT_SOCK, "SocketCreate host=%s:%ld", s.str, port));

   // set up sockAddr structure
   UNLOCK_MEMORY; // guich@300_22
   xmemzero((char *)&sockAddr, sizeof(sockAddr));
   sockAddr.family = netSocketAddrINET;
   sockAddr.port = NetHToNS((Word)port);
   // is the given str an ip address?
   sockAddr.addr = NetLibAddrAToIN(PALM_REFNUM(refnum) s.str);
   if (sockAddr.addr == (NetIPAddr)-1)
   {
      // no, not an ip, but a real address
      NetHostInfoBufType hostInfo;
      NetHostInfoPtr hostPtr;
      hostPtr = NetLibGetHostByName(PALM_REFNUM(refnum) s.str, &hostInfo, timeout, &err);
      if (hostPtr == NULL)
      {
         LOCK_MEMORY; // guich@300_22
         _LOG_ERROR((LOG_CAT_SOCK, "NetLibGetHostByName failed=0x%x", err));
         WOBJ_SocketLastErr(sock) = err;
         return;
      }
      sockAddr.addr = *((NetIPAddr *)hostInfo.hostInfo.addrListP[0]);
   }
   socketRef = NetLibSocketOpen(PALM_REFNUM(refnum) netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, timeout, &err);
   if (socketRef == -1)
   {
      LOCK_MEMORY; // guich@310_16
      _LOG_ERROR((LOG_CAT_SOCK, "NetLibSocketOpen failed=0x%x", err));
      WOBJ_SocketLastErr(sock) = err;
      return;
   }
   if (noLinger) // guich@330_18
   {
      NetSocketLingerType linger;
      linger.onOff = true;
      linger.time  = 0;
      NetLibSocketOptionSet(PALM_REFNUM(refnum) socketRef, netSocketOptLevelSocket, netSocketOptSockLinger, &linger, sizeof(linger),-1,&err);
   }
   //LOCK_MEMORY;
   status = NetLibSocketConnect(PALM_REFNUM(refnum) socketRef, (NetSocketAddrType *)&sockAddr, sizeof(sockAddr), timeout, &err);
   //UNLOCK_MEMORY;
   if (status == -1)
   {
      Err tempErr = err;
      NetLibSocketClose(PALM_REFNUM(refnum) socketRef, timeout, &err);
      LOCK_MEMORY; // guich@300_22
      _LOG_ERROR((LOG_CAT_SOCK, "NetLibSocketConnect failed=0x%x", tempErr));
      WOBJ_SocketLastErr(sock) = tempErr; // guich@310_16
      return;
   }

   LOCK_MEMORY; // guich@300_22
   WOBJ_SocketRef(sock) = socketRef;
}

/*----------------------------------------------------------------SocketClose-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketClose(Socket obj)
void SocketClose(Var stack[]) // (Lwaba/io/Socket;)Z
{
   WObject socket = stack[0].obj;
   WOBJ_SocketLastErr(socket) = 0;
   RETURN32(_SocketClose(socket))
}

/*---------------------------------------------------------------SocketIsOpen-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketIsOpen(Socket obj)
void SocketIsOpen(Var stack[]) // (Lwaba/io/Socket;)Z
{
   WObject socket = stack[0].obj;
   RETURN32((WOBJ_SocketRef(socket) == -1)?0:1)
}

/*-------------------------------------------------------SocketSetReadTimeout-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketSetReadTimeout(Socket obj, int millis)
void SocketSetReadTimeout(Var stack[]) // (Lwaba/io/Socket;I)Z
{
   WObject socket = stack[0].obj;
   int32 millis = stack[1].intValue;

   WOBJ_SocketLastErr(socket) = 0;
   if (millis < 0)
   {
      RETURN32(0)
   }
   WOBJ_SocketTimeout(socket) = millis;
   RETURN32(1)
}

/*------------------------------------------------------------SocketReadBytes-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static int socketReadBytes(Socket obj, byte buf[], int start, int count)
void SocketReadBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   SocketReadWriteBytes(stack, 1);
}

/*-----------------------------------------------------------SocketWriteBytes-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static int socketWriteBytes(Socket obj, byte buf[], int start, int count)
void SocketWriteBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   SocketReadWriteBytes(stack, 0);
}

/*-----------------------------------------------------------SocketDisconnect-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/NativeMethods boolean socketDisconnect()
void SocketDisconnect(Var stack[]) // ()Z
{
   Err err;
   int32 v = 0;

   if (vmGlobals->globalNetState == NET_IS_OPEN)
   {
      _unused_ uint16 refnum = vmGlobals->globalSocketLibRefNum;

      UNLOCK_MEMORY
      do
      {
         err = NetLibClose(PALM_REFNUM(refnum) true);
      } while (err == netErrStillOpen); // guich@512_4: keep calling until there's surely no more instances
      NetLibFinishCloseWait(PALM_REFNUM(refnum)); // guich@512_4: mortal bullet
      LOCK_MEMORY

      if (err == netErrNotOpen) err = 0;
      v = err == 0;
      if (err) (*vmGlobals->debug)((*vmGlobals->int2str)("err on SocketDisconnect",err));
      vmGlobals->globalNetState = NET_READY_FOR_OPEN;
      EvtEnqueueKey(0x030D,0x030D, commandKeyMask); // guich@300_38: add this to cut the line in the kyocera phone
   }
   RETURN32(v)
}

#else //!defined(PALMOS)

#if !defined(SOCKETS_DISABLED)
#include "SDL_net.h"
#endif

/*-------------------------------------------------------socketReadWriteBytes-+
|                                                                             |
+----------------------------------------------------------------------------*/
static int32 socketReadWriteBytes(Var stack[], int isRead)
{
#if defined(SOCKETS_DISABLED)
   return -1;
#else
   WObject socketObj = stack[0].obj;
   WObject byteArray = stack[1].obj;
   int32 start = stack[2].intValue;
   int32 count = stack[3].intValue;
   int32 timeout = WOBJ_SocketTimeout(socketObj);
   TCPsocket socket = (TCPsocket)WOBJ_SocketRef(socketObj);
   int32 gotten = 0;
   SDLNet_SocketSet set;

   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SocketLastErr(socketObj) = -14;
      return -1; // array null or range invalid
   }
   else if (!count || !socket)
   {
      return 0;
   }
   else
   {
      int32 tMax=0;
      uchar *bytes;
      WOBJ_SocketLastErr(socketObj) = 0;
      bytes = ((uchar *)WOBJ_arrayStart(byteArray)) + start;

      if (timeout > 0) {
         tMax = SDL_GetTicks() + timeout;
         set = 0;
         #if defined(__SYMBIAN32__) || (defined WIN32 && !defined WINCE) // pgr@520_3: timeout is buggy on win32.
         // no select() (SDLnetselect is useless)
            if (isRead) { //fdie@500 SDLNet_CheckSockets is for connection and read activity, not "write ability"
               set = SDLNet_AllocSocketSet(1);
               if (!set || (SDLNet_TCP_AddSocket(set, socket) == -1))
               {
                  SDLNet_FreeSocketSet(set);
                  return -1;  // just in case: out of memory
               }
            }
         #endif
      }
      else
      {
         set = 0;
         tMax = 0;  // don't care: we will wait for the default timeout.
      }
      for (;;) {
         int32 n;
         if (set && SDLNet_CheckSockets(set, timeout) < 1)
            break; // -1 is an error, 0 is timeout occured
         if (isRead)
            n = SDLNet_TCP_Recv(socket, bytes+gotten, count-gotten,timeout);
         else
            n = SDLNet_TCP_Send(socket, bytes+gotten, count-gotten);
         /*
         | Now, what? We may have read/write those `n' bytes in 1 microsec,
         | and the caller allowed us up to 5 seconds to read `count' bytes.
         | If we have enough time left, why not just to pursue at reading?
         */
         if ((n <= 0) || ((gotten += n) >= count) || ((tMax - (int32)SDL_GetTicks()) < 10)) // guich@552_35: changed to int32, to allow negative values
            break;
      }
      if (set) SDLNet_FreeSocketSet(set);
      return (gotten > 0)? gotten : -1;
   }
#endif
}

/*--------------------------------------------------------------socketDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void socketDestroy(WObject socketObj)
{
#if !defined(SOCKETS_DISABLED)
   TCPsocket socket = (TCPsocket)(WOBJ_SocketRef(socketObj));
   if (socket)
   {
      SDLNet_TCP_Close(socket);
   }
#endif
}

/*---------------------------------------------------------------SocketCreate-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/NativeMethods void socketCreate(Socket obj, String host, int port, int timeout, boolean noLinger)
void SocketCreate(Var stack[]) // (Lwaba/io/Socket;Ljava/lang/String;IIZ)V
{
#if !defined(SOCKETS_DISABLED)
   IPaddress serverIP;
   TCPsocket socket;
   WObject socketObj = stack[0].obj;
   UtfString server = (*vmGlobals->stringToUtf)(stack[1].obj, STU_NULL_TERMINATE | STU_USE_STATIC);
   Uint16 port = (Uint16)stack[2].intValue;
   // noLinger parameter ignored

   WOBJ_SocketTimeout(socketObj) = stack[3].intValue;
   WOBJ_SocketRef(socketObj) = (int)0;
   WOBJ_SocketLastErr(socketObj) = 4616;   // PalmOS: SocketNotOpen

   if (!vmGlobals->isNetStarted)
   {
      if (SDLNet_Init() < 0)
         return;
      else
         vmGlobals->isNetStarted = 1;
   }
   if (server.len == 0)
      return;
   SDLNet_ResolveHost(&serverIP, server.str, port);
   if (serverIP.host == INADDR_NONE)      // Couldn't resolve hostname
      return;
   socket = SDLNet_TCP_Open(&serverIP);
   if (socket)
   {
      WOBJ_SocketRef(socketObj) = (int)socket;
      WOBJ_SocketLastErr(socketObj) = 0;
   }
#endif
}

/*----------------------------------------------------------------SocketClose-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketClose(Socket obj)
void SocketClose(Var stack[]) // (Lwaba/io/Socket;)Z
{
#if !defined(SOCKETS_DISABLED)
   TCPsocket socket = (TCPsocket)(WOBJ_SocketRef(stack[0].obj));

   if (socket)
   {
      SDLNet_TCP_Close(socket);
      WOBJ_SocketRef(stack[0].obj) = 0;
   }
#endif
   RETURN32(1)
}

/*---------------------------------------------------------------SocketIsOpen-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketIsOpen(Socket obj)
void SocketIsOpen(Var stack[]) // (Lwaba/io/Socket;)Z
{
#if defined(SOCKETS_DISABLED)
   RETURN32(0);
#else
   RETURN32(((TCPsocket)(WOBJ_SocketRef(stack[0].obj)) == 0)? 0 : 1)
#endif
}

/*-------------------------------------------------------SocketSetReadTimeout-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static boolean socketSetReadTimeout(Socket obj, int millis)
void SocketSetReadTimeout(Var stack[]) // (Lwaba/io/Socket;I)Z
{
#ifndef PALMOS // fdie@ SDL_net not yet available for palmOS
   WOBJ_SocketTimeout(stack[0].obj) = stack[1].intValue;
   RETURN32(1)
#else
   RETURN32(0)
#endif
}

/*------------------------------------------------------------SocketReadBytes-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static int socketReadBytes(Socket obj, byte buf[], int start, int count)
void SocketReadBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   RETURN32(socketReadWriteBytes(stack, 1))
}

/*-----------------------------------------------------------SocketWriteBytes-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/Socket static int socketWriteBytes(Socket obj, byte buf[], int start, int count)
void SocketWriteBytes(Var stack[]) // (Lwaba/io/Socket;[BII)I
{
   RETURN32(socketReadWriteBytes(stack, 0))
}

#ifdef WINCE
static int32 hangUpRas(HRASCONN hRasConn) // guich@550_44: used below
{
   int32 err=0;
   DWORD dwRet;
   RASCONNSTATUS rStatus;
   xmemzero(&rStatus, sizeof(RASCONNSTATUS));
   rStatus.dwSize = sizeof(RASCONNSTATUS);
   dwRet = RasGetConnectStatus(hRasConn, &rStatus);
   if (dwRet != ERROR_INVALID_HANDLE)
   {
      MSG msg;
      err = RasHangUp(hRasConn);  // hRasConn - valid handle to the RAS connection
      while (dwRet != ERROR_INVALID_HANDLE)
      {
         while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
         {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
         }
         dwRet = RasGetConnectStatus(hRasConn, &rStatus); // guich@551_4: putted this inside the loop
      }
   }
   return err;
}
#endif

/*-----------------------------------------------------------SocketDisconnect-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/io/NativeMethods boolean socketDisconnect()
void SocketDisconnect(Var stack[]) // ()Z
{
   int32 bOk = 1;
#ifdef WINCE // guich@550_44: close all working connections
   RASCONN ras[20];
   DWORD dSize,dNumber,dCount;

   bOk = 0;
   ras[0].dwSize = sizeof(RASCONN);
   dSize = sizeof( ras );

   // Get active RAS - Connection
   if( RasEnumConnections( ras, &dSize, &dNumber ) == 0 )
   {
      bOk = 1;
      for( dCount = 0; dCount < dNumber; dCount++ )
      {
         // Hang up that connection
         if( hangUpRas(ras[dCount].hrasconn) != 0 )
         {
            bOk = 0;
            break;
         }
      }
   }
#endif
#if !defined(SOCKETS_DISABLED)
   SDLNet_Quit();
#endif
   vmGlobals->isNetStarted = 0;
   RETURN32(bOk)
}

#endif //!defined(PALMOS)

/*===========================================================================*/
