/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Sound.c,v 1.7 2005/07/13 23:13:18 fdie Exp $ */

// guich@511_3: enable for win32
// and use PalmOS syscalls until SDL_audio is ported on PalmOS
#if !defined(WIN32) && !defined(PALMOS)
#define DISABLE_SOUND
#endif
/* >>>PGR: ENHANCE (12/26/04)
| The Sound is actually disabled on Symbian, altho it "works".
|
| The main problem is "latency", this is because:
|
| 1) the AudioServer is started and stopped for each sound:
| SDL_OpenAudio creates the AudioServer, SDL_CloseAudio deletes it
| (via: EPOC_DeleteDevice(_THIS)  in SDL/src/audio/epoc/SDL_epocAudio.cpp)
| The reason is that I was not aware of SDL_BuildAudioCVT that build filters
| to convert any audio stream to the appropriate device format.
| => there is no need to Open / Close the audio: use SDL_BuildAudioCVT
| A good example can be found in SDL_Mixer/wavestream.c and also in the
| doc, for SDL_ConvertAudio
|
| 2) In this code, I wait for the Audio to be finished.  Filling the AudioBuffer
| is done by a thread: no need for waiting!
|
| 3) Load of the WAVE file (and conversion) should be done only once.
| This is what WOBJ_SoundClipRef(sound) should reference.
|
| I estimate these changes between 2 and 4 days of work.  LATER!
*/
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "superwaba.h"
#include "SW_natives.h"
#include "SW_palmDb.h"

#define WOBJ_SoundClipPath(o) (o)[1].obj
#define WOBJ_SoundClipRef(o)  (o)[2].refValue
#define WOBJ_SoundClipSize(o) (o)[3].intValue  // petrus@421_01

// Empty natives for the platforms without SDL_audio
#if defined(DISABLE_SOUND)

/*------------------------------------------------------------------soundTone-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundTone(_unused_ int freq, _unused_ int duration)
{
}

/*------------------------------------------------------------------SoundBeep-+
| // waba/fx/Sound static void soundBeep()                                    |
+----------------------------------------------------------------------------*/
void SoundBeep(_unused_ Var stack[]) // ()V
{
}

/*------------------------------------------------------------------SoundTone-+
| waba/fx/Sound static void soundTone(int freq, int duration)                 |
+----------------------------------------------------------------------------*/
void SoundTone(_unused_ Var stack[]) // (II)V
{
}

/*------------------------------------------------------------soundSetEnabled-+
| waba/fx/Sound static void soundSetEnabled(boolean on)                       |
+----------------------------------------------------------------------------*/
void SoundSetEnabled(_unused_ Var stack[]) // (Z)V
{
}

/*--------------------------------------------------------------soundClipPlay-+
| waba/fx/SoundClip static boolean soundClipPlay(SoundClip obj)               |
+----------------------------------------------------------------------------*/
void SoundClipPlay(_unused_ Var stack[]) // (Lwaba/fx/SoundClip;)Z
{
}

/*-----------------------------------------------------------soundClipDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundClipDestroy(_unused_ WObject sound)
{
}

/*--------------------------------------------------------------SoundClipFree-+
| waba/fx/NativeMethods void soundClipFree(SoundClip obj)                     |
+----------------------------------------------------------------------------*/
void SoundClipFree(_unused_ Var stack[]) // (Lwaba/fx/SoundClip;)V
{
}

// PalmOS native system calls until SDL audio is done
#elif defined PALMOS

#include <stdlib.h>

/*------------------------------------------------------------------soundTone-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundTone(int freq, int duration)
{
   SndCommandType sndCmd;
   sndCmd.cmd = sndCmdFreqDurationAmp;
   sndCmd.param1 = freq; // freq
   sndCmd.param2 = duration; // duration
   sndCmd.param3 = sndMaxAmp;
   UNLOCK_MEMORY // guich@421_29: try to make this work in Treo600
   SndDoCmd(NULL, &sndCmd, 0);
   LOCK_MEMORY
}

/*------------------------------------------------------------------SoundBeep-+
| // waba/fx/Sound static void soundBeep()                                    |
+----------------------------------------------------------------------------*/
void SoundBeep(_unused_ Var stack[]) // ()V
{
   UNLOCK_MEMORY
   SndPlaySystemSound(sndInfo);
   LOCK_MEMORY
}

/*------------------------------------------------------------------SoundTone-+
| waba/fx/Sound static void soundTone(int freq, int duration)                 |
+----------------------------------------------------------------------------*/
void SoundTone(Var stack[]) // (II)V
{
   (*vmGlobals->tone)(stack[0].intValue,stack[1].intValue);
}

/*------------------------------------------------------------soundSetEnabled-+
| waba/fx/Sound static void soundSetEnabled(boolean on)                       |
+----------------------------------------------------------------------------*/
void SoundSetEnabled(Var stack[]) // (Z)V
{
   UInt16 amp = stack[0].intValue ? vmGlobals->defSoundAmp : 0; // if turn on, set it to the default user selection
   SndSetDefaultVolume (NULL, &amp, NULL);
}

/*--------------------------------------------------------------soundClipPlay-+
| waba/fx/SoundClip static boolean soundClipPlay(SoundClip obj)               |
+----------------------------------------------------------------------------*/
void SoundClipPlay(Var stack[]) // (Lwaba/fx/SoundClip;)Z
{
   WObject sound, pathString;
   UtfString path;
   byte *wav;
   byte *rec;
   int32 v=0;
   // Until the sampled sound feature exists, check the trap address instead.
#if 0 //TODO SysGetTrapAddress is not available
   if (SysGetTrapAddress(sysTrapSndPlayResource) == SysGetTrapAddress(sysTrapSysUnimplemented))
   {
      RETURN32(0)
   }
#endif        
   sound = stack[0].obj;
   pathString = WOBJ_SoundClipPath(sound);
   path = (*vmGlobals->stringToUtf)(pathString, STU_NULL_TERMINATE | STU_USE_STATIC);
   rec = (*vmGlobals->lockWarpRec)(path.str, path.len, NULL);
   if (rec != NULL)
   {
      wav = rec;
      if ((int32)wav & 1) wav++; // guich@421_76: warp now aligns it for us.
      UNLOCK_MEMORY // guich@400_11
      v = SndPlayResource(wav, sndGameVolume, sndFlagSync);  // must use sync because we're going to free the buffer right away
      LOCK_MEMORY
      (*vmGlobals->unlockWarpRec)(rec, path.len);
   }
   RETURN32(v)
}

/*-----------------------------------------------------------soundClipDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundClipDestroy(_unused_ WObject sound)
{
}

/*--------------------------------------------------------------SoundClipFree-+
| waba/fx/NativeMethods void soundClipFree(SoundClip obj)                     |
+----------------------------------------------------------------------------*/
void SoundClipFree(_unused_ Var stack[]) // (Lwaba/fx/SoundClip;)V
{
}

#else

///////// POSIX /////////
#include "SDL.h"
#include <string.h>

typedef struct
{
   SDL_sem * playEnded;
   int isEnabled;
   Uint8 * buffer;
   Uint8 * bufMax;
   Uint8 * bufCurr;
} WaveCallbackData;

typedef struct
{
   SDL_sem * playEnded;
   int isEnabled;
   unsigned int timeCurr;
   unsigned int timeMax;
   char const * val;
} ToneCallbackData;

static short const sineVals[] =
{
   -9720, -9591, -9267, -8778, -8076, -7230, -6228, -5094, -3855, -2529, -1176,
   207, 1605, 2964, 4269, 5478, 6597, 7593, 8412, 9084, 9558, 9846, 9921, 9630,
   9342, 8778, 8127, 7299, 6291, 5202, 3999, 2700, 1350, -21, -1389, -2733, -4050,
   -5265, -6372, -7365, -8184, -8862, -9309, -9618
   /*
   | the reference for a 1000Hz wave at -20 dB is:
   | -3240, -3197, -3089, -2926, -2692, -2410, -2076, -1698,
   | -1285,  -843,  -392,   +69,  +535,  +988, +1423, +1826,
   | +2199, +2531, +2804, +3028, +3186, +3282, +3307, +3210,
   | +3114, +2926, +2709, +2433, +2097, +1734, +1333,  +900,
   |  +450,    -7,  -463,  -911, -1350, -1755, -2124, -2455,
   | -2728, -2954, -3103, -3206
   */
};

static char const * const sineValsMax = (char *)sineVals + sizeof(sineVals);

/*---------------------------------------------------------------waveCallback-+
| "Play a Wave" callback                                                      |
+----------------------------------------------------------------------------*/
static void waveCallback(void * udata, Uint8 * stream, int len)
{
   WaveCallbackData d = *(WaveCallbackData *)udata;
   if (d.isEnabled)
   {
      if (d.bufCurr < d.bufMax)
      {
         if (d.bufCurr+len > d.bufMax)
         {
            len = d.bufMax - d.bufCurr;
         }
         ((WaveCallbackData *)udata)->bufCurr += len;
         memcpy(stream, d.bufCurr, len);
//       SDL_MixAudio(stream, d.bufCurr, len, SDL_MIX_MAXVOLUME);
      }
      else
      {
         d.isEnabled = 0;   /* disable self */
         SDL_SemPost(d.playEnded);
      }
   }
}

/*------------------------------------------------------------------playSound-+
|                                                                             |
+----------------------------------------------------------------------------*/
static int playSound(SDL_RWops * file)
{
   SDL_AudioSpec waveSpec;
   Uint8 * buffer;
   Uint32 bufSize;
   int rc = 0;

   if (SDL_LoadWAV_RW(file, 1, &waveSpec, &buffer, &bufSize))
   {
      WaveCallbackData d;
      d.isEnabled = 0;                   // disable our waveCallback
      waveSpec.callback = waveCallback;  // "Play a Wave" callback
      waveSpec.userdata = &d;            // waveCallback data
      if (SDL_OpenAudio(&waveSpec, 0) >= 0)
      {
         d.playEnded = SDL_CreateSemaphore(0);
         d.buffer = buffer;
         d.bufCurr = d.buffer;
         d.bufMax = d.buffer + bufSize;
         d.isEnabled = 1;
         SDL_PauseAudio(0);            // make sure we play
         SDL_SemWait(d.playEnded);     // wait here until the callback finishes
         SDL_DestroySemaphore(d.playEnded);
         SDL_CloseAudio();
         rc = 1;
      }
      SDL_FreeWAV(buffer);
   }
   return rc;
}

/*---------------------------------------------------------------toneCallback-+
| "Play a Tone" callback                                                      |
+----------------------------------------------------------------------------*/
static void toneCallback(void * udata, Uint8 * stream, int len)
{
   ToneCallbackData d = *(ToneCallbackData *)udata;
   if (d.isEnabled)
   {
      if (d.timeCurr < d.timeMax)
      {
         if (d.timeCurr+len > d.timeMax)
            len = d.timeMax - d.timeCurr;
         ((ToneCallbackData *)udata)->timeCurr += len;
         len /= 2;
         while (len--)
         {
            if (d.val >= sineValsMax)
               d.val = (char const *)sineVals;
            *stream++ = *(d.val++);
            *stream++ = *(d.val++);
         }
         ((ToneCallbackData *)udata)->val = d.val;
      }
      else
      {
         d.isEnabled = 0;   /* disable self */
         SDL_SemPost(d.playEnded);
      }
   }
}

/*-------------------------------------------------------------------playTone-+
|                                                                             |
+----------------------------------------------------------------------------*/
static int playTone(int freq, int duration)
{
   SDL_AudioSpec waveSpec;
   ToneCallbackData d;
   double ratio = freq / 1000.0;
   int rc = 0;

   waveSpec.freq = (int)(44100 * ratio);  // Samples per sec (FM Radio quality)
   waveSpec.format = AUDIO_S16LSB;        // Signed 16-bit little-endian samples
   waveSpec.channels = 1;                 // Number of sound channels (stereo)
   waveSpec.silence = 0;
   waveSpec.samples = 4096;               // Size of the Audio buffer in samples
   waveSpec.padding = 0;
   waveSpec.size = 8192;
   waveSpec.callback = toneCallback;      // "Play a Tone" callback
   waveSpec.userdata = &d;                // waveCallback data

   d.isEnabled = 0;                       // disable our toneCallback
   if (SDL_OpenAudio(&waveSpec, 0) >= 0)
   {
      d.playEnded = SDL_CreateSemaphore(0);
      d.timeCurr = 0;
      d.timeMax = (int)(82 * duration * ratio);
      d.isEnabled = 1;
      d.val = (char const *)sineVals;
      SDL_PauseAudio(0);            // make sure we play
      SDL_SemWait(d.playEnded);     // wait here until the callback finishes
      SDL_DestroySemaphore(d.playEnded);
      SDL_CloseAudio();
   }
   return rc;
}

////// COMMON PART ///////

/*------------------------------------------------------------------soundTone-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundTone(int freq, int duration)
{
   playTone(freq, duration);
}

/*------------------------------------------------------------------SoundBeep-+
| // waba/fx/Sound static void soundBeep()                                    |
+----------------------------------------------------------------------------*/
void SoundBeep(Var stack[]) // ()V
{
#ifdef WIN32
   MessageBeep(MB_OK);
#else
   playTone(1000, 100); // guich@511_3: gosh, in win32 this is somewhat sloooowww! There's a huge gap between tones....
#endif
}

/*------------------------------------------------------------------SoundTone-+
| waba/fx/Sound static void soundTone(int freq, int duration)                 |
+----------------------------------------------------------------------------*/
void SoundTone(Var stack[]) // (II)V
{
   playTone(stack[0].intValue, stack[1].intValue);
}

/*------------------------------------------------------------soundSetEnabled-+
| waba/fx/Sound static void soundSetEnabled(boolean on)                       |
+----------------------------------------------------------------------------*/
void SoundSetEnabled(Var stack[]) // (Z)V
{
   int amp = stack[0].intValue ? 1 : 0; // if turn on, set it to the default user selection
   SDL_PauseAudio(amp);
}

/*--------------------------------------------------------------soundClipPlay-+
| waba/fx/SoundClip static boolean soundClipPlay(SoundClip obj)               |
+----------------------------------------------------------------------------*/
void SoundClipPlay(Var stack[]) // (Lwaba/fx/SoundClip;)Z
{
   WObject sound = stack[0].obj;

   // guich@421_40: try first to read from the attached pdbs
   // petrus@421-01: SDL_Audio needs a size

   byte * p = WOBJ_SoundClipRef(sound);
   uint32 size = WOBJ_SoundClipSize(sound);
   if (!p)
   {
      WObject pathString = WOBJ_SoundClipPath(sound);
      UtfString path = (*vmGlobals->stringToUtf)(
         pathString, STU_NULL_TERMINATE | STU_USE_STATIC
      );
      size = 0;   // so we get a full record
      p = (*vmGlobals->lockWarpRec)(path.str, (uint16)path.len, &size);
      if (!p)
      {
         RETURN32(playSound(SDL_RWFromFile(path.str, "rb")));
      }
      else
      {
         WOBJ_SoundClipRef(sound) = p;    // store the handle inside the file
         WOBJ_SoundClipSize(sound) = size; // store the size inside the file
      }
   }
   /*
   | WARNING:
   | "guich@421_76" add a 0x00 pad byte to align the data of .wav files
   | at an even boundary.  Therefore, the header is not recognized.
   | The "pad" is not counted in the pathLen, but is counted in the size.
   */
   if ((int)p & 1) 
   {
      p++; 
      size--;
   }
   RETURN32(playSound(SDL_RWFromMem(p, size)))
}

/*-----------------------------------------------------------soundClipDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void soundClipDestroy(WObject sound)
{
   byte * p = WOBJ_SoundClipRef(sound);
   if (p)
   {
      WObject pathString = WOBJ_SoundClipPath(sound);
      (*vmGlobals->unlockWarpRec)(p, WOBJ_arrayLen(WOBJ_StringCharArrayObj(WOBJ_SoundClipPath(sound))));
   }
   WOBJ_SoundClipSize(sound) = 0;
   WOBJ_SoundClipRef(sound) = 0;
}

/*--------------------------------------------------------------SoundClipFree-+
| waba/fx/NativeMethods void soundClipFree(SoundClip obj)                     |
+----------------------------------------------------------------------------*/
void SoundClipFree(Var stack[]) // (Lwaba/fx/SoundClip;)V
{
   soundClipDestroy(stack[0].obj);
}

#endif // else PALMOS
/*===========================================================================*/
