/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Window.c,v 1.7 2005/10/31 19:51:24 guich Exp $ */

#include "superwaba.h"
#include "SW_natives.h"
#include "SW_settings.h"
#include "SDL.h"

/////////// SIP MANAGEMENT FUNCTIONS //////////////

#ifdef HAS_SIP // guich@420_23: all these changes
CLSID keybSip;
int mustChangeSip;

/*-----------------------------------------------------------verifyCurrentsip-+
| detects which sips are available                                            |
+----------------------------------------------------------------------------*/
static int verifyCurrentSip(IMENUMINFO *info)
{
   CLSID current;
   SipGetCurrentIM(&current);
   if (keybSip.Data1 == 0 && lstrcmp(info->szName,TEXT("Keyboard")) == 0)   // store the first sip
      keybSip = info->clsid;
   if (info->clsid.Data1 == current.Data1) // is this the current one?
   {
      if (lstrcmp(info->szName,TEXT("Transcriber")) == 0) // is it the transcriber?
         mustChangeSip = 1;
   }
   return 1;
}

/*----------------------------------------------------------------initSipRect-+
| initialize the sip rectangle, to be able to restore when we exit            |
+----------------------------------------------------------------------------*/
static void initSipRect()
{
   // store the default rect
   SIPINFO si;
   int isNotDocked;
   xmemzero(&si, sizeof(SIPINFO));
   si.cbSize = sizeof (SIPINFO);
   SipGetInfo(&si); // must be before sip be hidden
   isNotDocked = (si.fdwFlags & SIPF_DOCKED) == 0; // guich@550_42: fixed sip getting bigger when it was visible on the sw app start

   SipShowIM(SIPF_OFF);
#ifdef __AYGSHELL_H__
   SHFullScreen(vmGlobals->mainHWnd, SHFS_HIDESIPBUTTON); // only supported in 3.0
#endif
   SipEnumIM(&verifyCurrentSip);
   if (mustChangeSip && keybSip.Data1 != 0)
   {
      SipSetCurrentIM(&keybSip); // go to the keyboard IM
      SipShowIM(SIPF_OFF);
      SipGetInfo(&si); // get info again
   }

   vmGlobals->sipRect = si.rcSipRect;
   if (isNotDocked) // a floating sip has a caption area and a thick border. the sip rect excludes this area, 
   {                // so we add it to sipRect. a docked sip does not contain such area.
      vmGlobals->sipRect.bottom += GetSystemMetrics(SM_CYCAPTION); 
      vmGlobals->sipRect.right -= vmGlobals->sipRect.left;
      vmGlobals->sipRect.left = 0;
   }
}
#endif

//////////////////////////////////////////////////////////////////////////
// waba/ui/Window static void windowCreate(Window obj)
void WindowCreate(Var stack[]) // (Lwaba/ui/Window;)V
{
   WObject win = stack[0].obj;
   WOBJ_ControlX(win) = 0;
   WOBJ_ControlY(win) = 0;
   WOBJ_ControlWidth(win) = vmGlobals->windowWidth;
   WOBJ_ControlHeight(win) = vmGlobals->windowHeight;
}
//////////////////////////////////////////////////////////////////////////
#define VK_HIDE 10000
#define VK_ONTOP 10001
#define VK_ONBOTTOM 10002
#define VK_SHOW 10003

// waba/ui/Window static void windowSetStatePosition(Window obj, int x, int y)
void WindowSetStatePosition(Var stack[]) // (Lwaba/ui/Window;II)V
{
   int x,y;
   x = stack[1].intValue;
   y = stack[2].intValue;
   // used to set the position of the virtual keyboard
   if ((x+y) > 6000)
      (*vmGlobals->showSip)((int16)((x == VK_HIDE || y == VK_HIDE)?0:1), (int16)((x == VK_ONTOP || y == VK_ONTOP)?1:(x == VK_SHOW || y == VK_SHOW)?2:0));
}
//////////////////////////////////////////////////////////////////////////
//waba/ui/NativeMethods static void windowUpdateScreen()
void WindowUpdateScreen(Var stack[]) // ()V  - guich@500_12
{
   SDL_UpdateRect(SDL_GetVideoSurface(), 0, 0, 0, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/MainWindow static void mainWindowCreate(MainWindow obj)
void MainWindowCreate(Var stack[]) // (Lwaba/ui/MainWindow;)V
{
   // window's bounds are set in WindowCreate
#ifdef HAS_SIP
   if (SipStatus() == SIP_STATUS_AVAILABLE)
      initSipRect();
#endif
#ifdef __AYGSHELL_H__
   if (!vmGlobals->useMinimizeButton) // guich@420_19
      SHDoneButton(vmGlobals->mainHWnd, SHDB_SHOW); // guich@320_48: only supported in 3.0
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/MainWindow static void mainWindowExit(int exitCode)
void MainWindowExit(Var stack[]) // (I)V
{
   (*vmGlobals->postStopEvent)(stack[0].intValue);
   vmGlobals->quitRightNow = 1; // guich@504_2
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/MainWindow static void mainWindowSetTimerInterval(int milliseconds)
void MainWindowSetTimerInterval(Var stack[]) // (I)V
{
   vmGlobals->globalTimerInterval = stack[0].intValue;
   vmGlobals->globalTimerStart = (*vmGlobals->getTimeStamp)();
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/MainWindow static String mainWindowGetCommandLine()
void MainWindowGetCommandLine(Var stack[]) // ()Ljava/lang/String;
{
   if (vmGlobals->commandLine && *vmGlobals->commandLine)
      RETURN32((*vmGlobals->createString)(vmGlobals->commandLine))
   else
      RETURN32(0)
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/NativeMethods void mainWindowSetThreadCount(int count)
void MainWindowSetThreadCount(Var stack[]) // (I)V
{
   vmGlobals->threadCount = stack[0].intValue;
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/NativeMethods boolean eventIsAvailable()
void EventIsAvailable(Var stack[]) // ()Z
{
   RETURN32((*vmGlobals->eventIsAvailable)())
}
//////////////////////////////////////////////////////////////////////////
// waba/ui/NativeMethods void windowPumpEvents()
void WindowPumpEvents(Var stack[]) // ()V
{
   if ((*vmGlobals->pumpEvents)(1) == -1)
   {
      // just break the current program execution and fail gracefully
      (*vmGlobals->throwException)("waba/sys/AppExitException", "");
      vmGlobals->quitRightNow = 1;   // guich@552_1: make sure that the program will quit
   }
}
